;;; Tests for AFP's primitive procedures.

;;; John David Stone
;;; Department of Computer Science
;;; Grinnell College

;;; created August 23, 2011
;;; last revised January 12, 2017

(import (afp primitives)
        (only (scheme base) guard error-object? raise rational?)
        (support testing)
        (support testing-procedures))

;;; Test the div-and-mod procedure.

(suite div-and-mod ()
  (test positive-dividend-positive-divisor
    (div-and-mod 123 10)
    2 (exact-integer? exact-integer?)
    (lambda (quot rem)
      (and (= quot 12)
           (= rem 3))))
  (test positive-dividend-positive-divisor-no-remainder
    (div-and-mod 123 41)
    2 (exact-integer? exact-integer?)
    (lambda (quot rem)
      (and (= quot 3)
           (zero? rem))))
  (test zero-dividend-positive-divisor
    (div-and-mod 0 10)
    2 (exact-integer? exact-integer?)
    (lambda (quot rem)
      (and (zero? quot)
           (zero? rem))))
  (test negative-dividend-positive-divisor
    (div-and-mod -123 10)
    2 (exact-integer? exact-integer?)
    (lambda (quot rem)
      (and (= quot -13)
           (= rem 7))))
  (test negative-dividend-positive-divisor-no-remainder
    (div-and-mod -123 41)
    2 (exact-integer? exact-integer?)
    (lambda (quot rem)
      (and (= quot -3)
           (zero? rem))))
  (test zero-divisor
    (guard (raised-object
             ((error-object? raised-object) #t)
             (else (raise raised-object)))
      (div-and-mod 123 0))
    1 (boolean?)
    true?)
  (test positive-dividend-negative-divisor
    (div-and-mod 123 -10)
    2 (exact-integer? exact-integer?)
    (lambda (quot rem)
      (and (= quot -12)
           (= rem 3))))
  (test positive-dividend-negative-divisor-no-remainder
    (div-and-mod 123 -41)
    2 (exact-integer? exact-integer?)
    (lambda (quot rem)
      (and (= quot -3)
           (zero? rem))))
  (test zero-dividend-negative-divisor
    (div-and-mod 0 -10)
    2 (exact-integer? exact-integer?)
    (lambda (quot rem)
      (and (zero? quot)
           (zero? rem))))
  (test negative-dividend-positive-divisor
    (div-and-mod -123 -10)
    2 (exact-integer? exact-integer?)
    (lambda (quot rem)
      (and (= quot 13)
           (= rem 7))))
  (test negative-dividend-negative-divisor-no-remainder
    (div-and-mod -123 -41)
    2 (exact-integer? exact-integer?)
    (lambda (quot rem)
      (and (= quot 3)
           (zero? rem)))))

;;; Test the AFP expt procedure.

(suite afp-expt ()
  (test positive-exponent
    (expt 3 5)
    1 (exact-integer?)
    (match? = 243))
  (test zero-exponent
    (expt 344 0)
    1 (exact-integer?)
    (match? = 1))
  (test zero-zero-special-case
    (expt 0 0)
    1 (exact-integer?)
    (match? = 1))
  (test negative-exponent
    (expt 3 -5)
    1 (exact-rational?)
    (lambda (result)
      (= result 1/243)))
  (test fractional-base
    (expt -3/4 5)
    1 (exact-rational?)
    (match? = -243/1024))
  (test fractional-exponent
    (guard (raised-object
             ((error-object? raised-object) #t)
             (else (raise raised-object)))
      (expt 170 1/199))
    1 (boolean?)
    true?))

;;; Test the delist procedure.

(suite delist ()
  (test null-list
    (delist (list))
    0 ())
  (test singleton
    (delist (list 'hpp))
    1 (symbol?)
    (match? symbol=? 'hpp))
  (test long-list
    (delist (list 'giiadn 'rpt 'dcaaitu 'aaerok 'oyticrt 'hwilns))
    6 (symbol? symbol? symbol? symbol? symbol? symbol?)
    (match? symbol=? 'giiadn 'rpt 'dcaaitu 'aaerok 'oyticrt 'hwilns)))

;;; Test the rec syntax.

(suite rec ()
  (test name-only
    (rec factorial (lambda (n)
                     (if (zero? n)
                         1
                         (* n (factorial (- n 1))))))
    1 (procedure?)
    (lambda (fact)
      (suite factorial-internal ()
        (test zero
          (fact 0)
          1 (natural-number?)
          (match? = 1))
        (test larger
          (fact 8)
          1 (natural-number?)
          (match? = 40320)))))
  (test fixed-arity-procedure
    (rec (termial n)
      (if (zero? n)
          0
          (+ n (termial (- n 1)))))
    1 (procedure?)
    (lambda (term)
      (suite termial-internal ()
        (test zero
          (term 0)
          1 (natural-number?)
          zero?)
        (test larger
          (term 100)
          1 (natural-number?)
          (match? = 5050)))))
  (test variable-arity-procedure
    (rec (sum . arguments)
      (if (null? arguments)
          0
          (+ (car arguments) (apply sum (cdr arguments)))))
    1 (procedure?)
    (lambda (summer)
      (suite sum-internal ()
        (test no-arguments
          (summer)
          1 (natural-number?)
          zero?)
        (test more-arguments
          (summer 323 314 159 890 714)
          1 (natural-number?)
          (match? = (+ 323 314 159 890 714)))))))
    
;;; Test the receive syntax.

(suite receive ()

  ;; receive correctly processes a multiple-valued expression that has no
  ;; values.

  (test no-values
    (receive () (values)
      (list))
    1 (null?))

  ;; receive correctly, if pointlessly, processes a single-valued
  ;; expression.

  (test one-value
    (receive (unua) 'aodoh
      unua)
    1 (symbol?)
    (match? symbol=? 'aodoh))

  ;; receive correctly processes a fixed number of values from a
  ;; multiple-valued expression.

  (test several-values
    (receive (unua dua tria kvara kvina)
             (values 'oeedr 'hgn 'ttei 'tmtend 'tillrt)
      (list unua dua tria kvara kvina))
    1 (list?)
    (match? equal? (list 'oeedr 'hgn 'ttei 'tmtend 'tillrt)))

  ;; The variable-arity version of receive correctly processes a
  ;; multiple-valued expression with no values.

  (test variable-arity-no-values
    (receive whatever (values)
      whatever)
    1 (null?))

  ;; The variable-arity version of receive correctly processes a
  ;; single-valued expression.

  (test variable-arity-one-value
    (receive whatever 'dasti
      whatever)
    1 (list?)
    (match? equal? (list 'dasti)))

  ;; The variable-arity version of receive correctly processes any number
  ;; of values from a multiple-valued expression.

  (test variable-arity-several-values
    (receive whatever
             (values 'dataner 'tedao 'lefe 'eoncogt 'eentg 'eedei)
      whatever)
    1 (list?)
    (match? equal? (list 'dataner 'tedao 'lefe 'eoncogt 'eentg 'eedei)))

  ;; The semi-variable-arity version of receive correctly processes a
  ;; multiple-valued expression that produces the minimum number of values.

  (test semi-variable-arity-minimum-values
    (receive (unua dua tria . others)
             (values 'aynrnt 'eanlhss 'irmo)
      (apply list unua dua tria others))
    1 (list?)
    (match? equal? (list 'aynrnt 'eanlhss 'irmo)))

  ;; The semi-variable-arity version of receive correctly
  ;; processes a multiple-valued expression that produces more
  ;; than the minimum number of values.

  (test semi-variable-arity-more-than-minimum
    (receive (unua dua tria kvara . others)
             (values 'swlahn 'suo 'frahnos 'atffho 'edeu 'arr)
      (apply list unua dua tria kvara others))
    1 (list?)
    (match? equal? (list 'swlahn 'suo 'frahnos 'atffho 'edeu 'arr)))

  ;; The semi-variable-arity-version of receive correctly processes a
  ;; single-valued expression if there is only one required formal.

  (test semi-variable-arity-single-value
    (receive (unua . others) 'smoitc
      (apply list unua others))
    1 (list?)
    (match? equal? (list 'smoitc))))

;;; Test the sect syntax.

(suite sect ()
  (test typical
    (sect expt <> 3)
    1 (procedure?)
    (lambda (cube)
      (suite sect-typical-internal ()
        (test cube
          (cube 8)
          1 (exact-integer?)
          (match? = 512)))))
  (test no-slots
    (sect expt 5 3)
    1 (procedure?)
    (lambda (thunk)
      (suite no-slots-internal ()
        (test thaw-thunk
          (thunk)
          1 (exact-integer?)
          (match? = 125)))))
  (test variable-arity
    (sect - <> 259 <...>)
    1 (procedure?)
    (lambda (sected)
      (suite sect-variable-arity ()
        (test no-extra-arguments
          (sected 821)
          1 (exact-integer?)
          (match? = (- 821 259)))
        (test some-extra-arguments
          (sected 723 482 805 263)
          1 (exact-integer?)
          (match? = (- 723 259 482 805 263))))))
  (test all-slots
    (sect <> <> <>)
    1 (procedure?)
    (lambda (binary-call)
      (suite all-slots-internal ()
        (test call
          (binary-call expt 291 880)
          1 (exact-integer?)
          (match? = (expt 291 880)))))))

;;; Test the natural-number? procedure.

(suite natural-number? ()
  (test zero
    (natural-number? 0)
    1 (boolean?)
    true?)
  (test successor
    (natural-number? 970)
    1 (boolean?)
    true?)
  (test non-number
    (natural-number? 'snflcr)
    1 (boolean?)
    not)
  (test non-integer
    (natural-number? 830/531)
    1 (boolean?)
    not)
  (test inexact
    (natural-number? 279.0)
    1 (boolean?)
    not)
  (test negative
    (natural-number? -493)
    1 (boolean?)
    not))    

;;; Make sure that null is null.

(suite null ()
  (test is-null-null?
    (null? null)
    1 (boolean?)
    true?))

;;; Test the source syntax.

(suite source ()
  (test all-zeroes
    ((rec all-zeroes (source (values 0 all-zeroes))))
    2 (natural-number? procedure?)
    (lambda (alpha beta)
      (and (zero? alpha)
           (suite all-zeroes-again ()
             (test second-tap
               (beta)
               2 (natural-number? procedure?)
               (lambda (gamma delta)
                 (and (zero? gamma)
                      (equal? delta beta)))))))))

;;; Test the end-of-source and end-of-source? procedures.

(suite end-of-source ((empty-source
                       (rec empty (source (values (end-of-source) empty))))
                      (empty-source?
                       (lambda (src)
                         (receive (element remainder) (src)
                           (end-of-source? element)))))
  (test in-isolation
    (end-of-source)
    1 (end-of-source?))
  (test as-terminator
    (source (values 'ioertdn (source (values 'eovhdh empty-source))))
    1 (procedure?)
    (lambda (finite-source)
      (suite first-of-finite-source ()
        (test first-tap
          (finite-source)
          2 (symbol? procedure?)
          (lambda (alpha beta)
            (and (symbol=? alpha 'ioertdn)
                 (suite second-of-finite-source ()
                   (test second-tap
                     (beta)
                     2 (symbol? procedure?)
                     (lambda (gamma delta)
                       (and (symbol=? gamma 'eovhdh)
                            (suite end-of-finite-source ()
                              (test third-tap
                                (delta)
                                2 (end-of-source? procedure?))))))))))))))


;;; copyright (C) 2011, 2017 John David Stone

;;; This program is free software.  You may redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation -- either version 3 of the License, or (at
;;; your option) any later version.  A copy of the GNU General Public
;;; License is available on the World Wide Web at
;;;
;;;                http://www.gnu.org/licenses/gpl.html
;;;
;;; This program is distributed in the hope that it will be useful, but
;;; WITHOUT ANY WARRANTY -- without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; General Public License for more details.
