;;; This is the test program for the procedures introduced in the "Sorting"
;;; chapter.

;;; John David Stone
;;; Department of Computer Science
;;; Grinnell College

;;; created September 27, 2011
;;; last revised January 6, 2017

(import (afp primitives)
        (afp arithmetic)
        (afp extras)
        (afp constant-procedures)
        (only (afp couplers) pipe)
        (only (afp adapters) identity converse ~each compare-by)
        (only (afp predicate-operations) ^et ^vel ^if)
        (only (afp natural-numbers)
              fold-natural lower-ply-natural conditionally-tally)
        (only (afp pairs) pair=? pair-of pair-of=)
        (only (afp boxes) box box-of box-of=)
        (only (afp lists)
              first rest empty-list? prepend list-of list-of= fold-list
              unfold-list extend-to-variable-arity ^or run every)
        (only (afp trees)
              make-empty-tree empty-tree? make-non-empty-tree
              singleton-tree tree? tree-of tree-of= rightmost leftmost
              extract-rightmost tree->list)
        (only (afp bushes) bush bush? empty-bush? bush-of bush-of=)
        (only (afp bags)
              bag? bag put-into-bag empty-bag? bag=? bag-of bag-of=
              bag-union bag-cardinality list->bag partition-bag-with-count)
        (only (afp sets) empty-set? set set? set=? set->bag)
        (afp ordering-relations)
        (rename (afp sorting insertion-sort) (sort sort-by-insertion))
        (rename (afp sorting selection-sort) (sort sort-by-selection))
        (rename (afp sorting quicksort) (sort quicksort))
        (rename (afp sorting mergesort) (sort mergesort))
        (afp binary-search-trees)
        (rename (afp sorting binary-search-tree-sort)
                (sort binary-search-tree-sort))
        (afp red-black-trees)
        (afp red-black-tables)
        (afp heaps)
        (rename (afp sorting heapsort) (sort heapsort))
        (afp order-statistics)
        (support testing)
        (support testing-procedures)
        (support random))

;;; Test procedures from the "Ordering relations" section.

(suite ordered? ()
  (test empty-list
    (ordered? string>=? (list))
    1 (boolean?)
    true?)
  (test singleton
    (ordered? string>=? (list "mooitef"))
    1 (boolean?)
    true?)
  (test correctly-ordered-doubleton
    (ordered? string<=? (list "dmr" "utse"))
    1 (boolean?)
    true?)
  (test incorrectly-ordered-doubleton
    (ordered? string>=? (list "bndta" "oermtli"))
    1 (boolean?)
    not)
  (test long-correctly-ordered-list
    (ordered? <= ((unfold-list (sect = <> 1000) identity add1) 0))
    1 (boolean?)
    true?)
  (test long-slightly-disordered-list
    (ordered? <= (append ((unfold-list (sect = <> 500) identity add1) 0)
                         (list 501 500)
                         ((unfold-list (sect = <> 1000) identity add1) 502)))
    1 (boolean?)
    not)
  (test long-entirely-disordered-list
    (ordered? <= ((unfold-list zero? identity sub1) 1000))
    1 (boolean?)
    not))

(suite prior-by ()
  (test lesser
    (prior-by <=)
    1 (procedure?)
    (lambda (lesser)
      (suite lesser-internal ()
        (test former-is-lesser
          (lesser 281 756)
          1 (exact-integer?)
          (match? = 281))
        (test latter-is-lesser
          (lesser 933 274)
          1 (exact-integer?)
          (match? = 274))
        (test equals
          (lesser 611 611)
          1 (exact-integer?)
          (match? = 611)))))
  (test subtler-ordering-relation
    (prior-by (pipe (~each cdr) string>=?))
    1 (procedure?)
    (lambda (follower)
      (suite subtler-internal ()
        (test former-is-follower
          (follower (cons 874 "hdosbh") (cons 486 "ecr"))
          1 (pair?)
          (match? pair=? (cons 874 "hdosbh")))
        (test latter-is-follower
          (follower (cons 481 "hsaltir") (cons 752 "ldsepi"))
          1 (pair?)
          (match? pair=? (cons 752 "ldsepi")))
        (test pseudo-equals
          (follower (cons 709 "uepwei") (cons 753 "uepwei"))
          1 (pair?)
          (match? pair=? (cons 709 "uepwei")))))))

(suite arrange-by ()
  (test lesser-and-greater
    (arrange-by <=)
    1 (procedure?)
    (lambda (arranger)
      (suite lesser-and-greater-internal ()
        (test as-is
          (arranger 825 861)
          2 (exact-integer? exact-integer?)
          (match? = 825 861))
        (test reversed
          (arranger 847 785)
          2 (exact-integer? exact-integer?)
          (match? = 785 847))
        (test equal
          (arranger 514 514)
          2 (exact-integer? exact-integer?)
          (match? = 514 514)))))
  (test subtler-comparator
    (arrange-by (pipe (~each (conditionally-tally empty-list?
                                                  (pipe car odd?)
                                                  cdr))
                      >=))
    1 (procedure?)
    (lambda (arranger)
      (suite subtler-internal ()
        (test as-is
          (arranger (list 175 499 323) (list 251 841 698 340))
          2 ((list-of exact-integer?) (list-of exact-integer?))
          (match? (list-of= =) (list 175 499 323) (list 251 841 698 340)))
        (test reversed
          (arranger (list 613 633 955 430)
                    (list 791 9 945 104 282 890 285))
          2 ((list-of exact-integer?) (list-of exact-integer?))
          (match? (list-of= =) (list 791 9 945 104 282 890 285)
                               (list 613 633 955 430)))
        (test equal
          (arranger (list 680 264) (list))
          2 ((list-of exact-integer?) (list-of exact-integer?))
          (^vel (match? (list-of= =) (list 680 264) (list))
                (match? (list-of= =) (list) (list 680 264))))))))

(suite extreme ()
  (test typical
    (extreme >=)
    1 (procedure?)
    (lambda (extremer)
      (suite extreme-internal ()
        (test singleton
          (extremer (bag 490))
          1 (exact-integer?)
          (match? = 490))
        (test larger-bag
          (extremer (bag 280 979 853 445 952 893 578 880 826 178 978 622
                         972 704 234 426 913 511 107 739 728 340 336 469
                         436 795 942 34 57 668 600 372 87 127 640 118 296
                         89 335 313 388 866 805 366 284 111 326 978 741 623
                         626 265 7 49 512 122 865 416 857 306 621 728 387 5
                         83 97 363 264 487 545 449 854 60 412 190 4 625 629
                         473 362 649 623 889 765 10 815 194 422 277 606 905
                         241 380 447 338 552 532 99 46 600 131 284 204 793
                         431 505 418 434 561 815 607 949 411 93 196 969 143
                         258 746 960 967 733 691 705 33 35 753 351 435 516
                         461 236 504 703 125 805 133 584 927 782 994 51 509
                         503 409 457 206 97 95 408 696 714 793 131 742 848
                         966 945 215 912 945 206 591 597 60 501 441 146 974
                         337 467 306 83 556 107 558 120 631 133 702 855 824
                         422 253 381 311 139 482 173 229 507 190 904 129
                         233 511 239 42 728 437 64 886 422 198 28 218 315
                         952 958 439 266 491 613 443 418 847 148 685 554
                         279 393 998 125 549 695 151 551 484 758 671 469
                         473 207 436 588 734 871 14 899 749 473 74 359 499
                         229 542 369 381 152 449 743 926 48 632 218 66 951
                         27 686 398 252 197 997 472 781 724 849 402 563 393
                         841 426 77 593 969 485 873 271 82))
          1 (exact-integer?)
          (match? = 998))))))

(suite extreme-and-others ()
  (test typical
    (extreme-and-others string>=?)
    1 (procedure?)
    (lambda (extremer)
      (suite extreme-and-others-internal ()
        (test singleton
          (extremer (bag "msl"))
          2 (string? (bag-of string?))
          (lambda (alpha beta)
            (and (string=? alpha "msl")
                 (empty-bag? beta))))
        (test larger-bag
          (extremer (bag "aiu" "npo" "relsl" "ocayf" "tti" "eidaf" "hdaa"
                         "ongtun" "hehat" "oakss" "ttd" "eetniar" "ysyrdlt"
                         "nhshndu" "tetlemy"))
          2 (string? (bag-of string?))
          (lambda (alpha beta)
            (and (string=? alpha "ysyrdlt")
                 ((bag-of= string=?) beta (bag "aiu" "npo" "relsl" "ocayf"
                                               "tti" "eidaf" "hdaa"
                                               "ongtun" "hehat" "oakss"
                                                "ttd" "eetniar" "nhshndu"
                                                "tetlemy")))))))))

(suite extreme-in-set ()
  (test typical
    (extreme-in-set string>=?)
    1 (procedure?)
    (lambda (extremer)
      (suite extreme-in-set-internal ()
        (test singleton
          (extremer (set "tigy"))
          1 (string?)
          (match? string=? "tigy"))
        (test larger-set
          (extremer (set "hoost" "eso" "toglhe" "eagrurr" "oow" "sossrua"))
          1 (string?)
          (match? string=? "toglhe"))))))

(suite extreme-in-list ()
  (test typical
    (extreme-in-list string>=?)
    1 (procedure?)
    (lambda (extremer)
      (suite extreme-in-list-internal ()
        (test singleton
          (extremer (list "asnlr"))
          1 (string?)
          (match? string=? "asnlr"))
        (test longer-list
          (extremer (list "csitcw" "ksanot" "eatoeen" "vtdatat" "aompeua"
                          "ultst" "emteova" "rirhh" "heeip"))
          1 (string?)
          (match? string=? "vtdatat"))))))

(suite extreme-and-others-in-list ()
  (test typical
    (extreme-and-others-in-list <=)
    1 (procedure?)
    (lambda (extremer)
      (suite extreme-and-others-in-list-internal ()
        (test singleton
          (extremer (list 808))
          2 (exact-integer? (list-of exact-integer?))
          (lambda (alpha beta)
            (and (= alpha 808)
                 (empty-list? beta))))
        (test longer-list
          (extremer (list 386 824 895 633 790 131 390 869 645))
          2 (exact-integer? (list-of exact-integer?))
          (lambda (alpha beta)
            (and (= alpha 131)
                 ((list-of= =) beta
                               (list 386 824 895 633 790 390 869 645)))))
        (test duplicates
          (extremer (list 248 248 248 248 248 248 248 248))
          2 (exact-integer? (list-of exact-integer?))
          (lambda (alpha beta)
            (and (= alpha 248)
                 ((list-of= =) beta
                               (list 248 248 248 248 248 248 248)))))))))

(suite compound-ordering ()
  (test no-arguments
    (compound-ordering)
    1 (procedure?)
    (lambda (indifferent)
      (suite no-arguments-internal ()
        (test one-way
          (indifferent 'solrwoe 'gkl)
          1 (boolean?)
          true?)
        (test the-other-way
          (indifferent 'glk 'solrwoe)
          1 (boolean?)
          true?)
        (test identical
          (indifferent 'ceumaa 'ceumaa)
          1 (boolean?)
          true?))))
  (test one-argument
    (compound-ordering >=)
    1 (procedure?)
    (lambda (ge)
      (suite one-argument-internal ()
        (test less
          (ge 750 766)
          1 (boolean?)
          not)
        (test equal
          (ge 257 257)
          1 (boolean?)
          true?)
        (test greater
          (ge 646 473)
          1 (boolean?)
          true?))))
  (test several-arguments
    (compound-ordering (pipe (~each (sect div <> 100)) <=)
                       (pipe (~each (sect mod <> 10)) <=)
                       (pipe (~each (sect mod <> 7)) <=))
    1 (procedure?)
    (lambda (ord)
      (suite several-arguments-internal ()
        (test primary-yes
          (ord 47 821)
          1 (boolean?)
          true?)
        (test primary-no
          (ord 816 167)
          1 (boolean?)
          not)
        (test secondary-yes
          (ord 465 407)
          1 (boolean?)
          true?)
        (test secondary-no
          (ord 568 577)
          1 (boolean?)
          not)
        (test tertiary-yes
          (ord 589 549)
          1 (boolean?)
          true?)
        (test tertiary-no
          (ord 264 224)
          1 (boolean?)
          not)
        (test tie-unbroken
          (ord 178 108)
          1 (boolean?)
          true?)))))

(suite pair-lex ()
  (test typical
    (pair-lex <= string>=?)
    1 (procedure?)
    (lambda (ord)
      (suite typical-internal ()
        (test primary-yes
          (ord (cons 467 "feromp") (cons 747 "fhelao"))
          1 (boolean?)
          true?)
        (test primary-no
          (ord (cons 308 "heyea") (cons 69 "areoao"))
          1 (boolean?)
          not)
        (test secondary-yes
          (ord (cons 28 "oaeet") (cons 28 "hhootdi"))
          1 (boolean?)
          true?)
        (test secondary-no
          (ord (cons 47 "eesjcer") (cons 47 "tky"))
          1 (boolean?)
          not)
        (test tie-unbroken
          (ord (cons 553 "riire") (cons 553 "riire"))
          1 (boolean?)
          true?)))))

(suite fixed-list-lex ()
  (test no-arguments
    (fixed-list-lex)
    1 (procedure?)
    (lambda (nullity)
      (suite no-arguments-internal ()
        (test only-case
          (nullity (list) (list))
          1 (boolean?)
          true?))))
  (test one-argument
    (fixed-list-lex <=)
    1 (procedure?)
    (lambda (le)
      (suite one-argument-internal ()
        (test less
          (le (list 42) (list 888))
          1 (boolean?)
          true?)
        (test equal
          (le (list 811) (list 811))
          1 (boolean?)
          true?)
        (test greater
          (le (list 924) (list 696))
          1 (boolean?)
          not))))
  (test multiple-arguments
    (fixed-list-lex string<=? string<=? <=)
    1 (procedure?)
    (lambda (ord)
      (suite multiple-arguments-internal ()
        (test primary-yes
          (ord (list "ebreae" "sais" 427)
               (list "iietfha" "yinodtt" 407))
          1 (boolean?)
          true?)
        (test primary-no
          (ord (list "tvohre" "iiam" 734)
               (list "ihrfu" "asn" 199))
          1 (boolean?)
          not)
        (test secondary-yes
          (ord (list "amoddd" "eghoe" 772)
               (list "amoddd" "oereudr" 600))
          1 (boolean?)
          true?)
        (test secondary-no
          (ord (list "ehtls" "hro" 713)
               (list "ehtls" "drttn" 991))
          1 (boolean?)
          not)
        (test tertiary-yes
          (ord (list "fvoeh" "ppteta" 474)
               (list "fvoeh" "ppteta" 745))
          1 (boolean?)
          true?)
        (test tertiary-no
          (ord (list "ihdhmd" "lknim" 959)
               (list "ihdhmd" "lknim" 201))
          1 (boolean?)
          not)
        (test tie-unbroken
          (ord (list "diwac" "rco" 19)
               (list "diwac" "rco" 19))
          1 (boolean?)
          true?)))))               

(suite star<=? ((sirius (alternative-make-star "Alpha Canis Majoris"
                                               -142/100 27/10 'A1Vm))
                (canopus (alternative-make-star "Alpha Carinae"
                                                -72/100 30 'F0Ib)))
  (test by-name-yes
    (star<=? sirius canopus)
    1 (boolean?)
    true?)
  (test by-name-no
    (star<=? canopus sirius)
    1 (boolean?)
    not)
  (test by-magnitude-yes
    (star<=? sirius
             (alternative-make-star "Alpha Canis Majoris"
                                    371/100 959 'A1Vm))
    1 (boolean?)
    true?)
  (test by-magnitude-no
    (star<=? canopus
             (alternative-make-star "Alpha Carinae"
                                    -352/100 7/10 'F0Ib))
    1 (boolean?)
    not)
  (test tie-unbroken
    (star<=? sirius sirius)
    1 (boolean?)
    true?))

(suite list-lex ()
  (test typical
    (list-lex <=)
    1 (procedure?)
    (lambda (ord)
      (suite typical-internal ()
        (test initial-yes
          (ord (list 410 453 379 471) (list 753 738 964))
          1 (boolean?)
          true?)
        (test initial-no
          (ord (list 548 997 661) (list 300 964 255 544 813))
          1 (boolean?)
          not)
        (test mid-yes
          (ord (list 163 134 395 60 100) (list 163 134 395 307))
          1 (boolean?)
          true?)
        (test mid-no
          (ord (list 382 502 942 756) (list 382 502 467 901))
          1 (boolean?)
          not)
        (test left-prefix
          (ord (list 346 867 575) (list 346 867 575 736 457 107))
          1 (boolean?)
          true?)
        (test right-prefix
          (ord (list 360 627 136 961) (list 360))
          1 (boolean?)
          not)
        (test identical
          (ord (list 458 554 671 251 848 675 963)
               (list 458 554 671 251 848 675 963))
          1 (boolean?)
          true?))))
  (test string<=?
    (list-lex char<=?)
    1 (procedure?)
    (lambda (sle)
      (suite string<=?-internal ()
        (test both-null
          (sle (list) (list))
          1 (boolean?)
          true?)
        (test left-null
          (sle (list) (list #\n #\r #\n #\s #\o #\i))
          1 (boolean?)
          true?)
        (test right-null
          (sle (list #\d #\r #\e #\t) (list))
          1 (boolean?)
          not)
        (test yes
          (sle (list #\h #\h #\e #\n #\a #\g #\w #\i #\v)
               (list #\h #\h #\m #\i #\e))
          1 (boolean?)
          true?)
        (test no
          (sle (list #\o #\a #\m #\a #\d #\c #\e)
               (list #\o #\a #\m #\a #\c #\n #\a #\e))
          1 (boolean?)
          not)
        (test left-prefix
          (sle (list #\h #\e #\a #\n) (list #\h #\e #\a #\n #\t #\d))
          1 (boolean?)
          true?)
        (test right-prefix
          (sle (list #\y #\a #\h #\o #\r #\a) (list #\y #\a))
          1 (boolean?)
          not)
        (test identical
          (sle (list #\t #\e #\s #\a) (list #\t #\e #\s #\a))
          1 (boolean?)
          true?)))))

(suite set->bag ()
  (test empty-set
    (set->bag (set))
    1 (bag?)
    empty-bag?)
  (test larger-set
    (set->bag (set 459 643 924 312 194 537 602))
    1 (bag?)
    (match? bag=? (bag 459 643 924 312 194 537 602))))

(suite list->bag ()
  (test empty-list
    (list->bag (list))
    1 (bag?)
    empty-bag?)
  (test longer-list
    (list->bag (list 278 536 954 438 261 751 385 929))
    1 (bag?)
    (match? bag=? (bag 278 536 954 438 261 751 385 929))))

(suite inserter ()
  (test minimal
    (inserter values?)
    1 (procedure?)
    (lambda (ins)
      (suite minimal-internal ()
        (test empty-list
          (ins 'lsl (list))
          1 ((list-of symbol?))
          (match? (list-of= symbol=?) (list 'lsl)))
        (test longer-list
          (ins 'uonaho (list 'tst 'eaenhv 'ihw))
          1 ((list-of symbol?))
          (^or (match? (list-of= symbol=?)
                       (list 'uonaho 'tst 'eaenhv 'ihw))
               (match? (list-of= symbol=?)
                       (list 'tst 'uonaho 'eaenhv 'ihw))
               (match? (list-of= symbol=?)
                       (list 'tst 'eaenhv 'uonaho 'ihw))
               (match? (list-of= symbol=?)
                       (list 'tst 'eaenhv 'ihw 'uonaho)))))))
  (test typical
    (inserter <=)
    1 (procedure?)
    (lambda (ins)
      (suite typical-internal
          ((iota (lambda (n)
                   ((unfold-list zero? (sect - n <>) sub1) n))))
        (test empty-list
          (ins 271 (list))
          1 ((list-of exact-integer?))
          (match? (list-of= =) (list 271)))
        (test at-front-on-singleton
          (ins 165 (list 577))
          1 ((list-of exact-integer?))
          (match? (list-of= =) (list 165 577)))
        (test at-rear-on-singleton
          (ins 861 (list 411))
          1 ((list-of exact-integer?))
          (match? (list-of= =) (list 411 861)))
        (test at-front-on-long-list
          (ins 0 (map add1 (iota 281)))
          1 ((list-of exact-integer?))
          (match? (list-of= =) (iota 282)))
        (test at-rear-on-long-list
          (ins 295 (iota 295)) 
          1 ((list-of exact-integer?))
          (match? (list-of= =) (iota 296)))
        (test in-middle-of-long-list
          (ins 45 (append (iota 45) (map (sect + <> 46) (iota 238))))
          1 ((list-of exact-integer?))
          (match? (list-of= =) (iota (+ 46 238))))))))

;; Test procedures in the "Sorting algorithms" section.

(suite sort-by-insertion
    ((iota (lambda (n)
             ((unfold-list zero? (sect - n <>) sub1) n)))
     (rng (random-number-generator 19981109))
     (sample (let ((rng2 (random-number-generator 201110043)))
               ((fold-natural bag (sect put-into-bag (rng2 128) <>)) 235))))
  (test empty-list
    (sort-by-insertion values? (bag))
    1 (list?)
    empty-list?)
  (test singleton
    (sort-by-insertion values? (bag 'eneoai))
    1 ((list-of symbol?))
    (match? (list-of= symbol=?) (list 'eneoai)))
  (test larger-bag
    (sort-by-insertion >= ((lower-ply-natural bag put-into-bag) 479))
    1 ((list-of exact-integer?))
    (match? (list-of= =) (reverse (iota 479))))
  (test shuffled
    (sort-by-insertion <= ((fold-list bag put-into-bag)
                           (shuffle rng (iota 382))))
    1 ((list-of exact-integer?))
    (match? (list-of= =) (iota 382)))
  (test all-equal
    (sort-by-insertion >= ((fold-natural bag (sect put-into-bag 423 <>)) 89))
    1 ((list-of exact-integer?))
    (match? (list-of= =) ((fold-natural list (sect prepend 423 <>)) 89)))
  (test random-elements
    (sort-by-insertion <= sample)
    1 ((list-of exact-integer?))
    (^et (sect ordered? <= <>)
         (run delist bag (match? (bag-of= =) sample)))))

(suite sort-by-selection
    ((iota (lambda (n)
             ((unfold-list zero? (sect - n <>) sub1) n)))
     (rng (random-number-generator 20000304))
     (sample (let ((rng2 (random-number-generator 201110044)))
               ((fold-natural bag (sect put-into-bag (rng2 128) <>)) 116))))
  (test null-list
    (sort-by-selection values? (bag))
    1 (list?)
    empty-list?)
  (test singleton
    (sort-by-selection values? (bag 'eneoai))
    1 ((list-of symbol?))
    (match? (list-of= symbol=?) (list 'eneoai)))
  (test larger-bag
    (sort-by-selection <= (apply bag (iota 534)))
    1 ((list-of natural-number?))
    (match? (list-of= =) (iota 534)))
  (test shuffled
    (sort-by-selection <= ((fold-list bag put-into-bag)
                           (shuffle rng (iota 677))))
    1 ((list-of natural-number?))
    (match? (list-of= =) (iota 677)))
  (test all-equal
    (sort-by-selection >= ((fold-natural bag (sect put-into-bag 812 <>)) 736))
    1 ((list-of natural-number?))
    (match? (list-of= =) ((fold-natural list (sect prepend 812 <>)) 736)))
  (test random-elements
    (sort-by-selection <= sample)
    1 ((list-of natural-number?))
    (^et (sect ordered? <= <>)
         (run delist bag (match? (bag-of= =) sample)))))

(suite quicksort 
    ((iota (lambda (n)
             ((unfold-list zero? (sect - n <>) sub1) n)))
     (rng (random-number-generator 19990421))
     (sample (let ((rng2 (random-number-generator 201110045)))
               ((fold-natural bag (sect put-into-bag (rng2 128) <>)) 912))))
  (test empty-list
    (quicksort values? (bag))
    1 (list?)
    empty-list?)
  (test singleton
    (quicksort values? (bag 'hwi))
    1 ((list-of symbol?))
    (match? (list-of= symbol=?) (list 'hwi)))
  (test larger-bag
    (quicksort >= ((lower-ply-natural bag put-into-bag) 324))
    1 ((list-of exact-integer?))
    (match? (list-of= =) (reverse (iota 324))))
  (test shuffled
    (quicksort <= ((fold-list bag put-into-bag) (shuffle rng (iota 366))))
    1 ((list-of exact-integer?))
    (match? (list-of= =) (iota 366)))
  (test all-equal
    (quicksort >= ((fold-natural bag (sect put-into-bag 974 <>)) 819))
    1 ((list-of exact-integer?))
    (match? (list-of= =) ((fold-natural list (sect prepend 974 <>)) 819)))
  (test random-elements
    (quicksort <= sample)
    1 ((list-of exact-integer?))
    (^et (sect ordered? <= <>)
         (pipe (fold-list bag put-into-bag)
               (match? (bag-of= =) sample)))))

(suite split-bag-evenly
    ((evenly-split? (lambda (equiv? aro)
                      (lambda (left right)
                        (let ((left-size (bag-cardinality left))
                              (right-size (bag-cardinality right)))
                          (and (or (= left-size right-size)
                                   (= left-size (add1 right-size)))
                               (equiv? (bag-union left right) aro)))))))
  (test null-bag
    (split-bag-evenly (bag))
    2 (bag? bag?)
    (every empty-bag?))
  (test singleton
    (split-bag-evenly (bag 'absnokm))
    2 ((bag-of symbol?) (bag-of symbol?))
    (match? (bag-of= symbol=?) (bag 'absnokm) (bag)))
  (test doubleton
    (split-bag-evenly (bag 'nicrnee 'eedsgo))
    2 ((bag-of symbol?) (bag-of symbol?))
    (evenly-split? (bag-of= symbol=?) (bag 'nicrnee 'eedsgo)))
  (test odd-bag
    (split-bag-evenly ((lower-ply-natural bag put-into-bag) 591))
    2 ((bag-of natural-number?) (bag-of natural-number?))
    (evenly-split? (bag-of= =) ((lower-ply-natural bag put-into-bag) 591)))
  (test even-bag
    (split-bag-evenly ((lower-ply-natural bag put-into-bag) 632))
    2 ((bag-of natural-number?) (bag-of natural-number?))
    (evenly-split? (bag-of= =) ((lower-ply-natural bag put-into-bag) 632))))

(suite merge-lists ()
  (test empty-lists
    (merge-lists string>=? (list) (list))
    1 (list?)
    empty-list?)
  (test left-empty
    (merge-lists <= (list) (list 62 233 265 653 838))
    1 ((list-of exact-integer?))
    (match? (list-of= =) (list 62 233 265 653 838)))
  (test right-empty
    (merge-lists <= (list 87 279 340 346) (list))
    1 ((list-of exact-integer?))
    (match? (list-of= =) (list 87 279 340 346)))
  (test singletons-left-first
    (merge-lists <= (list 94) (list 267))
    1 ((list-of exact-integer?))
    (match? (list-of= =) (list 94 267)))
  (test singletons-right-first
    (merge-lists <= (list 822) (list 662))
    1 ((list-of exact-integer?))
    (match? (list-of= =) (list 662 822)))
  (test longer-lists
    (merge-lists <=
                 (list 12 399 434 481 682 697 759 777 827 924 991) 
                 (list 1 8 60 479 513 599 702 924))
    1 ((list-of exact-integer?))
    (match? (list-of= =) (list 1 8 12 60 399 434 479 481 513 599 682 697
                               702 759 777 827 924 924 991))))

(suite mergesort
    ((iota (lambda (n)
             ((unfold-list zero? (sect - n <>) sub1) n)))
     (rng (random-number-generator 19990926))
     (sample (let ((rng2 (random-number-generator 201110046)))
               ((fold-natural bag (sect put-into-bag (rng2 24) <>)) 785))))
  (test empty-list
    (mergesort values? (bag))
    1 (list?)
    empty-list?)
  (test singleton
    (mergesort values? (bag 'rns))
    1 ((list-of symbol?))
    (match? (list-of= symbol=?) (list 'rns)))
  (test larger-bag
    (mergesort >= ((lower-ply-natural bag put-into-bag) 476))
    1 ((list-of exact-integer?))
    (match? (list-of= =) (reverse (iota 476))))
  (test shuffled
    (mergesort <= ((fold-list bag put-into-bag) (shuffle rng (iota 181))))
    1 ((list-of exact-integer?))
    (match? (list-of= =) (iota 181)))
  (test all-equal
    (mergesort >= ((fold-natural bag (sect put-into-bag 522 <>)) 730))
    1 ((list-of exact-integer?))
    (match? (list-of= =) ((fold-natural list (sect prepend 522 <>)) 730)))
  (test random-elements
    (mergesort <= sample)
    1 ((list-of exact-integer?))
    (^et (sect ordered? <= <>)
         (pipe (fold-list bag put-into-bag)
               (match? (bag-of= =) sample)))))

(suite put-into-binary-search-tree
    ((right-type? (tree-of (pair-of exact-integer? symbol?)))
     (equiv? (tree-of= (pair-of= = symbol=?))))
  (test typical
    (put-into-binary-search-tree (pipe (~each car) <=))
    1 (procedure?)
    (lambda (putter)
      (suite put-int-binary-search-tree-internal ()
        (test empty-tree
          (putter (cons 790 'aeehibt) (make-empty-tree))
          1 (right-type?)
          (match? equiv? (singleton-tree (cons 790 'aeehibt))))
        (test larger-bst
          (putter (cons 677 'nncr)
                  (make-non-empty-tree (cons 745 'loe)
                    (make-non-empty-tree (cons 736 'rcuess)
                      (make-non-empty-tree (cons 116 'uooha)
                        (make-empty-tree)
                        (make-non-empty-tree (cons 148 'occtoe)
                          (make-empty-tree)
                          (singleton-tree (cons 318 'eptpl))))
                      (make-empty-tree))
                    (make-non-empty-tree (cons 755 'ahes)
                      (make-empty-tree)
                      (singleton-tree (cons 812 'rhur)))))
          1 (right-type?)
          (match? equiv?
                  (make-non-empty-tree (cons 745 'loe)
                    (make-non-empty-tree (cons 736 'rcuess)
                      (make-non-empty-tree (cons 116 'uooha)
                        (make-empty-tree)
                        (make-non-empty-tree (cons 148 'occtoe)
                          (make-empty-tree)
                          (make-non-empty-tree (cons 318 'eptpl)
                            (make-empty-tree)
                            (singleton-tree (cons 677 'nncr)))))
                      (make-empty-tree))
                    (make-non-empty-tree (cons 755 'ahes)
                      (make-empty-tree)
                      (singleton-tree (cons 812 'rhur))))))))))

;;; Test the procedures in the "Binary search trees" section.

(suite search-binary-search-tree
    ((search-result? (^vel not (box-of (pair-of exact-integer? symbol?))))
     (equiv? (box-of= (pair-of= = symbol=?))))
  (test typical
    (search-binary-search-tree (pipe (~each car) <=))
    1 (procedure?)
    (lambda (seeker)
      (suite search-binary-search-tree-internal ()
        (test empty-tree
          (seeker (cons 511 null) (make-empty-tree))
          1 (search-result?)
          not)
        (test success-in-singleton-tree
          (seeker (cons 180 null) (singleton-tree (cons 180 'nnoumn)))
          1 (search-result?)
          (match? equiv? (box (cons 180 'nnoumn))))
        (test failure-in-singleton-tree
          (seeker (cons 120 null) (singleton-tree (cons 254 'nneo)))
          1 (search-result?)
          not)
        (test failure-in-larger-tree
          (seeker (cons 292 null)
                  (make-non-empty-tree (cons 515 'nheop)
                    (make-non-empty-tree (cons 122 'bneh)
                      (singleton-tree (cons 48 'votbb))
                      (singleton-tree (cons 226 'ish)))
                    (make-non-empty-tree (cons 863 'nsfsioi)
                      (singleton-tree (cons 660 'ltlfnav))
                      (make-empty-tree))))
          1 (search-result?)
          not)
        (test success-at-root
          (seeker (cons 535 null)
                  (make-non-empty-tree (cons 535 'nas)
                    (make-non-empty-tree (cons 117 'ote)
                      (make-empty-tree)
                      (make-non-empty-tree (cons 154 'usyres)
                        (make-empty-tree)
                        (make-non-empty-tree (cons 232 'chong)
                          (singleton-tree (cons 361 'onmt))
                          (make-empty-tree))))
                    (make-non-empty-tree (cons 725 'eesim)
                      (singleton-tree (cons 602 'cey))
                      (singleton-tree (cons 981 'doohh)))))
          1 (search-result?)
          (match? equiv? (box (cons 535 'nas))))
        (test leftmost
          (seeker (cons 70 null)
                  (make-non-empty-tree (cons 502 'uhudgor)
                    (make-non-empty-tree (cons 412 'pdyf)
                      (singleton-tree (cons 70 'ncnpdse))
                      (singleton-tree (cons 469 'htroh)))
                    (make-non-empty-tree (cons 578 'wwe)
                      (make-empty-tree)
                      (singleton-tree (cons 710 'odwinri)))))
          1 (search-result?)
          (match? equiv? (box (cons 70 'ncnpdse))))
        (test rightmost
          (seeker (cons 950 null)
                  (make-non-empty-tree (cons 864 'yswti)
                    (make-non-empty-tree (cons 182 'osi)
                      (singleton-tree (cons 135 'wsntai))
                      (singleton-tree (cons 477 'tscnhu)))
                    (make-non-empty-tree (cons 894 'esi)
                      (make-empty-tree)
                      (singleton-tree (cons 950 'ircinur)))))
          1 (search-result?)
          (match? equiv? (box (cons 950 'ircinur))))
        (test branches-both-ways
          (seeker (cons 362 null)
                  (make-non-empty-tree (cons 854 'swyte)
                    (make-non-empty-tree (cons 58 'dnlf)
                      (make-empty-tree)
                      (make-non-empty-tree (cons 566 'ahtn)
                        (make-non-empty-tree (cons 173 'nnhftn)
                          (make-empty-tree)
                          (make-non-empty-tree (cons 362 'oealeir)
                            (singleton-tree (cons 296 'naorto))
                            (make-empty-tree)))
                        (make-empty-tree)))
                    (singleton-tree (cons 897 'aaeoe))))
          1 (search-result?)
          (match? equiv? (box (cons 362 'oealeir))))))))

(suite naive-binary-search-tree-invariant? ()
  (test empty-tree
    (naive-binary-search-tree-invariant? string<=?
                                         (make-empty-tree))
    1 (boolean?)
    true?)
  (test singleton-tree
    (naive-binary-search-tree-invariant? <= (singleton-tree 265))
    1 (boolean?)
    true?)
  (test bad-order-in-subtree
    (naive-binary-search-tree-invariant? <=
      (make-non-empty-tree 64
        (singleton-tree 45)
        (make-non-empty-tree 930
          (singleton-tree 173)
          (singleton-tree 534))))
    1 (boolean?)
    not)
  (test bad-order-involving-root
    (naive-binary-search-tree-invariant? <=
      (make-non-empty-tree 667
        (singleton-tree 371)
        (make-non-empty-tree 831
          (make-non-empty-tree 697
            (singleton-tree 522)
            (make-empty-tree))
          (singleton-tree 836))))
    1 (boolean?)
    not)
  (test larger-tree-satisfying-invariant
    (naive-binary-search-tree-invariant? <=
      (make-non-empty-tree 371
        (make-non-empty-tree 131
          (singleton-tree 87)
          (singleton-tree 264))
        (make-non-empty-tree 745
          (make-empty-tree)
          (singleton-tree 907))))
    1 (boolean?)
    true?))

(suite rightmost ()
  (test singleton
    (rightmost (singleton-tree 'emndo))
    1 (symbol?)
    (match? symbol=? 'emndo))
  (test long-branch
    (rightmost (make-non-empty-tree 'hnuri
                 (make-empty-tree)
                 (make-non-empty-tree 'rgoge
                   (make-empty-tree)
                   (make-non-empty-tree 'liire
                     (make-empty-tree)
                     (make-non-empty-tree 'osenn
                       (make-empty-tree)
                       (singleton-tree 'aert))))))
    1 (symbol?)
    (match? symbol=? 'aert))
  (test rightmost-has-child
    (rightmost (make-non-empty-tree 'snwgrei
                 (make-empty-tree)
                 (make-non-empty-tree 'iaa
                   (make-non-empty-tree 'gna
                     (singleton-tree 'itpe)
                     (singleton-tree 'sefhcn))
                   (make-empty-tree))))
    1 (symbol?)
    (match? symbol=? 'iaa)))

(suite leftmost ()
  (test singleton
    (leftmost (singleton-tree 'bhletss))
    1 (symbol?)
    (match? symbol=? 'bhletss))
  (test long-branch
    (leftmost (make-non-empty-tree 'hdf
                (make-non-empty-tree 'hlaii
                  (make-non-empty-tree 'esnnca
                    (make-non-empty-tree 'elaveee
                      (singleton-tree 'hue)
                      (make-empty-tree))
                    (make-empty-tree))
                  (make-empty-tree))
                (make-empty-tree)))
    1 (symbol?)
    (match? symbol=? 'hue))
  (test leftmost-has-child
    (leftmost (make-non-empty-tree 'nrrhanm
                (make-non-empty-tree 'nroogx
                  (make-empty-tree)
                  (make-non-empty-tree 'onisba
                    (singleton-tree 'pnwhoo)
                    (singleton-tree 'hdhaod)))
                (make-empty-tree)))
    1 (symbol?)
    (match? symbol=? 'nroogx)))

(suite binary-search-tree-invariant? ()
  (test empty-tree
    (binary-search-tree-invariant? char>=? (make-empty-tree))
    1 (boolean?)
    true?)
  (test singleton-tree
    (binary-search-tree-invariant? >= (singleton-tree 534))
    1 (boolean?)
    true?)
  (test bad-order-in-subtree
    (binary-search-tree-invariant? >=
      (make-non-empty-tree 752
        (singleton-tree 899)
        (make-non-empty-tree 177
          (singleton-tree 657)
          (singleton-tree 491))))
    1 (boolean?)
    not)
  (test bad-order-involving-root
    (binary-search-tree-invariant? >=
      (make-non-empty-tree 926
        (singleton-tree 995)
        (make-non-empty-tree 633
          (make-non-empty-tree 820
            (singleton-tree 960)
            (make-empty-tree))
          (singleton-tree 329))))
    1 (boolean?)
    not)
  (test larger-tree-satisfying-invariant
    (binary-search-tree-invariant? >=
      (make-non-empty-tree 751
        (make-non-empty-tree 908
          (singleton-tree 927)
          (singleton-tree 835))
        (make-non-empty-tree 521
          (make-empty-tree)
          (singleton-tree 342))))
    1 (boolean?)
    true?))

(suite extract-rightmost ()
  (test singleton
    (extract-rightmost (singleton-tree 'bptw))
    2 (symbol? tree?)
    (lambda (alpha beta)
      (and (symbol=? alpha 'bptw)
           (empty-tree? beta))))
  (test long-branch
    (extract-rightmost (make-non-empty-tree 'buewsen
                         (make-empty-tree)
                         (make-non-empty-tree 'ohdi
                           (make-empty-tree)
                           (make-non-empty-tree 'bxsma
                             (make-empty-tree)
                             (make-non-empty-tree 'nty
                               (make-empty-tree)
                               (singleton-tree 'htfc))))))
    2 (symbol? (tree-of symbol?))
    (lambda (alpha beta)
      (and (symbol=? alpha 'htfc)
           ((tree-of= symbol=?) beta
                                (make-non-empty-tree 'buewsen
                                  (make-empty-tree)
                                  (make-non-empty-tree 'ohdi
                                    (make-empty-tree)
                                    (make-non-empty-tree 'bxsma
                                      (make-empty-tree)
                                      (singleton-tree 'nty))))))))
  (test rightmost-has-child
    (extract-rightmost (make-non-empty-tree 'thg
                         (make-empty-tree)
                         (make-non-empty-tree 'rvce
                           (make-non-empty-tree 'gon
                             (singleton-tree 'osd)
                             (singleton-tree 'lwbrs))
                           (make-empty-tree))))
    2 (symbol? (tree-of symbol?))
    (lambda (alpha beta)
      (and (symbol=? alpha 'rvce)
           ((tree-of= symbol=?) beta
                                (make-non-empty-tree 'thg
                                  (make-empty-tree)
                                  (make-non-empty-tree 'gon
                                    (singleton-tree 'osd)
                                    (singleton-tree 'lwbrs))))))))

(suite join-trees ()
  (test both-empty
    (join-trees (make-empty-tree) (make-empty-tree))
    1 (tree?)
    empty-tree?)
  (test left-empty
    (join-trees (make-empty-tree)
                (make-non-empty-tree 207
                  (singleton-tree 61)
                  (make-non-empty-tree 217
                    (make-empty-tree)
                    (make-non-empty-tree 410
                      (make-empty-tree)
                      (make-non-empty-tree 724
                        (make-empty-tree)
                        (singleton-tree 830))))))
    1 ((tree-of exact-integer?))
    (match? (tree-of= =) (make-non-empty-tree 207
                           (singleton-tree 61)
                           (make-non-empty-tree 217
                             (make-empty-tree)
                             (make-non-empty-tree 410
                               (make-empty-tree)
                               (make-non-empty-tree 724
                                 (make-empty-tree)
                                 (singleton-tree 830)))))))
  (test right-empty
    (join-trees (make-non-empty-tree 505
                  (make-empty-tree)
                  (singleton-tree 914))
                (make-empty-tree))
    1 ((tree-of exact-integer?))
    (match? (tree-of= =) (make-non-empty-tree 505
                           (make-empty-tree)
                           (singleton-tree 914))))
  (test empty-right-of-left
    (join-trees (make-non-empty-tree 536
                  (make-non-empty-tree 239
                    (singleton-tree 96)
                    (singleton-tree 392))
                  (make-empty-tree))
                (make-non-empty-tree 970
                  (singleton-tree 836)
                  (make-empty-tree)))
    1 ((tree-of exact-integer?))
    (match? (tree-of= =) (make-non-empty-tree 536
                           (make-non-empty-tree 239
                             (singleton-tree 96)
                             (singleton-tree 392))
                           (make-non-empty-tree 970
                             (singleton-tree 836)
                             (make-empty-tree)))))
  (test non-empty-right-of-left
    (join-trees (make-non-empty-tree 13
                  (make-empty-tree)
                  (make-non-empty-tree 693
                    (make-non-empty-tree 239
                      (make-empty-tree)
                      (make-non-empty-tree 606
                        (singleton-tree 545)
                        (make-empty-tree)))
                    (make-non-empty-tree 914
                      (make-empty-tree)
                      (singleton-tree 944))))
                (make-non-empty-tree 967
                  (make-empty-tree)
                  (singleton-tree 976)))
    1 ((tree-of exact-integer?))
    (match? (tree-of= =) (make-non-empty-tree 944
                           (make-non-empty-tree 13
                             (make-empty-tree)
                             (make-non-empty-tree 693
                               (make-non-empty-tree 239
                                 (make-empty-tree)
                                 (make-non-empty-tree 606
                                   (singleton-tree 545)
                                   (make-empty-tree)))
                               (singleton-tree 914)))
                           (make-non-empty-tree 967
                             (make-empty-tree)
                             (singleton-tree 976))))))

(suite extract-from-binary-search-tree
    ((right-type? (^vel not (box-of string?))))
  (test typical
    (extract-from-binary-search-tree string>=?)
    1 (procedure?)
    (lambda (extracter)
      (suite extract-from-binary-search-tree-internal ()
        (test empty-tree
          (extracter "inycoe" (make-empty-tree))
          2 (right-type? tree?)
          (lambda (alpha beta)
            (and (not alpha) (empty-tree? beta))))
        (test singleton-tree
          (extracter "orgnnyw" (singleton-tree "orgnnyw"))
          2 (right-type? tree?)
          (lambda (alpha beta)
            (and ((box-of= string=?) alpha (box "orgnnyw"))
                 (empty-tree? beta))))
        (test larger-tree
          (extracter "iea"
                     (make-non-empty-tree "ukta"
                       (singleton-tree "vef")
                       (make-non-empty-tree "nblr"
                         (make-non-empty-tree "toeo"
                           (make-empty-tree)
                           (make-non-empty-tree "smh"
                             (make-empty-tree)
                             (singleton-tree "nltonof")))
                         (make-non-empty-tree "iao"
                           (make-non-empty-tree "iea"
                             (singleton-tree "irtnoru")
                             (singleton-tree "idmkres"))
                           (make-non-empty-tree "hem"
                             (make-empty-tree)
                             (make-non-empty-tree "gocnldm"
                               (make-empty-tree)
                               (make-non-empty-tree "denuef"
                                 (make-non-empty-tree "eon"
                                   (singleton-tree "geeantc")
                                   (make-empty-tree))
                                 (singleton-tree "aeo"))))))))
          2 (right-type? (tree-of string?))
          (lambda (alpha beta)
            (and ((box-of= string=?) alpha (box "iea"))
                 ((tree-of= string=?)
                    beta
                    (make-non-empty-tree "ukta"
                     (singleton-tree "vef")
                     (make-non-empty-tree "nblr"
                       (make-non-empty-tree "toeo"
                         (make-empty-tree)
                         (make-non-empty-tree "smh"
                           (make-empty-tree)
                           (singleton-tree "nltonof")))
                       (make-non-empty-tree "iao"
                         (make-non-empty-tree "irtnoru"
                           (make-empty-tree)
                           (singleton-tree "idmkres"))
                         (make-non-empty-tree "hem"
                           (make-empty-tree)
                           (make-non-empty-tree "gocnldm"
                             (make-empty-tree)
                             (make-non-empty-tree "denuef"
                               (make-non-empty-tree "eon"
                                 (singleton-tree "geeantc")
                                 (make-empty-tree))
                               (singleton-tree "aeo")))))))))))
        (test failure-in-larger-tree
          (extracter "moeirh"
                     (make-non-empty-tree "vsiodig"
                       (singleton-tree "waiesr")
                       (make-non-empty-tree "nfrte"
                         (make-non-empty-tree "rhgcssr"
                           (singleton-tree "tiodei")
                           (make-empty-tree))
                         (make-non-empty-tree "ele"
                           (make-non-empty-tree "hadu"
                             (make-non-empty-tree "lawoe"
                               (make-empty-tree)
                               (singleton-tree "heedtrr"))
                             (make-empty-tree))
                           (make-non-empty-tree "aha"
                             (singleton-tree "arcss")
                             (singleton-tree "aeidtuf"))))))
          2 (right-type? (tree-of string?))
          (lambda (alpha beta)
            (and (not alpha)
                 ((tree-of= string=?)
                    beta
                    (make-non-empty-tree "vsiodig"
                      (singleton-tree "waiesr")
                      (make-non-empty-tree "nfrte"
                        (make-non-empty-tree "rhgcssr"
                          (singleton-tree "tiodei")
                          (make-empty-tree))
                        (make-non-empty-tree "ele"
                          (make-non-empty-tree "hadu"
                            (make-non-empty-tree "lawoe"
                              (make-empty-tree)
                              (singleton-tree "heedtrr"))
                            (make-empty-tree))
                          (make-non-empty-tree "aha"
                            (singleton-tree "arcss")
                            (singleton-tree "aeidtuf")))))))))))))

(suite tree->list ()
  (test empty-tree
    (tree->list (make-empty-tree))
    1 (list?)
    empty-list?)
  (test singleton-tree
    (tree->list (singleton-tree 'sieede))
    1 ((list-of symbol?))
    (match? (list-of= symbol=?) (list 'sieede)))
  (test larger-tree
    (tree->list (make-non-empty-tree 'lbdrhm
                  (make-non-empty-tree 'rtrang
                    (singleton-tree 'nae)
                    (singleton-tree 'rstgn))
                  (make-non-empty-tree 'owtnieo
                    (make-empty-tree)
                    (singleton-tree 'nes))))
    1 ((list-of symbol?))
    (match? (list-of= symbol=?)
            (list 'nae 'rtrang 'rstgn 'lbdrhm 'owtnieo 'nes))))

(suite binary-search-tree-sort
    ((iota (lambda (n)
             ((unfold-list zero? (sect - n <>) sub1) n)))
     (rng (random-number-generator 20000318))
     (sample (let ((rng2 (random-number-generator 20111006)))
               ((fold-natural bag (sect put-into-bag (rng2 31) <>)) 128))))
  (test empty-list
    (binary-search-tree-sort values? (bag))
    1 (list?)
    empty-list?)
  (test singleton
    (binary-search-tree-sort values? (bag 'rns))
    1 ((list-of symbol?))
    (match? (list-of= symbol=?) (list 'rns)))
  (test larger-bag
    (binary-search-tree-sort >= ((lower-ply-natural bag put-into-bag) 613))
    1 ((list-of exact-integer?))
    (match? (list-of= =) (reverse (iota 613))))
  (test shuffled
    (binary-search-tree-sort <= ((fold-list bag put-into-bag)
                                 (shuffle rng (iota 759))))
    1 ((list-of exact-integer?))
    (match? (list-of= =) (iota 759)))
  (test all-equal
    (binary-search-tree-sort >=
                             ((fold-natural bag (sect put-into-bag 840 <>))
                              358))
    1 ((list-of exact-integer?))
    (match? (list-of= =) ((fold-natural list (sect prepend 840 <>)) 358)))
  (test random-elements
    (binary-search-tree-sort <= sample)
    1 ((list-of exact-integer?))
    (^et (sect ordered? <= <>)
         (pipe (fold-list bag put-into-bag)
               (match? (bag-of= =) sample)))))

;;; Test the procedures in the "Red-black trees" section.

(suite color? ()
  (test red
    (color? 'red)
    1 (boolean?)
    true?)
  (test black
    (color? 'black)
    1 (boolean?)
    true?)
  (test bad-symbol
    (color? 'green)
    1 (boolean?)
    not)
  (test unicorn
    (color? -643/272)
    1 (boolean?)
    not))

(suite color=? ()
  (test reds
    (color=? 'red 'red)
    1 (boolean?)
    true?)
  (test blacks
    (color=? 'black 'black)
    1 (boolean?)
    true?)
  (test mismatch
    (color=? 'black 'red)
    1 (boolean?)
    not))

(suite red-black-tree? ()
  (test empty-tree
    (red-black-tree? (make-empty-tree))
    1 (boolean?)
    true?)
  (test good-singleton
    (red-black-tree? (singleton-tree (cons 'red 361)))
    1 (boolean?)
    true?)
  (test bad-color
    (red-black-tree? (singleton-tree (cons 'purple 939)))
    1 (boolean?)
    not)
  (test non-pair
    (red-black-tree? (singleton-tree 278))
    1 (boolean?)
    not)
  (test unicorn
    (red-black-tree? "unicorn")
    1 (boolean?)
    not)
  (test good-larger-tree
    (red-black-tree? (make-non-empty-tree (cons 'black 547)
                       (make-non-empty-tree (cons 'red 468)
                         (singleton-tree (cons 'black 41))
                         (singleton-tree (cons 'black 501)))
                       (make-non-empty-tree (cons 'red 677)
                         (singleton-tree (cons 'black 626))
                         (make-non-empty-tree (cons 'black 801)
                           (make-empty-tree)
                           (singleton-tree (cons 'red 923))))))
    1 (boolean?)
    true?)
  (test embedded-defect
    (red-black-tree? (make-non-empty-tree (cons 'red 613)
                       (singleton-tree (cons 'black 443))
                       (make-non-empty-tree (cons 'black 700)
                         (singleton-tree (cons 'green 647))
                         (singleton-tree (cons 'red 991)))))
    1 (boolean?)
    not))

(suite red-black-tree=? ()
  (test both-empty
    (red-black-tree=? (make-empty-tree) (make-empty-tree))
    1 (boolean?)
    true?)
  (test left-empty
    (red-black-tree=? (make-empty-tree)
                      (singleton-tree (cons 'black 538)))
    1 (boolean?)
    not)
  (test right-empty
    (red-black-tree=? (make-non-empty-tree (cons 'black 548)
                        (singleton-tree (cons 'black 343))
                        (make-non-empty-tree (cons 'red 759)
                          (singleton-tree (cons 'black 586))
                          (singleton-tree (cons 'black 976))))
                      (make-empty-tree))
    1 (boolean?)
    not)
  (test equal-non-empty-trees
    (red-black-tree=? (make-non-empty-tree (cons 'black 329)
                        (make-non-empty-tree (cons 'black 272)
                          (singleton-tree (cons 'red 70))
                          (make-empty-tree))
                        (make-non-empty-tree (cons 'black 990)
                          (singleton-tree (cons 'red 581))
                          (make-empty-tree)))
                      (make-non-empty-tree (cons 'black 329)
                        (make-non-empty-tree (cons 'black 272)
                          (singleton-tree (cons 'red 70))
                          (make-empty-tree))
                        (make-non-empty-tree (cons 'black 990)
                          (singleton-tree (cons 'red 581))
                          (make-empty-tree))))
    1 (boolean?)
    true?)
  (test same-elements-in-same-inorder
    (red-black-tree=? (make-non-empty-tree (cons 'black 364)
                        (make-non-empty-tree (cons 'black 202)
                          (singleton-tree (cons 'black 47))
                          (singleton-tree (cons 'black 266)))
                        (make-non-empty-tree (cons 'red 819)
                          (make-non-empty-tree (cons 'black 608)
                            (singleton-tree (cons 'black 530))
                            (make-non-empty-tree (cons 'red 665)
                              (make-empty-tree)
                              (singleton-tree (cons 'black 763))))
                          (make-non-empty-tree (cons 'black 928)
                            (singleton-tree (cons 'black 867))
                            (singleton-tree (cons 'black 944)))))
                      (make-non-empty-tree (cons 'black 608)
                        (make-non-empty-tree (cons 'red 266)
                          (make-non-empty-tree (cons 'black 202)
                            (singleton-tree (cons 'red 47))
                            (make-empty-tree))
                          (make-non-empty-tree (cons 'black 530)
                            (singleton-tree (cons 'red 364))
                            (make-empty-tree)))
                        (make-non-empty-tree (cons 'red 867)
                          (make-non-empty-tree (cons 'black 763)
                            (singleton-tree (cons 'red 665))
                            (singleton-tree (cons 'red 819)))
                          (make-non-empty-tree (cons 'black 944)
                            (singleton-tree (cons 'red 928))
                            (make-empty-tree)))))
    1 (boolean?)
    not)
  (test different-elements-in-same-shape
    (red-black-tree=? (make-non-empty-tree (cons 'black 208)
                        (make-non-empty-tree (cons 'black 102)
                          (singleton-tree (cons 'red 24))
                          (make-empty-tree))
                        (make-non-empty-tree (cons 'red 687)
                          (singleton-tree (cons 'black 574))
                          (singleton-tree (cons 'black 689))))
                      (make-non-empty-tree (cons 'black 678)
                        (make-non-empty-tree (cons 'black 515)
                          (singleton-tree (cons 'red 293))
                          (make-empty-tree))
                        (make-non-empty-tree (cons 'red 836)
                          (singleton-tree (cons 'black 812))
                          (singleton-tree (cons 'black 909)))))
    1 (boolean?)
    not)
  (test color-mismatch
    (red-black-tree=? (singleton-tree (cons 'red 807))
                      (singleton-tree (cons 'black 807)))
    1 (boolean?)
    not))

(suite red? ()
  (test empty-tree
    (red? (make-empty-tree))
    1 (boolean?)
    not)
  (test red-tree
    (red? (singleton-tree (cons 'red 714)))
    1 (boolean?)
    true?)
  (test black-tree
    (red? (make-non-empty-tree (cons 'black 515)
            (make-empty-tree)
            (singleton-tree (cons 'red 788))))
    1 (boolean?)
    not))

(suite red-black-tree-element ()
  (test minimal
    (red-black-tree-element (singleton-tree (cons 'black 869)))
    1 (exact-integer?)
    (match? = 869))
  (test larger-tree
    (red-black-tree-element (make-non-empty-tree (cons 'red 560)
                              (make-non-empty-tree (cons 'black 133)
                                (make-empty-tree)
                                (singleton-tree (cons 'red 385)))
                              (singleton-tree (cons 'black 976))))
    1 (exact-integer?)
    (match? = 560)))

(suite satisfies-red-black-invariant? ()
  (test empty-tree
    (satisfies-red-black-invariant? (make-empty-tree))
    1 (boolean?)
    true?)
  (test singleton-red
    (satisfies-red-black-invariant? (singleton-tree (cons 'red 957)))
    1 (boolean?)
    true?)
  (test singleton-black
    (satisfies-red-black-invariant? (singleton-tree (cons 'black 263)))
    1 (boolean?)
    true?)
  (test red-over-red
    (satisfies-red-black-invariant? (make-non-empty-tree (cons 'red 734)
                                      (make-empty-tree)
                                      (singleton-tree (cons 'red 761))))
    1 (boolean?)
    not)
  (test mismatched-m-values
    (satisfies-red-black-invariant? (make-non-empty-tree (cons 'black 514)
                                      (singleton-tree (cons 'black 289))
                                      (make-empty-tree)))
    1 (boolean?)
    not)
  (test embedded-failure
    (satisfies-red-black-invariant?
      (make-non-empty-tree (cons 'black 299)
        (make-non-empty-tree (cons 'black 280)
          (singleton-tree (cons 'red 180))
          (make-non-empty-tree (cons 'red 295)
            (singleton-tree (cons 'red 289))
            (make-empty-tree)))
        (make-non-empty-tree (cons 'red 514)
          (singleton-tree (cons 'black 342))
          (singleton-tree (cons 'black 519)))))
    1 (boolean?)
    not))

(suite red-black-search-tree? ()
  (test empty-tree
    (red-black-search-tree? char<=? (make-empty-tree))
    1 (boolean?)
    true?)
  (test bad-structure
    (red-black-search-tree? <= (singleton-tree 363))
    1 (boolean?)
    not)
  (test good-singleton
    (red-black-search-tree? >= (singleton-tree (cons 'black 918)))
    1 (boolean?)
    true?)
  (test unicorn
    (red-black-search-tree? values? 'ierlrh)
    1 (boolean?)
    not)
  (test violates-red-black-invariant
    (red-black-search-tree? <= (make-non-empty-tree (cons 'red 541)
                                 (singleton-tree (cons 'red 166))
                                 (make-empty-tree)))
    1 (boolean?)
    not)
  (test violates-binary-search-tree-invariant
    (red-black-search-tree? >= (make-non-empty-tree (cons 'red 142)
                                 (singleton-tree (cons 'black 425))
                                 (make-non-empty-tree (cons 'black 111)
                                   (singleton-tree (cons 'red 342))
                                   (make-empty-tree))))
    1 (boolean?)
    not)
  (test good-larger-tree
    (red-black-search-tree? <= (make-non-empty-tree (cons 'black 342)
                                 (make-non-empty-tree (cons 'red 224)
                                   (make-non-empty-tree (cons 'black 209)
                                     (singleton-tree (cons 'red 134))
                                     (make-empty-tree))
                                   (make-non-empty-tree (cons 'black 260)
                                     (make-empty-tree)
                                     (singleton-tree (cons 'red 270))))
                                 (make-non-empty-tree (cons 'red 723)
                                   (make-non-empty-tree (cons 'black 523)
                                     (singleton-tree (cons 'red 467))
                                     (singleton-tree (cons 'red 699)))
                                   (make-non-empty-tree (cons 'black 916)
                                     (singleton-tree (cons 'red 835))
                                     (singleton-tree (cons 'red 924))))))
    1 (boolean?)
    true?))

(suite color-flip ()
  (test black-to-red
    (color-flip (make-non-empty-tree (cons 'black 439)
                  (singleton-tree (cons 'red 315))
                  (singleton-tree (cons 'red 543))))
    1 (red-black-tree?)
    (match? red-black-tree=? (make-non-empty-tree (cons 'red 439)
                               (singleton-tree (cons 'black 315))
                               (singleton-tree (cons 'black 543)))))
  (test red-to-black
    (color-flip (make-non-empty-tree (cons 'red 607)
                  (make-non-empty-tree (cons 'black 541)
                    (make-non-empty-tree (cons 'black 103)
                      (make-empty-tree)
                      (singleton-tree (cons 'red 430)))
                    (singleton-tree (cons 'black 541)))
                  (make-non-empty-tree (cons 'black 863)
                    (make-non-empty-tree (cons 'black 828)
                      (singleton-tree (cons 'red 645))
                      (make-empty-tree))
                    (singleton-tree (cons 'black 957)))))
    1 (red-black-tree?)
    (match? red-black-tree=? (make-non-empty-tree (cons 'black 607)
                               (make-non-empty-tree (cons 'red 541)
                                 (make-non-empty-tree (cons 'black 103)
                                   (make-empty-tree)
                                   (singleton-tree (cons 'red 430)))
                                 (singleton-tree (cons 'black 541)))
                               (make-non-empty-tree (cons 'red 863)
                                 (make-non-empty-tree (cons 'black 828)
                                   (singleton-tree (cons 'red 645))
                                   (make-empty-tree))
                                 (singleton-tree (cons 'black 957)))))))

(suite rotate-left ()
  (test minimal
    (rotate-left (make-non-empty-tree (cons 'black 611)
                   (make-empty-tree)
                   (singleton-tree (cons 'red 942))))
    1 (red-black-tree?)
    (match? red-black-tree=? (make-non-empty-tree (cons 'black 942)
                               (singleton-tree (cons 'red 611))
                               (make-empty-tree))))
  (test larger
    (rotate-left (make-non-empty-tree (cons 'black 132)
                   (make-non-empty-tree (cons 'black 67)
                     (make-empty-tree)
                     (singleton-tree (cons 'red 106)))
                   (make-non-empty-tree (cons 'red 152)
                     (singleton-tree (cons 'black 140))
                     (make-non-empty-tree (cons 'black 989)
                       (singleton-tree (cons 'red 892))
                       (make-empty-tree)))))
    1 (red-black-tree?)
    (match? red-black-tree=? (make-non-empty-tree (cons 'black 152)
                               (make-non-empty-tree (cons 'red 132)
                                 (make-non-empty-tree (cons 'black 67)
                                   (make-empty-tree)
                                   (singleton-tree (cons 'red 106)))
                                 (singleton-tree (cons 'black 140)))
                               (make-non-empty-tree (cons 'black 989)
                                 (singleton-tree (cons 'red 892))
                                 (make-empty-tree))))))

(suite rotate-right ()
  (test minimal
    (rotate-right (make-non-empty-tree (cons 'black 933)
                    (singleton-tree (cons 'red 545))
                    (make-empty-tree)))
    1 (red-black-tree?)
    (match? red-black-tree=? (make-non-empty-tree (cons 'black 545)
                               (make-empty-tree)
                               (singleton-tree (cons 'red 933)))))
  (test larger
    (rotate-right (make-non-empty-tree (cons 'black 536)
                    (make-non-empty-tree (cons 'red 262)
                      (singleton-tree (cons 'black 109))
                      (make-non-empty-tree (cons 'black 403)
                        (singleton-tree (cons 'red 307))
                        (make-empty-tree)))
                    (make-non-empty-tree (cons 'black 715)
                      (singleton-tree (cons 'red 630))
                      (singleton-tree (cons 'red 984)))))
    1 (red-black-tree?)
    (match? red-black-tree=? (make-non-empty-tree (cons 'black 262)
                               (singleton-tree (cons 'black 109))
                               (make-non-empty-tree (cons 'red 536)
                                 (make-non-empty-tree (cons 'black 403)
                                   (singleton-tree (cons 'red 307))
                                   (make-empty-tree))
                                 (make-non-empty-tree (cons 'black 715)
                                   (singleton-tree (cons 'red 630))
                                   (singleton-tree (cons 'red 984))))))))

(suite lift-red ()
  (test lift-needed
    (lift-red (make-non-empty-tree (cons 'black 563)
                (singleton-tree (cons 'red 192))
                (singleton-tree (cons 'red 574))))
    1 (red-black-tree?)
    (match? red-black-tree=? (make-non-empty-tree (cons 'red 563)
                               (singleton-tree (cons 'black 192))
                               (singleton-tree (cons 'black 574)))))
  (test one-black-subtree
    (lift-red (make-non-empty-tree (cons 'black 270)
                (make-non-empty-tree (cons 'black 69)
                  (make-empty-tree)
                  (singleton-tree (cons 'red 120)))
                (make-non-empty-tree (cons 'red 458)
                  (singleton-tree (cons 'black 277))
                  (singleton-tree (cons 'black 927)))))
    1 (red-black-tree?)
    (match? red-black-tree=? (make-non-empty-tree (cons 'black 270)
                               (make-non-empty-tree (cons 'black 69)
                                 (make-empty-tree)
                                 (singleton-tree (cons 'red 120)))
                               (make-non-empty-tree (cons 'red 458)
                                 (singleton-tree (cons 'black 277))
                                 (singleton-tree (cons 'black 927))))))
  (test minimal
    (lift-red (singleton-tree (cons 'black 708)))
    1 (red-black-tree?)
    (match? red-black-tree=? (singleton-tree (cons 'black 708))))
  (test two-black-subtrees
    (lift-red (make-non-empty-tree (cons 'black 785)
                (singleton-tree (cons 'black 289))
                (make-non-empty-tree (cons 'black 862)
                  (singleton-tree (cons 'red 804))
                  (make-empty-tree))))
    1 (red-black-tree?)
    (match? red-black-tree=? (make-non-empty-tree (cons 'black 785)
                               (singleton-tree (cons 'black 289))
                               (make-non-empty-tree (cons 'black 862)
                                 (singleton-tree (cons 'red 804))
                                 (make-empty-tree))))))

(suite left-leaning? ()
  (test empty-tree
    (left-leaning? (make-empty-tree))
    1 (boolean?)
    true?)
  (test singleton-black
    (left-leaning? (singleton-tree (cons 'black 239)))
    1 (boolean?)
    true?)
  (test singleton-red
    (left-leaning? (singleton-tree (cons 'red 602)))
    1 (boolean?)
    true?)
  (test empty-and-red
    (left-leaning? (make-non-empty-tree (cons 'black 775)
                     (make-empty-tree)
                     (singleton-tree (cons 'red 912))))
    1 (boolean?)
    not)
  (test red-and-empty
    (left-leaning? (make-non-empty-tree (cons 'black 658)
                     (singleton-tree (cons 'red 401))
                     (make-empty-tree)))
    1 (boolean?)
    true?)
  (test red-and-red
    (left-leaning? (make-non-empty-tree (cons 'black 307)
                     (singleton-tree (cons 'red 107))
                     (singleton-tree (cons 'red 454))))
    1 (boolean?)
    true?)
  (test red-and-black
    (left-leaning? (make-non-empty-tree (cons 'black 725)
                     (make-non-empty-tree (cons 'red 267)
                       (singleton-tree (cons 'black 193))
                       (singleton-tree (cons 'black 575)))
                     (singleton-tree (cons 'black 775))))
    1 (boolean?)
    true?)
  (test black-and-red
    (left-leaning? (make-non-empty-tree (cons 'black 621)
                     (singleton-tree (cons 'black 388))
                     (make-non-empty-tree (cons 'red 854)
                       (singleton-tree (cons 'black 641))
                       (singleton-tree (cons 'black 797)))))
    1 (boolean?)
    not)
  (test black-and-black
    (left-leaning? (make-non-empty-tree (cons 'black 318)
                     (singleton-tree (cons 'black 133))
                     (singleton-tree (cons 'black 588))))
    1 (boolean?)
    true?)
  (test embedded-right-lean
    (left-leaning? (make-non-empty-tree (cons 'black 434)
                     (make-non-empty-tree (cons 'red 110)
                       (singleton-tree (cons 'black 73))
                       (singleton-tree (cons 'black 178)))
                     (make-non-empty-tree (cons 'black 609)
                       (make-empty-tree)
                       (singleton-tree (cons 'red 700)))))
    1 (boolean?)
    not))

(suite lean-left ()
  (test rotation-needed
    (lean-left (make-non-empty-tree (cons 'black 289)
                 (singleton-tree (cons 'black 69))
                 (make-non-empty-tree (cons 'red 804)
                   (singleton-tree (cons 'black 785))
                   (singleton-tree (cons 'black 862)))))
    1 (red-black-tree?)
    (match? red-black-tree=? (make-non-empty-tree (cons 'black 804)
                               (make-non-empty-tree (cons 'red 289)
                                 (singleton-tree (cons 'black 69))
                                 (singleton-tree (cons 'black 785)))
                               (singleton-tree (cons 'black 862)))))
  (test left-subtree-red
    (lean-left (make-non-empty-tree (cons 'black 342)
                 (make-non-empty-tree (cons 'red 239)
                   (singleton-tree (cons 'black 120))
                   (singleton-tree (cons 'black 281)))
                 (singleton-tree (cons 'black 916))))
    1 (red-black-tree?)
    (match? red-black-tree=? (make-non-empty-tree (cons 'black 342)
                               (make-non-empty-tree (cons 'red 239)
                                 (singleton-tree (cons 'black 120))
                                 (singleton-tree (cons 'black 281)))
                               (singleton-tree (cons 'black 916)))))
  (test minimal
    (lean-left (singleton-tree (cons 'black 860)))
    1 (red-black-tree?)
    (match? red-black-tree=? (singleton-tree (cons 'black 860))))
  (test both-subtrees-black
    (lean-left (make-non-empty-tree (cons 'red 173)
                 (singleton-tree (cons 'black 136))
                 (singleton-tree (cons 'black 455))))
    1 (red-black-tree?)
    (match? red-black-tree=? (make-non-empty-tree (cons 'red 173)
                               (singleton-tree (cons 'black 136))
                               (singleton-tree (cons 'black 455))))))

(suite rebalance-reds ()
  (test minimal
    (rebalance-reds (make-non-empty-tree (cons 'black 680)
                      (make-non-empty-tree (cons 'red 301)
                        (singleton-tree (cons 'red 16))
                        (make-empty-tree))
                      (make-empty-tree)))
    1 (red-black-tree?)
    (match? red-black-tree=? (make-non-empty-tree (cons 'black 301)
                               (singleton-tree (cons 'red 16))
                               (singleton-tree (cons 'red 680)))))
  (test larger-rebalancing-case
    (rebalance-reds (make-non-empty-tree (cons 'black 670)
                      (make-non-empty-tree (cons 'red 314)
                        (make-non-empty-tree (cons 'red 148)
                          (singleton-tree (cons 'black 121))
                          (singleton-tree (cons 'black 174)))
                        (singleton-tree (cons 'black 531)))
                      (make-non-empty-tree (cons 'black 686)
                        (make-empty-tree)
                        (singleton-tree (cons 'red 924)))))
    1 (red-black-tree?)
    (match? red-black-tree=? (make-non-empty-tree (cons 'black 314)
                               (make-non-empty-tree (cons 'red 148)
                                 (singleton-tree (cons 'black 121))
                                 (singleton-tree (cons 'black 174)))
                               (make-non-empty-tree (cons 'red 670)
                                 (singleton-tree (cons 'black 531))
                                 (make-non-empty-tree (cons 'black 686)
                                   (make-empty-tree)
                                   (singleton-tree (cons 'red 924)))))))
  (test singleton-tree
    (rebalance-reds (singleton-tree (cons 'black 801)))
    1 (red-black-tree?)
    (match? red-black-tree=? (singleton-tree (cons 'black 801))))
  (test singleton-left-subtree
    (rebalance-reds (make-non-empty-tree (cons 'black 507)
                      (singleton-tree (cons 'red 349))
                      (make-empty-tree)))
    1 (red-black-tree?)
    (match? red-black-tree=? (make-non-empty-tree (cons 'black 507)
                               (singleton-tree (cons 'red 349))
                               (make-empty-tree))))
  (test red-left-subtree-with-black-left-subtree
    (rebalance-reds (make-non-empty-tree (cons 'black 744)
                      (make-non-empty-tree (cons 'red 539)
                        (singleton-tree (cons 'black 339))
                        (make-non-empty-tree (cons 'black 565)
                          (make-empty-tree)
                          (singleton-tree (cons 'red 656))))
                      (make-non-empty-tree (cons 'black 846)
                        (singleton-tree (cons 'red 953))
                        (make-empty-tree))))
    1 (red-black-tree?)
    (match? red-black-tree=? (make-non-empty-tree (cons 'black 744)
                               (make-non-empty-tree (cons 'red 539)
                                 (singleton-tree (cons 'black 339))
                                 (make-non-empty-tree (cons 'black 565)
                                   (make-empty-tree)
                                   (singleton-tree (cons 'red 656))))
                               (make-non-empty-tree (cons 'black 846)
                                 (singleton-tree (cons 'red 953))
                                 (make-empty-tree))))))

(suite force-color ()
  (test singleton
    (force-color 'red (singleton-tree (cons 'black 573)))
    1 (red-black-tree?)
    (match? red-black-tree=? (singleton-tree (cons 'red 573))))
  (test no-change
    (force-color 'black (make-non-empty-tree (cons 'black 749)
                          (singleton-tree (cons 'red 382))
                          (make-empty-tree)))
    1 (red-black-tree?)
    (match? red-black-tree=? (make-non-empty-tree (cons 'black 749)
                               (singleton-tree (cons 'red 382))
                               (make-empty-tree))))
  (test larger-tree
    (force-color 'black (make-non-empty-tree (cons 'red 347)
                          (make-non-empty-tree (cons 'black 226)
                            (singleton-tree (cons 'red 27))
                            (make-empty-tree))
                          (singleton-tree (cons 'black 874))))
    1 (red-black-tree?)
    (match? red-black-tree=? (make-non-empty-tree (cons 'black 347)
                               (make-non-empty-tree (cons 'black 226)
                                 (singleton-tree (cons 'red 27))
                                 (make-empty-tree))
                               (singleton-tree (cons 'black 874))))))

(suite put-into-red-black-tree ()
  (test ascending-numerical
    (put-into-red-black-tree <=)
    1 (procedure?)
    (lambda (putter)
      (suite ascending-numerical-internal ()
        (test empty-tree
          (putter 180 (make-empty-tree))
          1 (red-black-tree?)
          (match? red-black-tree=? (singleton-tree (cons 'black 180))))
        (test singleton-tree
          (putter 786 (singleton-tree (cons 'black 654)))
          1 (red-black-tree?)
          (match? red-black-tree=? (make-non-empty-tree (cons 'black 786)
                                     (singleton-tree (cons 'red 654))
                                     (make-empty-tree))))
        (test no-rotation
          (putter 114 (make-non-empty-tree (cons 'black 411)
                        (make-non-empty-tree (cons 'red 344)
                          (singleton-tree (cons 'black 28))
                          (singleton-tree (cons 'black 376)))
                        (make-non-empty-tree (cons 'black 591)
                          (singleton-tree (cons 'red 427))
                          (singleton-tree (cons 'red 695)))))
          1 (red-black-tree?)
          (match? red-black-tree=?
                  (make-non-empty-tree (cons 'black 411)
                    (make-non-empty-tree (cons 'red 344)
                      (make-non-empty-tree (cons 'black 114)
                        (singleton-tree (cons 'red 28))
                        (make-empty-tree))
                      (singleton-tree (cons 'black 376)))
                    (make-non-empty-tree (cons 'black 591)
                      (singleton-tree (cons 'red 427))
                      (singleton-tree (cons 'red 695))))))
        (test clockwise
          (putter 441 (make-non-empty-tree (cons 'black 848)
                        (singleton-tree (cons 'red 523))
                        (make-empty-tree)))
          1 (red-black-tree?)
          (match? red-black-tree=? (make-non-empty-tree (cons 'black 523)
                                     (singleton-tree (cons 'red 441))
                                     (singleton-tree (cons 'red 848)))))
        (test counterclockwise
          (putter 708 (make-non-empty-tree (cons 'black 748)
                        (singleton-tree (cons 'black 574))
                        (singleton-tree (cons 'black 890))))
          1 (red-black-tree?)
          (match? red-black-tree=?
                  (make-non-empty-tree (cons 'black 748)
                    (make-non-empty-tree (cons 'black 708)
                      (singleton-tree (cons 'red 574))
                      (make-empty-tree))
                    (singleton-tree (cons 'black 890)))))
        (test counterclockwise-then-clockwise
          (putter 983 (make-non-empty-tree (cons 'black 995)
                        (singleton-tree (cons 'red 574))
                        (make-empty-tree)))
          1 (red-black-tree?)
          (match? red-black-tree=? (make-non-empty-tree (cons 'black 983)
                                     (singleton-tree (cons 'red 574))
                                     (singleton-tree (cons 'red 995)))))
        (test color-flip
          (putter 536 (make-non-empty-tree (cons 'black 480)
                        (singleton-tree (cons 'red 436))
                        (singleton-tree (cons 'red 558))))
          1 (red-black-tree?)
          (match? red-black-tree=?
                  (make-non-empty-tree (cons 'black 480)
                    (singleton-tree (cons 'black 436))
                    (make-non-empty-tree (cons 'black 558)
                      (singleton-tree (cons 'red 536))
                      (make-empty-tree)))))))))

(suite search-red-black-tree
    ((search-result? (^vel not (box-of exact-integer?)))
     (equiv? (box-of= =)))
  (test typical
    (search-red-black-tree <=)
    1 (procedure?)
    (lambda (seeker)
      (suite search-red-black-tree-internal ()
        (test empty-tree
          (seeker 300 (make-empty-tree))
          1 (search-result?)
          not)
        (test success-in-singleton-tree
          (seeker 558 (singleton-tree (cons 'black 558)))
          1 (search-result?)
          (match? equiv? (box 558)))
        (test failure-in-singleton-tree
          (seeker 533 (singleton-tree (cons 'black 669)))
          1 (search-result?)
          not)
        (test failure-in-larger-tree
          (seeker 755 (make-non-empty-tree (cons 'black 784)
                        (make-non-empty-tree (cons 'black 682)
                          (singleton-tree (cons 'red 80))
                          (singleton-tree (cons 'red 708)))
                        (make-non-empty-tree (cons 'black 978)
                          (singleton-tree (cons 'red 809))
                          (make-empty-tree))))
          1 (search-result?)
          not)
        (test success-at-root
          (seeker 852 (make-non-empty-tree (cons 'black 852)
                        (make-non-empty-tree (cons 'black 769)
                          (singleton-tree (cons 'red 649))
                          (make-empty-tree))
                        (singleton-tree (cons 'black 994))))
          1 (search-result?)
          (match? equiv? (box 852)))
        (test leftmost
          (seeker 114 (make-non-empty-tree (cons 'black 876)
                        (make-non-empty-tree (cons 'black 128)
                          (singleton-tree (cons 'red 114))
                          (singleton-tree (cons 'red 813)))
                        (make-non-empty-tree (cons 'black 993)
                          (singleton-tree (cons 'red 944))
                          (make-empty-tree))))
          1 (search-result?)
          (match? equiv? (box 114)))
        (test rightmost
          (seeker 902 (make-non-empty-tree (cons 'black 821)
                        (make-non-empty-tree (cons 'black 390)
                          (singleton-tree (cons 'red 192))
                          (singleton-tree (cons 'red 577)))
                        (make-non-empty-tree (cons 'black 902)
                          (singleton-tree (cons 'red 837))
                          (make-empty-tree))))
          1 (search-result?)
          (match? equiv? (box 902)))
        (test branches-both-ways
          (seeker 382 (make-non-empty-tree (cons 'black 573)
                        (make-non-empty-tree (cons 'red 347)
                          (make-non-empty-tree (cons 'black 226)
                            (singleton-tree (cons 'red 27))
                            (make-empty-tree))
                          (singleton-tree (cons 'black 382)))
                        (make-non-empty-tree (cons 'black 749)
                          (make-empty-tree)
                          (singleton-tree (cons 'red 874)))))
          1 (search-result?)
          (match? equiv? (box 382)))))))

(suite safe-red-black-tree? ()
  (test empty-tree
    (safe-red-black-tree? (make-empty-tree))
    1 (boolean?)
    true?)
  (test singleton-black
    (safe-red-black-tree? (singleton-tree (cons 'black 989)))
    1 (boolean?)
    not)
  (test singleton-red
    (safe-red-black-tree? (singleton-tree (cons 'red 835)))
    1 (boolean?)
    true?)
  (test black-over-red
    (safe-red-black-tree? (make-non-empty-tree (cons 'black 938)
                            (singleton-tree (cons 'red 821))
                            (make-empty-tree)))
    1 (boolean?)
    true?)
  (test black-over-blacks
    (safe-red-black-tree? (make-non-empty-tree (cons 'black 642)
                            (singleton-tree (cons 'black 401))
                            (singleton-tree (cons 'black 663))))
    1 (boolean?)
    not)
  (test black-over-reds
    (safe-red-black-tree? (make-non-empty-tree (cons 'black 602)
                            (singleton-tree (cons 'red 124))
                            (singleton-tree (cons 'red 738))))
    1 (boolean?)
    true?)
  (test red-over-blacks
    (safe-red-black-tree? (make-non-empty-tree (cons 'red 599)
                            (singleton-tree (cons 'black 321))
                            (singleton-tree (cons 'black 844))))
    1 (boolean?)
    true?))

(suite avoid-unsafe-right ()
  (test singleton
    (avoid-unsafe-right (singleton-tree (cons 'red 944)))
    1 (red-black-tree?)
    (match? red-black-tree=? (singleton-tree (cons 'red 944))))
  (test red-left-red-right
    (avoid-unsafe-right (make-non-empty-tree (cons 'black 380)
                          (singleton-tree (cons 'red 266))
                          (singleton-tree (cons 'red 695))))
    1 (red-black-tree?)
    (match? red-black-tree=? (make-non-empty-tree (cons 'black 380)
                               (singleton-tree (cons 'red 266))
                               (singleton-tree (cons 'red 695)))))
  (test red-left-black-right
    (avoid-unsafe-right (make-non-empty-tree (cons 'black 556)
                          (make-non-empty-tree (cons 'red 257)
                            (singleton-tree (cons 'black 122))
                            (singleton-tree (cons 'black 271)))
                          (singleton-tree (cons 'black 690))))
    1 (red-black-tree?)
    (match? red-black-tree=? (make-non-empty-tree (cons 'black 257)
                               (singleton-tree (cons 'black 122))
                               (make-non-empty-tree (cons 'red 556)
                                 (singleton-tree (cons 'black 271))
                                 (singleton-tree (cons 'black 690))))))
  (test no-change
    (avoid-unsafe-right (make-non-empty-tree (cons 'red 759)
                          (singleton-tree (cons 'black 470))
                          (make-non-empty-tree (cons 'black 961)
                            (singleton-tree (cons 'red 794))
                            (make-empty-tree))))
    1 (red-black-tree?)
    (match? red-black-tree=? (make-non-empty-tree (cons 'red 759)
                               (singleton-tree (cons 'black 470))
                               (make-non-empty-tree (cons 'black 961)
                                 (singleton-tree (cons 'red 794))
                                 (make-empty-tree)))))
  (test color-flip-only
    (avoid-unsafe-right (make-non-empty-tree (cons 'red 284)
                          (make-non-empty-tree (cons 'black 55)
                            (singleton-tree (cons 'black 40))
                            (singleton-tree (cons 'black 154)))
                          (make-non-empty-tree (cons 'black 338)
                            (singleton-tree (cons 'black 302))
                            (singleton-tree (cons 'black 794)))))
    1 (red-black-tree?)
    (match? red-black-tree=? (make-non-empty-tree (cons 'black 284)
                               (make-non-empty-tree (cons 'red 55)
                                 (singleton-tree (cons 'black 40))
                                 (singleton-tree (cons 'black 154)))
                               (make-non-empty-tree (cons 'red 338)
                                 (singleton-tree (cons 'black 302))
                                 (singleton-tree (cons 'black 794))))))
  (test rotation-needed
    (avoid-unsafe-right (make-non-empty-tree (cons 'red 723)
                          (make-non-empty-tree (cons 'black 494)
                            (make-non-empty-tree (cons 'red 341)
                              (singleton-tree (cons 'black 184))
                              (singleton-tree (cons 'black 472)))
                            (singleton-tree (cons 'black 618)))
                          (make-non-empty-tree (cons 'black 975)
                            (singleton-tree (cons 'black 861))
                            (singleton-tree (cons 'black 999)))))
    1 (red-black-tree?)
    (match? red-black-tree=? (make-non-empty-tree (cons 'red 494)
                               (make-non-empty-tree (cons 'black 341)
                                 (singleton-tree (cons 'black 184))
                                 (singleton-tree (cons 'black 472)))
                               (make-non-empty-tree (cons 'black 975)
                                 (make-non-empty-tree (cons 'red 723)
                                   (singleton-tree (cons 'black 618))
                                   (singleton-tree (cons 'black 861)))
                                 (singleton-tree (cons 'black 999)))))))

(suite avoid-unsafe-left ()
  (test empty-left
    (avoid-unsafe-left (singleton-tree (cons 'black 949)))
    1 (red-black-tree?)
    (match? red-black-tree=? (singleton-tree (cons 'black 949))))
  (test safe-left
    (avoid-unsafe-left (make-non-empty-tree (cons 'red 372)
                         (make-non-empty-tree (cons 'black 224)
                           (singleton-tree (cons 'red 59))
                           (make-empty-tree))
                         (singleton-tree (cons 'black 944))))
    1 (red-black-tree?)
    (match? red-black-tree=? (make-non-empty-tree (cons 'red 372)
                               (make-non-empty-tree (cons 'black 224)
                                 (singleton-tree (cons 'red 59))
                                 (make-empty-tree))
                               (singleton-tree (cons 'black 944)))))
  (test color-flip-only
    (avoid-unsafe-left (make-non-empty-tree (cons 'red 531)
                         (make-non-empty-tree (cons 'black 261)
                           (singleton-tree (cons 'black 218))
                           (singleton-tree (cons 'black 457)))
                         (make-non-empty-tree (cons 'black 845)
                           (singleton-tree (cons 'black 653))
                           (singleton-tree (cons 'black 911)))))
    1 (red-black-tree?)
    (match? red-black-tree=? (make-non-empty-tree (cons 'black 531)
                               (make-non-empty-tree (cons 'red 261)
                                 (singleton-tree (cons 'black 218))
                                 (singleton-tree (cons 'black 457)))
                               (make-non-empty-tree (cons 'red 845)
                                 (singleton-tree (cons 'black 653))
                                 (singleton-tree (cons 'black 911))))))
  (test red-shift
    (avoid-unsafe-left (make-non-empty-tree (cons 'red 217)
                         (make-non-empty-tree (cons 'black 172)
                           (singleton-tree (cons 'black 153))
                           (singleton-tree (cons 'black 208)))
                         (make-non-empty-tree (cons 'black 788)
                           (make-non-empty-tree (cons 'red 495)
                             (singleton-tree (cons 'black 308))
                             (singleton-tree (cons 'black 570)))
                           (singleton-tree (cons 'black 997)))))
    1 (red-black-tree?)
    (match? red-black-tree=? (make-non-empty-tree (cons 'red 495)
                               (make-non-empty-tree (cons 'black 217)
                                 (make-non-empty-tree (cons 'red 172)
                                   (singleton-tree (cons 'black 153))
                                   (singleton-tree (cons 'black 208)))
                                 (singleton-tree (cons 'black 308)))
                               (make-non-empty-tree (cons 'black 788)
                                 (singleton-tree (cons 'black 570))
                                 (singleton-tree (cons 'black 997)))))))

(suite extract-leftmost-from-red-black-tree ()
  (test singleton
    (extract-leftmost-from-red-black-tree
      (singleton-tree (cons 'black 832)))
    2 (exact-integer? red-black-tree?)
    (lambda (alpha beta)
      (and (= alpha 832)
           (empty-tree? beta))))
  (test black-over-red
    (extract-leftmost-from-red-black-tree
      (make-non-empty-tree (cons 'black 689)
        (singleton-tree (cons 'red 30))
        (make-empty-tree)))
    2 (exact-integer? red-black-tree?)
    (lambda (alpha beta)
      (and (= alpha 30)
           (red-black-tree=? beta (singleton-tree (cons 'black 689))))))
  (test red-over-black-and-black
    (extract-leftmost-from-red-black-tree
      (make-non-empty-tree (cons 'red 460)
        (singleton-tree (cons 'black 54))
        (singleton-tree (cons 'black 735))))
    2 (exact-integer? red-black-tree?)
    (lambda (alpha beta)
      (and (= alpha 54)
           (red-black-tree=? beta
                             (make-non-empty-tree (cons 'black 735)
                               (singleton-tree (cons 'red 460))
                               (make-empty-tree))))))
  (test black-over-red-and-red
    (extract-leftmost-from-red-black-tree
      (make-non-empty-tree (cons 'black 386)
        (singleton-tree (cons 'red 184))
        (singleton-tree (cons 'red 483))))
    2 (exact-integer? red-black-tree?)
    (lambda (alpha beta)
      (and (= alpha 184)
           (red-black-tree=? beta
                             (make-non-empty-tree (cons 'black 483)
                               (singleton-tree (cons 'red 386))
                               (make-empty-tree))))))
  (test harder-case
    (extract-leftmost-from-red-black-tree
      (make-non-empty-tree (cons 'black 783)
        (make-non-empty-tree (cons 'red 327)
          (singleton-tree (cons 'black 185))
          (make-non-empty-tree (cons 'black 781)
            (singleton-tree (cons 'red 380))
            (make-empty-tree)))
        (singleton-tree (cons 'black 923))))
    2 (exact-integer? red-black-tree?)
    (lambda (alpha beta)
      (and (= alpha 185)
           (red-black-tree=? beta
                             (make-non-empty-tree (cons 'black 783)
                               (make-non-empty-tree (cons 'red 380)
                                 (singleton-tree (cons 'black 327))
                                 (singleton-tree (cons 'black 781)))
                               (singleton-tree (cons 'black 923))))))))

(suite join-red-black-trees ()
  (test no-elements
    (join-red-black-trees 'red (make-empty-tree) (make-empty-tree))
    1 (red-black-tree?)
    empty-tree?)
  (test one-element
    (join-red-black-trees 'red
                          (singleton-tree (cons 'red 760))
                          (make-empty-tree))
    1 (red-black-tree?)
    (match? red-black-tree=? (singleton-tree (cons 'black 760))))
  (test minimal-red
    (join-red-black-trees 'red
                           (singleton-tree (cons 'black 26))
                           (make-non-empty-tree (cons 'black 792)
                             (singleton-tree (cons 'red 790))
                             (make-empty-tree)))
    1 (red-black-tree?)
    (match? red-black-tree=? (make-non-empty-tree (cons 'red 790)
                               (singleton-tree (cons 'black 26))
                               (singleton-tree (cons 'black 792)))))
  (test larger
    (join-red-black-trees 'black
                          (make-non-empty-tree (cons 'black 515)
                            (make-non-empty-tree (cons 'red 251)
                              (singleton-tree (cons 'black 178))
                              (singleton-tree (cons 'black 277)))
                            (singleton-tree (cons 'black 674)))
                          (make-non-empty-tree (cons 'red 749)
                            (make-non-empty-tree (cons 'black 721)
                              (singleton-tree (cons 'black 676))
                              (singleton-tree (cons 'black 727)))
                            (make-non-empty-tree (cons 'black 956)
                              (singleton-tree (cons 'black 812))
                              (singleton-tree (cons 'black 997)))))
    1 (red-black-tree?)
    (match? red-black-tree=?
            (make-non-empty-tree (cons 'black 676)
              (make-non-empty-tree (cons 'black 515)
                (make-non-empty-tree (cons 'red 251)
                  (singleton-tree (cons 'black 178))
                  (singleton-tree (cons 'black 277)))
                (singleton-tree (cons 'black 674)))
              (make-non-empty-tree (cons 'black 956)
                (make-non-empty-tree (cons 'red 749)
                  (make-non-empty-tree (cons 'black 727)
                    (singleton-tree (cons 'red 721))
                    (make-empty-tree))
                  (singleton-tree (cons 'black 812)))
                (singleton-tree (cons 'black 997)))))))

(suite extract-from-red-black-tree ()
  (test typical
    (extract-from-red-black-tree <=)
    1 (procedure?)
    (lambda (extracter)
      (suite extract-from-red-black-tree-internal
          ((ok? (^vel not (box-of exact-integer?)))
           (numbox? (box-of exact-integer?))
           (equiv? (box-of= =)))
        (test empty-tree
          (extracter 9 (make-empty-tree))
          2 (ok? red-black-tree?)
          (lambda (alpha beta)
            (and (not alpha)
                 (red-black-tree=? beta (make-empty-tree)))))
        (test singleton-success
          (extracter 139 (singleton-tree (cons 'black 139)))
          2 (ok? red-black-tree?)
          (lambda (alpha beta)
            (and (numbox? alpha)
                 (equiv? alpha (box 139))
                 (red-black-tree=? beta (make-empty-tree)))))
        (test singleton-failure
          (extracter 718 (singleton-tree (cons 'black 734)))
          2 (ok? red-black-tree?)
          (lambda (alpha beta)
            (and (not alpha)
                 (red-black-tree=? beta
                                   (singleton-tree (cons 'black 734))))))
        (test black-over-red-lesser
          (extracter 642 (make-non-empty-tree (cons 'black 811)
                           (singleton-tree (cons 'red 642))
                           (make-empty-tree)))
          2 (ok? red-black-tree?)
          (lambda (alpha beta)
            (and (numbox? alpha)
                 (equiv? alpha (box 642))
                 (red-black-tree=? beta
                                   (singleton-tree (cons 'black 811))))))
        (test black-over-red-greater
          (extracter 355 (make-non-empty-tree (cons 'black 355)
                           (singleton-tree (cons 'red 98))
                           (make-empty-tree)))
          2 (ok? red-black-tree?)
          (lambda (alpha beta)
            (and (numbox? alpha)
                 (equiv? alpha (box 355))
                 (red-black-tree=? beta
                                   (singleton-tree (cons 'black 98))))))
        (test black-over-blacks-least
          (extracter 72 (make-non-empty-tree (cons 'black 136)
                          (singleton-tree (cons 'black 72))
                          (singleton-tree (cons 'black 217))))
          2 (ok? red-black-tree?)
          (lambda (alpha beta)
            (and (numbox? alpha)
                 (equiv? alpha (box 72))
                 (red-black-tree=? beta
                                   (make-non-empty-tree (cons 'black 217)
                                     (singleton-tree (cons 'red 136))
                                     (make-empty-tree))))))
        (test black-over-blacks-mid
          (extracter 822 (make-non-empty-tree (cons 'black 822)
                          (singleton-tree (cons 'black 453))
                          (singleton-tree (cons 'black 990))))
          2 (ok? red-black-tree?)
          (lambda (alpha beta)
            (and (numbox? alpha)
                 (equiv? alpha (box 822))
                 (red-black-tree=? beta
                                   (make-non-empty-tree (cons 'black 990)
                                     (singleton-tree (cons 'red 453))
                                     (make-empty-tree))))))
        (test black-over-blacks-greatest
          (extracter 686 (make-non-empty-tree (cons 'black 638)
                          (singleton-tree (cons 'black 573))
                          (singleton-tree (cons 'black 686))))
          2 (ok? red-black-tree?)
          (lambda (alpha beta)
            (and (numbox? alpha)
                 (equiv? alpha (box 686))
                 (red-black-tree=? beta
                                   (make-non-empty-tree (cons 'black 638)
                                     (singleton-tree (cons 'red 573))
                                     (make-empty-tree))))))
        (test black-over-reds-least
          (extracter 126 (make-non-empty-tree (cons 'black 701)
                           (singleton-tree (cons 'red 126))
                           (singleton-tree (cons 'red 855))))
          2 (ok? red-black-tree?)
          (lambda (alpha beta)
            (and (numbox? alpha)
                 (equiv? alpha (box 126))
                 (red-black-tree=? beta
                                   (make-non-empty-tree (cons 'black 855)
                                     (singleton-tree (cons 'red 701))
                                     (make-empty-tree))))))
        (test black-over-reds-mid
          (extracter 574 (make-non-empty-tree (cons 'black 574)
                           (singleton-tree (cons 'red 63))
                           (singleton-tree (cons 'red 933))))
          2 (ok? red-black-tree?)
          (lambda (alpha beta)
            (and (numbox? alpha)
                 (equiv? alpha (box 574))
                 (red-black-tree=? beta
                                   (make-non-empty-tree (cons 'black 933)
                                     (singleton-tree (cons 'red 63))
                                     (make-empty-tree))))))
        (test black-over-reds-greatest
          (extracter 726 (make-non-empty-tree (cons 'black 413)
                           (singleton-tree (cons 'red 295))
                           (singleton-tree (cons 'red 726))))
          2 (ok? red-black-tree?)
          (lambda (alpha beta)
            (and (numbox? alpha)
                 (equiv? alpha (box 726))
                 (red-black-tree=? beta
                                   (make-non-empty-tree (cons 'black 413)
                                     (singleton-tree (cons 'red 295))
                                     (make-empty-tree))))))
        (test BBdxk-0
          (extracter 128 (make-non-empty-tree (cons 'black 340)
                           (make-non-empty-tree (cons 'black 237)
                             (singleton-tree (cons 'red 128))
                             (make-empty-tree))
                           (singleton-tree (cons 'black 985))))
          2 (ok? red-black-tree?)
          (lambda (alpha beta)
            (and (numbox? alpha)
                 (equiv? alpha (box 128))
                 (red-black-tree=? beta
                                   (make-non-empty-tree (cons 'black 340)
                                     (singleton-tree (cons 'black 237))
                                     (singleton-tree
                                       (cons 'black 985)))))))
        (test BBdxk-1
          (extracter 462 (make-non-empty-tree (cons 'black 583)
                           (make-non-empty-tree (cons 'black 462)
                             (singleton-tree (cons 'red 344))
                             (make-empty-tree))
                           (singleton-tree (cons 'black 721))))
          2 (ok? red-black-tree?)
          (lambda (alpha beta)
            (and (numbox? alpha)
                 (equiv? alpha (box 462))
                 (red-black-tree=? beta
                                   (make-non-empty-tree (cons 'black 583)
                                     (singleton-tree (cons 'black 344))
                                     (singleton-tree
                                       (cons 'black 721)))))))
        (test BBdxk-2
          (extracter 494 (make-non-empty-tree (cons 'black 494)
                           (make-non-empty-tree (cons 'black 124)
                             (singleton-tree (cons 'red 2))
                             (make-empty-tree))
                           (singleton-tree (cons 'black 875))))
          2 (ok? red-black-tree?)
          (lambda (alpha beta)
            (and (numbox? alpha)
                 (equiv? alpha (box 494))
                 (red-black-tree=? beta
                                   (make-non-empty-tree (cons 'black 124)
                                     (singleton-tree (cons 'black 2))
                                     (singleton-tree
                                       (cons 'black 875)))))))
        (test BBdxk-3
          (extracter 776 (make-non-empty-tree (cons 'black 628)
                           (make-non-empty-tree (cons 'black 576)
                             (singleton-tree (cons 'red 96))
                             (make-empty-tree))
                           (singleton-tree (cons 'black 776))))
          2 (ok? red-black-tree?)
          (lambda (alpha beta)
            (and (numbox? alpha)
                 (equiv? alpha (box 776))
                 (red-black-tree=? beta
                                   (make-non-empty-tree (cons 'black 576)
                                     (singleton-tree (cons 'black 96))
                                     (singleton-tree
                                       (cons 'black 628)))))))
        (test BkBdx-0
          (extracter 44 (make-non-empty-tree (cons 'black 574)
                          (singleton-tree (cons 'black 44))
                          (make-non-empty-tree (cons 'black 682)
                            (singleton-tree (cons 'red 612))
                            (make-empty-tree))))
          2 (ok? red-black-tree?)
          (lambda (alpha beta)
            (and (numbox? alpha)
                 (equiv? alpha (box 44))
                 (red-black-tree=? beta
                                   (make-non-empty-tree (cons 'black 612)
                                     (singleton-tree (cons 'black 574))
                                     (singleton-tree
                                       (cons 'black 682)))))))
        (test BkBdx-1
          (extracter 345 (make-non-empty-tree (cons 'black 345)
                           (singleton-tree (cons 'black 57))
                           (make-non-empty-tree (cons 'black 909)
                             (singleton-tree (cons 'red 498))
                             (make-empty-tree))))
          2 (ok? red-black-tree?)
          (lambda (alpha beta)
            (and (numbox? alpha)
                 (equiv? alpha (box 345))
                 (red-black-tree=? beta
                                   (make-non-empty-tree (cons 'black 498)
                                     (singleton-tree (cons 'black 57))
                                     (singleton-tree
                                       (cons 'black 909)))))))
        (test BkBdx-2
          (extracter 391 (make-non-empty-tree (cons 'black 147)
                           (singleton-tree (cons 'black 103))
                           (make-non-empty-tree (cons 'black 697)
                             (singleton-tree (cons 'red 391))
                             (make-empty-tree))))
          2 (ok? red-black-tree?)
          (lambda (alpha beta)
            (and (numbox? alpha)
                 (equiv? alpha (box 391))
                 (red-black-tree=? beta
                                   (make-non-empty-tree (cons 'black 147)
                                     (singleton-tree (cons 'black 103))
                                     (singleton-tree
                                       (cons 'black 697)))))))
        (test BkBdx-3
          (extracter 854 (make-non-empty-tree (cons 'black 663)
                           (singleton-tree (cons 'black 18))
                           (make-non-empty-tree (cons 'black 854)
                             (singleton-tree (cons 'red 813))
                             (make-empty-tree))))
          2 (ok? red-black-tree?)
          (lambda (alpha beta)
            (and (numbox? alpha)
                 (equiv? alpha (box 854))
                 (red-black-tree=? beta
                                   (make-non-empty-tree (cons 'black 663)
                                     (singleton-tree (cons 'black 18))
                                     (singleton-tree
                                       (cons 'black 813)))))))
        (test BRkkk-0
          (extracter 383 (make-non-empty-tree (cons 'black 808)
                           (make-non-empty-tree (cons 'red 611)
                             (singleton-tree (cons 'black 383))
                             (singleton-tree (cons 'black 642)))
                           (singleton-tree (cons 'black 965))))
          2 (ok? red-black-tree?)
          (lambda (alpha beta)
            (and (numbox? alpha)
                 (equiv? alpha (box 383))
                 (red-black-tree=? beta
                                   (make-non-empty-tree (cons 'black 808)
                                     (make-non-empty-tree (cons 'black 642)
                                       (singleton-tree (cons 'red 611))
                                       (make-empty-tree))
                                     (singleton-tree
                                      (cons 'black 965)))))))
        (test BRkkk-1
          (extracter 276 (make-non-empty-tree (cons 'black 658)
                           (make-non-empty-tree (cons 'red 276)
                             (singleton-tree (cons 'black 60))
                             (singleton-tree (cons 'black 557)))
                           (singleton-tree (cons 'black 927))))
          2 (ok? red-black-tree?)
          (lambda (alpha beta)
            (and (numbox? alpha)
                 (equiv? alpha (box 276))
                 (red-black-tree=? beta
                                   (make-non-empty-tree (cons 'black 658)
                                     (make-non-empty-tree (cons 'black 557)
                                       (singleton-tree (cons 'red 60))
                                       (make-empty-tree))
                                     (singleton-tree
                                      (cons 'black 927)))))))
        (test BRkkk-2
          (extracter 703 (make-non-empty-tree (cons 'black 709)
                           (make-non-empty-tree (cons 'red 235)
                             (singleton-tree (cons 'black 223))
                             (singleton-tree (cons 'black 703)))
                           (singleton-tree (cons 'black 980))))
          2 (ok? red-black-tree?)
          (lambda (alpha beta)
            (and (numbox? alpha)
                 (equiv? alpha (box 703))
                 (red-black-tree=? beta
                                   (make-non-empty-tree (cons 'black 709)
                                     (make-non-empty-tree (cons 'black 235)
                                       (singleton-tree (cons 'red 223))
                                       (make-empty-tree))
                                     (singleton-tree
                                      (cons 'black 980)))))))
        (test BRkkk-3
          (extracter 578 (make-non-empty-tree (cons 'black 578)
                           (make-non-empty-tree (cons 'red 195)
                             (singleton-tree (cons 'black 59))
                             (singleton-tree (cons 'black 258)))
                           (singleton-tree (cons 'black 709))))
          2 (ok? red-black-tree?)
          (lambda (alpha beta)
            (and (numbox? alpha)
                 (equiv? alpha (box 578))
                 (red-black-tree=? beta
                                   (make-non-empty-tree (cons 'black 195)
                                     (singleton-tree (cons 'black 59))
                                     (make-non-empty-tree (cons 'black 709)
                                       (singleton-tree (cons 'red 258))
                                       (make-empty-tree)))))))
        (test BRkkk-4
          (extracter 902 (make-non-empty-tree (cons 'black 841)
                           (make-non-empty-tree (cons 'red 501)
                             (singleton-tree (cons 'black 434))
                             (singleton-tree (cons 'black 739)))
                           (singleton-tree (cons 'black 902))))
          2 (ok? red-black-tree?)
          (lambda (alpha beta)
            (and (numbox? alpha)
                 (equiv? alpha (box 902))
                 (red-black-tree=? beta
                                   (make-non-empty-tree (cons 'black 501)
                                     (singleton-tree (cons 'black 434))
                                     (make-non-empty-tree (cons 'black 841)
                                       (singleton-tree (cons 'red 739))
                                       (make-empty-tree)))))))
        (test BBdxBdx-1
          (extracter 490 (make-non-empty-tree (cons 'black 680)
                           (make-non-empty-tree (cons 'black 490)
                             (singleton-tree (cons 'red 211))
                             (make-empty-tree))
                           (make-non-empty-tree (cons 'black 941)
                             (singleton-tree (cons 'red 840))
                             (make-empty-tree))))
          2 (ok? red-black-tree?)
          (lambda (alpha beta)
            (and (numbox? alpha)
                 (equiv? alpha (box 490))
                 (red-black-tree=? beta
                                   (make-non-empty-tree (cons 'black 680)
                                     (singleton-tree (cons 'black 211))
                                     (make-non-empty-tree (cons 'black 941)
                                       (singleton-tree (cons 'red 840))
                                       (make-empty-tree)))))))
        (test BBdxBdx-2
          (extracter 632 (make-non-empty-tree (cons 'black 632)
                           (make-non-empty-tree (cons 'black 150)
                             (singleton-tree (cons 'red 9))
                             (make-empty-tree))
                           (make-non-empty-tree (cons 'black 776)
                             (singleton-tree (cons 'red 671))
                             (make-empty-tree))))
          2 (ok? red-black-tree?)
          (lambda (alpha beta)
            (and (numbox? alpha)
                 (equiv? alpha (box 632))
                 (red-black-tree=? beta
                                   (make-non-empty-tree (cons 'black 671)
                                     (make-non-empty-tree (cons 'black 150)
                                       (singleton-tree (cons 'red 9))
                                       (make-empty-tree))
                                     (singleton-tree
                                       (cons 'black 776)))))))
        (test BBdxBdx-4
          (extracter 934 (make-non-empty-tree (cons 'black 791)
                           (make-non-empty-tree (cons 'black 618)
                             (singleton-tree (cons 'red 101))
                             (make-empty-tree))
                           (make-non-empty-tree (cons 'black 934)
                             (singleton-tree (cons 'red 807))
                             (make-empty-tree))))
          2 (ok? red-black-tree?)
          (lambda (alpha beta)
            (and (numbox? alpha)
                 (equiv? alpha (box 934))
                 (red-black-tree=? beta
                                   (make-non-empty-tree (cons 'black 791)
                                     (make-non-empty-tree (cons 'black 618)
                                       (singleton-tree (cons 'red 101))
                                       (make-empty-tree))
                                     (singleton-tree
                                       (cons 'black 807)))))))
        (test BBddk-0
          (extracter 66 (make-non-empty-tree (cons 'black 730)
                          (make-non-empty-tree (cons 'black 159)
                            (singleton-tree (cons 'red 66))
                            (singleton-tree (cons 'red 247)))
                          (singleton-tree (cons 'black 848))))
          2 (ok? red-black-tree?)
          (lambda (alpha beta)
            (and (numbox? alpha)
                 (equiv? alpha (box 66))
                 (red-black-tree=? beta
                                   (make-non-empty-tree (cons 'black 730)
                                     (make-non-empty-tree (cons 'black 247)
                                       (singleton-tree (cons 'red 159))
                                       (make-empty-tree))
                                     (singleton-tree
                                       (cons 'black 848)))))))
        (test BBddk-1
          (extracter 221 (make-non-empty-tree (cons 'black 636)
                           (make-non-empty-tree (cons 'black 221)
                             (singleton-tree (cons 'red 104))
                             (singleton-tree (cons 'red 385)))
                           (singleton-tree (cons 'black 673))))
          2 (ok? red-black-tree?)
          (lambda (alpha beta)
            (and (numbox? alpha)
                 (equiv? alpha (box 221))
                 (red-black-tree=? beta
                                   (make-non-empty-tree (cons 'black 636)
                                     (make-non-empty-tree (cons 'black 385)
                                       (singleton-tree (cons 'red 104))
                                       (make-empty-tree))
                                     (singleton-tree
                                       (cons 'black 673)))))))
        (test BBddk-2
          (extracter 251 (make-non-empty-tree (cons 'black 558)
                           (make-non-empty-tree (cons 'black 234)
                             (singleton-tree (cons 'red 159))
                             (singleton-tree (cons 'red 251)))
                           (singleton-tree (cons 'black 655))))
          2 (ok? red-black-tree?)
          (lambda (alpha beta)
            (and (numbox? alpha)
                 (equiv? alpha (box 251))
                 (red-black-tree=? beta
                                   (make-non-empty-tree (cons 'black 558)
                                     (make-non-empty-tree (cons 'black 234)
                                       (singleton-tree (cons 'red 159))
                                       (make-empty-tree))
                                     (singleton-tree
                                       (cons 'black 655)))))))
        (test BBddk-3
          (extracter 915 (make-non-empty-tree (cons 'black 915)
                           (make-non-empty-tree (cons 'black 628)
                             (singleton-tree (cons 'red 80))
                             (singleton-tree (cons 'red 829)))
                           (singleton-tree (cons 'black 976))))
          2 (ok? red-black-tree?)
          (lambda (alpha beta)
            (and (numbox? alpha)
                 (equiv? alpha (box 915))
                 (red-black-tree=? beta
                                   (make-non-empty-tree (cons 'black 628)
                                     (singleton-tree (cons 'black 80))
                                     (make-non-empty-tree (cons 'black 976)
                                       (singleton-tree (cons 'red 829))
                                       (make-empty-tree)))))))
        (test BBddk-4
          (extracter 868 (make-non-empty-tree (cons 'black 778)
                           (make-non-empty-tree (cons 'black 474)
                             (singleton-tree (cons 'red 85))
                             (singleton-tree (cons 'red 500)))
                           (singleton-tree (cons 'black 868))))
          2 (ok? red-black-tree?)
          (lambda (alpha beta)
            (and (numbox? alpha)
                 (equiv? alpha (box 868))
                 (red-black-tree=? beta
                                   (make-non-empty-tree (cons 'black 474)
                                     (singleton-tree (cons 'black 85))
                                     (make-non-empty-tree (cons 'black 778)
                                       (singleton-tree (cons 'red 500))
                                       (make-empty-tree)))))))
        (test BRBdxBB-2
          (extracter 646 (make-non-empty-tree (cons 'black 811)
                           (make-non-empty-tree (cons 'red 646)
                             (make-non-empty-tree (cons 'black 25)
                               (singleton-tree (cons 'red 1))
                               (make-empty-tree))
                             (singleton-tree (cons 'black 769)))
                           (singleton-tree (cons 'black 873))))
          2 (ok? red-black-tree?)
          (lambda (alpha beta)
            (and (numbox? alpha)
                 (equiv? alpha (box 646))
                 (red-black-tree=? beta
                                   (make-non-empty-tree (cons 'black 811)
                                     (make-non-empty-tree (cons 'red 25)
                                       (singleton-tree (cons 'black 1))
                                       (singleton-tree (cons 'black 769)))
                                     (singleton-tree
                                       (cons 'black 873)))))))
        (test BRBdxBB-3
          (extracter 463 (make-non-empty-tree (cons 'black 802)
                           (make-non-empty-tree (cons 'red 273)
                             (make-non-empty-tree (cons 'black 168)
                               (singleton-tree (cons 'red 140))
                               (make-empty-tree))
                             (singleton-tree (cons 'black 463)))
                           (singleton-tree (cons 'black 839))))
          2 (ok? red-black-tree?)
          (lambda (alpha beta)
            (and (numbox? alpha)
                 (equiv? alpha (box 463))
                 (red-black-tree=? beta
                                   (make-non-empty-tree (cons 'black 802)
                                     (make-non-empty-tree (cons 'red 168)
                                       (singleton-tree (cons 'black 140))
                                       (singleton-tree (cons 'black 273)))
                                     (singleton-tree
                                       (cons 'black 839)))))))
        (test BRBdxBB-5
          (extracter 760 (make-non-empty-tree (cons 'black 713)
                           (make-non-empty-tree (cons 'red 579)
                             (make-non-empty-tree (cons 'black 578)
                               (singleton-tree (cons 'red 98))
                               (make-empty-tree))
                             (singleton-tree (cons 'black 691)))
                           (singleton-tree (cons 'black 760))))
          2 (ok? red-black-tree?)
          (lambda (alpha beta)
            (and (numbox? alpha)
                 (equiv? alpha (box 760))
                 (red-black-tree=? beta
                                   (make-non-empty-tree (cons 'black 579)
                                     (make-non-empty-tree (cons 'black 578)
                                       (singleton-tree (cons 'red 98))
                                       (make-empty-tree))
                                     (make-non-empty-tree (cons 'black 713)
                                       (singleton-tree (cons 'red 691))
                                       (make-empty-tree)))))))
        (test elaborate-failure
          (extracter 524 (make-non-empty-tree (cons 'black 808)
                           (make-non-empty-tree (cons 'red 193)
                             (singleton-tree (cons 'black 63))
                             (make-non-empty-tree (cons 'black 459)
                               (singleton-tree (cons 'red 415))
                               (make-empty-tree)))
                           (singleton-tree (cons 'black 983))))
          2 (ok? red-black-tree?)
          (lambda (alpha beta)
            (and (not alpha)
                 (red-black-tree=? beta
                                   (make-non-empty-tree (cons 'black 808)
                                     (make-non-empty-tree (cons 'red 193)
                                       (singleton-tree (cons 'black 63))
                                       (make-non-empty-tree
                                         (cons 'black 459)
                                         (singleton-tree (cons 'red 415))
                                         (make-empty-tree)))
                                     (singleton-tree
                                       (cons 'black 983)))))))))))

(suite rbtable-of
    ((less-or-equal-last-digit? (pipe (~each (sect mod <> 10)) <=)))
  (test specific
    (rbtable-of exact-integer? symbol? less-or-equal-last-digit?)
    1 (procedure?)
    (lambda (tab?)
      (suite rbtable-of-internal ()
        (test empty-table
          (tab? (make-empty-tree))
          1 (boolean?)
          true?)
        (test singleton
          (tab? (singleton-tree (cons 'black (cons 684 'iheanas))))
          1 (boolean?)
          true?)
        (test larger-tree
          (tab? (make-non-empty-tree (cons 'black (cons 396 'anrrdt))
                  (make-non-empty-tree (cons 'black (cons 511 'eoogod))
                    (singleton-tree (cons 'red (cons 70 'fil)))
                    (singleton-tree (cons 'red (cons 362 'rismian))))
                  (singleton-tree (cons 'black (cons 797 'eouoe)))))
          1 (boolean?)
          true?)
        (test random-value
          (tab? 947/955)
          1 (boolean?)
          not)
        (test non-pair-in-tree
          (tab? (make-non-empty-tree (cons 'black (cons 381 'ndrw))
                  (singleton-tree (cons 'black (cons 380 'dyddo)))
                  (make-non-empty-tree 1987/2
                    (singleton-tree (cons 'red (cons 992 'lhrt)))
                    (make-empty-tree))))
          1 (boolean?)
          not)
        (test no-color-in-pair
          (tab? (make-non-empty-tree (cons 'black (cons 395 'eprlrst))
                  (singleton-tree (cons 11 'tntbse))
                  (singleton-tree (cons 'red (cons 538 'ieh)))))
          1 (boolean?)
          not)
        (test wrong-type-key
          (tab? (make-non-empty-tree (cons 'black (cons 388 'hsrl))
                  (singleton-tree (cons 'red (cons 627 'henelo)))
                  (singleton-tree (cons 'red (cons 'hmt 'nehrtae)))))

          1 (boolean?)
          not)
        (test wrong-type-entry
          (tab? (make-non-empty-tree (cons 'black (cons 494 'eghmj))
                  (singleton-tree (cons 'black (cons 391 'taonont)))
                  (make-non-empty-tree (cons 'black (cons 345 883))
                    (singleton-tree (cons 'red (cons 6 'lshol)))
                    (make-empty-tree))))
          1 (boolean?)
          not)
        (test red
          (tab? (singleton-tree (cons 'red (cons 241 'enteh))))
          1 (boolean?)
          not)
        (test fails-color-constraint
          (tab? (make-non-empty-tree (cons 'black (cons 586 'ses))
                  (make-non-empty-tree (cons 'red (cons 384 'hietaeh))
                    (singleton-tree (cons 'red (cons 491 'hfleu)))
                    (make-empty-tree))
                  (singleton-tree (cons 'red (cons 817 'yboohn)))))
          1 (boolean?)
          not)
        (test fails-shape-constraint
          (tab? (make-non-empty-tree (cons 'black (cons 486 'sbweii))
                  (singleton-tree (cons 'black (cons 463 'rooree)))
                  (make-empty-tree)))
          1 (boolean?)
          not)
        (test fails-binary-search-tree-invariant
          (tab? (make-non-empty-tree (cons 'black (cons 323 'knn))
                  (singleton-tree (cons 'red (cons 247 'tenwngn)))
                  (make-empty-tree)))
          1 (boolean?)
          not)
        (test fails-left-leaning-invariant
          (tab? (make-non-empty-tree (cons 'black (cons 930 'tdtkg))
                  (make-empty-tree)
                  (singleton-tree (cons 'red (cons 262 'actta)))))
          1 (boolean?)
          not)
        (test duplicate-key
          (tab? (make-non-empty-tree (cons 'black (cons 503 'eyuwf))
                  (make-non-empty-tree (cons 'red (cons 492 'inhe))
                    (singleton-tree (cons 'black (cons 311 'nra)))
                    (singleton-tree (cons 'black (cons 663 'mll))))
                  (make-non-empty-tree (cons 'black (cons 647 'ioyfor))
                    (singleton-tree (cons 'red (cons 454 'mrnefno)))
                    (make-empty-tree))))
          1 (boolean?)
          not)
        (test duplicate-association
          (tab? (make-non-empty-tree (cons 'black (cons 884 'asspu))
                  (singleton-tree (cons 'black (cons 9 'vtv)))
                  (make-non-empty-tree (cons 'black (cons 835 'veenaui))
                    (singleton-tree (cons 'red (cons 715 'veenaui)))
                    (make-empty-tree))))
          1 (boolean?)
          not)))))

(suite fast-rbtable-adjoiner ()
  (test specific
    (fast-rbtable-adjoiner <=)
    1 (procedure?)
    (lambda (adj)
      (suite fast-rbtable-adjoiner-internal
          ((good-table? (rbtable-of exact-integer? symbol? <=))
           (matcher (compare-by (pipe tree->list (sect map cdr <>))
                                (list-of= (pair-of= = symbol=?)))))
        (test empty-table
          (adj 80 'daptmf (make-empty-tree))
          1 (good-table?)
          (match? matcher
                  (singleton-tree (cons 'black (cons 80 'daptmf)))))
        (test singleton-table
          (adj 410 'dpa (singleton-tree (cons 'black (cons 984 'cntifa))))
          1 (good-table?)
          (match? matcher
                  (make-non-empty-tree (cons 'black (cons 984 'cntifa))
                    (singleton-tree (cons 'red (cons 410 'dpa)))
                    (make-empty-tree))))
        (test larger-table
          (adj 990
               'eira
               (make-non-empty-tree (cons 'black (cons 421 'datw))
                 (singleton-tree (cons 'red (cons 191 'iioeew)))
                 (singleton-tree (cons 'red (cons 779 'omermet)))))
          1 (good-table?)
          (match? matcher
                  (make-non-empty-tree (cons 'black (cons 421 'datw))
                    (singleton-tree (cons 'black (cons 191 'iioeew)))
                    (make-non-empty-tree (cons 'black (cons 990 'eira))
                      (singleton-tree (cons 'red (cons 779 'omermet)))
                      (make-empty-tree)))))))))

(suite rbtable-builder
    ((less-or-equal-last-digit? (pipe (~each (sect mod <> 10)) <=)))
  (test specific
    (rbtable-builder less-or-equal-last-digit?)
    1 (procedure?)
    (lambda (rbtable)
      (suite rbtable-builder-internal
          ((good-table?
            (rbtable-of exact-integer? symbol? less-or-equal-last-digit?))
           (matcher (compare-by 
                      (pipe tree->list (sect map cdr <>))
                      (list-of= (pair-of= (pipe (~each (sect mod <> 10)) =)
                                          symbol=?)))))
        (test no-arguments
          (rbtable)
          1 (good-table?)
          empty-tree?)
        (test several-arguments
          (rbtable (cons 839 'drl)
                   (cons 850 'rhpwlc)
                   (cons 737 'aatrrs)
                   (cons 1 'nha))
          1 (good-table?)
          (match? matcher
                  (make-non-empty-tree (cons 'black (cons 1 'nha))
                    (singleton-tree (cons 'black (cons 0 'rhpwlc)))
                    (make-non-empty-tree (cons 'black (cons 9 'drl))
                      (singleton-tree (cons 'red (cons 7 'aatrrs)))
                      (make-empty-tree)))))))))

(suite rbtable-searcher
    ((less-or-equal-last-digit? (pipe (~each (sect mod <> 10)) <=)))
  (test specific
    (rbtable-searcher less-or-equal-last-digit?)
    1 (procedure?)
    (lambda (searcher)
      (suite rbtable-searcher-internal
          ((rbtable (rbtable-builder less-or-equal-last-digit?)))
        (test minimal
          (searcher 164 (rbtable))
          1 (boolean?)
          not)
        (test empty-table-with-default
          (searcher 394 (rbtable) 'moknnep)
          1 (symbol?)
          (match? symbol=? 'moknnep))
        (test entry-from-singleton
          (searcher 927 (rbtable (cons 157 'rsh)) 'nup)
          1 (symbol?)
          (match? symbol=? 'rsh))
        (test entry-from-larger-table
          (searcher 918
                    (rbtable (cons 306 'othdo)
                             (cons 475 'adag)
                             (cons 78 'dasgoio)
                             (cons 797 'ttmaeti))
                    'hohaao)
          1 (symbol?)
          (match? symbol=? 'dasgoio))
        (test not-found
          (searcher 390
                    (rbtable (cons 155 'tnlne)
                             (cons 264 'ieneau)
                             (cons 499 'rasn)))
          1 (boolean?)
          not)
        (test not-found-with-default
          (searcher 188
                    (rbtable (cons 980 'iof)
                             (cons 961 'tde)
                             (cons 447 'ifepaex)
                             (cons 405 'ertr))
                    'oianap)
          1 (symbol?)
          (match? symbol=? 'oianap))))))

(suite rbtable-keys ((rbtable (rbtable-builder <=)))
  (test empty-table
    (rbtable-keys (rbtable))
    1 (set?)
    empty-set?)
  (test non-empty-table
    (rbtable-keys (rbtable (cons 26 'hteegec)
                           (cons 75 'rodptiy)
                           (cons 347 'vgf)))
    1 (set?)
    (match? set=? (set 26 75 347))))

(suite rbtable-entries ((rbtable (rbtable-builder <=)))
  (test empty-table
    (rbtable-entries (rbtable))
    1 (bag?)
    empty-bag?)
  (test non-empty-table
    (rbtable-entries (rbtable (cons 496 'enurti)
                              (cons 756 'llsee)
                              (cons 16 'lbt)
                              (cons 28 'mefbed)
                              (cons 700 'oanb)
                              (cons 843 'thmv)))
    1 (bag?)
    (match? bag=? (bag 'enurti 'llsee 'lbt 'mefbed 'oanb 'thmv)))
  (test duplicate-entries
    (rbtable-entries (rbtable (cons 25 'mnalao)
                              (cons 363 'tmseyes)
                              (cons 64 'reotoep)
                              (cons 100 'tmseyes)))
    1 (bag?)
    (match? bag=? (bag 'mnalao 'tmseyes 'reotoep 'tmseyes))))

(suite rbtable-deleter
    ((less-or-equal-last-digit? (pipe (~each (sect mod <> 10)) <=)))
  (test specific
    (rbtable-deleter less-or-equal-last-digit?)
    1 (procedure?)
    (lambda (del)
      (suite rb-table-deleter-internal
          ((good-table?
            (rbtable-of exact-integer? symbol? less-or-equal-last-digit?))
           (matcher (compare-by
                      (pipe tree->list (sect map cdr <>))
                      (list-of= (pair-of=
                                  (pipe (~each (sect mod <> 10)) =)
                                  symbol=?))))
           (rbtable (rbtable-builder less-or-equal-last-digit?)))
        (test empty-table
          (del 486 (rbtable))
          1 (good-table?)
          (match? matcher (rbtable)))
        (test singleton-table
          (del 997 (rbtable (cons 357 'otud)))
          1 (good-table?)
          (match? matcher (rbtable)))
        (test larger-table
          (del 611 (rbtable (cons 768 'hntol)
                            (cons 710 'ritas)
                            (cons 41 'pteolw)
                            (cons 543 'aene)
                            (cons 842 'aaeoti)))
          1 (good-table?)
          (match? matcher (rbtable (cons 8 'hntol)
                                   (cons 0 'ritas)
                                   (cons 3 'aene)
                                   (cons 2 'aaeoti))))
        (test unmatched-key
          (del 739 (rbtable (cons 632 'nlg)
                            (cons 238 'tseamr)
                            (cons 670 'weeo)))
          1 (good-table?)
          (match? matcher (rbtable (cons 2 'nlg)
                                   (cons 8 'tseamr)
                                   (cons 0 'weeo))))))))

(suite rbtable-adjoiner
     ((less-or-equal-last-digit? (pipe (~each (sect mod <> 10)) <=)))
  (test specific
    (rbtable-adjoiner less-or-equal-last-digit?)
    1 (procedure?)
    (lambda (adj)
      (suite rbtable-adjoiner-internal
          ((good-table?
            (rbtable-of exact-integer? symbol? less-or-equal-last-digit?))
           (matcher (compare-by 
                      (pipe tree->list (sect map cdr <>))
                      (list-of= (pair-of= (pipe (~each (sect mod <> 10)) =)
                                          symbol=?)))))
        (test empty-table
          (adj 907 'tsnnetr (make-empty-tree))
          1 (good-table?)
          (match? matcher
                  (singleton-tree (cons 'black (cons 7 'tsnnetr)))))
        (test larger-table
          (adj 751 'rmpit (make-non-empty-tree (cons 'black (cons 4 'btif))
                            (singleton-tree (cons 'red (cons 0 'isdtvh)))
                            (singleton-tree (cons 'red (cons 9 'gnso)))))
          1 (good-table?)
          (match? matcher
                  (make-non-empty-tree (cons 'black (cons 4 'btif))
                    (make-non-empty-tree (cons 'black (cons 1 'rmpit))
                      (singleton-tree (cons 'red (cons 0 'isdtvh)))
                      (make-empty-tree))
                    (singleton-tree (cons 'black (cons 9 'gnso))))))
        (test duplicate-key
          (adj 848
               'reroete
               (make-non-empty-tree (cons 'black (cons 118 'hre))
                 (make-non-empty-tree (cons 'black (cons 743 'gcshha))
                   (singleton-tree (cons 'red (cons 712 'twiaaus)))
                   (make-empty-tree))
                 (singleton-tree (cons 'black (cons 19 'dokrnvs)))))
          1 (good-table?)
          (match? matcher
                  (make-non-empty-tree (cons 'black (cons 3 'gcshha))
                    (singleton-tree (cons 'black (cons 2 'twiaaus)))
                    (make-non-empty-tree (cons 'black (cons 9 'dokrnvs))
                      (singleton-tree (cons 'red (cons 8 'reroete)))
                      (make-empty-tree)))))))))

(suite rbtable-size ((rbtable (rbtable-builder <=)))
  (test zero
    (rbtable-size (rbtable))
    1 (natural-number?)
    zero?)
  (test larger
    (rbtable-size (rbtable (cons 496 'sohodiy)
                           (cons 756 'hdxetod)
                           (cons 16 'melrn)
                           (cons 28 'trmhnhy)
                           (cons 700 'rnho)
                           (cons 843 'rseple)
                           (cons 25 'etbu)))
    1 (natural-number?)
    (match? = 7)))

(suite rbtable=? ((rbtable (rbtable-builder <=))
                  (sample ((rbtable-builder <=) (cons 363 'sneye)
                                                (cons 64 'mset)
                                                (cons 100 'dea))))
  (test both-empty
    (rbtable=? (rbtable) (rbtable))
    1 (boolean?)
    true?)
  (test empty-and-non-empty
    (rbtable=? (rbtable) (rbtable (cons 110 'unmsi)))
    1 (boolean?)
    not)
  (test identical-tables
    (rbtable=? sample sample)
    1 (boolean?)
    true?)
  (test duplicate-tables
    (rbtable=? sample (rbtable (cons 363 'sneye)
                               (cons 64 'mset)
                               (cons 100 'dea)))
    1 (boolean?)
    true?)
  (test permuted-tables
    (rbtable=? (rbtable (cons 145 'mal)
                        (cons 678 'liiir)
                        (cons 11 'afbn))
               (rbtable (cons 11 'afbn)
                        (cons 145 'mal)
                        (cons 678 'liiir)))
    1 (boolean?)
    true?)
  (test swapped-entries
    (rbtable=? (rbtable (cons 467 'snthsto)
                        (cons 624 'inwdert)
                        (cons 457 'usyw))
               (rbtable (cons 467 'inwdert)
                        (cons 624 'snthsto)
                        (cons 457 'usyw)))
    1 (boolean?)
    not)
  (test different-entry
    (rbtable=? (rbtable (cons 25 'htdi)
                        (cons 345 'oid)
                        (cons 121 'hdsh))
               (rbtable (cons 25 'htdi)
                        (cons 345 'hhesetd)
                        (cons 121 'hdsh)))
    1 (boolean?)
    not)
  (test extra-key
    (rbtable=? (rbtable (cons 827 'fhhpc)
                        (cons 594 'iaihsr)
                        (cons 126 'nroa)
                        (cons 936 'elh))
               (rbtable (cons 827 'fhhpc)
                        (cons 594 'iaihsr)
                        (cons 936 'elh)))
    1 (boolean?)
    not)
  (test missing-key
    (rbtable=? (rbtable (cons 152 'idse)
                        (cons 120 'ftllye)
                        (cons 766 'ddn))
               (rbtable (cons 152 'idse)
                        (cons 420 'ndatcf)
                        (cons 120 'ftllye)
                        (cons 766 'ddn)))
    1 (boolean?)
    not)
  (test unrelated-key
    (rbtable=? (rbtable (cons 903 'ixutnpw)
                        (cons 4 'esatehd)
                        (cons 16 'hron))
               (rbtable (cons 903 'ixutnpw)
                        (cons 642 'esatehd)
                        (cons 16 'hron)))
    1 (boolean?)
    not)
  (test different-shape
    (rbtable=? (make-non-empty-tree (cons 'black (cons 642 'soyi))
                 (singleton-tree (cons 'black (cons 272 'hand)))
                 (make-non-empty-tree (cons 'black (cons 952 'tsbtihi))
                   (singleton-tree (cons 'red (cons 652 'uas)))
                   (make-empty-tree)))
               (make-non-empty-tree (cons 'black (cons 652 'uas))
                 (make-non-empty-tree (cons 'black (cons 642 'soyi))
                   (singleton-tree (cons 'red (cons 272 'hand)))
                   (make-empty-tree))
                 (singleton-tree (cons 'black (cons 952 'tsbtihi)))))
    1 (boolean?)
    true?)
  (test different-colors
    (rbtable=? (make-non-empty-tree (cons 'black (cons 419 'altrn))
                 (singleton-tree (cons 'black (cons 411 'oiaw)))
                 (singleton-tree (cons 'black (cons 805 'yenonle))))
               (make-non-empty-tree (cons 'black (cons 419 'altrn))
                 (singleton-tree (cons 'red (cons 411 'oiaw)))
                 (singleton-tree (cons 'red (cons 805 'yenonle)))))
    1 (boolean?)
    true?))

(suite rbtable-of=
    ((less-or-equal-last-digit? (pipe (~each (sect mod <> 10)) <=))
     (rbtable (rbtable-builder (pipe (~each (sect mod <> 10)) <=))))
  (test specific
    (rbtable-of= less-or-equal-last-digit? symbol=?)
    1 (procedure?)
    (lambda (tab=?)
      (suite rbtable-of=-internal ((sample (rbtable (cons 419 'etniah)
                                                    (cons 805 'tti))))
        (test both-empty
          (tab=? (rbtable) (rbtable))
          1 (boolean?)
          true?)
        (test empty-and-non-empty
          (tab=? (rbtable (cons 324 'ouoetfa) (cons 411 'eatbnoo))
                 (rbtable))
          1 (boolean?)
          not)
        (test identical-tables
          (tab=? sample sample)
          1 (boolean?)
          true?)
        (test permuted-tables
          (tab=? sample (rbtable (cons 885 'tti) (cons 779 'etniah)))
          1 (boolean?)
          true?)
        (test swapped-entries
          (tab=? (rbtable (cons 33 'ahh)
                          (cons 394 'oapttrr)
                          (cons 880 'tlines))
                 (rbtable (cons 954 'tlines)
                          (cons 750 'oapttrr)
                          (cons 93 'ahh)))
          1 (boolean?)
          not)
        (test different-entry
          (tab=? (rbtable (cons 959 'mpe)
                          (cons 116 'tsi))
                 (rbtable (cons 789 'oemu)
                          (cons 76 'tsi)))
          1 (boolean?)
          not)
        (test extra-key
          (tab=? (rbtable (cons 763 'alt)
                          (cons 995 'utra)
                          (cons 596 'sos))
                 (rbtable (cons 125 'utra)
                          (cons 936 'sos)))
          1 (boolean?)
          not)
        (test unrelated-key
          (tab=? (rbtable (cons 827 'ytnrto)
                          (cons 646 'tsaireg)
                          (cons 984 'fcwt)
                          (cons 92 'etht))
                 (rbtable (cons 16 'tsaireg)
                          (cons 132 'etht)
                          (cons 963 'ahnrt)
                          (cons 387 'ytnrto)))
          1 (boolean?)
          not)))))

(suite rbtable-updater
    ((less-or-equal-last-digit? (pipe (~each (sect mod <> 10)) <=))
     (equal-last-digit? (pipe (~each (sect mod <> 10)) =)))
  (test specific
    (rbtable-updater less-or-equal-last-digit?)
    1 (procedure?)
    (lambda (upd)
      (suite rbtable-updater-internal
          ((rbtable (rbtable-builder less-or-equal-last-digit?))
           (good-table? (rbtable-of exact-integer?
                                    exact-integer?
                                    less-or-equal-last-digit?))
           (matcher (rbtable-of= equal-last-digit? =)))
        (test empty-table
          (upd 980 add1 (rbtable) 878)
          1 (good-table?)
          (match? matcher (rbtable (cons 880 (add1 878)))))
        (test empty-no-default
          (upd 341
               (^if exact-integer? add1 (constant 612))
               (rbtable))
          1 (good-table?)
          (match? matcher (rbtable (cons 361 612))))
        (test replacement
          (upd 278 square (rbtable (cons 758 74)) 795)
          1 (good-table?)
          (match? matcher (rbtable (cons 438 (square 74)))))
        (test default-in-larger-table
          (upd 203
               sub1
               (rbtable (cons 378 277)
                        (cons 799 117)
                        (cons 422 741)
                        (cons 807 486))
               901)
          1 (good-table?)
          (match? matcher (rbtable (cons 378 277)
                                   (cons 799 117)
                                   (cons 422 741)
                                   (cons 807 486)
                                   (cons 203 900))))
        (test replacement-in-larger-table
          (upd 279
               double
               (rbtable (cons 279 447)
                        (cons 882 317)
                        (cons 298 362)
                        (cons 800 733)))
          1 (good-table?)
          (match? matcher (rbtable (cons 279 (double 447))
                                   (cons 882 317)
                                   (cons 298 362)
                                   (cons 800 733))))))))

;;; Test the procedures in the "Heaps" section.

(suite empty-heap ()
  (test check-it
    (empty-heap)
    1 (bush?)
    empty-bush?))

(suite empty-heap? ()
  (test empty-heap
    (empty-heap? (empty-heap))
    1 (boolean?)
    true?)
  (test singleton-heap
    (empty-heap? (bush 'oioaabt))
    1 (boolean?)
    not)
  (test larger-heap
    (empty-heap? (bush 369
                   (bush 461
                     (bush 526 (bush 557) (bush 786) (bush 631)
                               (bush 595)))))
    1 (boolean?)
    not))

(suite satisfies-heap-invariants? ()
  (test typical
    (satisfies-heap-invariants? <=)
    1 (procedure?)
    (lambda (tester)
      (suite satisfies-heap-invariants?-internal ()
        (test singleton-bush
          (tester (bush 587))
          1 (boolean?)
          true?)
        (test good-doubleton-bush
          (tester (bush 112 (bush 647)))
          1 (boolean?)
          true?)
        (test disordered-doubleton-bush
          (tester (bush 640 (bush 509)))
          1 (boolean?)
          not)
        (test contains-empty-bush
          (tester (bush 615 (bush) (bush 939)))
          1 (boolean?)
          not)
        (test larger-bush
          (tester (bush 272
                        (bush 822)
                        (bush 333 (bush 812) (bush 349))
                        (bush 288 (bush 845))))
          1 (boolean?)
          true?)
        (test nested-disorder
          (tester (bush 142
                        (bush 417 (bush 846))
                        (bush 362
                              (bush 728)
                              (bush 613 (bush 380)))))
          1 (boolean?)
          not)
        (test nested-null-bush
          (tester (bush 211
                        (bush 251
                              (bush 511
                                    (bush 518
                                          (bush 663)
                                          (bush)
                                          (bush 702))
                                    (bush 636)))))
          1 (boolean?)
          not)))))

(suite merge-heaps ()
  (test typical
    (merge-heaps <=)
    1 (procedure?)
    (lambda (merger)
      (suite merge-heaps-internal ()
        (test both-empty
          (merger (empty-heap) (empty-heap))
          1 (bush?)
          empty-bush?)
        (test left-empty
          (merger (empty-heap) (bush 286 
                                     (bush 698)
                                     (bush 831)
                                     (bush 891)
                                     (bush 749)
                                     (bush 704)))
          1 ((bush-of exact-integer?))
          (match? (bush-of= =) (bush 286 
                                     (bush 698)
                                     (bush 831)
                                     (bush 891)
                                     (bush 749)
                                     (bush 704))))
        (test right-empty
          (merger (bush 509
                        (bush 679
                              (bush 930))
                        (bush 656)
                        (bush 830))
                  (empty-heap))
          1 ((bush-of exact-integer?))
          (match? (bush-of= =) (bush 509
                                     (bush 679
                                           (bush 930))
                                     (bush 656)
                                     (bush 830))))
        (test neither-empty
          (merger (bush 18
                        (bush 49)
                        (bush 754)
                        (bush 27)
                        (bush 565)
                        (bush 262)
                        (bush 199)
                        (bush 720)
                        (bush 654)
                        (bush 345)
                        (bush 622)
                        (bush 727))
                  (bush 70
                        (bush 316
                              (bush 384)
                              (bush 754))
                        (bush 669)
                        (bush 727)
                        (bush 454)
                        (bush 92)
                        (bush 971)
                        (bush 720)))
          1 ((bush-of exact-integer?))
          (match? (bush-of= =) (bush 18
                                     (bush 70
                                           (bush 316
                                                 (bush 384)
                                                 (bush 754))
                                           (bush 669)
                                           (bush 727)
                                           (bush 454)
                                           (bush 92)
                                           (bush 971)
                                           (bush 720))
                                     (bush 49)
                                     (bush 754)
                                     (bush 27)
                                     (bush 565)
                                     (bush 262)
                                     (bush 199)
                                     (bush 720)
                                     (bush 654)
                                     (bush 345)
                                     (bush 622)
                                     (bush 727))))))))

(suite heap-of ()
  (test typical
    (heap-of <= exact-integer?)
    1 (procedure?)
    (lambda (heap?)
      (suite typical-heap-internal ()
        (test empty-heap
          (heap? (empty-heap))
          1 (boolean?)
          true?)
        (test non-empty-heap
          (heap? (bush 185
                       (bush 338
                             (bush 487)
                             (bush 363)
                             (bush 802)
                             (bush 953)
                             (bush 553)
                             (bush 514
                                   (bush 733)
                                   (bush 897)))))
          1 (boolean?)
          true?)
        (test unicorn
          (heap? 259)
          1 (boolean?)
          not)
        (test non-bush
          (heap? (list 127 (bush 435) 335))
          1 (boolean?)
          not)
        (test not-heap-ordered
          (heap? (bush 40
                       (bush 93
                             (bush 126 (bush 118))
                             (bush 435))
                       (bush 710)))
          1 (boolean?)
          not)
        (test has-empty-child
          (heap? (bush 306
                       (bush 816
                             (bush 187)
                             (bush 171))
                       (bush 11)
                       (bush)
                       (bush 618 (bush 48))))
          1 (boolean?)
          not))))
  (test specialty
    (heap-of (pipe (~each car) string>=?)
             (pair-of string? natural-number?)) 
    1 (procedure?)
    (lambda (heap?)
      (suite specialty-heap-internal ()
        (test yes
          (heap? (bush (cons "uiabo" 54)
                       (bush (cons "owlntlu" 542)
                             (bush (cons "dtrrwd" 179))
                             (bush (cons "nthele" 175))
                             (bush (cons "bmoye" 107))
                             (bush (cons "ceer" 801)))))
          1 (boolean?)
          true?)
        (test bad-ordering
          (heap? (bush (cons "slew" 875)
                       (bush (cons "oroet" 414)
                             (bush (cons "edndet" 905)
                                   (bush (cons "lanv" 744)
                                         (bush (cons "ehacni" 757)))))))
          1 (boolean?)
          not)
        (test bad-value
          (heap? (bush (cons "tei" 424)
                       (bush (cons "ssoett" 49))
                       (bush (cons "nbsmopg" 861)
                             (bush (cons "dmap" 'uep))
                             (bush (cons "nbmehf" 194)))))
          1 (boolean?)
          not)))))

(suite heap-adjoiner ((natural-heap? (heap-of <= natural-number?)))
  (test typical
    (heap-adjoiner <=)
    1 (procedure?)
    (lambda (put-into-heap)
      (suite heap-adjoiner-internal ()
        (test empty-heap
          (put-into-heap 212 (empty-heap))
          1 (natural-heap?)
          (match? (bush-of= =) (bush 212)))
        (test non-empty-heap
          (put-into-heap 593 (bush 324 (bush 789)))
          1 (natural-heap?)
          (match? (bush-of= =) (bush 324 (bush 593) (bush 789))))
        (test raising-multiplicity
          (put-into-heap 306 (bush 306 (bush 887)))
          1 (natural-heap?)
          (match? (bush-of= =) (bush 306 (bush 306 (bush 887)))))))))

(suite heap-list-folder
    ((typical-heap? (heap-of <= exact-integer?))
     (heap (extend-to-variable-arity (empty-heap) (heap-adjoiner <=))))
  (test typical
    (heap-list-folder <=)
    1 (procedure?)
    (lambda (folder)
      (suite heap-list-folder-internal ()
        (test no-heaps
          (folder (list))
          1 (typical-heap?)
          empty-heap?)
        (test one-heap
          (folder (list (heap 968 598 954 610 747)))
          1 (typical-heap?)
          (match? (bush-of= =) (bush 598
                                     (bush 968)
                                     (bush 610
                                           (bush 954)
                                           (bush 747)))))
        (test general-case
          (folder (list (heap 281 756)
                        (heap 611 933 274 874 486)
                        (heap 752)
                        (heap 481 709 753)
                        (heap 772 935 709 874)
                        (heap 563 33)
                        (heap 135)
                        (heap 679)
                        (heap 362)))
          1 (typical-heap?)
          (match? (bush-of= =)
                  (bush 33
                        (bush 281 (bush 756))
                        (bush 274
                              (bush 611)
                              (bush 933)
                              (bush 486 (bush 874)))
                        (bush 752)
                        (bush 481 (bush 709 (bush 753)))
                        (bush 709
                              (bush 772)
                              (bush 935)
                              (bush 874))
                        (bush 135 (bush 362 (bush 679)))
                        (bush 563))))))))

(suite heap-list-catenator
    ((typical-heap? (heap-of <= exact-integer?))
     (heap (extend-to-variable-arity (empty-heap) (heap-adjoiner <=))))
  (test typical
    (heap-list-catenator <=)
    1 (procedure?)
    (lambda (merger)
      (suite heap-list-catenator-internal ()
        (test no-heaps
          (merger (list))
          1 (typical-heap?)
          empty-heap?)
        (test one-heap
          (merger (list (heap 794 305 62 155 808 898)))
          1 (typical-heap?)
          (match? (bush-of= =) (bush 62
                                     (bush 794)
                                     (bush 305)
                                     (bush 155 (bush 808 (bush 898))))))
        (test general-case
          (merger (list (heap 220 554 287 945 620)
                        (heap 778)
                        (heap 242 652 419 422 67 56 841)
                        (heap 316 433 976)
                        (heap 694)
                        (heap 525 746 320)))
          1 (typical-heap?)
          (match? (bush-of= =)
                  (bush 56
                        (bush 242)
                        (bush 652)
                        (bush 419)
                        (bush 422)
                        (bush 67)
                        (bush 841)
                        (bush 220
                              (bush 287
                                    (bush 554)
                                    (bush 620 (bush 945))))
                        (bush 778)
                        (bush 316 (bush 433 (bush 976)))
                        (bush 694)
                        (bush 320 (bush 525) (bush 746)))))))))

(suite heap-list-merger
    ((typical-heap? (heap-of <= exact-integer?))
     (heap (extend-to-variable-arity (empty-heap) (heap-adjoiner <=))))
  (test typical
    (heap-list-merger <=)
    1 (procedure?)
    (lambda (merger)
      (suite heap-list-merger-internal ()
        (test no-heaps
          (merger (list))
          1 (typical-heap?)
          empty-heap?)
        (test one-heap
          (merger (list (heap 864 251 137 262 552 764)))
          1 (typical-heap?)
          (match? (bush-of= =) (bush 137
                                     (bush 864)
                                     (bush 251)
                                     (bush 262
                                           (bush 552 (bush 764))))))
        (test general-case
          (merger (list (heap 344 138)
                        (heap 307 803 44 4 224 953)
                        (heap 768 161 250)
                        (heap 887 811 768 612 777 517 436)
                        (heap 442)
                        (heap 212)
                        (heap 969 351 519 404)))
          1 (typical-heap?)
          (match? (bush-of= =)
                  (bush 4
                        (bush 161
                              (bush 212
                                    (bush 351
                                          (bush 969)
                                          (bush 404 (bush 519)))
                                    (bush 442))
                              (bush 436
                                    (bush 887)
                                    (bush 811)
                                    (bush 768)
                                    (bush 612)
                                    (bush 777)
                                    (bush 517))
                              (bush 768)
                              (bush 250))
                        (bush 138 (bush 344))
                        (bush 307)
                        (bush 803)
                        (bush 44)
                        (bush 224 (bush 953)))))))))

(suite heap-extractor
    ((typical-heap? (heap-of <= exact-integer?))
     (heap (extend-to-variable-arity (empty-heap) (heap-adjoiner <=))))
  (test typical
    (heap-extractor <=)
    1 (procedure?)
    (lambda (extract-from-heap)
      (suite heap-extractor-internal ()
        (test singleton
          (extract-from-heap (heap 306))
          2 (exact-integer? typical-heap?)
          (lambda (alpha beta)
            (and (= alpha 306)
                 (empty-heap? beta))))
        (test larger-heap
          (extract-from-heap (heap 887 43 862 716 299 508)) 
          2 (exact-integer? typical-heap?)
          (lambda (alpha beta)
            (and (= alpha 43)
                 ((bush-of= =) beta
                               (bush 299
                                     (bush 887)
                                     (bush 862)
                                     (bush 716)
                                     (bush 508))))))
        (test duplicate-values
          (extract-from-heap (heap 443 217 145 835 883 145))
          2 (exact-integer? typical-heap?)
          (lambda (alpha beta)
            (and (= alpha 145)
                 ((bush-of= =) beta
                               (bush 145
                                     (bush 217 (bush 443))
                                     (bush 835)
                                     (bush 883))))))))))

(suite heap-of= ((lesser-last-digit (pipe (~each (sect mod <> 10)) <=)))
  (test typical
    (heap-of= string>=?)
    1 (procedure?)
    (lambda (equiv?)
      (suite string-heaps
          ((heap (extend-to-variable-arity (empty-heap)
                                           (heap-adjoiner string>=?))))
        (test empty-heaps
          (equiv? (empty-heap) (empty-heap))
          1 (boolean?)
          true?)
        (test left-heap-empty
          (equiv? (empty-heap) (heap "esu" "hsahtt" "rcwrrpr"))
          1 (boolean?)
          not)
        (test right-heap-empty
          (equiv? (heap "entoe" "oskt") (empty-heap))
          1 (boolean?)
          not)
        (let ((sample (heap "eiu" "lhs" "eere" "eiiss" "eue" "hptw")))
          (test identical-heaps
            (equiv? sample sample)
            1 (boolean?)
            true?))
        (let ((sample (heap "ncwhiol" "daerb" "uame" "oesofo" "eras")))
          (test duplicate-heaps
            (equiv? sample (heap "ncwhiol" "daerb" "uame" "oesofo" "eras"))
            1 (boolean?)
            true?))
        (test different-order
          (equiv? (heap "eee" "dta" "emywsc" "eapodnd" "syhfte" "dtdhsr"
                        "nshtd")
                  (heap "syhfte" "nshtd" "dta" "eee" "dtdhsr" "emywsc"
                        "eapodnd"))
          1 (boolean?)
          true?)
        (test different-elements
          (equiv? (heap "nnawh" "areeaat" "abn" "htdbn")
                  (heap "oyiwts" "iteande" "andgith" "nmcal"))
          1 (boolean?)
          not)
        (test subheap
          (equiv? (heap "atau" "alc" "ieluu" "sle" "hsswirs")
                  (heap "atau" "ieluu" "sle" "hsswirs"))
          1 (boolean?)
          not))))
  (test unusual
    (heap-of= lesser-last-digit)
    1 (procedure?)
    (lambda (equiv?)
      (suite last-digit-heaps
          ((heap (extend-to-variable-arity
                  (empty-heap)
                  (heap-adjoiner lesser-last-digit))))
        (test yes
          (equiv? (heap 777 681 558 18 472 120 450 951 432)
                  (heap 122 582 551 31 647 600 8 650 908))
          1 (boolean?)
          true?)
        (test no
          (equiv? (heap 999 862 955 125 394 876)
                  (heap 9 345 142 732 6 514))
          1 (boolean?)
          not)))))

(suite fold-heap ((nanc (lambda (left right)
                          (* left (add1 right))))
                  (tally-by-parity (lambda (element evens odds)
                                     (if (even? element)
                                         (values (add1 evens) odds)
                                         (values evens (add1 odds)))))
                  (greater-entry (pipe (~each cdr) >=)))
  (test do-nothing
    (fold-heap <= (create) black-hole)
    1 (procedure?)
    (lambda (idler)
      (suite do-nothing-internal
          ((heap (extend-to-variable-arity (empty-heap)
                                           (heap-adjoiner <=))))
        (test stop-at-once
          (idler (empty-heap))
          0 ())
        (test do-nothing-gracefully
          (idler (heap 73 41 387 40 240 551 297))
          0 ()))))
  (test non-associative-non-commutative
    (fold-heap >= (create 801) nanc)
    1 (procedure?)
    (lambda (nancer)
      (suite non-associative-non-commutative-internal
          ((heap (extend-to-variable-arity (empty-heap)
                                           (heap-adjoiner >=))))
        (test empty-heap
          (nancer (empty-heap))
          1 (exact-integer?)
          (match? = 801))
        (test singleton-heap
          (nancer (heap 836))
          1 (exact-integer?)
          (match? = (nanc 836 801)))
        (test larger-heap
          (nancer (heap 13 951 855 495 455))
          1 (exact-integer?)
          (match? = (nanc 951
                      (nanc 855
                        (nanc 495
                          (nanc 455 (nanc 13 801))))))))))
  (test tallying-by-parity
    (fold-heap <= (create 0 0) tally-by-parity)
    1 (procedure?)
    (lambda (tallier)
      (suite tallying-by-parity-internal
          ((heap (extend-to-variable-arity (empty-heap)
                                           (heap-adjoiner <=))))
        (test empty-heap
          (tallier (empty-heap))
          2 (natural-number? natural-number?)
          (match? = 0 0))
        (test several-elements-tally-by-parity
          (tallier (heap 997 118 293 882 519 562 51))
          2 (natural-number? natural-number?)
          (match? = 3 4)))))
  (test complicated-orderer
    (fold-heap greater-entry list prepend)
    1 (procedure?)
    (lambda (orderer)
      (suite complicated-orderer-internal
          ((right-type? (list-of (pair-of symbol? exact-integer?)))
           (equiv? (list-of= (pair-of= symbol=? =)))
           (heap (extend-to-variable-arity (empty-heap)
                                           (heap-adjoiner greater-entry))))
        (test empty-heap
          (orderer (empty-heap))
          1 (right-type?)
          (match? equiv? (list)))
        (test larger-heap
          (orderer (heap (cons 'ddy 382)
                         (cons 'olo 505)
                         (cons 'sso 77)
                         (cons 'aan 978)
                         (cons 'avucmiv 585)))
          1 (right-type?)
          (match? equiv? (list (cons 'aan 978)
                               (cons 'avucmiv 585)
                               (cons 'olo 505)
                               (cons 'ddy 382)
                               (cons 'sso 77))))))))

(suite unfold-heap ((reverse-digits>=? (pipe (~each (lambda (n)
                                                      (+ (* 10 (mod n 10))
                                                         (div n 10))))
                                             >=)))
  (test stop-at-once
    (unfold-heap <= values? list values)
    1 (procedure?)
    (lambda (stopper)
      (suite stop-at-once-internal ()
        (test no-arguments
          (stopper)
          1 ((heap-of values? symbol?))
          empty-heap?)
        (test many-arguments
          (stopper 'rllhflt 'eltesu 'rtmdtd 'ayre)
          1 ((heap-of values? symbol?))
          empty-heap?))))
  (test square-last-digits
    (unfold-heap <= zero? (pipe square (sect mod <> 10)) sub1)
    1 (procedure?)
    (lambda (squarer)
      (suite square-last-digits-internal
          ((heap (extend-to-variable-arity (empty-heap)
                                           (heap-adjoiner <=)))
           (right-type? (heap-of <= natural-number?))
           (equiv? (heap-of= <=)))
        (test none
          (squarer 0)
          1 (right-type?)
          empty-heap?)
        (test more
          (squarer 12)
          1 (right-type?)
          (match? equiv? (heap 1 4 9 6 5 6 9 4 1 0 1 4))))))
  (test multiargument
    (unfold-heap <=
                 (lambda (a b c)
                   (< a 1))
                 *
                 (lambda (a b c)
                   (if (< 1 c)
                       (values a b (sub1 c))
                       (if (< 1 b)
                           (values a (sub1 b) 2)
                           (values (sub1 a) 3 2)))))
    1 (procedure?)
    (lambda (arguer)
      (suite multiargument-internal
          ((heap (extend-to-variable-arity (empty-heap)
                                           (heap-adjoiner <=)))
           (right-type? (heap-of <= natural-number?))
           (equiv? (heap-of= <=)))
        (test run-it
          (arguer 2 3 2)
          1 (right-type?)
          (match? equiv? (heap 12 6 8 4 4 2 6 3 4 2 2 1))))))
  (test reversing-digits
    (unfold-heap reverse-digits>=? empty-list? first rest)
    1 (procedure?)
    (lambda (unfolder)
      (suite reversing-digits-internal
          ((heap (extend-to-variable-arity
                  (empty-heap)
                  (heap-adjoiner reverse-digits>=?)))
           (right-type? (heap-of reverse-digits>=? natural-number?))
           (equiv? (heap-of= reverse-digits>=?)))
        (test empty-list
          (unfolder (list))
          1 (right-type?)
          empty-heap?)
        (test longer-list
          (unfolder (list 89 27 18 37 64 98 46))
          1 (right-type?)
          (match? equiv? (heap 89 27 18 37 64 98 46)))))))

(suite process-heap ((nanc (lambda (left right)
                             (* left (add1 right)))))
  (test minimal
    (process-heap values? (create) (constant))
    1 (procedure?)
    (lambda (idler)
      (suite minimal-internal
          ((heap (extend-to-variable-arity (empty-heap)
                                           (heap-adjoiner values?))))
        (test empty-heap
          (idler (empty-heap))
          0 ())
        (test larger-heap
          (idler (heap 'ohw 'ciomm 'ttw 'yirgakt 'opr 'ewatsaa 'ufh))
          0 ()))))
  (test reverse
    (process-heap <= list prepend)
    1 (procedure?)
    (lambda (reverser)
      (suite reverse-internal
          ((heap (extend-to-variable-arity (empty-heap)
                                           (heap-adjoiner <=)))
           (right-type? (list-of exact-integer?))
           (equiv? (list-of= =)))
        (test empty-heap
          (reverser (heap))
          1 (right-type?)
          empty-list?)
        (test singleton-heap
          (reverser (heap 962))
          1 (right-type?)
          (match? equiv? (list 962)))
        (test doubleton
          (reverser (heap 429 980))
          1 (right-type?)
          (match? equiv? (list 980 429)))
        (test larger-heap
          (reverser (heap 121 151 415 463 301 837 899 286))
          1 (right-type?)
          (match? equiv? (list 899 837 463 415 301 286 151 121))))))
  (test nanc
    (process-heap <= (create 72) nanc)
    1 (procedure?)
    (lambda (nancer)
      (suite nanc-internal
          ((heap (extend-to-variable-arity (empty-heap)
                                           (heap-adjoiner <=))))
        (test empty-heap
          (nancer (empty-heap))
          1 (exact-integer?)
          (match? = 72))
        (test singleton
          (nancer (heap 784))
          1 (exact-integer?)
          (match? = (nanc 784 72)))
        (test larger-heap
          (nancer (heap 98 835 345 602 760))
          1 (exact-integer?)
          (match? = (nanc 835
                      (nanc 760
                        (nanc 602
                          (nanc 345 (nanc 98 72))))))))))
  (test multiargument
    (process-heap <=
                  (create 964 936 776)
                  (lambda (alpha beta gamma delta)
                    (values (+ alpha beta)
                            (+ beta gamma)
                            (+ gamma delta))))
    1 (procedure?)
    (lambda (processor)
      (suite multiargument-internal
          ((heap (extend-to-variable-arity (empty-heap)
                                           (heap-adjoiner <=))))
        (test empty-heap
          (processor (empty-heap))
          3 (exact-integer? exact-integer? exact-integer?)
          (match? = 964 936 776))
        (test larger-heap
          (processor (heap 390 466 610 852))
          3 (exact-integer? exact-integer? exact-integer?)
          (match? = 3282 7504 11940))))))

(suite heapsort
    ((iota (lambda (n)
             ((unfold-list zero? (sect - n <>) sub1) n)))
     (rng (random-number-generator 19990414))
     (sample (let ((rng2 (random-number-generator 20111010)))
               ((fold-natural bag (sect put-into-bag (rng2 128) <>))
                353))))
  (test empty-bag
    (heapsort values? (bag))
    1 (list?)
    empty-list?)
  (test singleton
    (heapsort values? (bag 'sgo))
    1 ((list-of symbol?))
    (match? (list-of= symbol=?) (list 'sgo)))
  (test larger-bag
    (heapsort <= ((lower-ply-natural bag put-into-bag) 509))
    1 ((list-of natural-number?))
    (match? (list-of= =) (iota 509)))
  (test shuffled
    (heapsort <= (shuffle rng ((fold-list bag put-into-bag) (iota 452))))
    1 ((list-of natural-number?))
    (match? (list-of= =) (iota 452)))
  (test all-equal
    (heapsort >= ((fold-natural bag (sect put-into-bag 825 <>)) 959))
    1 ((list-of natural-number?))
    (match? (list-of= =) ((fold-natural list (sect prepend 825 <>)) 959)))
  (test random-elements
    (heapsort <= sample)
    1 ((list-of natural-number?))
    (match? (list-of= =) (mergesort <= sample))))

(suite order-statistic-by-sorting
    ((rng (random-number-generator 20000319))
     (shuffled-bag (lambda (rand size)
                     ((fold-list bag put-into-bag)
                      (shuffle rand ((unfold-list zero? sub1 sub1)
                                     size))))))
  (test singleton-bag
    (order-statistic-by-sorting values? (bag 'tdiyyw) 0)
    1 (symbol?)
    (match? symbol=? 'tdiyyw))
  (test second-to-front
    (order-statistic-by-sorting <= (shuffled-bag rng 86) 1) 
    1 (exact-integer?)
    (match? = 1))
  (test second-from-rear
    (order-statistic-by-sorting <= (shuffled-bag rng 909) 907)
    1 (exact-integer?)
    (match? = 907))
  (test median
    (order-statistic-by-sorting <= (shuffled-bag rng 731) 365)
    1 (exact-integer?)
    (match? = 365))
  (test random-index
    (order-statistic-by-sorting >= (shuffled-bag rng 330) 166)
    1 (exact-integer?)
    (match? = 163))
  (test with-duplicates
    (order-statistic-by-sorting >= (bag 2 7 1 8 2 8 1 8 2 8 5) 7)
    1 (exact-integer?)
    (match? = 2)))

(suite partition-bag-with-count ()
  (test empty-bag
    (partition-bag-with-count values? (bag))
    3 (natural-number? bag? bag?)
    (lambda (alpha beta gamma)
      (and (zero? alpha)
           (empty-bag? beta)
           (empty-bag? gamma))))
  (test singleton-bag-satisfying
    (partition-bag-with-count even? (bag 912))
    3 (natural-number? (bag-of exact-integer?) (bag-of exact-integer?))
    (lambda (alpha beta gamma)
      (and (= alpha 1)
           ((bag-of= =) beta (bag 912))
           (empty-bag? gamma))))
  (test singleton-bag-unsatisfying
    (partition-bag-with-count even? (bag 659))
    3 (natural-number? (bag-of exact-integer?) (bag-of exact-integer?))
    (lambda (alpha beta gamma)
      (and (zero? alpha)
           (empty-bag? beta)
           ((bag-of= =) gamma (bag 659)))))
  (test all-elements-out
    (partition-bag-with-count (constant #f)
                              (bag 'tkm 'dneir 'plnn 'beeb 'rssshtn))
    3 (natural-number? (bag-of symbol?) (bag-of symbol?))
    (lambda (alpha beta gamma)
      (and (zero? alpha)
           (empty-bag? beta)
           ((bag-of= symbol=?) gamma
                               (bag 'tkm 'dneir 'plnn 'beeb 'rssshtn)))))
  (test distributed
    (partition-bag-with-count even?
                              (bag 641 68 704 312 230 150 852 381 675))
    3 (natural-number? (bag-of exact-integer?) (bag-of exact-integer?))
    (lambda (alpha beta gamma)
      (and (= alpha 6)
           ((bag-of= =) beta (bag 68 704 312 230 150 852))
           ((bag-of= =) gamma (bag 641 381 675)))))
  (test all-elements-in
    (partition-bag-with-count values?
                              (bag 'hamiad 'taerar 'udwmshw 'omegpw))
    3 (natural-number? (bag-of symbol?) (bag-of symbol?))
    (lambda (alpha beta gamma)
      (and (= alpha 4)
           ((bag-of= symbol=?) beta
                               (bag 'hamiad 'taerar 'udwmshw 'omegpw))
           (empty-bag? gamma)))))

(suite order-statistic
    ((rng (random-number-generator 19990427))
     (shuffled-bag (lambda (rand size)
                     ((fold-list bag put-into-bag)
                      (shuffle rand ((unfold-list zero? sub1 sub1)
                                     size))))))
  (test singleton-bag
    (order-statistic values? (bag 'fwp) 0)
    1 (symbol?)
    (match? symbol=? 'fwp))
  (test second-to-front
    (order-statistic <= (shuffled-bag rng 792) 1) 
    1 (exact-integer?)
    (match? = 1))
  (test second-from-rear
    (order-statistic <= (shuffled-bag rng 647) 645)
    1 (exact-integer?)
    (match? = 645))
  (test median
    (order-statistic <= (shuffled-bag rng 627) 313)
    1 (exact-integer?)
    (match? = 313))
  (test random-index
    (order-statistic >= (shuffled-bag rng 718) 566)
    1 (exact-integer?)
    (match? = 151))
  (test with-duplicates
    (order-statistic >= (bag 3 1 4 1 5 9 2 6 5 3 5 8 9 7) 7)
    1 (exact-integer?)
    (match? = 5)))

;;; copyright (C) 2011, 2017 John David Stone

;;; This program is free software.  You may redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation -- either version 3 of the License, or (at
;;; your option) any later version.  A copy of the GNU General Public
;;; License is available on the World Wide Web at
;;;
;;;                http://www.gnu.org/licenses/gpl.html
;;;
;;; This program is distributed in the hope that it will be useful, but
;;; WITHOUT ANY WARRANTY -- without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; General Public License for more details.
