;;; This is the test program for the procedures introduced in the chapter
;;; on sublist search.

;;; John David Stone
;;; Department of Computer Science
;;; Grinnell College

;;; created November 2, 2011
;;; last revised January 9, 2017

(import (afp primitives)
        (only (afp arithmetic) add1 sub1)
        (only (afp extras)
              nucleotide-value dna-numeric-value dna-modular-numeric-value)
        (only (afp constant-procedures) constant)
        (only (afp couplers) pipe cross)
        (only (afp lists)
              empty-list? list-of list-of= unfold-list matches-prefix?)
        (rename (afp sublist-search simple-slow-sublist-search)
                (sublist-search simple-slow-sublist-search)
                (substring-search simple-slow-substring-search))
        (afp strings)
        (rename (afp sublist-search Knuth-Morris-Pratt)
                (sublist-search KMP-sublist-search)
                (substring-search KMP-substring-search))
        (rename (afp sublist-search Boyer-Moore)
                (substring-search Boyer-Moore-substring-search))
        (rename (afp sublist-search Rabin-Karp)
                (substring-search Rabin-Karp-substring-search))
        (support testing)
        (support testing-procedures))

;;; Test the procedures in the section on the simple, slow sublist search
;;; algorithm.

(suite matches-prefix? ()
  (test minimal
    (matches-prefix? (list) (list))
    1 (boolean?)
    true?)
  (test yes
    (matches-prefix? (list 'lndeoen 'ewfr 'enueg)
                     (list 'lndeoen 'ewfr 'enueg 'hlhtet 'peow 'etso))
    1 (boolean?)
    true?)
  (test mismatch-at-beginning
    (matches-prefix? (list 'lost 'iiaf 'yooat 'tnroom)
                     (list 'nldehd 'iiaf 'yooat 'tnroom 'tlkeuae))
    1 (boolean?)
    not)
  (test mismatch-in-middle
    (matches-prefix? (list 'othee 'iben 'atfrmr 'aegaob)
                     (list 'othee 'iben 'aveht 'aegaob 'hin 'yduns))
    1 (boolean?)
    not)
  (test mismatch-at-end
    (matches-prefix? (list 'ntyie 'iths 'rtetphi 'otsnmok)
                     (list 'ntyie 'iths 'rtetphi 'roa 'roobh 'eooltn))
    1 (boolean?)
    not)
  (test whole-match
    (matches-prefix? (list 'dtiuds 'srnt 'fonr 'utgil 'yehnc)
                     (list 'dtiuds 'srnt 'fonr 'utgil 'yehnc))
    1 (boolean?)
    true?))

(suite simple-slow-sublist-search ((numlist? (list-of natural-number?))
                                   (numlist=? (list-of= =)))
  (test minimal
    (simple-slow-sublist-search (list) (list))
    1 (numlist?)
    (match? numlist=? (list 0)))
  (test no-matches
    (simple-slow-sublist-search
      (list 'bidpsi 'omhthra)
      (list 'morepe 'arnim 'wsnerl 'mso 'gmirkd 'ota))
    1 (numlist?)
    empty-list?)
  (test partial-matches-only
    (simple-slow-sublist-search
      (list 'entmt 'ige 'ssdr)
      (list 'lgri 'tome 'entmt 'unbsl 'ige 'ssdr 'ution 'entmt 'ige 'dnsph
            'entmt 'entmt 'ige 'sgrerm 'ige 'ssdr 'entmt 'eyioe 'oerwapr
            'liora 'entmt 'ige))
    1 (numlist?)
    empty-list?)
  (test unique-match-at-front
    (simple-slow-sublist-search
      (list 'floihbs 'atolhsd 'rfyelhi 'ooiwrhh)
      (list 'floihbs 'atolhsd 'rfyelhi 'ooiwrhh 'wtdatn 'dutaen 'ohtoa))
    1 (numlist?)
    (match? numlist=? (list 0)))
  (test unique-match-at-end
    (simple-slow-sublist-search
      (list 'wsesu 'oholrg 'shs 'teh)
      (list 'othl 'ont 'one 'botele 'wsesu 'oholrg 'shs 'teh))
    1 (numlist?)
    (match? numlist=? (list 4)))
  (test unique-match-in-middle
    (simple-slow-sublist-search
      (list 'epeamn 'eti 'sose 'eoooae)
      (list 'tetnat 'ileiodi 'epeamn 'eti 'sose 'eoooae 'linotrn))
    1 (numlist?)
    (match? numlist=? (list 2)))
  (test overlapping-matches  ;; example from the text
    (simple-slow-sublist-search
      (list 'foo 'bar 'foo)
      (list 'foo 'foo 'foo 'bar 'foo 'bar 'foo 'foo 'bar 'bar 'foo 'bar
            'foo 'foo 'bar))
    1 (numlist?)
    (match? numlist=? (list 2 4 10)))
  (test match-everywhere
    (simple-slow-sublist-search
      (list 'ioh 'ioh)
      (list 'ioh 'ioh 'ioh 'ioh 'ioh 'ioh 'ioh 'ioh 'ioh))
    1 (numlist?)
    (match? numlist=? (list 0 1 2 3 4 5 6 7)))
  (test whole-list-match
    (simple-slow-sublist-search
      (list 'daeeiov 'aiseian 'tdar 'fmrfntn 'nmtotn)
      (list 'daeeiov 'aiseian 'tdar 'fmrfntn 'nmtotn))
    1 (numlist?)
    (match? numlist=? (list 0)))
  (test empty-pattern
    (simple-slow-sublist-search (list)
                                (list 'wld 'atcsta 'drm 'ngen 'ouofr))
    1 (numlist?)
    (match? numlist=? (list 0 1 2 3 4 5))))

(suite occurs-at? ()
  (test empty-in-empty
    (occurs-at? "" "" 0)
    1 (boolean?)
    true?)
  (test empty-at-beginning
    (occurs-at? "" "odntnea" 0)
    1 (boolean?)
    true?)
  (test empty-in-middle
    (occurs-at? "" "rhn"  2)
    1 (boolean?)
    true?)
  (test empty-at-end
    (occurs-at? "" "sees" 4)
    1 (boolean?)
    true?)
  (test non-empty-at-beginning
    (occurs-at? "ieih" "ieihariefinaorn" 0)
    1 (boolean?)
    true?)
  (test non-empty-in-middle
    (occurs-at? "osrin" "gtbosrinaah" 3)
    1 (boolean?)
    true?)
  (test non-empty-at-end
    (occurs-at? "anooa" "anoehhmhianooa" 9)
    1 (boolean?)
    true?)
  (test mismatch-at-beginning
    (occurs-at? "otthhw" "telereahrmtthhwrenwwl" 9)
    1 (boolean?)
    not)
  (test mismatch-at-end
    (occurs-at? "rdhesi" "tdoslerdhesttoo" 6)
    1 (boolean?)
    not)
  (test mismatch-in-middle
    (occurs-at? "roduenn" "atnuriieoarodkennasl" 10)
    1 (boolean?)
    not)
  (test wrong-position
    (occurs-at? "wbwp" "yosboewbwpshocs" 5)
    1 (boolean?)
    not)
  (test no-occurrence
    (occurs-at? "eocro" "truotefuuenamgf" 7)
    1 (boolean?)
    not))

(suite simple-slow-substring-search ((numlist? (list-of natural-number?))
                                     (numlist=? (list-of= =)))
  (test minimal
    (simple-slow-substring-search "" "")
    1 (numlist?)
    (match? numlist=? (list 0)))
  (test no-matches
    (simple-slow-substring-search "uome" "eaoebls")
    1 (numlist?)
    empty-list?)
  (test partial-only
    (simple-slow-substring-search
      "loitr"
      "loieratyloitgenhauloloitloidkenglooitrsealehlodgwy")
    1 (numlist?)
    empty-list?)
  (test match-at-beginning
    (simple-slow-substring-search "rfwoups" "rfwoupstaeiipcsiwphlhhrig")
    1 (numlist?)
    (match? numlist=? (list 0)))
  (test match-at-end
    (simple-slow-substring-search "abee"
                                  "ddhoeiaetkrlouswhnsemontnrbeulansceabee")
    1 (numlist?)
    (match? numlist=? (list 35)))
  (test match-in-middle
    (simple-slow-substring-search "qoaeo"
                                  "ainmocoeoeoswmnaeeyqoaeobttresdeoio")
    1 (numlist?)
    (match? numlist=? (list 19)))
  (test overlapping
  (simple-slow-substring-search "eanheeanhe"
                                "eanheeanheeanheeanheeanheeanheeanhe")
    1 (numlist?)
    (match? numlist=? (list 0 5 10 15 20 25)))
  (test match-everywhere
    (simple-slow-substring-search "aa" "aaaaaaaaa")
    1 (numlist?)
    (match? numlist=? (list 0 1 2 3 4 5 6 7)))
  (test irregularly-spaced-matches
    (simple-slow-substring-search
      "nokihmnok"
      "eeesnokihmnokhndehmnokihmnokihmnokhoaayuoldenokihmnokhor")
    1 (numlist?)
    (match? numlist=? (list 4 19 25 44)))
  (test near-misses
    (simple-slow-substring-search
      "aaaab"
      "aaaaaaaacaaaadaaeaaaaaabaaabaafaaabbaaaa")
    1 (numlist?)
    (match? numlist=? (list 19)))
  (test empty-string-everywhere
    (simple-slow-substring-search "" "hcdeee")
    1 (numlist?)
    (match? numlist=? (list 0 1 2 3 4 5 6))))

;;; Test the procedures in the section on the Knuth-Morris-Pratt
;;; algorithm.

(suite next-prefix-length ((tricky-case (list 'a 'b 'a 'c 'a 'b 'a 'd 'a 'b
                                              'a 'c 'a 'b 'a 'c)))
  (test minimal
    (next-prefix-length (list 'a 'a) 1 (list 0))
    1 (natural-number?)
    (match? = 1))
  (test minimal-non-matching
    (next-prefix-length (list 'a 'b) 1 (list 0))
    1 (natural-number?)
    zero?)
  (test first
    (next-prefix-length tricky-case 1 (list 0))
    1 (natural-number?)
    zero?)
  (test second
    (next-prefix-length tricky-case 2 (list 0 0))
    1 (natural-number?)
    (match? = 1))
  (test third
    (next-prefix-length tricky-case 3 (list 1 0 0))
    1 (natural-number?)
    zero?)
  (test fifth
    (next-prefix-length tricky-case 5 (list 1 0 1 0 0)) 
    1 (natural-number?)
    (match? = 2))
  (test seventh
    (next-prefix-length tricky-case 7 (list 3 2 1 0 1 0 0))
    1 (natural-number?)
    zero?)
  (test eighth
    (next-prefix-length tricky-case 8 (list 0 3 2 1 0 1 0 0))
    1 (natural-number?)
    (match? = 1))
  (test thirteenth
    (next-prefix-length tricky-case 13 (list 5 4 3 2 1 0 3 2 1 0 1 0 0))
    1 (natural-number?)
    (match? = 6))
  (test fourteenth
    (next-prefix-length tricky-case 14 (list 6 5 4 3 2 1 0 3 2 1 0 1 0 0))
    1 (natural-number?)
    (match? = 7))
  (test fifteenth
    (next-prefix-length tricky-case 15 (list 7 6 5 4 3 2 1 0 3 2 1 0 1 0 0))
    1 (natural-number?)
    (match? = 4)))

(suite prefix-function ((iota (pipe (unfold-list zero? sub1 sub1) reverse)))
  (test singleton-list
    (prefix-function (list 'ayir))
    1 (procedure?)
    (lambda (numberer)
      (suite singleton-list-internal ()
        (test only
          (numberer 1)
          1 (natural-number?)
          zero?))))
  (test two-non-matching
    (prefix-function (list 'nbob 'eaii))
    1 (procedure?)
    (lambda (numberer)
      (suite two-non-matching-internal ()
        (test two-non-matching-first
          (numberer 1)
          1 (natural-number?)
          zero?)
        (test two-non-matching-second
          (numberer 2)
          1 (natural-number?)
          zero?))))
  (test two-matching
    (prefix-function (list 'seyul 'seyul))
    1 (procedure?)
    (lambda (numberer)
      (suite two-matching ()
        (test two-non-matching-first
          (numberer 1)
          1 (natural-number?)
          zero?)
        (test two-non-matching-second
          (numberer 2)
          1 (natural-number?)
          (match? = 1)))))
  (test intricate
    (prefix-function (list 'a 'b 'a 'c 'a 'b 'a 'd 'a 'b 'a 'c 'a 'b 'a 'c))
    1 (procedure?)
    (lambda (numberer)
      ((list-of= =) (map (pipe add1 numberer) (iota 16))
                    (list 0 0 1 0 1 2 3 0 1 2 3 4 5 6 7 4)))))

(suite KMP-stepper ()
  (test minimal
    (KMP-stepper (list 'penhdth))
    1 (procedure?)
    (lambda (stepper)
      (suite minimal-internal ()
        (test minimal-match
          (stepper 'penhdth 0)
          2 (boolean? natural-number?)
          (lambda (alpha beta)
            (and alpha (zero? beta))))
        (test minimal-non-match
          (stepper 'enh 0)
          2 (boolean? natural-number?)
          (lambda (alpha beta)
            (and (not alpha) (zero? beta)))))))
  (test various
    (KMP-stepper (list 'rosnlda 'rtha 'rosnlda 'aathr 'rosnlda 'rtha
                       'rosnlda))
    1 (procedure?)
    (lambda (stepper)
      (suite various-internal ()
        (test partial-match
          (stepper 'rosnlda 4)
          2 (boolean? natural-number?)
          (lambda (alpha beta)
            (and (not alpha) (= beta 5))))
        (test success-without-backtrack
          (stepper 'rosnlda 6)
          2 (boolean? natural-number?)
          (lambda (alpha beta)
            (and alpha (= beta 3))))
        (test success-with-backtrack
          (stepper 'rtha 3)
          2 (boolean? natural-number?)
          (lambda (alpha beta)
            (and (not alpha) (= beta 2))))
        (test no-match
          (stepper 'tdl 0)
          2 (boolean? natural-number?)
          (lambda (alpha beta)
            (and (not alpha) (zero? beta))))
        (test run-down-to-zero
          (stepper 'etaett 6)
          2 (boolean? natural-number?)
          (lambda (alpha beta)
            (and (not alpha) (zero? beta))))))))

(suite KMP-sublist-search ((numlist? (list-of natural-number?))
                           (numlist=? (list-of= =)))
  (test minimal
    (KMP-sublist-search (list) (list))
    1 (numlist?)
    (match? numlist=? (list 0)))
  (test no-matches
    (KMP-sublist-search
      (list 'ayhynoe 'nehsid 'noersti 'mel)
      (list 'fhdtal 'rdhnvu 'nee 'eserl 'drrkgh 'rsri 'apb 'ewfghs 'iii))
    1 (numlist?)
    empty-list?)
  (test partial-matches-only
    (KMP-sublist-search
      (list 'sve 'omsoi 'dnwasa)
      (list 'usta 'oheehn 'sve 'cututab 'omsoi 'dnwasa 'edabe 'sve 'omsoi
            'taniag 'sve 'sve 'omsoi 'gmwgeko 'omsoi 'dnwasa 'sve 'oean
            'aseho 'sqt 'sve 'omsoi))
    1 (numlist?)
    empty-list?)
  (test unique-match-at-front
    (KMP-sublist-search
      (list 'aiuetho 'getefe 'eiyyo)
      (list 'aiuetho 'getefe 'eiyyo 'ifrsat 'oaun 'chyeuoe 'tvuend 'diw))
    1 (numlist?)
    (match? numlist=? (list 0)))
  (test unique-match-at-end
    (KMP-sublist-search
      (list 'noiaawk 'aie 'ftaiim 'assh 'void)
      (list 'gsnsha 'esf 'nnronaq 'noiaawk 'aie 'ftaiim 'assh 'void))
    1 (numlist?)
    (match? numlist=? (list 3)))
  (test unique-match-in-middle
    (KMP-sublist-search
      (list 'ednoee 'ffrehrr 'cdeagty)
      (list 'aneudl 'awno 'grra 'numl 'dvjaa  'ednoee 'ffrehrr 'cdeagty
            'nhe 'ryoclt 'nme 'cwet))
    1 (numlist?)
    (match? numlist=? (list 5)))
  (test overlapping-matches  ;; example from the text
    (KMP-sublist-search
      (list 'foo 'bar 'foo)
      (list 'foo 'foo 'foo 'bar 'foo 'bar 'foo 'foo 'bar 'bar 'foo 'bar
            'foo 'foo 'bar))
    1 (numlist?)
    (match? numlist=? (list 2 4 10)))
  (test match-everywhere
    (KMP-sublist-search
      (list 'ndtey 'ndtey 'ndtey)
      (list 'ndtey 'ndtey 'ndtey 'ndtey 'ndtey 'ndtey 'ndtey 'ndtey
            'ndtey))
    1 (numlist?)
    (match? numlist=? (list 0 1 2 3 4 5 6)))
  (test whole-list-match
    (KMP-sublist-search
      (list 'osr 'hel 'oli 'snmi 'aotiam 'rpui 'frain)
      (list 'osr 'hel 'oli 'snmi 'aotiam 'rpui 'frain))
    1 (numlist?)
    (match? numlist=? (list 0)))
  (test empty-pattern
    (KMP-sublist-search (list)
                        (list 'eitu 'geciah 'ddhh 'akm 'dotpgin 'nshnrig))
    1 (numlist?)
    (match? numlist=? (list 0 1 2 3 4 5 6))))

(suite next-string-prefix-length ((tricky-case "abacabadabacabac"))
  (test minimal
    (next-string-prefix-length "aa" 1 (list 0))
    1 (natural-number?)
    (match? = 1))
  (test minimal-non-matching
    (next-string-prefix-length "ab" 1 (list 0))
    1 (natural-number?)
    zero?)
  (test first
    (next-string-prefix-length tricky-case 1 (list 0))
    1 (natural-number?)
    zero?)
  (test second
    (next-string-prefix-length tricky-case 2 (list 0 0))
    1 (natural-number?)
    (match? = 1))
  (test third
    (next-string-prefix-length tricky-case 3 (list 1 0 0))
    1 (natural-number?)
    zero?)
  (test fifth
    (next-string-prefix-length tricky-case 5 (list 1 0 1 0 0)) 
    1 (natural-number?)
    (match? = 2))
  (test seventh
    (next-string-prefix-length tricky-case 7 (list 3 2 1 0 1 0 0))
    1 (natural-number?)
    zero?)
  (test eighth
    (next-string-prefix-length tricky-case 8 (list 0 3 2 1 0 1 0 0))
    1 (natural-number?)
    (match? = 1))
  (test thirteenth
    (next-string-prefix-length
      tricky-case 13 (list 5 4 3 2 1 0 3 2 1 0 1 0 0))
    1 (natural-number?)
    (match? = 6))
  (test fourteenth
    (next-string-prefix-length
      tricky-case 14 (list 6 5 4 3 2 1 0 3 2 1 0 1 0 0))
    1 (natural-number?)
    (match? = 7))
  (test fifteenth
    (next-string-prefix-length
      tricky-case 15 (list 7 6 5 4 3 2 1 0 3 2 1 0 1 0 0))
    1 (natural-number?)
    (match? = 4)))

(suite string-prefix-function
    ((iota (pipe (unfold-list zero? sub1 sub1) reverse)))
  (test singleton-string
    (string-prefix-function "v")
    1 (procedure?)
    (lambda (numberer)
      (suite singleton-string-internal ()
        (test only
          (numberer 1)
          1 (natural-number?)
          zero?))))
  (test two-non-matching
    (string-prefix-function "ei")
    1 (procedure?)
    (lambda (numberer)
      (suite two-non-matching-internal ()
        (test two-non-matching-first
          (numberer 1)
          1 (natural-number?)
          zero?)
        (test two-non-matching-second
          (numberer 2)
          1 (natural-number?)
          zero?))))
  (test two-matching
    (string-prefix-function "hh")
    1 (procedure?)
    (lambda (numberer)
      (suite two-matching ()
        (test two-non-matching-first
          (numberer 1)
          1 (natural-number?)
          zero?)
        (test two-non-matching-second
          (numberer 2)
          1 (natural-number?)
          (match? = 1)))))
  (test intricate
    (string-prefix-function "abacabadabacabac")
    1 (procedure?)
    (lambda (numberer)
      ((list-of= =) (map (pipe add1 numberer) (iota 16))
                    (list 0 0 1 0 1 2 3 0 1 2 3 4 5 6 7 4)))))

(suite KMP-string-stepper ()
  (test minimal
    (KMP-string-stepper "a")
    1 (procedure?)
    (lambda (stepper)
      (suite minimal-internal ()
        (test minimal-match
          (stepper #\a 0)
          2 (boolean? natural-number?)
          (lambda (alpha beta)
            (and alpha (zero? beta))))
        (test minimal-non-match
          (stepper #\o 0)
          2 (boolean? natural-number?)
          (lambda (alpha beta)
            (and (not alpha) (zero? beta)))))))
  (test various
    (KMP-string-stepper "abacaba")
    1 (procedure?)
    (lambda (stepper)
      (test partial-match
        (stepper #\a 4)
        2 (boolean? natural-number?)
        (lambda (alpha beta)
          (and (not alpha) (= beta 5))))
      (test success-without-backtrack
        (stepper #\a 6)
        2 (boolean? natural-number?)
        (lambda (alpha beta)
          (and alpha (= beta 3))))
      (test success-with-backtrack
        (stepper #\b 3)
        2 (boolean? natural-number?)
        (lambda (alpha beta)
          (and (not alpha) (= beta 2))))
      (test no-match
        (stepper #\n 0)
        2 (boolean? natural-number?)
        (lambda (alpha beta)
          (and (not alpha) (zero? beta))))
      (test run-down-to-zero
        (stepper #\r 6)
        2 (boolean? natural-number?)
        (lambda (alpha beta)
          (and (not alpha) (zero? beta)))))))

(suite KMP-substring-search ((numlist? (list-of natural-number?))
                             (numlist=? (list-of= =)))
  (test null-string-in-null-string
    (KMP-substring-search "" "")
    1 (numlist?)
    (match? numlist=? (list 0)))
  (test null-string-in-non-null-string
    (KMP-substring-search "" "ise")
    1 (numlist?)
    (match? numlist=? (list 0 1 2 3)))
  (test no-matches
    (KMP-substring-search "ire" "nweah")
    1 (numlist?)
    empty-list?)
  (test match-at-beginning
    (KMP-substring-search "dhn" "dhnpesrsomfaiithtaerhwh")
    1 (numlist?)
    (match? numlist=? (list 0)))
  (test match-at-end
    (KMP-substring-search "oee" "pfhelsiwimabeyspumcterarehtoee")
    1 (numlist?)
    (match? numlist=? (list 27)))
  (test match-in-middle
    (KMP-substring-search "saary"
                          "eiuidornblbvspewttasshotasaaryanndrtbtwbtia")
    1 (numlist?)
    (match? numlist=? (list 25)))
  (test several-matches
    (KMP-substring-search "giosnh" "giosnhgiosnhgiosnhgiosnhgiosnh")
    1 (numlist?)
    (match? numlist=? (list 0 6 12 18 24)))
  (test match-everywhere
    (KMP-substring-search "w" "wwwwwwwww")
    1 (numlist?)
    (match? numlist=? (list 0 1 2 3 4 5 6 7 8)))
  (test overlapping-matches
    (KMP-substring-search "nnolnnol" "nnolnnolnnolnnolnnolnnolnnol")
    1 (numlist?)
    (match? numlist=? (list 0 4 8 12 16 20)))
  (test irregularly-spaced-matches
    (KMP-substring-search
       "deahtsrardea"
       "yiynfedeahtsrardeahtsrardeaarambdeahtsrardeaonoqdeahtsrardeaefdelr")
    1 (numlist?)
    (match? numlist=? (list 6 15 32 48)))
  (test near-misses
    (KMP-substring-search "aaaab"
                          "aaaaaaaacaaaadaaeaaaaaabaaabaafaaabbaaaa")
    1 (numlist?)
    (match? numlist=? (list 19)))
  (test case-sensitive
    (KMP-substring-search "eardrt"
                          "inevniheardrteeayaEARDRTrineteeardrtnosetfn")
    1 (numlist?)
    (match? numlist=? (list 7 30)))
  (test stepper
    (KMP-substring-search "abacaba" "dabacabaddabacabdababacabacaba")
    1 (numlist?)
    (match? numlist=? (list 1 19 23))))

;;; Test the procedures in the section on the Boyer-Moore algorithm.

(suite rightmost-occurrence-finder ()
  (test empty-pattern
    (rightmost-occurrence-finder "")
    1 (procedure?)
    (lambda (finder)
      (suite empty-pattern-internal ()
        (test typical-sought
          (finder #\n)
          1 (exact-integer?)
          (match? = -1))
        (test nul-sought
          (finder (integer->char 0))
          1 (exact-integer?)
          (match? = -1)))))
  (test trivial-pattern
    (rightmost-occurrence-finder "r")
    1 (procedure?)
    (lambda (finder)
      (suite trivial-pattern-internal ()
        (test seek-and-find
          (finder #\r)
          1 (exact-integer?)
          zero?)
        (test case-variant
          (finder #\R)
          1 (exact-integer?)
          (match? = -1))
        (test seek-and-fail
          (finder #\;)
          1 (exact-integer?)
          (match? = -1)))))
  (test repeated-values
    (rightmost-occurrence-finder "ahephba")
    1 (procedure?)
    (lambda (finder)
      (suite repeated-values-internal ()
        (test first-and-last
          (finder #\a)
          1 (exact-integer?)
          (match? = 6))
        (test repeated-internal
          (finder #\h)
          1 (exact-integer?)
          (match? = 4))
        (test unique
          (finder #\p)
          1 (exact-integer?)
          (match? = 3))
        (test absent
          (finder #\n)
          1 (exact-integer?)
          (match? = -1)))))
  (test no-repeats
    (rightmost-occurrence-finder "yhbo")
    1 (procedure?)
    (lambda (finder)
      (suite no-repeats-internal ()
        (test try-all
          (map finder (list #\y #\h #\b #\o #\s))
          1 ((list-of exact-integer?))
          (match? (list-of= =) (list 0 1 2 3 -1)))))))

(suite Boyer-Moore-stepper ()
  (test empty-pattern
    (Boyer-Moore-stepper "" "bohmen")
    1 (procedure?)
    (lambda (step)
      (suite empty-pattern-internal ()
        (test beginning
          (step 0)
          1 (exact-integer?)
          (match? = -1))
        (test middle
          (step 3)
          1 (exact-integer?)
          (match? = -1))
        (test end
          (step 6)
          1 (exact-integer?)
          (match? = -1)))))
  (test pattern-equals-text
    (Boyer-Moore-stepper "fnaei" "fnaei")
    1 (procedure?)
    (lambda (step)
      (suite pattern-equals-text-internal ()
        (test successful-match
          (step 0)
          1 (exact-integer?)
          (match? = -1)))))
  (test pattern-partly-matches
    (Boyer-Moore-stepper "uhoeec" "ntoeec")
    1 (procedure?)
    (lambda (step)
      (suite pattern-partly-matches-internal ()
        (test partial-match
          (step 0)
          1 (exact-integer?)
          (match? = 1)))))
  (test mismatch-at-end
    (Boyer-Moore-stepper "nowmadc" "nowmade")
    1 (procedure?)
    (lambda (step)
      (suite mismatch-at-end-internal ()
        (test no-match
          (step 0)
          1 (exact-integer?)
          (match? = 6)))))
  (test variety-of-positions
    (Boyer-Moore-stepper "rheo" "rheoarorhentgelheorandnrheoeitwprheo")
    1 (procedure?)
    (lambda (step)
      (suite variety-of-positions ()
        (test at-beginning
          (step 0)
          1 (exact-integer?)
          (match? = -1))
        (test last-char-only
          (step 3)
          1 (exact-integer?)
          (match? = 2))
        (test first-three-chars
          (step 7)
          1 (exact-integer?)
          (match? = 3))
        (test last-three-chars
          (step 14)
          1 (exact-integer?)
          zero?)
        (test in-middle
          (step 23)
          1 (exact-integer?)
          (match? = -1))
        (test at-end
          (step 32)
          1 (exact-integer?)
          (match? = -1)))))

  ;; Sedgewick, Figure 19.7

  (test sedgewick
    (Boyer-Moore-stepper "STING"
                         "A STRING SEARCHING EXAMPLE CONSISTING OF ...")
    1 (procedure?)
    (lambda (step)
      (suite sedgewick-internal ()
        (test one
          (step 0)
          1 (exact-integer?)
          (match? = 4))
        (test two
          (step 5)
          1 (exact-integer?)
          (match? = 4))
        (test three
          (step 9)
          1 (exact-integer?)
          (match? = 4))
        (test four
          (step 14)
          1 (exact-integer?)
          (match? = 4))
        (test five
          (step 19)
          1 (exact-integer?)
          (match? = 4))
        (test six
          (step 24)
          1 (exact-integer?)
          (match? = 4))
        (test seven
          (step 29)
          1 (exact-integer?)
          (match? = 4))
        (test eight
          (step 32)
          1 (exact-integer?)
          (match? = -1))))))

(suite Boyer-Moore-substring-search ()
  (test empty-string-in-empty-string
    (Boyer-Moore-substring-search "" "")
    1 ((list-of natural-number?))
    (match? (list-of= =) (list 0)))
  (test non-empty-string-in-empty-string
    (Boyer-Moore-substring-search "wfgl" "")
    1 ((list-of natural-number?))
    empty-list?)
  (test empty-string-in-non-empty-string
    (Boyer-Moore-substring-search "" "hcdeee")
    1 ((list-of natural-number?))
    (match? (list-of= =) (list 0 1 2 3 4 5 6)))
  (test no-matches
    (Boyer-Moore-substring-search "uome" "eaoebls")
    1 ((list-of natural-number?))
    (match? (list-of= =) (list)))
  (test match-at-beginning
    (Boyer-Moore-substring-search "rfwoups" "rfwoupstaeiipcsiwphlhhrig")
    1 ((list-of natural-number?))
    (match? (list-of= =) (list 0)))
  (test match-at-end
    (Boyer-Moore-substring-search "abee"
                         "ddhoeiaetkrlouswhnsemontnrbeulansceabee")
    1 ((list-of natural-number?))
    (match? (list-of= =) (list 35)))
  (test match-in-middle
    (Boyer-Moore-substring-search "qoaeo"
                                  "ainmocoeoeoswmnaeeyqoaeobttresdeoio")
    1 ((list-of natural-number?))
    (match? (list-of= =) (list 19)))
  (test several-matches
    (Boyer-Moore-substring-search "hdaa"
                                  "hdaahdaahdaahdaahdaahdaahdaahdaa")
    1 ((list-of natural-number?))
    (match? (list-of= =) (list 0 4 8 12 16 20 24 28)))
  (test match-everywhere
    (Boyer-Moore-substring-search "a" "aaaaaaaaa")
    1 ((list-of natural-number?))
    (match? (list-of= =) (list 0 1 2 3 4 5 6 7 8)))
  (test overlapping-matches
    (Boyer-Moore-substring-search "eanheeanhe"
                                  "eanheeanheeanheeanheeanheeanheeanhe")
    1 ((list-of natural-number?))
    (match? (list-of= =) (list 0 5 10 15 20 25)))
  (test irregularly-spaced-matches
    (Boyer-Moore-substring-search
      "nokihmnok"
      "eeesnokihmnokhndehmnokihmnokihmnokhoaayuoldenokihmnokhor")
    1 ((list-of natural-number?))
    (match? (list-of= =) (list 4 19 25 44)))
  (test near-misses
    (Boyer-Moore-substring-search
      "aaaab"
      "aaaaaaaacaaaadaaeaaaaaabaaabaafaaabbaaaa")
    1 ((list-of natural-number?))
    (match? (list-of= =) (list 19)))
  (test case-sensitive
    (Boyer-Moore-substring-search
      "hsoead"
      "aittarehsoeadbhhlatHSOEADnoaahsoeadprehhth")
    1 ((list-of natural-number?))
    (match? (list-of= =) (list 7 29)))

  ;; Sedgewick, Figure 19.7

  (test sedgewick
    (Boyer-Moore-substring-search
      "STING"
      "A STRING SEARCHING EXAMPLE CONSISTING OF ...")
    1 ((list-of natural-number?))
    (match? (list-of= =) (list 32))))

;;; Tests for procedures in the section on the Rabin-Karp algorithm.

(suite nucleotide-value ()
  (test least
    (nucleotide-value 'A)
    1 (natural-number?)
    zero?)
  (test greatest
    (nucleotide-value 'T)
    1 (natural-number?)
    (match? = 3))
  (test middle
    (nucleotide-value 'G)
    1 (natural-number?)
    (match? = 2)))

(suite dna-numeric-value ()
  (test empty-list
    (dna-numeric-value (list))
    1 (natural-number?)
    zero?)
  (test singleton-list
    (dna-numeric-value (list 'C))
    1 (natural-number?)
    (match? = 1))
  (test longer-list
    (dna-numeric-value (list 'T 'A 'A 'C 'G 'T))  ;; example from text
    1 (natural-number?)
    (match? = 3099))
  (test low-end
    (dna-numeric-value (list 'A 'A 'A 'A 'A 'A 'A 'A 'A))
    1 (natural-number?)
    zero?)
  (test high-end
    (dna-numeric-value (list 'T 'T 'T 'T 'T 'T 'T 'T 'T 'T 'T))
    1 (natural-number?)
    (match? = 4194303)))

(suite dna-modular-numeric-value ()
  (test empty-list
    (dna-modular-numeric-value (list))
    1 (natural-number?)
    zero?)
  (test singleton-list
    (dna-modular-numeric-value (list 'C))
    1 (natural-number?)
    (match? = 1))
  (test low-end
    (dna-modular-numeric-value (list 'A 'A 'A 'A 'A 'A 'A 'A 'A 'A))
    1 (natural-number?)
    zero?)
  (test high-end
    (dna-modular-numeric-value (list 'T 'T 'T 'T 'T 'T 'T 'T 'T))
    1 (natural-number?)
    (match? = 262143))
  (test equal-to-modulus
    (dna-modular-numeric-value
      (list 'C 'T 'T 'T 'T 'T 'T 'T 'T 'T 'T 'C 'G 'C))
    1 (natural-number?)
    zero?)
  (test just-past-modulus
    (dna-modular-numeric-value
      (list 'C 'T 'T 'T 'T 'T 'T 'T 'T 'T 'T 'C 'G 'G))
    1 (natural-number?)
    (match? = 1))
  (test long-random
    (dna-modular-numeric-value
      (list 'G 'G 'A 'G 'G 'T 'T 'A 'G 'G 'C 'G 'T 'C 'A 'A 'G 'T 'A 'A))
    1 (natural-number?)
    (match? = 78244337)))

(suite modular-numeric-evaluator ()
  (test minimal
    (modular-numeric-evaluator 1 (constant 0) 1)
    1 (procedure?)
    (lambda (evaluator)
      (suite minimal-internal ()
        (test empty-list
          (evaluator (list))
          1 (natural-number?)
          zero?)
        (test longer-list
          (evaluator (list null null null null null null null null))
          1 (natural-number?)
          zero?))))
  (test dna
    (modular-numeric-evaluator 4 nucleotide-value 134217689)
    1 (procedure?)
    (lambda (evaluator)
      (suite dna-internal ()
        (test empty-list
          (evaluator (list))
          1 (natural-number?)
          zero?)
        (test longer-list
          (evaluator (list 'G 'G 'A 'G 'G 'T 'T 'A 'G 'G 'C 'G 'T 'C 'A 'A
                           'G 'T 'A 'A))
          1 (natural-number?)
          (match? = 78244337)))))
  (test ascii
    (modular-numeric-evaluator 128 char->integer 45539)
    1 (procedure?)
    (lambda (evaluator)
      (suite ascii-internal ()
        (test singleton
          (evaluator (list #\g))
          1 (natural-number?)
          (match? = 103))
        (test longer-list
          (evaluator (list #\d #\e #\f #\i))
          1 (natural-number?)
          (match? = 36646))))))

(suite slide-text-window ()
  (test minimal
    (slide-text-window 1 1 1)
    1 (procedure?)
    (lambda (slider)
      (suite minimal-internal ()
        (test also-minimal
          (slider 0 0 0)
          1 (natural-number?)
          zero?))))
  (test bitstring
    (slide-text-window 5 2 13)
    1 (procedure?)
    (lambda (slider)
      (suite bitstring-internal ()
        (test all-zeroes
          (slider 0 0 0)
          1 (natural-number?)
          zero?)
        (test all-ones
          (slider 5 1 1)
          1 (natural-number?)
          (match? = 5))
        (test typical
          (slider 11 1 0)
          1 (natural-number?)
          (match? = 3))
        (test after-typical
          (slider 3 1 1)
          1 (natural-number?)
          (match? = 1)))))
  (test ascii
    (slide-text-window 4 128 45539)
    1 (procedure?)
    (lambda (slider)
      (suite ascii-internal ()
        (test all-nuls
          (slider 0 0 0)
          1 (natural-number?)
          zero?)
        (test all-dels
          (slider 28589 127 127)
          1 (natural-number?)
          (match? = 28589))
        (test typical
          (slider 36646 100 110)
          1 (natural-number?)
          (match? = 10238))))))

(suite Rabin-Karp-sublist-searcher ((numlist? (list-of natural-number?))
                                    (numlist=? (list-of= =)))
  (test minimal
    (Rabin-Karp-sublist-searcher 1 (constant 0) 1)
    1 (procedure?)
    (lambda (searcher)
      (suite minimal-internal ()
        (test empty-pattern-empty-text
          (searcher (list) (list))
          1 (numlist?)
          (match? numlist=? (list 0)))
        (test empty-pattern-long-text
          (searcher (list) (list null null null null null))
          1 (numlist?)
          (match? numlist=? (list 0 1 2 3 4 5)))
        (test long-pattern-long-text
          (searcher (list null null null null null)
                    (list null null null null null null null null))
          1 (numlist?)
          (match? numlist=? (list 0 1 2 3))))))
  (test dna
    (Rabin-Karp-sublist-searcher 4 nucleotide-value 134217689)
    1 (procedure?)
    (lambda (searcher)
      (suite dna-internal ()
        (test empty-pattern-empty-text
          (searcher (list) (list))
          1 (numlist?)
          (match? numlist=? (list 0)))
        (test empty-pattern-long-text
          (searcher (list) (list 'C 'A 'G 'A 'T 'C 'A))
          1 (numlist?)
          (match? numlist=? (list 0 1 2 3 4 5 6 7)))
        (test pattern-equals-text
          (searcher (list 'A 'C 'T 'A 'A) (list 'A 'C 'T 'A 'A))
          1 (numlist?)
          (match? numlist=? (list 0)))
        (test singleton-pattern-repeated
          (searcher (list 'C) (list 'C 'C 'C 'C 'C 'G 'C))
          1 (numlist?)
          (match? numlist=? (list 0 1 2 3 4 6)))
        (test no-match
          (searcher (list 'A 'T 'C 'A 'G)
                    (list 'C 'T 'C 'G 'G 'A 'C 'T 'T 'A))
          1 (numlist?)
          empty-list?)
        (test match-at-beginning
          (searcher (list 'A 'T 'C 'A 'G)
                    (list 'A 'T 'C 'A 'G 'C 'A 'T 'C 'A))
          1 (numlist?)
          (match? numlist=? (list 0)))
        (test match-at-end
          (searcher (list 'A 'T 'C 'A 'G)
                    (list 'A 'A 'T 'C 'T 'C 'T 'A 'C 'T 'A 'T 'C 'A 'G))
          1 (numlist?)
          (match? numlist=? (list 10)))
        (test match-in-middle
          (searcher (list 'A 'T 'C 'A 'G)
                    (list 'A 'G 'C 'T 'A 'A 'T 'C 'A 'G 'C 'G 'C 'A 'A))
          1 (numlist?)
          (match? numlist=? (list 5)))
        (test several-matches
          (searcher (list 'T 'T 'G 'C 'G)
                    (list 'T 'T 'G 'C 'G 'T 'T 'G 'C 'G 'T 'T 'G 'C 'G 'T
                          'T 'G 'C 'G 'T 'T 'G 'C 'G 'T 'T 'G 'C 'G 'T 'T
                          'G 'C 'G 'T 'T 'G 'C 'G))
          1 (numlist?)
          (match? numlist=? (list 0 5 10 15 20 25 30 35)))
        (test overlapping-matches
          (searcher (list 'C 'T 'T 'A 'T 'C)
                    (list 'C 'T 'T 'A 'T 'C 'T 'T 'A 'T 'C 'T 'T 'A 'T 'C))
          1 (numlist?)
          (match? numlist=? (list 0 5 10)))
        (test irregularly-spaced-matches
          (searcher (list 'T 'C 'C 'T 'A 'T 'C)
                    (list 'G 'A 'T 'C 'C 'T 'A 'T 'C 'A 'T 'C 'C 'T 'A 'T
                          'C 'C 'T 'A 'T 'C 'C 'G 'T 'C 'C 'T 'C 'C 'T 'A
                          'T 'C 'A 'T))
          1 (numlist?)
          (match? numlist=? (list 2 10 15 27)))
        (test many-near-misses
          (searcher (list 'C 'C 'C 'C 'G)
                    (list 'C 'C 'C 'C 'C 'C 'C 'C 'C 'T 'C 'C 'C 'C 'C 'A
                          'C 'C 'C 'G 'C 'C 'C 'C 'C 'C 'G 'C 'C 'C 'G 'G
                          'C 'C 'C 'C))
          1 (numlist?)
          (match? numlist=? (list 22)))
        (test false-positive-signature
          (searcher (list 'C 'C 'A 'T 'A 'T 'A 'C 'G 'C 'A 'T 'C 'T 'T)
                    (list 'G 'C 'C 'A 'T 'A 'T 'A 'C 'G 'C 'A 'T 'C 'T 'T
                          'C 'T 'A 'T 'A 'T 'A 'C 'G 'C 'A 'T 'T 'G 'A 'T
                          'C 'C 'A 'T 'A 'T 'A 'C 'G 'C 'A 'T 'C 'T 'T))
          1 (numlist?)
          (match? numlist=? (list 1 32))))))
  (test iso-latin-1
    (Rabin-Karp-sublist-searcher 256 char->integer 2097413)
    1 (procedure?)
    (lambda (searcher)
      (suite iso-latin-1-internal ()
        (test empty-pattern-empty-text
          (searcher (list) (list))
          1 (numlist?)
          (match? numlist=? (list 0)))
        (test empty-pattern-long-text
          (searcher (list) (list #\g #\h #\t #\i #\l #\a))
          1 (numlist?)
          (match? numlist=? (list 0 1 2 3 4 5 6)))
        (test pattern-equals-text
          (searcher (list #\a #\y #\d #\a #\a) (list #\a #\y #\d #\a #\a))
          1 (numlist?)
          (match? numlist=? (list 0)))
        (test singleton-pattern-repeated
          (searcher (list #\w) (list #\w #\w #\w #\r #\w #\w))
          1 (numlist?)
          (match? numlist=? (list 0 1 2 4 5)))
        (test no-match
          (searcher (list #\h #\o #\i #\g #\t)
                    (list #\a #\t #\u #\e #\a #\l #\a #\h #\a #\s #\o))
          1 (numlist?)
          empty-list?)
        (test match-at-beginning
          (searcher (list #\g #\v #\n #\k #\f #\e)
                    (list #\g #\v #\n #\k #\f #\e #\e #\h #\a))
          1 (numlist?)
          (match? numlist=? (list 0)))
        (test match-at-end
          (searcher (list #\h #\f #\f)
                    (list #\d #\r #\a #\h #\f #\f))
          1 (numlist?)
          (match? numlist=? (list 3)))
        (test match-in-middle
          (searcher (list #\d #\a #\a #\t #\r #\o)
                    (list #\i #\e #\a #\i #\u #\d #\a #\a #\t #\r #\o #\i
                          #\i #\n))
          1 (numlist?)
          (match? numlist=? (list 5)))
        (test several-matches
          (searcher (list #\e #\n #\e #\l #\e)
                    (list #\e #\n #\e #\l #\e #\e #\n #\e #\l #\e #\e #\n
                          #\e #\l #\e #\e #\n #\e #\l #\e #\e #\n #\e #\l
                          #\e #\e #\n #\e #\l #\e))
          1 (numlist?)
          (match? numlist=? (list 0 5 10 15 20 25)))
        (test overlapping-matches
          (searcher (list #\c #\r #\p #\t #\a #\f #\c #\r)
                    (list #\c #\r #\p #\t #\a #\f #\c #\r #\p #\t #\a #\f
                          #\c #\r #\p #\t #\a #\f #\c #\r #\p #\t #\a #\f
                          #\c #\r #\p #\t #\a #\f #\c #\r #\p #\t #\a #\f
                          #\c #\r))
          1 (numlist?)
          (match? numlist=? (list 0 6 12 18 24 30)))
        (test irregularly-spaced-matches
          (searcher (list #\o #\a #\e #\s #\a)
                    (list #\a #\n #\p #\o #\a #\e #\s #\a #\o #\a #\o #\a
                          #\e #\s #\a #\o #\a #\e #\s #\a #\s #\e #\o #\a
                          #\h #\o #\o #\o #\a #\e #\s #\a #\i #\u))
          1 (numlist?)
          (match? numlist=? (list 3 10 15 27)))
        (test many-near-misses
          (searcher (list #\u #\u #\u #\r)
                    (list #\u #\u #\u #\u #\u #\u #\u #\h #\u #\u #\u #\n
                          #\u #\u #\u #\u #\u #\u #\r #\u #\u #\r #\u #\u
                          #\u #\u #\o #\u #\u #\r #\r #\u #\u #\u))
          1 (numlist?)
          (match? numlist=? (list 15)))
        (test false-positive-signature
          (searcher (list #\d #\i #\m #\i #\n #\i #\s #\h)
                    (list #\r #\d #\i #\m #\i #\n #\i #\s #\h #\e #\t #\t
                          #\c #\a #\s #\e #\w #\o #\r #\k #\n #\d #\i #\d
                          #\i #\m #\i #\n #\i #\s #\h #\h))
          1 (numlist?)
          (match? numlist=? (list 1 23)))))))

(suite string-evaluate ()
  (test empty-string
    (string-evaluate "" 0)
    1 (natural-number?)
    zero?)
  (test singleton-string
    (string-evaluate "p" 1)
    1 (natural-number?)
    (match? = 112))
  (test string-of-nuls
    (string-evaluate "\x0;\x0;\x0;\x0;\x0;\x0;\x0;\x0;\x0;\x0;\x0;" 11)
    1 (natural-number?)
    zero?)
  (test string-of-maxchars
    (string-evaluate "\x10ffff;\x10ffff;\x10ffff;\x10ffff;\x10ffff;\x10ffff;"
                     6)
    1 (natural-number?)
    (match? = 219))
  (test sample
    (string-evaluate "tstvs" 5)
    1 (natural-number?)
    (match? = 431))
  (test another-sample
    (string-evaluate "ecrek" 5)
    1 (natural-number?)
    (match? = 123))
  (test stop-short
    (string-evaluate "tstvs" 3)
    1 (natural-number?)
    (match? = 167)))

(suite Rabin-Karp-substring-search ((numlist? (list-of natural-number?))
                                    (numlist=? (list-of= =)))
  (test empty-pattern-empty-text
    (Rabin-Karp-substring-search "" "")
    1 (numlist?)
    (match? numlist=? (list 0)))
  (test empty-pattern-long-text
    (Rabin-Karp-substring-search "" "htg")
    1 (numlist?)
    (match? numlist=? (list 0 1 2 3)))
  (test singleton-pattern-equals-text
    (Rabin-Karp-substring-search "n" "n")
    1 (numlist?)
    (match? numlist=? (list 0)))
  (test singleton-pattern-repeated
    (Rabin-Karp-substring-search "u" "uuuuuu")
    1 (numlist?)
    (match? numlist=? (list 0 1 2 3 4 5)))
  (test no-matches
    (Rabin-Karp-substring-search "uome" "eaoebls")
    1 (numlist?)
    empty-list?)
  (test match-at-beginning
    (Rabin-Karp-substring-search "rfwoups" "rfwoupstaeiipcsiwphlhhrig")
    1 (numlist?)
    (match? numlist=? (list 0)))
  (test match-at-end
    (Rabin-Karp-substring-search "abee"
                                 "ddhoeiaetkrlouswhnsemontnrbeulansceabee")
    1 (numlist?)
    (match? numlist=? (list 35)))
  (test match-in-middle
    (Rabin-Karp-substring-search "qoaeo"
                                 "ainmocoeoeoswmnaeeyqoaeobttresdeoio")
    1 (numlist?)
    (match? numlist=? (list 19)))
  (test several-matches
    (Rabin-Karp-substring-search "hdaa" "hdaahdaahdaahdaahdaahdaahdaahdaa")
    1 (numlist?)
    (match? numlist=? (list 0 4 8 12 16 20 24 28)))
  (test overlapping-matches
    (Rabin-Karp-substring-search "eanheeanhe"
                                 "eanheeanheeanheeanheeanheeanheeanhe")
    1 (numlist?)
    (match? numlist=? (list 0 5 10 15 20 25)))
  (test irregularly-spaced-matches
    (Rabin-Karp-substring-search
      "nokihmnok"
      "eeesnokihmnokhndehmnokihmnokihmnokhoaayuoldenokihmnokhor")
    1 (numlist?)
    (match? numlist=? (list 4 19 25 44)))
  (test near-misses
    (Rabin-Karp-substring-search "aaaab"
                                 "aaaaaaaacaaaadaaeaaaaaabaaabaafaaabbaaaa")
    1 (numlist?)
    (match? numlist=? (list 19)))
  (test case-sensitive
    (Rabin-Karp-substring-search
      "hsoead"
      "aittarehsoeadbhhlatHSOEADnoaahsoeadprehhth")
    1 (numlist?)
    (match? numlist=? (list 7 29)))
  (test unicode
    (Rabin-Karp-substring-search "déjà vu"
                                 "Je n'ai déjà vu les États-Unis.")
    1 (numlist?)
    (match? numlist=? (list 8)))
  (test collision
    (Rabin-Karp-substring-search "aa" "iaaibōhaat")
    1 (numlist?)
    (match? numlist=? (list 1 7))))

;;; copyright (C) 2011 John David Stone

;;; This program is free software.  You may redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation -- either version 3 of the License, or (at
;;; your option) any later version.  A copy of the GNU General Public
;;; License is available on the World Wide Web at
;;;
;;;                http://www.gnu.org/licenses/gpl.html
;;;
;;; This program is distributed in the hope that it will be useful, but
;;; WITHOUT ANY WARRANTY -- without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; General Public License for more details.
