;;; Testing the code from the "Data structures" chapter

;;; John David Stone
;;; Department of Computer Science
;;; Grinnell College

;;; created August 31, 2011
;;; last revised January 12, 2017

(import (afp primitives)
        (only (afp arithmetic)
              lesser add1 sub1 double halve divisible-by? mod)
        (afp extras)
        (afp constant-procedures)
        (only (afp couplers) pipe cross dispatch)
        (only (afp adapters)
              >initial >next >all-but-initial identity ~each)
        (only (afp predicate-operations)
              ^not ^et ^vel ^if)
        (only (afp natural-numbers)
              ply-natural lower-ply-natural conditionally-tally)
        (afp pairs)
        (afp boxes)
        (afp lists)
        (afp sources)
        (afp trees)
        (afp bushes)
        (afp bags)
        (afp sets)
        (afp tables)
        (afp buffers)
        (support testing)
        (support testing-procedures))

;;; Test the procedures in the "Sum types" section.

(suite suit? ()
  (test diamond-suit
    (suit? 'diamonds)
    1 (boolean?)
    true?)
  (test non-suit-symbol
    (suit? 'hippogryphs)
    1 (boolean?)
    not)
  (test non-symbol
    (suit? 1)
    1 (boolean?)
    not))

(suite suit=? ()
  (test yes
    (suit=? 'hearts 'hearts)
    1 (boolean?)
    true?)
  (test no
    (suit=? 'clubs 'spades)
    1 (boolean?)
    not))

(suite character-or-Boolean? ()
  (test character
    (character-or-Boolean? #\v)
    1 (boolean?)
    true?)
  (test Boolean
    (character-or-Boolean? #f)
    1 (boolean?)
    true?)
  (test other
    (character-or-Boolean? 'bah)
    1 (boolean?)
    not))

(suite character-or-Boolean=? ()
  (test both-characters-yes
    (character-or-Boolean=? #\n #\n)
    1 (boolean?)
    true?)
  (test both-characters-no
    (character-or-Boolean=? #\a #\e)
    1 (boolean?)
    not)
  (test both-Booleans-yes
    (character-or-Boolean=? #f #f)
    1 (boolean?)
    true?)
  (test both-Booleans-no
    (character-or-Boolean=? #f #t)
    1 (boolean?)
    not)
  (test mismatched-source-types
    (character-or-Boolean=? #\t #f)
    1 (boolean?)
    not))

(suite survey-response
    ((yes! (make-confident-survey-response #t))
     (no! (make-confident-survey-response #f))
     (maybe-so (make-diffident-survey-response #t))
     (maybe-not (make-diffident-survey-response #f)))
  (test constructor-yes!
    yes!
    1 (natural-number?)
    (match? = 1))
  (test constructor-no!
    no!
    1 (natural-number?)
    zero?)
  (test constructor-maybe-so
    maybe-so
    1 (natural-number?)
    (match? = 3))
  (test constructor-maybe-not
    maybe-not
    1 (natural-number?)
    (match? = 2))
  (test classification-predicate-low
    (survey-response? -1)
    1 (boolean?)
    not)
  (test classification-predicate-high
    (survey-response? 4)
    1 (boolean?)
    not)
  (test classification-predicate-sneak
    (survey-response? 3/2)
    1 (boolean?)
    not)
  (test classification-predicate-unicorn
    (survey-response? 'unicorn)
    1 (boolean?)
    not)
  (test classification-predicate-ok
    (survey-response? maybe-so)
    1 (boolean?)
    true?)
  (test equality-predicate-ok
    (survey-response=? maybe-not maybe-not)
    1 (boolean?)
    true?)
  (test equality-predicate-different-confidence-level
    (survey-response=? yes! maybe-so)
    1 (boolean?)
    not)
  (test equality-predicate-different-polarity
    (survey-response=? yes! no!)
    1 (boolean?)
    not)
  (test no!-confident?
    (confident? no!)
    1 (boolean?)
    true?)
  (test yes!-confident?
    (confident? yes!)
    1 (boolean?)
    true?)
  (test maybe-not-confident?
    (confident? maybe-not)
    1 (boolean?)
    not)
  (test maybe-so-confident?
    (confident? maybe-so)
    1 (boolean?)
    not)
  (test no!-diffident?
    (diffident? no!)
    1 (boolean?)
    not)
  (test yes!-diffident?
    (diffident? yes!)
    1 (boolean?)
    not)
  (test maybe-not-diffident?
    (diffident? maybe-not)
    1 (boolean?)
    true?)
  (test maybe-so-diffident?
    (diffident? maybe-so)
    1 (boolean?)
    true?)
  (test no!-projection
    (survey-response->Boolean no!)
    1 (boolean?)
    not)
  (test yes!-projection
    (survey-response->Boolean yes!)
    1 (boolean?)
    true?)
  (test maybe-not-projection
    (survey-response->Boolean maybe-not)
    1 (boolean?)
    not)
  (test maybe-so-affirmative?
    (survey-response->Boolean maybe-so)
    1 (boolean?)
    true?)
  (test no!-negation
    (negate no!)
    1 (survey-response?)
    (match? survey-response=? yes!))
  (test yes!-negation
    (negate yes!)
    1 (survey-response?)
    (match? survey-response=? no!))
  (test maybe-not-negation
    (negate maybe-not)
    1 (survey-response?)
    (match? survey-response=? maybe-so))
  (test maybe-so-negation
    (negate maybe-so)
    1 (survey-response?)
    (match? survey-response=? maybe-not)))

;;; Test the procedures from the "Pairs" section.

(suite decons ()
  (test pair-of-symbols
    (decons (cons 'anil 'eis))
    2 (symbol? symbol?)
    (match? symbol=? 'anil 'eis))

  ;; A pair with a null cdr is treated like any other.

  (test null-cdr
    (decons (cons 'rhanal null))
    2 (symbol? null?)
    (pipe >initial (match? symbol=? 'rhanal)))

  ;; It makes no difference if the cdr is itself a pair.

  (test pair-cdr
    (decons (cons 'nln (cons 'dcmdlle (cons 'eaune null))))
    2 (symbol? (pair-of symbol? (pair-of symbol? null?)))
    (lambda (alpha beta)
      (symbol=? alpha 'nln)
      (receive (left right) (decons beta)
        (and (symbol? left)
             (symbol=? left 'dcmdlle)
             (pair? right)
             (receive (inner-left inner-right) (decons right)
               (and (symbol? inner-left)
                    (symbol=? inner-left 'eaune)
                    (null? inner-right))))))))

(suite pair=? ()
  (test yes
    (pair=? (cons 933 385) (cons 933 385))
    1 (boolean?)
    true?)
  (test different-components
    (pair=? (cons 883 612) (cons 626 1))
    1 (boolean?)
    not)
  (test different-order
    (pair=? (cons 513 622) (cons 622 513))
    1 (boolean?)
    not))

(suite character-and-Boolean ((sample (make-character-and-Boolean #\o #t)))
  (test classification-predicate-non-pair
    (character-and-Boolean? 'tyero)
    1 (boolean?)
    not)
  (test classification-predicate-swapped-values
    (character-and-Boolean? (cons #t #\e))
    1 (boolean?)
    not)
  (test classification-predicate-mistype
    (character-and-Boolean? (cons #\p 156))
    1 (boolean?)
    not)
  (test classification-predicate-ok
    (character-and-Boolean? sample)
    1 (boolean?)
    true?)
  (test select-character
    (select-character sample)
    1 (char?)
    (match? char=? #\o))
  (test select-Boolean
    (select-Boolean sample)
    1 (boolean?)
    true?))

(suite alternative-survey-response
    ((yes! (alternative-make-confident-survey-response #t))
     (no! (alternative-make-confident-survey-response #f))
     (maybe-so (alternative-make-diffident-survey-response #t))
     (maybe-not (alternative-make-diffident-survey-response #f)))
  (test constructor-yes!
    yes!
    1 ((pair-of symbol? boolean?))
    (match? (pair-of= symbol=? boolean=?) (cons 'confident #t)))
  (test constructor-no!
    no!
    1 ((pair-of symbol? boolean?))
    (match? (pair-of= symbol=? boolean=?) (cons 'confident #f)))
  (test constructor-maybe-so
    maybe-so
    1 ((pair-of symbol? boolean?))
    (match? (pair-of= symbol=? boolean=?) (cons 'diffident #t)))
  (test constructor-maybe-not
    maybe-not
    1 ((pair-of symbol? boolean?))
    (match? (pair-of= symbol=? boolean=?) (cons 'diffident #f)))
  (test classification-predicate-wrong-symbol
    (alternative-survey-response? (cons 'affident #t))
    1 (boolean?)
    not)
  (test classification-predicate-wrong-type
    (alternative-survey-response? (cons 'confident 692))
    1 (boolean?)
    not)
  (test classification-predicate-unicorn
    (alternative-survey-response? 'unicorn)
    1 (boolean?)
    not)
  (test classification-predicate-ok
    (alternative-survey-response? maybe-so)
    1 (boolean?)
    true?)
  (test equality-predicate-ok
    (alternative-survey-response=? maybe-not maybe-not)
    1 (boolean?)
    true?)
  (test equality-predicate-construction-ok
    (alternative-survey-response=?
     no!
     (alternative-make-confident-survey-response #f))
    1 (boolean?)
    true?)
  (test equality-predicate-different-confidence-level
    (alternative-survey-response=? yes! maybe-so)
    1 (boolean?)
    not)
  (test equality-predicate-different-polarity
    (alternative-survey-response=? yes! no!)
    1 (boolean?)
    not)
  (test no!-confident?
    (alternative-confident? no!)
    1 (boolean?)
    true?)
  (test yes!-confident?
    (alternative-confident? yes!)
    1 (boolean?)
    true?)
  (test maybe-not-confident?
    (alternative-confident? maybe-not)
    1 (boolean?)
    not)
  (test maybe-so-confident?
    (alternative-confident? maybe-so)
    1 (boolean?)
    not)
  (test no!-diffident?
    (alternative-diffident? no!)
    1 (boolean?)
    not)
  (test yes!-diffident?
    (alternative-diffident? yes!)
    1 (boolean?)
    not)
  (test maybe-not-diffident?
    (alternative-diffident? maybe-not)
    1 (boolean?)
    true?)
  (test maybe-so-diffident?
    (alternative-diffident? maybe-so)
    1 (boolean?)
    true?)
  (test no!-projection
    (alternative-survey-response->Boolean no!)
    1 (boolean?)
    not)
  (test yes!-projection
    (alternative-survey-response->Boolean yes!)
    1 (boolean?)
    true?)
  (test maybe-not-projection
    (alternative-survey-response->Boolean maybe-not)
    1 (boolean?)
    not)
  (test maybe-so-affirmative?
    (alternative-survey-response->Boolean maybe-so)
    1 (boolean?)
    true?)
  (test no!-negation
    (alternative-negate no!)
    1 (alternative-survey-response?)
    (match? alternative-survey-response=? yes!))
  (test yes!-negation
    (alternative-negate yes!)
    1 (alternative-survey-response?)
    (match? alternative-survey-response=? no!))
  (test maybe-not-negation
    (alternative-negate maybe-not)
    1 (alternative-survey-response?)
    (match? alternative-survey-response=? maybe-so))
  (test maybe-so-negation
    (alternative-negate maybe-so)
    1 (alternative-survey-response?)
    (match? alternative-survey-response=? maybe-not)))

(suite make-new-zero ()
  (test construction
    (make-new-zero)
    1 (pair?)
    (pipe decons (match? symbol=? 'z 'z))))

(suite make-new-successor ()
  (test one
    (make-new-successor (make-new-zero))
    1 (pair?)
    (lambda (pr)
      (receive (left right) (decons pr)
        (and (symbol? left)
             (symbol=? left 'nn)
             (pair? right)
             ((pipe decons (match? symbol=? 'z 'z)) right)))))
  (test four
    (make-new-successor
     (make-new-successor
      (make-new-successor (make-new-successor (make-new-zero)))))
    1 (pair?)
    (lambda (pr)
      (receive (left4 right4) (decons pr)
        (and (symbol? left4)
             (symbol=? left4 'nn)
             (pair? right4)
             (receive (left3 right3) (decons right4)
               (and (symbol? left3)
                    (symbol=? left3 'nn)
                    (pair? right3)
                    (receive (left2 right2) (decons right3)
                      (and (symbol? left2)
                           (symbol=? left2 'nn)
                           (pair? right2)
                           (receive (left1 right1) (decons right2)
                             (and (symbol? left1)
                                  (symbol=? left1 'nn)
                                  (pair? right1)
                                  ((pipe decons (match? symbol=? 'z 'z))
                                   right1))))))))))))

(suite new-natural-numbers
    ((new-0 (make-new-zero))
     (new-1 (make-new-successor (make-new-zero)))
     (new-2 (make-new-successor (make-new-successor (make-new-zero))))
     (new-3 (make-new-successor
             (make-new-successor (make-new-successor (make-new-zero)))))
     (new-4 (make-new-successor
             (make-new-successor
              (make-new-successor (make-new-successor (make-new-zero))))))
     (new-5 (make-new-successor
             (make-new-successor
              (make-new-successor
               (make-new-successor
                (make-new-successor (make-new-zero))))))) 
     (new-6 (make-new-successor
             (make-new-successor
              (make-new-successor
               (make-new-successor
                (make-new-successor
                 (make-new-successor (make-new-zero))))))))
     (new-7 (make-new-successor
             (make-new-successor
              (make-new-successor
               (make-new-successor
                (make-new-successor
                 (make-new-successor
                  (make-new-successor (make-new-zero)))))))))
     (new-8 (make-new-successor
             (make-new-successor
              (make-new-successor
               (make-new-successor
                (make-new-successor
                 (make-new-successor
                  (make-new-successor
                   (make-new-successor (make-new-zero)))))))))))
  (test zero-is-natural
    (new-natural-number? new-0)
    1 (boolean?)
    true?)
  (test one-is-natural
    (new-natural-number? new-1)
    1 (boolean?)
    true?)
  (test five-is-natural
    (new-natural-number? new-5)
    1 (boolean?)
    true?)
  (test phony-is-unnatural
    (new-natural-number? (cons 'nn (cons 0 null)))
    1 (boolean?)
    not)
  (test non-pair-is-unnatural
    (new-natural-number? 'aot)
    1 (boolean?)
    not)
  (test old-natural-is-unnatural
    (new-natural-number? 260)
    1 (boolean?)
    not)
  (test is-zero-zero?
    (new-zero? new-0)
    1 (boolean?)
    true?)
  (test is-positive-zero?
    (new-zero? new-4)
    1 (boolean?)
    not)
  (test is-zero-positive?
    (new-positive? new-0)
    1 (boolean?)
    not)
  (test is-seven-positive?
    (new-positive? new-7)
    1 (boolean?)
    true?)
  (test successor-of-zero
    (new-add1 new-0)
    1 (new-natural-number?)
    (lambda (pr)
      (receive (left right) (decons pr)
        (and (symbol=? left 'nn)
             (new-zero? right)))))
  (test successor-of-three
    (new-add1 new-3)
    1 (new-natural-number?)
    (lambda (pr)
      (receive (left4 right4) (decons pr)
        (and (symbol=? left4 'nn)
             (receive (left3 right3) (decons right4)
               (and (symbol=? left3 'nn)
                    (receive (left2 right2) (decons right3)
                      (and (symbol=? left2 'nn)
                           (receive (left1 right1) (decons right2)
                             (and (symbol=? left1 'nn)
                                  (new-zero? right1)))))))))))
  (test predecessor-of-one
    (new-sub1 new-1)
    1 (new-natural-number?)
    new-zero?)
  (test predecessor-of-five
    (new-sub1 new-5)
    1 (new-natural-number?)
    (lambda (pr)
      (receive (left4 right4) (decons pr)
        (and (symbol=? left4 'nn)
             (receive (left3 right3) (decons right4)
               (and (symbol=? left3 'nn)
                    (receive (left2 right2) (decons right3)
                      (and (symbol=? left2 'nn)
                           (receive (left1 right1) (decons right2)
                             (and (symbol=? left1 'nn)
                                  (new-zero? right1)))))))))))
  (test are-zeroes-equal?
    (new-= new-0 new-0)
    1 (boolean?)
    true?)
  (test does-arithmetic-work?
    (new-= new-5 (new-+ new-1 new-4))
    1 (boolean?)
    true?)
  (test are-errors-detected?
    (new-= new-8 (new-+ new-5 new-2))
    1 (boolean?)
    not)
  (test left-unequal-to-zero
    (new-= new-3 new-0)
    1 (boolean?)
    not)
  (test right-unequal-to-zero
    (new-= new-0 new-7)
    1 (boolean?)
    not)
  (test do-nothing-fold
    (new-fold-natural (create) (create))
    1 (procedure?)
    (lambda (do-nothing)
      (suite do-nothing-internal ()
        (test do-nothing-gracefully
          (do-nothing new-8)
          0 ()))))
  (test doubler
    (new-fold-natural make-new-zero (pipe new-add1 new-add1))
    1 (procedure?)
    (lambda (double)
      (suite doubler-internal ()
        (test double-zero
          (double new-0)
          1 (new-natural-number?)
          new-zero?)
        (test double-three
          (double new-3)
          1 (new-natural-number?)
          (match? new-= new-6)))))
  (test power-of-two-by-doubling
    (new-fold-natural (create new-1)
                      (new-fold-natural make-new-zero
                                        (pipe new-add1 new-add1)))
    1 (procedure?)
    (lambda (power-of-two)
      (suite power-of-two-by-doubling-internal ()
        (test zero
          (power-of-two new-0)
          1 (new-natural-number?)
          (match? new-= new-1))
        (test positive-power
          (power-of-two new-3)
          1 (new-natural-number?)
          (match? new-= new-8)))))
  (test zero-addend
    (new-+ new-3 new-0)
    1 (new-natural-number?)
    (match? new-= new-3))
  (test zero-augend
    (new-+ new-0 new-4)
    1 (new-natural-number?)
    (match? new-= new-4))
  (test non-zero-addends
    (new-+ new-2 new-3)
    1 (new-natural-number?)
    (match? new-+ new-5))
  (test is-zero-even?
    (new-even? new-0)
    1 (boolean?)
    true?)
  (test is-seven-even?
    (new-even? new-7)
    1 (boolean?)
    not)
  (test is-eight-even?
    (new-even? new-8)
    1 (boolean?)
    true?)
  (test zero-dividend
    (new-div-and-mod new-0 new-3)
    2 (new-natural-number? new-natural-number?)
    (match? new-= new-0 new-0))
  (test unit-divisor
    (new-div-and-mod new-4 new-1)
    2 (new-natural-number? new-natural-number?)
    (match? new-= new-4 new-0))
  (test general-division
    (new-div-and-mod new-7 new-3)
    2 (new-natural-number? new-natural-number?)
    (match? new-= new-2 new-1)))

;;; Test procedures from the "Boxes" section.

(suite box ()
  (test does-it-work?
    (box 'dmhaho)
    1 ((box-of symbol?))
    (lambda (pr)
      (and (symbol? (car pr))
           (symbol=? (car pr) 'dmhaho)
           (null? (cdr pr))))))

(suite boxes ((box0 (box 'gsow))
              (box1 (box 'eodhn))
              (box2 (box 'gsow)))
  (test classification-predicate-yes
    (box? box0)
    1 (boolean?)
    true?)
  (test classification-predicate-no
    (box? 'nsirf)
    1 (boolean?)
    not)
  (test classification-predicate-sneak
    (box? (cons 'dooi 'amcole))
    1 (boolean?)
    not)
  (test equality-predicate-yes
    (box=? box0 box2)
    1 (boolean?)
    true?)
  (test equality-predicate-no
    (box=? box0 box1)
    1 (boolean?)
    not)
  (test debox
    (debox box1)
    1 (symbol?)
    (match? symbol=? 'eodhn)))

;;; Test the procedures in the "Lists" section.

(suite alternative-first ()
  (test singleton
    (alternative-first (list 'nri))
    1 (symbol?)
    (match? symbol=? 'nri))
  (test longer-list
    (alternative-first (list 'pdoe 'ilao 'uaw 'aohrust 'elhb 'sehdadm))
    1 (symbol?)
    (match? symbol=? 'pdoe)))

(suite first ()
  (test singleton
    (alternative-first (list 'thlnred))
    1 (symbol?)
    (match? symbol=? 'thlnred))
  (test longer-list
    (alternative-first (list 'lahyl 'lbis 'euhohre 'smauts 'ubut))
    1 (symbol?)
    (match? symbol=? 'lahyl)))

(suite rest ()
  (test singleton
    (rest (list 'rbat))
    1 (list?)
    (pipe delist (match? symbol=?)))
  (test longer-list
    (rest (list 'laretrf 'msmhms 'diehico 'tlceawl 'ugp 'ftlatsw))
    1 (list?)
    (pipe delist (match? symbol=? 'msmhms 'diehico 'tlceawl 'ugp
                                  'ftlatsw))))

(suite empty-list? ()
  (test empty
    (empty-list? (list))
    1 (boolean?)
    true?)
  (test singleton
    (empty-list? (list 'lor))
    1 (boolean?)
    not)
  (test longer-list
    (empty-list? (list 'nuueaht 'lmcr 'olm 'ndl 'nenf 'ret 'sadi 'tno)) 
    1 (boolean?)
    not))

(suite non-empty-list? ()
  (test empty
    (non-empty-list? (list))
    1 (boolean?)
    not)
  (test singleton
    (non-empty-list? (list 'ttnde))
    1 (boolean?)
    true?)
  (test longer-list
    (non-empty-list? (list 'indityi 'ern 'gos 'aeas 'aelhf 'eyhnhwr))
    1 (boolean?)
    true?))

(suite prepend ()
  (test empty-list
    (prepend 'rthecu (list))
    1 (list?)
    (pipe delist (match? symbol=? 'rthecu)))
  (test longer-list
    (prepend 'dytftir (list 'hdifqtr 'stni 'eeh 'rldlwd 'tgne))
    1 (list?)
    (pipe delist (match? symbol=? 'dytftir 'hdifqtr 'stni 'eeh 'rldlwd
                                  'tgne))))

(suite deprepend ()
  (test singleton
    (deprepend (list 'tnbii))
    2 (symbol? list?)
    (^et (pipe >initial (sect symbol=? <> 'tnbii))
         (pipe >next (pipe delist (match? symbol=?)))))
  (test longer-list
    (deprepend (list 'euwoah 'chu 'tep 'ibry 'teet 'esllnh))
    2 (symbol? list?)
    (^et (pipe >initial (sect symbol=? <> 'euwoah))
         (pipe >next (pipe delist
                           (match? symbol=? 'chu 'tep 'ibry 'teet
                                   'esllnh))))))

(suite list=? ((sample-list (list 'ael 'hlown 'hihem 'kddboh)))
  (test empty-lists
    (list=? (list) (list))
    1 (boolean?)
    true?)
  (test left-list-empty
    (list=? (list) (list 'eweo 'lewofh 'hets 'pnka))
    1 (boolean?)
    not)
  (test right-list-empty
    (list=? (list 'ckel 'hrddeaa 'oto 'nrhrb 'tnoedne) (list))
    1 (boolean?)
    not)
  (test identical-list
    (list=? sample-list sample-list)
    1 (boolean?)
    true?)
  (test duplicate-list
    (list=? sample-list (list 'ael 'hlown 'hihem 'kddboh))
    1 (boolean?)
    true?)
  (test different-order
    (list=? (list 'rhnko 'oerdo 'tdkatg) (list 'oerdo 'tdkatg 'rhnko))
    1 (boolean?)
    not)
  (test different-lengths
    (list=? (list 'iuoisth 'iuoisth 'iuoisth)
            (list 'iuoisth 'iuoisth 'iuoisth 'iuoisth))
    1 (boolean?)
    not)
  (test different-duplicate-elements
    (list=? (list 'eea 'dnslira 'dnslira) (list 'eea 'eea 'dnslira))
    1 (boolean?)
    not))

(suite list-ref ()
  (test singleton
    (alternative-list-ref (list 'iuaarf) 0)
    1 (symbol?)
    (match? symbol=? 'iuaarf))
  (test first-of-longer-list
    (alternative-list-ref (list 'oeecdga 'toi 'fte 'eebat 'ocor 'chdueb) 0)
    1 (symbol?)
    (match? symbol=? 'oeecdga))
  (test last-of-longer-list
    (alternative-list-ref (list 'umirdda 'end 'aeamk 'yosdn 'lyiwft) 4)
    1 (symbol?)
    (match? symbol=? 'lyiwft))
  (test middle-of-longer-list
    (alternative-list-ref (list 'rihhww 'wldc 'oepo 'lemanr 'wporntm
                                'sncaeir 'npsuha 'eranc)
                          5)
    1 (symbol?)
    (match? symbol=? 'sncaeir)))

(suite list-of-numbers? ()
  (test empty
    (list-of-numbers? (list))
    1 (boolean?)
    true?)
  (test good-singleton
    (list-of-numbers? (list 308))
    1 (boolean?)
    true?)
  (test bad-singleton
    (list-of-numbers? (list 'isitrna))
    1 (boolean?)
    not)
  (test good-long-list
    (list-of-numbers? (list 217 277 482 700 228 712 729))
    1 (boolean?)
    true?)
  (test one-wrong-un
    (list-of-numbers? (list 29 871 222 746 'naa 287 619))
    1 (boolean?)
    not)
  (test unicorn
    (list-of-numbers? 833)
    1 (boolean?)
    not))

(suite list-of-numbers=? ((sample (list 109 162 605 206)))
  (test both-empty
    (list-of-numbers=? (list) (list))
    1 (boolean?)
    true?)
  (test left-empty
    (list-of-numbers=? (list) (list 599 794 249))
    1 (boolean?)
    not)
  (test right-empty
    (list-of-numbers=? (list 513 675 747 391) (list))
    1 (boolean?)
    not)
  (test identical-lists
    (list-of-numbers=? sample sample)
    1 (boolean?)
    true?)
  (test like-lists
    (list-of-numbers=? sample (list 109 162 605 206))
    1 (boolean?)
    true?)
  (test length-mismatch
    (list-of-numbers=? (list 121 498 566 782) (list 121 498 566))
    1 (boolean?)
    not)
  (test value-mismatch 
    (list-of-numbers=? (list 885 62 63) (list 885 62 497))
    1 (boolean?)
    not))

(suite list-of ()
  (test generic
    (list-of values?)
    1 (procedure?)
    (lambda (alternative-list?)
      (suite generic-internal ()
        (test empty-list
          (alternative-list? (list))
          1 (boolean?)
          true?)
        (test longer-list
          (alternative-list? (list 'sroi 'ogea 'ahremw 'arefona))
          1 (boolean?)
          true?)
        (test sneak
          (alternative-list? (cons 'ieniahe (cons 'uetcm 'kme)))
          1 (boolean?)
          not)
        (test unicorn
          (alternative-list? 811)
          1 (boolean?)
          not))))
  (test numbers
    (list-of number?)
    1 (procedure?)
    (lambda (alternative-list-of-numbers?)
      (suite numbers-internal ()
        (test empty
          (alternative-list-of-numbers? (list))
          1 (boolean?)
          true?)
        (test good-singleton
          (alternative-list-of-numbers? (list 172))
          1 (boolean?)
          true?)
        (test bad-singleton
          (alternative-list-of-numbers? (list 'atsdtr))
          1 (boolean?)
          not)
        (test good-long-list
          (alternative-list-of-numbers? (list 172 493 174 535))
          1 (boolean?)
          true?)
        (test one-wrong-un
          (alternative-list-of-numbers? (list 2 966 645 305 'hae 822))
          1 (boolean?)
          not)))))

(suite list-of= ()
  (test generic-equality
    (list-of= equal?)
    1 (procedure?)
    (lambda (alternative-list=?)
      (suite generic-equality-internal ((sample (list 'eags 'eohnbl 'gsdni
                                                      'ern 'hpe 'sdiar)))
        (test empty-lists
          (alternative-list=? (list) (list))
          1 (boolean?)
          true?)
        (test left-list-empty
          (alternative-list=? (list) (list 'istbhek))
          1 (boolean?)
          not)
        (test right-list-empty
          (alternative-list=? (list 'ieatna 'fsoho 'sawhhm) (list))
          1 (boolean?)
          not)
        (test identical-list
          (alternative-list=? sample sample)
          1 (boolean?)
          true?)
        (test duplicate-list
          (alternative-list=? sample (list 'eags 'eohnbl 'gsdni 'ern 'hpe
                                           'sdiar))
          1 (boolean?)
          true?)
        (test different-order
          (alternative-list=? (list 'ttnn 'oeoese 'bdm 'hyhe)
                              (list 'bdm 'ttnn 'hyhe 'oeoese))
          1 (boolean?)
          not)
        (test different-lengths
          (alternative-list=? (list 'rastdr 'rastdr 'rastdr 'rastdr)
                              (list 'rastdr 'rastdr))
          1 (boolean?)
          not)
        (test different-duplicate-elements
          (alternative-list=? (list 'haidaym 'old 'old 'smrys 'posn)
                              (list 'haidaym 'old 'smrys 'smrys 'posn))
          1 (boolean?)
          not))))
  (test number-equality
    (list-of= =)
    1 (procedure?)
    (lambda (alternative-list-of-numbers=?)
      (suite number-equality-internal ((sample (list 585 534 294 558 640)))
        (test both-empty
          (alternative-list-of-numbers=? (list) (list))
          1 (boolean?)
          true?)
        (test left-empty
          (alternative-list-of-numbers=? (list) (list 266 993 651 958))
          1 (boolean?)
          not)
        (test right-empty
          (alternative-list-of-numbers=? (list 638 588 467) (list))
          1 (boolean?)
          not)
        (test identical-lists
          (alternative-list-of-numbers=? sample sample)
          1 (boolean?)
          true?)
        (test like-lists
          (alternative-list-of-numbers=? sample
                                         (list 585 534 294 558 640))
          1 (boolean?)
          true?)
        (test length-mismatch
          (alternative-list-of-numbers=? (list 785 193 951 899 785)
                                         (list 844 623))
          1 (boolean?)
          not)
        (test value-mismatch 
          (alternative-list-of-numbers=? (list 640 525 982 17)
                                         (list 640 525 928 17))
          1 (boolean?)
          not)))))

(suite adapter ()
  (test no-indices
    (adapter)
    1 (procedure?)
    (lambda (black-hole)
      (suite no-indices-internal ()
        (test black-hole-nullary
          (black-hole)
          0 ())
        (test black-hole-multary
          (black-hole 'wbtey 'ibtfreo 'agpet 'ifo 'fyerb 'cii)
          0 ()))))
  (test selection
    (adapter 0 2 3 6)
    1 (procedure?)
    (lambda (selector)
      (suite selection-internal ()
        (test selecting
          (selector 'omef 'ebltm 'otrta 'ideu 'torhtr 'oenh 'ekln)
          4 (symbol? symbol? symbol? symbol?)
          (match? symbol=? 'omef 'otrta 'ideu 'ekln)))))
  (test permutation
    (adapter 3 1 0 4 2)
    1 (procedure?)
    (lambda (permuter)
      (suite permutation-internal ()
        (test permuting
          (permuter 'eeldua 'sdxd 'eudtyfa 'eesdst 'rtqosa)
          5 (symbol? symbol? symbol? symbol? symbol?)
          (match? symbol=? 'eesdst 'sdxd 'eeldua 'rtqosa 'eudtyfa)))))
  (test duplication
    (adapter 2 2 2 2 2 2)
    1 (procedure?)
    (lambda (duplicator)
      (suite duplication-internal ()
        (test duplicating
          (duplicator 'smnee 'teova 'etrde 'tttnoo)
          6 (symbol? symbol? symbol? symbol? symbol? symbol?)
          (match? symbol=? 'etrde 'etrde 'etrde 'etrde 'etrde 'etrde))))))

(suite alternative-length ()
  (test empty-list
    (alternative-length (list))
    1 (natural-number?)
    zero?)
  (test singleton
    (alternative-length (list 'ssehgo))
    1 (natural-number?)
    (match? = 1))
  (test longer-list
    (alternative-length (list 'sthy 'atsnir 'etddet 'ieih 'affeegn 'wshaoe
                              'wmlurym 'dncioms 'mtyeflu 'lynn 'soaha
                              'hpela))
    1 (natural-number?)
    (match? = 12)))

(suite sum ()
  (test empty-list
    (sum (list))
    1 (exact-integer?)
    zero?)
  (test singleton
    (sum (list 976))
    1 (exact-integer?)
    (match? = 976))
  (test longer-list
    (sum (list 294 11 468 250 281 408 71))
    1 (exact-integer?)
    (match? = (+ 294 11 468 250 281 408 71))))

(suite catenate ()
  (test both-arguments-empty
    (catenate (list) (list))
    1 (list?)
    empty-list?)
  (test first-argument-empty
    (catenate (list) (list 'ehd 'lan 'aesrgeh 'sdef))
    1 ((list-of symbol?))
    (match? (list-of= symbol=?) (list 'ehd 'lan 'aesrgeh 'sdef)))
  (test second-argument-empty
    (catenate (list 'hewrt 'iifupe 'vteveas 'sre 'eurp) (list))
    1 ((list-of symbol?))
    (match? (list-of= symbol=?) (list 'hewrt 'iifupe 'vteveas 'sre 'eurp)))
  (test singleton-left
    (catenate (list 'oiweet) (list 'opoy 'oet 'stotte))
    1 ((list-of symbol?))
    (match? (list-of= symbol=?) (list 'oiweet 'opoy 'oet 'stotte)))
  (test singleton-right
    (catenate (list 'snfi 'lcidcvl 'hie 'dri 'uynhr) (list 'wcuari))
    1 ((list-of symbol?))
    (match? (list-of= symbol=?)
            (list 'snfi 'lcidcvl 'hie 'dri 'uynhr 'wcuari)))
  (test longer-lists
    (catenate (list 'efneo 'yriagt 'hea 'hwoao 'art)
              (list 'tirnwi 'eoohlvc 'derdea 'sgo))
    1 ((list-of symbol?))
    (match? (list-of= symbol=?)
            (list 'efneo 'yriagt 'hea 'hwoao 'art 'tirnwi 'eoohlvc 'derdea
                  'sgo))))

(suite fold-list
    ((nanc (lambda (left right)
             (* left (add1 right))))
     (tally-by-parity (lambda (element evens odds)
                        (if (even? element)
                            (values (add1 evens) odds)
                            (values evens (add1 odds))))))
  (test do-nothing
    (fold-list (create) black-hole)
    1 (procedure?)
    (lambda (idler)
      (suite do-nothing-internal ()
        (test do-nothing-at-all
          (idler (list))
          0 ())
        (test do-nothing-at-length
          (idler (list 'eulos 'oeirwdu 'ahtwbrt 'awg 'hgine))
          0 ()))))

  ;; Fold some lists together, using a non-associative and non-commutative
  ;; procedure so that any mistakes in the direction or order of folding
  ;; are visible.

  (test fold-nanc
    (fold-list (create 887) nanc)
    1 (procedure?)
    (lambda (folder)
      (suite fold-nanc-internal ()
        (test empty-list
          (folder (list))
          1 (exact-integer?)
          (match? = 887))
        (test singleton
          (folder (list 821))
          1 (exact-integer?)
          (match? = (nanc 821 887)))
        (test several-elements
          (folder (list 902 743 482 301 849))
          1 (exact-integer?)
          (match? = (nanc 902
                          (nanc 743
                                (nanc 482
                                      (nanc 301 (nanc 849 887))))))))))
  (test fold-tally-by-parity
    (fold-list (create 0 0) tally-by-parity)
    1 (procedure?)
    (lambda (folder)
      (suite fold-tally-by-parity-internal ()
        (test empty-list
          (folder (list))
          2 (natural-number? natural-number?)
          (match? = 0 0))
        (test several-elements
          (folder (list 80 972 428 721 473 359 558))
          2 (natural-number? natural-number?)
          (match? = 4 3))))))

(suite length-ufl ()
  (test empty-list
    (length-using-fold-list (list))
    1 (natural-number?)
    zero?)
  (test singleton
    (length-using-fold-list (list 'hesookw))
    1 (natural-number?)
    (match? = 1))
  (test longer-list
    (length-using-fold-list (list 'ofc 'oulee 'dmhnv 'tieiiac 'onibtu
                                  'etiro 'monlys 'lnrtwo 'fira))
    1 (natural-number?)
    (match? = 9)))

(suite sum-ufl ()
  (test empty-list
    (sum-using-fold-list (list))
    1 (exact-integer?)
    zero?)
  (test singleton
    (sum-using-fold-list (list 566))
    1 (exact-integer?)
    (match? = 566))
  (test longer-list
    (sum (list 826 27 714 79 730))
    1 (exact-integer?)
    (match? = (+  826 27 714 79 730))))

(suite catenate-ufl ()
  (test both-arguments-empty
    (catenate-using-fold-list (list) (list))
    1 (list?)
    empty-list?)
  (test first-argument-empty
    (catenate-using-fold-list (list) (list  'hiowsi 'enhtee))
    1 ((list-of symbol?))
    (match? (list-of= symbol=?) (list 'hiowsi 'enhtee)))
  (test second-argument-empty
    (catenate-using-fold-list (list 'dte 'tteoeas 'fehougi 'tci 'fielhn
                                    'tnnfn)
                              (list))
    1 ((list-of symbol?))
    (match? (list-of= symbol=?)
            (list 'dte 'tteoeas 'fehougi 'tci 'fielhn 'tnnfn)))
  (test singleton-left
    (catenate-using-fold-list (list 'wder)
                              (list 'tsperr 'gtrnh 'hbrehne 'rbrah))
    1 ((list-of symbol?))
    (match? (list-of= symbol=?)
            (list 'wder 'tsperr 'gtrnh 'hbrehne 'rbrah))) 
  (test singleton-right
    (catenate-using-fold-list (list 'atweft 'dseoeoe 'trimna) (list 'eind))
    1 ((list-of symbol?))
    (match? (list-of= symbol=?) (list 'atweft 'dseoeoe 'trimna 'eind)))
  (test longer-lists
    (catenate-using-fold-list (list 'fkndk 'vhrswdu 'iiabs)
                              (list 'sthss 'teferet 'tsae 'iphie))
    1 ((list-of symbol?))
    (match? (list-of= symbol=?)
            (list 'fkndk 'vhrswdu 'iiabs 'sthss 'teferet 'tsae 'iphie))))

(suite alternative-fold-list
    ((nanc (lambda (left right)
             (* left (add1 right))))
     (tally-by-parity (lambda (element evens odds)
                        (if (even? element)
                            (values (add1 evens) odds)
                            (values evens (add1 odds))))))
  (test do-nothing
    (alternative-fold-list (create) black-hole)
    1 (procedure?)
    (lambda (idler)
      (suite do-nothing-internal ()
        (test do-nothing-at-all
          (idler (list))
          0 ())
        (test do-nothing-at-length
          (idler (list 'eulos 'oeirwdu 'ahtwbrt 'awg 'hgine))
          0 ()))))
  (test fold-nanc
    (fold-list (create 948) nanc)
    1 (procedure?)
    (lambda (folder)
      (suite fold-nanc-internal ()
        (test empty-list
          (folder (list))
          1 (exact-integer?)
          (match? = 948))
        (test singleton
          (folder (list 396))
          1 (exact-integer?)
          (match? = (nanc 396 948)))
        (test several-elements
          (folder (list 953 28 980 648))
          1 (exact-integer?)
          (match? = (nanc 953
                          (nanc 28
                                (nanc 980 (nanc 648 948)))))))))
  (test fold-tally-by-parity
    (fold-list (create 0 0) tally-by-parity)
    1 (procedure?)
    (lambda (folder)
      (suite fold-tally-by-parity-internal ()
        (test empty-list
          (folder (list))
          2 (natural-number? natural-number?)
          (match? = 0 0))
        (test several-elements
          (folder (list 826 722 780 448 933 774 178))
          2 (natural-number? natural-number?)
          (match? = 6 1))))))

(suite process-list ((nanc (lambda (left right)
                             (* left (add1 right)))))
  (test minimal
    (process-list (create) (constant))
    1 (procedure?)
    (lambda (idler)
      (suite minimal-internal ()
        (test empty-list
          (idler (list))
          0 ())
        (test longer-list
          (idler (list 'oiwola 'raetee 'oiuweg))
          0 ()))))
  (test reverse
    (process-list list prepend)
    1 (procedure?)
    (lambda (reverser)
      (suite reverse-internal ()
        (test empty-list
          (reverser (list))
          1 (list?)
          empty-list?)
        (test singleton
          (reverser (list 'aglc))
          1 (list?)
          (match? list=? (list 'aglc)))
        (test doubleton
          (reverser (list 'eae 'noceetc))
          1 (list?)
          (match? list=? (list 'noceetc 'eae)))
        (test longer-list
          (reverser (list 'ecrir 'wens 'pntedrb 'cpidfkd 'oss 'ter
                          'weaggnc))
          1 (list?)
          (match? list=? (list 'weaggnc 'ter 'oss 'cpidfkd 'pntedrb 'wens
                               'ecrir))))))
  (test nanc
    (process-list (create 0) nanc)
    1 (procedure?)
    (lambda (nancer)
      (suite nanc-internal ()
        (test empty-list
          (nancer (list))
          1 (exact-integer?)
          zero?)
        (test singleton
          (nancer (list 567))
          1 (exact-integer?)
          (match? = (* 567 (add1 0))))
        (test longer-list
          (nancer (list 805 192 873 243 892))
          1 (exact-integer?)
          (match? = (* 892
                       (add1
                        (* 243
                           (add1
                            (* 873
                               (add1 (* 192
                                        (add1 (* 805
                                                 (add1 0)))))))))))))))
  (test multiargument
    (process-list (create 950 590 198)
                  (lambda (alpha beta gamma delta)
                    (values (+ alpha beta)
                            (+ beta gamma)
                            (+ gamma delta))))
    1 (procedure?)
    (lambda (processor)
      (suite multiargument-internal ()
        (test empty-list
          (processor (list))
          3 (exact-integer? exact-integer? exact-integer?)
          (match? = 950 590 198))
        (test longer-list
          (processor (list 646 74 353))
          3 (exact-integer? exact-integer? exact-integer?)
          (match? = 2023 4806 5464))))))

(suite postpend ()
  (test empty-list
    (postpend 'stra (list))
    1 (list?)
    (match? list=? (list 'stra)))
  (test longer-list
    (postpend 'ttc (list 'miky 'ichh 'lydrah 'iohi 'etom))
    1 (list?)
    (match? list=? (list 'miky 'ichh 'lydrah 'iohi 'etom 'ttc))))

(suite slow-reverse ()
  (test empty-list
    (slow-reverse (list))
    1 (list?)
    empty-list?)
  (test singleton
    (slow-reverse (list 'hbedmdn))
    1 (list?)
    (match? list=? (list 'hbedmdn)))
  (test doubleton
    (slow-reverse (list 'aenn 'eye))
    1 (list?)
    (match? list=? (list 'eye 'aenn)))
  (test longer-list
    (slow-reverse (list 'saeiae 'oft 'iesr 'nte 'paea))
    1 (list?)
    (match? list=? (list 'paea 'nte 'iesr 'oft 'saeiae))))

(suite alternative-reverse ()
  (test empty-list
    (alternative-reverse (list))
    1 (list?)
    empty-list?)
  (test singleton
    (alternative-reverse (list 'igtiao))
    1 (list?)
    (match? list=? (list 'igtiao)))
  (test doubleton
    (alternative-reverse (list 'deocu 'ddii))
    1 (list?)
    (match? list=? (list 'ddii 'deocu)))
  (test longer-list
    (alternative-reverse (list 'nny 'toeenc 'hriaoe 'dfdst 'nisl 'duignf
                               'nkhuin 'yul))
    1 (list?)
    (match? list=? (list 'yul 'nkhuin 'duignf 'nisl 'dfdst 'hriaoe 'toeenc
                         'nny))))

(suite unfold-list ()
  (test stops-immediately
    (unfold-list values? list values)
    1 (procedure?)
    (lambda (stopper)
      (suite stops-immediately-internal ()
        (test minimal
          (stopper)
          1 (list?)
          empty-list?))))
  (test typical-case
    (unfold-list odd? square halve)
    1 (procedure?)
    (lambda (unfolder)
      (suite typical-case-internal ()
        (test does-it-work?
          (unfolder 40)
          1 ((list-of exact-integer?))
          (match? (list-of= =) (list 1600 400 100))))))
  (test first-ten-squares
    (unfold-list (sect = <> 10) square add1)
    1 (procedure?)
    (lambda (unfolder)
      (suite first-ten-squares-internal ()
        (test does-it-work?
          (unfolder 0)
          1 ((list-of exact-integer?))
          (match? (list-of= =) (list 0 1 4 9 16 25 36 49 64 81))))))
  (test Collatz
    (unfold-list (sect = <> 1)
                 identity
                 (^if odd? (pipe (sect * <> 3) add1) halve))
    1 (procedure?)
    (lambda (unfolder)
      (suite Collatz-internal ()
        (test short
          (unfolder 19)
          1 ((list-of exact-integer?))
          (match? (list-of= =)
                  (list 19 58 29 88 44 22 11 34 17 52 26 13 40 20 10 5 16 8
                        4 2)))
        (test long
          (unfolder 27)
          1 ((list-of exact-integer?))
          (match? (list-of= =)
                  (list 27 82 41 124 62 31 94 47 142 71 214 107 322 161 484
                        242 121 364 182 91 274 137 412 206 103 310 155 466
                        233 700 350 175 526 263 790 395 1186 593 1780 890
                        445 1336 668 334 167 502 251 754 377 1132 566 283
                        850 425 1276 638 319 958 479 1438 719 2158 1079
                        3238 1619 4858 2429 7288 3644 1822 911 2734 1367
                        4102 2051 6154 3077 9232 4616 2308 1154 577 1732
                        866 433 1300 650 325 976 488 244 122 61 184 92 46
                        23 70 35 106 53 160 80 40 20 10 5 16 8 4 2))))))
  (test many-arguments
    (unfold-list (lambda (a b c) (< a 1))
                 *
                 (lambda (a b c)
                   (if (< 1 c)
                       (values a b (sub1 c))
                       (if (< 1 b)
                           (values a (sub1 b) 5)
                           (values (sub1 a) 4 5)))))
    1 (procedure?)
    (lambda (unfolder)
      (suite many-arguments-internal ()
        (test multiples
          (unfolder 3 4 5)
          1 ((list-of exact-integer?))
          (match? (list-of= =)
                  (list 60 48 36 24 12 45 36 27 18 9 30 24 18 12 6 15 12 9
                        6 3 40 32 24 16 8 30 24 18 12 6 20 16 12 8 4 10 8 6
                        4 2 20 16 12 8 4 15 12 9 6 3 10 8 6 4 2 5 4 3 2
                        1)))))))

(suite pairs-with-given-sum ()
  (test zero
    (pairs-with-given-sum 0)
    1 ((list-of (pair-of exact-integer? exact-integer?)))
    (match? (list-of= (pair-of= = =)) (list (cons 0 0))))
  (test larger-number
    (pairs-with-given-sum 14)
    1 ((list-of (pair-of exact-integer? exact-integer?)))
    (match? (list-of= (pair-of= = =))
            (list (cons 0 14) (cons 1 13) (cons 2 12) (cons 3 11)
                  (cons 4 10) (cons 5 9) (cons 6 8) (cons 7 7)
                  (cons 8 6) (cons 9 5) (cons 10 4) (cons 11 3)
                  (cons 12 2) (cons 13 1) (cons 14 0)))))

(suite adjacent-pairs ()
  (test singleton
    (adjacent-pairs (list 'dwieya))
    1 ((list-of (pair-of symbol? symbol?)))
    empty-list?)
  (test long-list
    (adjacent-pairs (list 'dos 'ehiirgw 'nusa 'ioonde 'eciel 'ahens))
    1 ((list-of (pair-of symbol? symbol?)))
    (match? (list-of= (pair-of= symbol=? symbol=?))
            (list (cons 'dos 'ehiirgw)
                  (cons 'ehiirgw 'nusa)
                  (cons 'nusa 'ioonde)
                  (cons 'ioonde 'eciel)
                  (cons 'eciel 'ahens)))))

(suite square-listers ()
  (test pn-zero
    (square-lister-with-ply-natural 0)
    1 ((list-of exact-integer?))
    empty-list?)
  (test pn-one
    (square-lister-with-ply-natural 1)
    1 ((list-of exact-integer?))
    (match? (list-of= =) (list 1)))
  (test pn-larger
    (square-lister-with-ply-natural 6)
    1 ((list-of exact-integer?))
    (match? (list-of= =) (list 36 25 16 9 4 1)))
  (test ul-zero
    (square-lister-with-unfold-list 0)
    1 ((list-of exact-integer?))
    empty-list?)
  (test ul-one
    (square-lister-with-unfold-list 1)
    1 ((list-of exact-integer?))
    (match? (list-of= =) (list 1)))
  (test ul-larger
    (square-lister-with-unfold-list 6)
    1 ((list-of exact-integer?))
    (match? (list-of= =) (list 36 25 16 9 4 1)))
  (test n-zero
    (neutral-square-lister 0)
    1 ((list-of exact-integer?))
    empty-list?)
  (test n-one
    (neutral-square-lister 1)
    1 ((list-of exact-integer?))
    (match? (list-of= =) (list 1)))
  (test n-larger
    (neutral-square-lister 6)
    1 ((list-of exact-integer?))
    (match? (list-of= =) (list 36 25 16 9 4 1))))

(suite conditionally-unfold-list ()
  (test stops-immediately
    (conditionally-unfold-list values? values? list values)
    1 (procedure?)
    (lambda (stopper)
      (suite stops-immediately-internal ()
        (test stopper
          (stopper)
          1 (list?)
          empty-list?))))
  (test pass-all
    (conditionally-unfold-list odd? values? square halve)
    1 (procedure?)
    (lambda (unfolder)
      (suite pass-all-internal ()
        (test does-it-work?
          (unfolder 40)
          1 ((list-of exact-integer?))
          (match? (list-of= =) (list 1600 400 100))))))
  (test fail-all
    (conditionally-unfold-list zero? (constant #f) identity sub1)
    1 (procedure?)
    (lambda (unfolder)
      (suite fail-all-internal ()
        (test does-it-exclude?
          (unfolder 29)
          1 ((list-of exact-integer?))
          empty-list?))))
  (test squares-of-small-odds  ;; example from the text
    ((conditionally-unfold-list zero? odd? square sub1) 10)
    1 ((list-of exact-integer?))
    (match? (list-of= =) (list 81 49 25 9 1)))
  (test many-arguments
    ((conditionally-unfold-list
       (lambda (a b c) (< a 1))
       (lambda (a b c)
         (and (odd? a) (odd? b) (odd? c))) 
       *
       (lambda (a b c)
         (if (< 1 c)
             (values a b (sub1 c))
             (if (< 1 b)
                 (values a (sub1 b) 5)
                 (values (sub1 a) 4 5)))))
     3 4 5)
    1 ((list-of exact-integer?))
    (match? (list-of= =) (list 45 27 9 15 9 3 15 9 3 5 3 1))))

;; Test the procedures from the "List algorithms" section.

(suite append ()
  (test no-arguments
    (alternative-append)
    1 (list?)
    empty-list?)
  (test one-argument
    (alternative-append (list 'pte 'iiuit 'emnrl 'thoos 'nbe))
    1 ((list-of symbol?))
    (match? (list-of= symbol=?) (list 'pte 'iiuit 'emnrl 'thoos 'nbe)))
  (test variety
    (alternative-append (list 'viot 'ytche 'aeorao)
                        (list)
                        (list 'lspcai 'eswa)
                        (list)
                        (list)
                        (list 'hutmtnh)
                        (list 'auwrnht 'hoo)
                        (list)
                        (list 'sohsa 'dln 'uep 'oedsu 'eyrnuao)
                        (list))
  1 ((list-of symbol?))
  (match? (list-of= symbol=?)
          (list 'viot 'ytche 'aeorao 'lspcai 'eswa 'hutmtnh 'auwrnht 'hoo
                'sohsa 'dln 'uep 'oedsu 'eyrnuao))))

(suite ^and ()
  (test no-arguments
    (^and)
    1 (procedure?)
    (lambda (conjoined)
      (suite no-arguments-internal ()
        (test run-it
          (conjoined 'hoemwwe 'wier 'fhrf 'effsdi 'ebwtd)
          1 (boolean?)
          true?))))
  (test one-argument
    (^and even?)
    1 (procedure?)
    (lambda (conjoined)
      (suite no-arguments-internal ()
        (test even
          (conjoined 324)
          1 (boolean?)
          true?)
        (test odd
          (conjoined 717)
          1 (boolean?)
          not))))
  (test multiple-arguments
    (^and > divisible-by? (pipe + even?))
    1 (procedure?)
    (lambda (conjoined)
      (suite mulitple-arguments-internal ()
        (test not-greater
          (conjoined 735 735)
          1 (boolean?)
          not)
        (test not-evenly-divisible
          (conjoined 792 130)
          1 (boolean?)
          not)
        (test sum-not-even
          (conjoined 4 1)
          1 (boolean?)
          not)
        (test ok
          (conjoined 385 5)
          1 (boolean?)
          true?)))))

(suite extend-to-variable-arity ()
  (test append
    (extend-to-variable-arity (list) catenate)
    1 (procedure?)
    (lambda (appender)
      (suite append-internal ()
        (test no-arguments
          (appender)
          1 (list?)
          empty-list?)
        (test one-argument
          (appender (list 'eeeue 'hah 'noetsns))
          1 ((list-of symbol?))
          (match? (list-of= symbol=?) (list 'eeeue 'hah 'noetsns)))
        (test variety
          (appender (list)
                    (list 'ereero 'uoslr)
                    (list)
                    (list)
                    (list 'ylrose 'ele 'trbw 'esaysa)
                    (list 'bngop 'aypbo 'okur)
                    (list)
                    (list 'ngeahdr))
          1 ((list-of symbol?))
          (match? (list-of= symbol=?)
                  (list 'ereero 'uoslr 'ylrose 'ele 'trbw 'esaysa 'bngop
                        'aypbo 'okur 'ngeahdr))))))
  (test nanc
    (extend-to-variable-arity 0 (pipe (cross double add1) +))
    1 (procedure?)
    (lambda (nancer)
      (suite nanc-internal ()
        (test no-arguments
          (nancer)
          1 (exact-integer?)
          zero?)
        (test one-argument
          (nancer 682)
          1 (exact-integer?)
          (match? = 1365))
        (test multi-argument
          (nancer 833 561 755 123)
          1 (exact-integer?)
          (match? = (+ (double 833)
                       (add1 (+ (double 561)
                                (add1 (+ (double 755)
                                         (add1 (+ (double 123)
                                                  (add1 0)))))))))))))
  (test extended-expt
    ((extend-to-variable-arity 1 expt) 2 3 3)  ;; example from the text
    1 (exact-integer?)
    (match? = 134217728))) ;;  = 2^27 -- (expt 2 (expt 3 3)),
                           ;;            not (expt (expt 2 3) 3)

(suite ^or ()
  (test no-arguments
    (^or)
    1 (procedure?)
    (lambda (disjoined)
      (suite no-arguments-internal ()
        (test run-it
          (disjoined 'afnhr 'eeiaaic 'antmi 'mtth 'oaateb 'hmvood)
          1 (boolean?)
          not))))
  (test one-argument
    (^or even?)
    1 (procedure?)
    (lambda (disjoined)
      (suite one-argument-internal ()
        (test even
          (disjoined 694)
          1 (boolean?)
          true?)
        (test odd
          (disjoined 293)
          1 (boolean?)
          not))))
  (test several-arguments
    (^or < divisible-by? (pipe + even?))
    1 (procedure?)
    (lambda (disjoined)
      (suite several-arguments-internal ()
        (test less-only
          (disjoined 157 866)
          1 (boolean?)
          true?)
        (test evenly-divisible-only
          (disjoined 462 231)
          1 (boolean?)
          true?)
        (test even-sum-only
          (disjoined 989 671)
          1 (boolean?)
          true?)
        (test fail-all
          (disjoined 420 303)
          1 (boolean?)
          not)))))

(suite run ((sum-and-difference (dispatch + -)))
  (test no-procedures
    (run)
    1 (procedure?)
    (lambda (result)
      (suite no-procedures-internal ()
        (test no-arguments
          (result)
          0 ())
        (test some-arguments
          (result 'dhhee 'eov 'nesrdar 'tsblae)
          4 (symbol? symbol? symbol? symbol?)
          (match? symbol=? 'dhhee 'eov 'nesrdar 'tsblae)))))
  (test one-procedure
    (run sum-and-difference)
    1 (procedure?)
    (lambda (result)
      (suite one-procedure-internal ()
        (test sum-and-difference
          (result 611 181)
          2 (exact-integer? exact-integer?)
          (match? = 792 430)))))
  (test several-procedures
    (run halve square even? not)
    1 (procedure?)
    (lambda (result)
      (suite several-procedures-internal ()
        (test odd-square-half
          (result 842)
          1 (boolean?)
          true?)
        (test even-square-half
          (result 353)
          1 (boolean?)
          not)))))

(suite extend-to-positive-arity ()
  (test min
    (extend-to-positive-arity lesser)
    1 (procedure?)
    (lambda (min)
      (suite min-internal ()
        (test one-argument
          (min 693)
          1 (exact-integer?)
          (match? = 693))
        (test many-arguments
          (min 889 927 87 598 457)
          1 (exact-integer?)
          (match? = 87)))))
  (test nanc
    (extend-to-positive-arity (pipe (cross double add1) +))
    1 (procedure?)
    (lambda (nancer)
      (suite nanc-internal ()
        (test one-argument
          (nancer 323)
          1 (exact-integer?)
          (match? = 323))
        (test multiple-arguments
          (nancer 1 840 302 527)
          1 (exact-integer?)
          (match? = (+ (double 1)
                       (add1 (+ (double 840)
                                (add1 (+ (double 302)
                                         (add1 527))))))))))))

(suite extend-like-subtraction ()
  (test neo-division
    (extend-like-subtraction (sect / 1 <>) /)
    1 (procedure?)
    (lambda (divide)
      (suite neo-division-internal ()
        (test one-argument
          (divide 82)
          1 (exact-rational?)
          (match? = 1/82))
        (test two-arguments
          (divide 863 324)
          1 (exact-rational?)
          (match? = 863/324))
        (test multiple-arguments
          (divide 634 118 501 915 25)
          1 (exact-rational?)
          (match? = (* 634/118 1/501 1/915 1/25))))))
  (test nanc
    (extend-like-subtraction square (pipe (cross double add1) +))
    1 (procedure?)
    (lambda (nancer)
      (suite nanc-internal ()
        (test one-argument
          (nancer 888)
          1 (exact-integer?)
          (match? = (square 888)))
        (test two-arguments
          (nancer 969 611)
          1 (exact-integer?)
          (match? = (+ (double 969) (add1 611))))
        (test multiple-arguments
          (nancer 832 284 592 276)
          1 (exact-integer?)
          (match? = (+ (double (+ (double (+ (double 832) (add1 284)))
                                  (add1 592)))
                       (add1 276))))))))                            

(suite extract ()
  (test empty-list
    (extract values? (list))
    2 ((^or box? boolean?) list?)
    (lambda (alpha beta)
      (and (not alpha) (empty-list? beta))))
  (test middle-of-larger-list
    (extract odd? (list 458 363 522))
    2 ((^vel (box-of number?) boolean?) (list-of number?))
    (lambda (alpha beta)
      (and (box? alpha)
           (= (debox alpha) 363)
           ((list-of= =) beta (list 458 522)))))
  (test multiple-matches
    (extract odd? (list 81 990 81 715))
    2 ((^vel (box-of number?) boolean?) (list-of number?))
    (lambda (alpha beta)
      (and (box? alpha)
           (= (debox alpha) 81)
           ((list-of= =) beta (list 990 81 715)))))
  (test no-matches
    (extract (constant #f) (list 'itrhaeh 'onohn 'rihnre))
    2 ((^vel (box-of symbol?) boolean?) (list-of symbol?))
    (lambda (alpha beta)
      (and (not alpha)
           ((list-of= symbol=?) beta (list 'itrhaeh 'onohn 'rihnre))))))

(suite map ()
  (test simple-empty
    (alternative-map square (list))
    1 ((list-of exact-integer?))
    empty-list?)
  (test simple-singleton
    (alternative-map square (list 893))
    1 ((list-of exact-integer?))
    (match? (list-of= =) (list (square 893))))
  (test simple-long-list
    (alternative-map square (list 546 597 42 745 740))
    1 ((list-of exact-integer?))
    (match? (list-of= =)
            (list (square 546) (square 597) (square 42) (square 745)
                  (square 740))))
  (test many-empty-lists
    (alternative-map * (list) (list) (list))
    1 ((list-of exact-integer?))
    empty-list?)
  (test many-singletons
    (alternative-map * (list 565) (list 532) (list 81) (list 488))
    1 ((list-of exact-integer?))
    (match? (list-of= =) (list (* 565 532 81 488))))
  (test many-long-lists
    (alternative-map * (list 125 986 984)
                       (list 855 979 295)
                       (list 186 153 138)
                       (list 295 290 976)
                       (list 876 269 206))
    1 ((list-of exact-integer?))
    (match? (list-of= =) (list (* 125 855 186 295 876)
                               (* 986 979 153 290 269)
                               (* 984 295 138 976 206)))))

(suite collect-map ()
  (test empty-list
    (collect-map identity (list))
    1 (list?)
    empty-list?)
  (test black-hole
    (collect-map black-hole
                 (list 'omjd 'eitbt 'ensdurc 'ratemga 'plsetih))
    1 (list?)
    empty-list?)
  (test like-map
    (collect-map square (list 133 681 64 96))
    1 ((list-of exact-integer?))
    (match? (list-of= =)
            (list (square 133) (square 681) (square 64) (square 96))))
  (test triplicator
    (collect-map (lambda (something)
                   (values something something something))
                 (list 'bnpyo 'ykl 'hnueuos))
    1 ((list-of symbol?))
    (match? (list-of= symbol=?)
            (list 'bnpyo 'bnpyo 'bnpyo 'ykl 'ykl 'ykl 'hnueuos 'hnueuos
                  'hnueuos)))
  (test disperse-reverse
    (collect-map (pipe (sect cons 'oel <>) delist)
                 (list (list 'lpadn 'oafs)
                       (list 'gsrvoge 'odwr 'nsi 'leoahec 'nta)
                       (list)
                       (list 'rigonsr 'lcncgye 'oil)))
    1 ((list-of symbol?))
    (match? (list-of= symbol=?)
            (list 'oel 'lpadn 'oafs 'oel 'gsrvoge 'odwr 'nsi 'leoahec 'nta
                  'oel 'oel 'rigonsr 'lcncgye 'oil)))
  (test higher-arity
    (collect-map (dispatch + - *) (list 249 63 307 180)
                                  (list 707 25 903 315))
    1 ((list-of exact-integer?))
    (match? (list-of= =)
            (list (+ 249 707) (- 249 707) (* 249 707)
                  (+ 63 25) (- 63 25) (* 63 25)
                  (+ 307 903) (- 307 903) (* 307 903)
                  (+ 180 315) (- 180 315) (* 180 315))))) 

(suite dispatch-return-all ()
  (test no-arguments
    (dispatch-return-all)
    1 (procedure?)
    (lambda (ebony-aperture)
      (suite no-arguments-internal ()
        (test really-nothing
          (ebony-aperture)
          0 ())
        (test something-to-vanish
          (ebony-aperture 'sdopfm 'nnehka 'efegiha 'vscrrqd)
          0 ()))))
  (test no-results
    (dispatch-return-all (constant) black-hole)
    1 (procedure?)
    (lambda (vanisher)
      (suite no-results-internal ()
        (test try-it
          (vanisher 'omtr 'aonrhee 'deiah)
          0 ()))))
  (test multiple-results
    (dispatch-return-all decons (dispatch cdr car))
    1 (procedure?)
    (lambda (dispatcher)
      (suite multiple-results-internal ()
        (test try-it
          (dispatcher (cons 'toioatr 'lueaoen))
          4 (symbol? symbol? symbol? symbol?)
          (match? symbol=? 'toioatr 'lueaoen 'lueaoen 'toioatr)))))
  (test varying-number-of-results
    (dispatch-return-all decons car black-hole
                         (dispatch car cdr car car cdr))
    1 (procedure?)
    (lambda (dispatcher)
      (suite varying-number-of-results-internal ()
        (test try-it
          (dispatcher (cons 'eeteht 'ttocree))
          8 (symbol? symbol? symbol? symbol? symbol? symbol? symbol?
             symbol?)
          (match? symbol=? 'eeteht 'ttocree 'eeteht 'eeteht 'ttocree
                           'eeteht 'eeteht 'ttocree))))))

(suite cross-return-all ()
  (test no-arguments
    (cross-return-all)
    1 (procedure?)
    (lambda (crosser)
      (suite no-arguments-internal ()
        (test do-nothing
          (crosser)
          0 ()))))
  (test several-arguments
    (cross-return-all decons (dispatch square halve))
    1 (procedure?)
    (lambda (crosser)
      (suite several-arguments-internal ()
        (test run-it
          (crosser (cons 'hul 'rnie) 566)
          4 (symbol? symbol? exact-integer? exact-integer?)
          (lambda (alpha beta gamma delta)
            (and (symbol=? alpha 'hul)
                 (symbol=? beta 'rnie)
                 (= gamma (square 566))
                 (= delta (halve 566))))))))
  (test variable-number-of-results
    (cross-return-all delist
                      (pipe cdr delist)
                      (dispatch double square halve)
                      black-hole
                      deprepend)
    1 (procedure?)
    (lambda (crosser)
      (suite variable-number-of-results ()
        (test run-vnr
          (crosser (list 'cpdd 'eanedep 'tetflur)
                   (list 'obo 'tdli 'lhrra 'lbgarvr 'iomm)
                   718
                   'lrtshl
                   (list 'heitgr 'fndo 'aehcai))
          12 (symbol? symbol? symbol? symbol? symbol? symbol? symbol?
              exact-integer? exact-integer? exact-integer? symbol?
              (list-of symbol?))
          (lambda (alpha beta gamma delta epsilon zeta eta theta iota kappa
                   mu nu)
            (and (symbol=? alpha 'cpdd)
                 (symbol=? beta 'eanedep)
                 (symbol=? gamma 'tetflur)
                 (symbol=? delta 'tdli)
                 (symbol=? epsilon 'lhrra)
                 (symbol=? zeta 'lbgarvr)
                 (symbol=? eta 'iomm)
                 (= theta (double 718))
                 (= iota (square 718))
                 (= kappa (halve 718))
                 (symbol=? mu 'heitgr)
                 ((list-of= symbol=?) nu (list 'fndo 'aehcai)))))))))

(suite filter ()
  (test empty-list
    (filter (constant #t) (list))
    1 (list?)
    empty-list?)
  (test singleton-list-satisfying
    (filter even? (list 112))
    1 ((list-of exact-integer?))
    (match? (list-of= =) (list 112)))
  (test singleton-list-unsatisfying
    (filter even? (list 423))
    1 ((list-of exact-integer?))
    empty-list?)
  (test all-elements-removed
    (filter (constant #f)
                        (list 'ewhlrum 'toetl 'trrhpa 'eue 'oern))
    1 ((list-of symbol?))
    empty-list?)
  (test some-elements-removed
    (filter even? (list 399 542 442 789 897 126 486 908))
    1 ((list-of exact-integer?))
    (match? (list-of= =) (list 542 442 126 486 908)))
  (test no-elements-removed
    (filter values? (list 'witel 'ried 'asla 'toiev 'erfeh))
    1 ((list-of symbol?))
    (match? (list-of= symbol=?) (list 'witel 'ried 'asla 'toiev 'erfeh))))

(suite remp ()
  (test empty-list
    (remp (constant #f) (list))
    1 (list?)
    empty-list?)
  (test singleton-list-satisfying
    (remp even? (list 260))
    1 ((list-of exact-integer?))
    empty-list?)
  (test singleton-list-unsatisfying
    (remp even? (list 415))
    1 ((list-of exact-integer?))
    (match? (list-of= =) (list 415)))
  (test all-elements-removed
    (remp values? (list 'dmw 'otimt 'gsht 'hhchs 'oitr 'rad))
    1 ((list-of symbol?))
    empty-list?)
  (test some-elements-removed
    (remp even? (list 290 617 837 29 214 903 996 257)) 
    1 ((list-of exact-integer?))
    (match? (list-of= =) (list 617 837 29 903 257)))
  (test no-elements-removed
    (remp (constant #f) (list 'cdsst 'anv 'rte 'rdhtia 'doe))
    1 ((list-of symbol?))
    (match? (list-of= symbol=?) (list 'cdsst 'anv 'rte 'rdhtia 'doe))))

(suite partition ()
  (test empty-list
    (partition values? (list))
    2 (list? list?)
    (lambda (alpha beta)
      (and (empty-list? alpha)
           (empty-list? beta))))
  (test singleton-list-satisfying
    (partition even? (list 434))
    2 ((list-of exact-integer?) (list-of exact-integer?))
    (match? (list-of= =) (list 434) (list)))
  (test singleton-list-unsatisfying
    (partition even? (list 639))
    2 ((list-of exact-integer?) (list-of exact-integer?))
    (match? (list-of= =) (list) (list 639)))
  (test all-elements-out
    (partition (constant #f)
                           (list 'htote 'acor 'vrd 'rten 'sstosi 'stdseil))
    2 ((list-of symbol?) (list-of symbol?))
    (match? (list-of= symbol=?)
            (list)
            (list 'htote 'acor 'vrd 'rten 'sstosi 'stdseil)))
  (test distributed
    (partition even? (list 424 308 829 946 572 656 742 71))
    2 ((list-of exact-integer?) (list-of exact-integer?))
    (match? (list-of= =)
            (list 424 308 946 572 656 742)
            (list 829 71)))
  (test all-elements-in
    (partition values?
                           (list 'ldliieo 'imonii 'itt 'onrafd 'eng))
    2 ((list-of symbol?) (list-of symbol?))
    (match? (list-of= symbol=?)
            (list 'ldliieo 'imonii 'itt 'onrafd 'eng)
            (list))))

(suite drop ()
  (test all-of-list
    (drop (list 'npcbkie 'nhh 'oethoer 'cchooa 'eaoif) 5)
    1 ((list-of symbol?))
    empty-list?)
  (test some-of-list
    (drop (list 'las 'aaleri 'nmo 'fnoe 'oasn 'odh 'amcc) 4) 
    1 ((list-of symbol?))
    (match? (list-of= symbol=?) (list 'oasn 'odh 'amcc)))
  (test none-of-list
    (drop (list 'lrrw 'frbfun 'nrgte) 0)
    1 ((list-of symbol?))
    (match? (list-of= symbol=?) (list 'lrrw 'frbfun 'nrgte)))
  (test empty-list
    (drop (list) 0)
    1 (list?)
    empty-list?))

(suite take ()
  (test all-of-list
    (take (list 'ree 'anoni 'trrri 'emht 'iwaom) 5)
    1 ((list-of symbol?))
    (match? (list-of= symbol=?) (list 'ree 'anoni 'trrri 'emht 'iwaom)))
  (test some-of-list
    (take (list 'rhirfht 'tre 'dgs 'gisn 'urshlom 'hun 'box) 4)
    1 ((list-of symbol?))
    (match? (list-of= symbol=?) (list 'rhirfht 'tre 'dgs 'gisn)))
  (test none-of-list
    (take (list 'crleat 'fmthbl 'tilei) 0)
    1 ((list-of symbol?))
    empty-list?)
  (test empty-list
    (take (list) 0)
    1 (list?)
    empty-list?))

(suite sublist ()
  (test empty-list
    (sublist (list) 0 0)
    1 (list?)
    empty-list?)
  (test extract-all
    (sublist (list 'teaas 'ahawl 'asvap 'pclgrfl 'kwosde 'dkhlate) 0 6)
    1 ((list-of symbol?))
    (match? (list-of= symbol=?)
            (list 'teaas 'ahawl 'asvap 'pclgrfl 'kwosde 'dkhlate)))
  (test extract-at-front
    (sublist (list 'eewhbm 'nsr 'tais 'ehno 'osysnra 'reo 'veire 'esceh
                   'tora)
             0
             4)
    1 ((list-of symbol?))
    (match? (list-of= symbol=?) (list 'eewhbm 'nsr 'tais 'ehno)))
  (test extract-at-rear
    (sublist (list 'ryer 'teef 'wfehme 'utlrhr 'ldiuw 'thh 'rifihn 
                   'needfnl 'dwaherh 'toothba)
             3
             10)
    1 ((list-of symbol?))
    (match? (list-of= symbol=?)
            (list 'utlrhr 'ldiuw 'thh 'rifihn 'needfnl 'dwaherh 'toothba)))
  (test extract-from-middle
    (sublist (list 'rtsgdy 'nhh 'eeh 'iiasfel 'tsda 'njae 'hsimyng
                   'ere 'oigudu 'pset 'utuoau)
             5
             8)
    1 ((list-of symbol?))
    (match? (list-of= symbol=?) (list 'njae 'hsimyng 'ere)))
  (test extract-empty-sublist-at-front
    (sublist (list 'wues 'hne 'netieng) 0 0)
    1 ((list-of symbol?))
    empty-list?)
  (test extract-empty-sublist-at-rear
    (sublist (list 'snm 'eviha 'oayp 'eow 'eyct 'sliyy 'hwdw 'hlnohy 'enp)
             9
             9)
    1 ((list-of symbol?))
    empty-list?)
  (test extract-empty-sublist-from-middle
    (sublist (list 'edh 'res 'isec 'ery 'eheuteo 'iyroee 'oidns
                   'rrvetk 'oeelbhi 'eehun 'egeso)
             6
             6)
    1 ((list-of symbol?))
    empty-list?))

(suite all-even? ()
  (test empty-list
    (all-even? (list))
    1 (boolean?)
    true?)
  (test yes
    (all-even? (list 34 914 60 600 870))
    1 (boolean?)
    true?)
  (test none
    (all-even? (list 901 883 987))
    1 (boolean?)
    not)
  (test early-out
    (all-even? (list 533 382 622 564))
    1 (boolean?)
    not)
  (test late-out
    (all-even? (list 558 598 622 533))
    1 (boolean?)
    not)
  (test mixed
    (all-even? (list 872 849 214 671 405))
    1 (boolean?)
    not))

(suite for-all? ()
  (test null-list
    (for-all? even? null)
    1 (boolean?)
    true?)
  (test all-satisfying
    (for-all? even? (list 882 298 800 196))
    1 (boolean?)
    true?)
  (test some-satisfying
    (for-all? even? (list 279 586 984 212 841))
    1 (boolean?)
    not)
  (test none-satisfying
    (for-all? even? (list 575 939 87))
    1 (boolean?)
    not))

(suite exists? ()
  (test null-list
    (exists? odd? null)
    1 (boolean?)
    not)
  (test all-satisfying
    (exists? odd? (list 575 939 87))
    1 (boolean?)
    true?)
  (test some-satisfying
    (exists? odd? (list 586 984 279 212 841))
    1 (boolean?)
    true?)
  (test none-satisfying
    (exists? odd? (list 882 298 800 196))
    1 (boolean?)
    not))

(suite every ()
  (test just-the-one
    (every even?)
    1 (procedure?)
    (lambda (all-even?)
      (suite every-internal ()
        (test no-arguments
          (all-even?)
          1 (boolean?)
          true?)
        (test all-even
          (all-even? 6 2 0 4 8)  ;; example from the text
          1 (boolean?)
          true?)
        (test some-even
          (all-even? 233 765 159 324 185)
          1 (boolean?)
          not)
        (test none-even
          (all-even? 897 831 875 465)
          1 (boolean?)
          not)))))

(suite at-least-one ()
  (test just-another
    (at-least-one even?)
    1 (procedure?)
    (lambda (any-even?)
      (suite any-internal ()
        (test no-arguments
          (any-even?)
          1 (boolean?)
          not)
        (test all-even
          (any-even? 250 854 710 228)
          1 (boolean?)
          true?)
        (test some-even
          (any-even? 887 607 717 236 764)
          1 (boolean?)
          true?)
        (test none-even
          (any-even? 979 993 675)
          1 (boolean?)
          not)))))

(suite transpose ()
  (test singleton-list
    (transpose (list (list 'nirana 'aoeot 'ahht)))
    1 ((list-of (list-of symbol?)))
    (match? (list-of= (list-of= symbol=?)) (list (list 'nirana)
                                                 (list 'aoeot)
                                                 (list 'ahht))))
  (test two-lists
    (transpose (list (list 'etedox 'aeloaa 'eteefr 'tedln 'ennah)
                     (list 'itc 'tae 'tdr 'mcdera 'lulr)))
    1 ((list-of (list-of symbol?)))
    (match? (list-of= (list-of= symbol=?))
            (list (list 'etedox 'itc)
                  (list 'aeloaa 'tae)
                  (list 'eteefr 'tdr)
                  (list 'tedln 'mcdera)
                  (list 'ennah 'lulr))))
  (test list-of-singletons
    (transpose (list (list 'eha)
                     (list 'ddpdist)
                     (list 'twguc)
                     (list 'sos)))
    1 ((list-of (list-of symbol?)))
    (match? (list-of= (list-of= symbol=?))
            (list (list 'eha 'ddpdist 'twguc 'sos))))
  (test several-lists
    (transpose (list (list 'amoaoo 'oheme 'lee 'lsiot 'eac)
                     (list 'fgmhse 'ashgem 'uoh 'seahl 'elt)
                     (list 'irci 'yerw 'esah 'thhntte 'eotrhad)
                     (list 'ooea 'hwroe 'nrtchoo 'aahnitg 'hhttos)
                     (list 'oesonpp 'enfedn 'egtnio 'uhmi 'iye)
                     (list 'orod 'ltkiid 'iaeenen 'egnke 'toa)
                     (list 'lsen 'onljwa 'bocmi 'cahhyva 'ehdt)))
    1 ((list-of (list-of symbol?)))
    (match? (list-of= (list-of= symbol=?))
            (list (list 'amoaoo 'fgmhse 'irci 'ooea 'oesonpp 'orod 'lsen)
                  (list 'oheme 'ashgem 'yerw 'hwroe 'enfedn 'ltkiid
                        'onljwa)
                  (list 'lee 'uoh 'esah 'nrtchoo 'egtnio 'iaeenen 'bocmi)
                  (list 'lsiot 'seahl 'thhntte 'aahnitg 'uhmi 'egnke
                        'cahhyva)
                  (list 'eac 'elt 'eotrhad 'hhttos 'iye 'toa 'ehdt)))))

(suite zip ()
  (test one-list
    (zip (list 'rsmaaet 'ttoia 'lcyc))
    1 ((list-of (list-of symbol?)))
    (match? (list-of= (list-of= symbol=?))
            (list (list 'rsmaaet) (list 'ttoia) (list 'lcyc))))
  (test two-lists
    (zip (list 'lerh 'hmrsmpg 'neto 'vlducn 'ihdcska)
         (list 'dcln 'tloieu 'lnt 'edw 'tots))
    1 ((list-of (list-of symbol?)))
    (match? (list-of= (list-of= symbol=?))
            (list (list 'lerh 'dcln)
                  (list 'hmrsmpg 'tloieu)
                  (list 'neto 'lnt)
                  (list 'vlducn 'edw)
                  (list 'ihdcska 'tots))))
  (test singletons
    (zip (list 'twrto) (list 'tat) (list 'oua) (list 'wee))
    1 ((list-of (list-of symbol?)))
    (match? (list-of= (list-of= symbol=?))
            (list (list 'twrto 'tat 'oua 'wee))))
  (test several-lists
    (zip (list 'osgitr 'opdgr 'eaoeiid 'mvr 'paslnf)
         (list 'nero 'imldfe 'ieel 'ooer 'ilp)
         (list 'nod 'iiety 'atshue 'haitt 'fnt)
         (list 'loeoo 'elerii 'dhep 'reerhh 'stoehwl)
         (list 'aaaph 'ottdco 'ahlesan 'tla 'wan)
         (list 'ura 'sob 'itheocw 'oahwahe 'vefhdo) 
         (list 'lnahma 'seh 'hag 'rfr 'enpo))
    1 ((list-of (list-of symbol?)))
    (match? (list-of= (list-of= symbol=?))
            (list (list 'osgitr 'nero 'nod 'loeoo 'aaaph 'ura 'lnahma)
                  (list 'opdgr 'imldfe 'iiety 'elerii 'ottdco 'sob 'seh)
                  (list 'eaoeiid 'ieel 'atshue 'dhep 'ahlesan 'itheocw
                        'hag)
                  (list 'mvr 'ooer 'haitt 'reerhh 'tla 'oahwahe 'rfr)
                  (list 'paslnf 'ilp 'fnt 'stoehwl 'wan 'vefhdo 'enpo)))))

(suite unzip ()
  (test singleton-list
    (unzip (list (list 'dlfa 'isod 'oanamoo)))
    3 ((list-of symbol?) (list-of symbol?) (list-of symbol?))
    (match? (list-of= symbol=?) (list 'dlfa) (list 'isod) (list 'oanamoo)))
  (test two-lists
    (unzip (list (list 'bnto 'esuha 'ixai 'eansgs 'msnyol)
                 (list 'odnat 'dnto 'dnfo 'tas 'ionssh)))
    5 ((list-of symbol?) (list-of symbol?) (list-of symbol?)
       (list-of symbol?) (list-of symbol?))
    (match? (list-of= symbol=?)
            (list 'bnto 'odnat)
            (list 'esuha 'dnto)
            (list 'ixai 'dnfo)
            (list 'eansgs 'tas)
            (list 'msnyol 'ionssh)))
  (test singletons
    (unzip (list (list 'nlh) (list 'owehii) (list 'thea) (list 'teiu)))
    1 ((list-of symbol?))
    (match? (list-of= symbol=?) (list 'nlh 'owehii 'thea 'teiu)))
  (test several-lists
    (unzip (list (list 'giadf 'aahgl 'mdtg 'tahron 'aotcnee 'deti 'uewdi)
                 (list 'ceepav 'fgrue 'erm 'esc 'hsiuabn 'siowowh 'eiesao)
                 (list 'oytatoc 'iha 'foef 'ffi 'yery 'dataae 'bnt)
                 (list 'togt 'esaipi 'ualy 'othoh 'lts 'etet 'fnislal)
                 (list 'tinrore 'soniit 'mee 'edoer 'dli 'oesorda 'lrnaew)))
    7 ((list-of symbol?) (list-of symbol?) (list-of symbol?)
       (list-of symbol?) (list-of symbol?) (list-of symbol?)
       (list-of symbol?))
    (match? (list-of= symbol=?)
            (list 'giadf 'ceepav 'oytatoc 'togt 'tinrore)
            (list 'aahgl 'fgrue 'iha 'esaipi 'soniit)
            (list 'mdtg 'erm 'foef 'ualy 'mee)
            (list 'tahron 'esc 'ffi 'othoh 'edoer)
            (list 'aotcnee 'hsiuabn 'yery 'lts 'dli)
            (list 'deti 'siowowh 'dataae 'etet 'oesorda)
            (list 'uewdi 'eiesao 'bnt 'fnislal 'lrnaew))))

;;; Test the procedures from the "Sources" section.

(suite all-zeroes ()
  (test tap-it
    (all-zeroes)
    2 (natural-number? source?)
    (lambda (alpha beta)
      (zero? alpha))))              ;; There's no way to check beta.

(suite natural-number-source ()
  (test try-it
    ((pipe source->list >initial) natural-number-source 240)
    1 ((list-of natural-number?))
    (match? (list-of= =) ((unfold-list zero? (sect - 240 <>) sub1) 240))))

(suite tap ()
  (test try-it
    (tap all-zeroes)
    2 (natural-number? source?)
    (lambda (alpha beta)
      (zero? alpha))))             ;; There's no way to check beta.

(suite source-ref
    ((even-source ((rec (evener base)
                     (source (values base (evener (+ base 2)))))
                   0)))
  (test position-zero
    (source-ref even-source 0)
    1 (exact-integer?)
    zero?)
  (test later-position
    (source-ref even-source 291)
    1 (exact-integer?)
    (match? = (double 291))))

(suite source->list
    ((even-source ((rec (evener base)
                     (source (values base (evener (+ base 2)))))
                   0))
     (iota (lambda (n)
             ((unfold-list zero? (sect - n <>) sub1) n))))
  (test take-none
    (source->list even-source 0)
    2 ((list-of exact-integer?) source?)
    (lambda (alpha beta)
      (and (empty-list? alpha) (equal? beta even-source))))
  (test take-several
    (source->list even-source 335)
    2 ((list-of exact-integer?) source?)
    (lambda (alpha beta)
      (and ((list-of= =) alpha (map double (iota 335)))
           (= (source-ref beta 0) (double 335))
           (= (source-ref beta 1) (double 336))))))

(suite source-drop
    ((even-source ((rec (evener base)
                     (source (values base (evener (+ base 2)))))
                   0))
     (iota (lambda (n)
             ((unfold-list zero? (sect - n <>) sub1) n))))
  (test drop-none
    (source-drop even-source 0)
    1 (source?)
    (match? equal? even-source))
  (test drop-one
    (source-drop even-source 1)
    1 (source?)
    (run (sect source->list <> 5)
         >initial
         (match? (list-of= =) (list 2 4 6 8 10))))
  (test drop-many
    (source-drop even-source 586)
    1 (source?)
    (run (sect source->list <> 5)
         >initial
         (match? (list-of= =) (list 1172 1174 1176 1178 1180)))))

(suite constant-source ()
  (test try-it
    (constant-source 'meetfes)
    1 (source?)
    (run (sect source->list <> 6)
         >initial
         (match? (list-of= symbol=?)
                 (list 'meetfes 'meetfes 'meetfes 'meetfes 'meetfes
                       'meetfes)))))

(suite cyclic-source ()
  (test single-argument
    (cyclic-source 'sts)
    1 (source?)
    (run (sect source->list <> 7)
         >initial
         (match? (list-of= symbol=?)
                 (list 'sts 'sts 'sts 'sts 'sts 'sts 'sts))))
  (test multiple-arguments
    (cyclic-source 'naetne 'heret 'gatttti)
    1 (source?)
    (run (sect source->list <> 11)
         >initial
         (match? (list-of= symbol=?)
                 (list 'naetne 'heret 'gatttti 'naetne 'heret 'gatttti
                       'naetne 'heret 'gatttti 'naetne 'heret)))))

(suite interleave ()
  (test try-it
    (interleave (cyclic-source 'uetid 'snoieit 'gfs)
                (cyclic-source 'tusliv 'mslhnle))
    1 (source?)
    (run (sect source->list <> 15)
         >initial
         (match? (list-of= symbol=?)
                 (list 'uetid 'tusliv 'snoieit 'mslhnle 'gfs 'tusliv 'uetid
                       'mslhnle 'snoieit 'tusliv 'gfs 'mslhnle 'uetid
                       'tusliv 'snoieit)))))

(suite alternator ()
  (test try-it
    (alternator (cyclic-source 'ihveno 'eearaft 'alnogay 'eremb 'unfatn
                               'lslw 'rrud 'lae 'doueeti))
    1 (source?)
    (run (sect source->list <> 11)
         >initial
         (match? (list-of= symbol=?)
                 (list 'ihveno 'alnogay 'unfatn 'rrud 'doueeti 'eearaft
                       'eremb 'lslw 'lae 'ihveno 'alnogay)))))

(suite disinterleave ()
  (test try-it
    (disinterleave (cyclic-source 'oglmc 'fnhid 'pftor 'aehn 'iet))
    2 (source? source?)
    (lambda (alpha beta)
      (let ((alpha-list ((pipe source->list >initial) alpha 9))
            (beta-list ((pipe source->list >initial) beta 7)))
        (and ((list-of= symbol=?) alpha-list
                                  (list 'oglmc 'pftor 'iet 'fnhid 'aehn
                                        'oglmc 'pftor 'iet 'fnhid))
             ((list-of= symbol=?) beta-list
                                  (list 'fnhid 'aehn 'oglmc 'pftor 'iet
                                        'fnhid 'aehn)))))))

(suite integer-source ()
  (test collect-eighteen
    ((pipe source->list >initial) integer-source 18)
    1 ((list-of exact-integer?))
    (match? (list-of= =)
            (list 0 -1 1 -2 2 -3 3 -4 4 -5 5 -6 6 -7 7 -8 8 -9))))

(suite map-source ()
  (test single-source
    (map-source halve (constant-source 44))
    1 (source?)
    (run (sect source->list <> 6)
         >initial
         (match? (list-of= =) (let ((half (halve 44)))
                                (list half half half half half half)))))
  (test multiple-sources
    (map-source list (constant-source 787)
                     (constant-source 526)
                     (constant-source 511))
    1 (source?)
    (run (sect source->list <> 3)
         >initial
         (match? (list-of= (list-of= =))
                 (let ((ls (list 787 526 511)))
                   (list ls ls ls)))))
  (test example-0-through-7
    (map-source + (cyclic-source 0 1)
                  (cyclic-source 0 0 2 2)
                  (cyclic-source 0 0 0 0 4 4 4 4))
    1 (source?)
    (run (sect source->list <> 18)
         >initial
         (match? (list-of= =)
                 (list 0 1 2 3 4 5 6 7 0 1 2 3 4 5 6 7 0 1)))))

(suite alternative-natural-number-source ()
  (test try-it
    ((pipe source->list >initial) alternative-natural-number-source 524)
    1 ((list-of natural-number?))
    (match? (list-of= =) ((unfold-list zero? (sect - 524 <>) sub1) 524))))

(suite process-source ()
  (test minimal
    (process-source list prepend)
    1 (procedure?)
    (lambda (prepender)
      (suite minimal-internal ()
        (test prepend-indefinitely
          (prepender (constant-source 'guela))
          1 (source?)
          (run (sect source->list <> 4)
               >initial
               (match? (list-of= (list-of= symbol=?))
                       (list (list)
                             (list 'guela)
                             (list 'guela 'guela)
                             (list 'guela 'guela 'guela))))))))
  (test partial-products
    (process-source (create 1) *)
    1 (procedure?)
    (lambda (multiplier)
      (suite partial-products-internal
          ((positive-integer-source
            (map-source add1 natural-number-source)))
        (test factorial-source
          (multiplier positive-integer-source)
          1 (source?)
          (run (sect source->list <> 11)
               >initial
               (match? (list-of= =)
                       (list 1 1 2 6 24 120 720 5040 40320 362880
                             3628800))))
        (test powers-of-three
          (multiplier (constant-source 3))
          1 (source?)
          (run (sect source->list <> 8)
               >initial
               (match? (list-of= =) (list 1 3 9 27 81 243 729 2187))))))))

(suite unfold-source ((iota (lambda (n)
                              ((unfold-list zero? (sect - n <>) sub1) n))))
  (test minimal
    (unfold-source (create 'dne) (create))
    1 (procedure?)
    (lambda (unfolder)
      (suite minimal-internal ()
        (test run-it
          (unfolder)
          1 (source?)
          (run (sect source->list <> 6)
               >initial
               (match? (list-of= symbol=?)
                       (list 'dne 'dne 'dne 'dne 'dne 'dne)))))))
  (test square-unfolder
    (unfold-source square add1)
    1 (procedure?)
    (lambda (unfolder)
      (suite square-unfolder-internal ()
        (test try-it
          (unfolder 0)
          1 (source?)
          (run (sect source->list <> 638)
               >initial
               (match? (list-of= =) (map square (iota 638))))))))
  (test multiple-guides
    (unfold-source >initial (dispatch >next +))
    1 (procedure?)
    (lambda (unfolder)
      (suite fibonacci ()
        (test try-it
          (unfolder 0 1)
          1 (source?)
          (run (sect source->list <> 12)
               >initial
               (match? (list-of= =)
                       (list 0 1 1 2 3 5 8 13 21 34 55 89))))))))

(suite Fibonacci ()
  (test collect-thirteen
    ((pipe source->list >initial) Fibonacci 13)
    1 ((list-of exact-integer?))
    (match? (list-of= =)
            (list 0 1 1 2 3 5 8 13 21 34 55 89 144))))

(suite empty-source? ()
  (test yes
    (empty-source? (constant-source (end-of-source)))
    1 (boolean?)
    true?)
  (test yes-despite-deviousness
    (empty-source? (source (values (end-of-source) natural-number-source)))
    1 (boolean?)
    true?)
  (test no
    (empty-source? (source (values 0 (constant-source (end-of-source)))))
    1 (boolean?)
    not)
  (test infinite-source
    (empty-source? natural-number-source)
    1 (boolean?)
    not))

(suite list->finite-source ()
  (test empty-list
    (list->finite-source (list))
    1 (source?)
    empty-source?)
  (test singleton
    (list->finite-source (list 'obekar))
    1 (source?)
    (pipe (sect source->list <> 1)
          (lambda (alpha beta)
            (and ((list-of= symbol=?) alpha (list 'obekar))
                 (empty-source? beta)))))
  (test longer-list
    (list->finite-source (list 'dysh 'oidt 'amtbbnw 'eoe 'gihnte))
    1 (source?)
    (pipe (sect source->list <> 5)
          (lambda (alpha beta)
            (and ((list-of= symbol=?) alpha  (list 'dysh 'oidt 'amtbbnw
                                                   'eoe 'gihnte))
                 (empty-source? beta))))))

(suite finite-source ()
  (test no-arguments
    (finite-source)
    1 (source?)
    empty-source?)
  (test many-arguments
    (finite-source 'ecsae 'delb 'hdrp 'oiipa)
    1 (source?)
    (pipe (sect source->list <> 4)
          (lambda (alpha beta)
            (and ((list-of= symbol=?) alpha (list 'ecsae 'delb 'hdrp
                                                  'oiipa))
                 (empty-source? beta))))))

(suite finite-source->list ()
  (test empty-source
    (finite-source->list (finite-source))
    1 (list?)
    empty-list?)
  (test singleton-source
    (finite-source->list (finite-source 'ebr))
    1 (list?)
    (match? (list-of= symbol=?) (list 'ebr)))
  (test longer-source
    (finite-source->list (finite-source 'hbcg 'cfhon 'nttyytt 'lstsyv
                                        'mtwno 'evan 'wudh))
    1 (list?)
    (match? (list-of= symbol=?) (list 'hbcg 'cfhon 'nttyytt 'lstsyv 'mtwno
                                      'evan 'wudh))))

(suite catenate-sources ()
  (test both-arguments-empty
    (catenate-sources (finite-source) (finite-source))
    1 (source?)
    empty-source?)
  (test first-argument-empty
    (catenate-sources (finite-source)
                      (finite-source 'yyr 'rgne 'ytth 'enli 'otpmswh))
    1 (source?)
    (pipe finite-source->list
          (match? (list-of= symbol=?)
                  (list 'yyr 'rgne 'ytth 'enli 'otpmswh))))
  (test second-argument-empty
    (catenate-sources (finite-source 'noaia 'itre 'mto 'thtdai)
                      (finite-source))
    1 (source?)
    (pipe finite-source->list
          (match? (list-of= symbol=?)
                  (list 'noaia 'itre 'mto 'thtdai))))
  (test longer-sources
    (catenate-sources (finite-source 'ghn 'yaotwot 'ytel 'sonh 'hoseve)
                      (finite-source 'eac 'elaiwf 'spiddeu))
    1 (source?)
    (pipe finite-source->list
          (match? (list-of= symbol=?)
                  (list 'ghn 'yaotwot 'ytel 'sonh 'hoseve 'eac 'elaiwf
                        'spiddeu)))))

(suite append-sources ()
  (test no-arguments
    (append-sources)
    1 (source?)
    empty-source?)
  (test one-argument
    (append-sources (finite-source 'fasttdp 'buergt 'eoco 'okltoof))
    1 (source?)
    (pipe finite-source->list
          (match? (list-of= symbol=?)
                  (list 'fasttdp 'buergt 'eoco 'okltoof))))
  (test variety
    (append-sources (finite-source)
                    (finite-source 'tgrhor 'geewda 'hepsh)
                    (finite-source 'eeaicn)
                    (finite-source)
                    (finite-source 'irsosk 'nesui 'ers 'heua))
    1 (source?)
    (pipe finite-source->list
          (match? (list-of= symbol=?)
                  (list 'tgrhor 'geewda 'hepsh 'eeaicn 'irsosk 'nesui 'ers
                        'heua)))))

(suite fold-finite-source
    ((nanc (lambda (left right)
             (* left (add1 right))))
     (tally-by-parity (lambda (element evens odds)
                        (if (even? element)
                            (values (add1 evens) odds)
                            (values evens (add1 odds))))))
  (test minimal
    (fold-finite-source (create) black-hole)
    1 (procedure?)
    (lambda (idler)
      (suite minimal-internal ()
        (test do-nothing-at-all
          (idler (finite-source))
          0 ())
        (test do-nothing-at-length
          (idler (finite-source 'uthoi 'eam 'fwnas 'solsaet 'neo 'erofut
                                'tosi 'mnth 'ldnsath))
          0 ()))))
  (test fold-nanc
    (fold-finite-source (create 810) nanc)
    1 (procedure?)
    (lambda (folder)
      (suite fold-nanc-internal ()
        (test empty-source
          (folder (finite-source))
          1 (exact-integer?)
          (match? = 810))
        (test singleton-source
          (folder (finite-source 540))
          1 (exact-integer?)
          (match? = (nanc 540 810)))
        (test several-elements
          (folder (finite-source 730 55 512 17 315))
          1 (exact-integer?)
          (match? = (nanc 730
                          (nanc 55
                                (nanc 512
                                      (nanc 17
                                            (nanc 315 810))))))))))
  (test fold-tally-by-parity
    (fold-finite-source (create 0 0) tally-by-parity)
    1 (procedure?)
    (lambda (folder)
      (suite fold-tally-by-parity-internal ()
        (test empty-source
          (folder (finite-source))
          2 (natural-number? natural-number?)
          (match? = 0 0))
        (test several-elements
          (folder (finite-source 216 619 181 618 816 299 448 304))
          2 (natural-number? natural-number?)
          (match? = 5 3))))))

(suite map-finite-source ()
  (test empty-source
    (map-finite-source square (finite-source))
    1 (source?)
    empty-source?)
  (test singleton-source
    (map-finite-source add1 (finite-source 434))
    1 (source?)
    (pipe finite-source->list (match? (list-of= =) (list 435))))
  (test larger-source
    (map-finite-source halve
                       (finite-source 747 225 502 211 918 91 701 953))
    1 (source?)
    (pipe finite-source->list
          (match? (list-of= =) (list 373 112 251 105 459 45 350 476)))))

;;; Test the procedures from the "Tuples" section.

(suite brighter? ((sirius (make-star "Alpha Canis Majoris" -142/100 27/10
                                       'A1Vm))
                  (canopus (make-star "Alpha Carinae" -72/100 30 'F0Ib)))
  (test less-bright
    (brighter? canopus sirius)
    1 (boolean?)
    not)
  (test equally-bright
    (brighter? canopus canopus)
    1 (boolean?)
    not)
  (test more-bright
    (brighter? sirius canopus)
    1 (boolean?)
    true?))

(suite filter-by-spectral-class ()
  (test empty-catalog
    (filter-by-spectral-class 'A1Vm (list))
    1 ((list-of star?))
    empty-list?)
  (test real-filtering
    (filter-by-spectral-class 'G2V 
      (list (make-star "Sol" -2672/100 0 'G2V)
            (make-star "Alpha Canis Majoris" -142/100 27/10 'A1Vm)
            (make-star "Alpha Carinae" -72/100 30 'F0Ib)
            (make-star "Alpha Boötis" -5/100 112/10 'K2IIIp)
            (make-star "Alpha-1 Centauri" -1/100 13/10 'G2V)
            (make-star "Alpha Lyrae" 3/100 77/10 'A0Vvar)))
    1 ((list-of star?))
    (match? (list-of= star=?)
            (list (make-star "Sol" -2672/100 0 'G2V)
                  (make-star "Alpha-1 Centauri" -1/100 13/10 'G2V)))))

(suite nearest-star ()
  (test singleton-catalog
    (nearest-star (make-star "Sol" -2672/100 0 'G2V))
    1 (star?)
    (match? star=? (make-star "Sol" -2672/100 0 'G2V)))
  (test longer-catalog
    (nearest-star (make-star "Alpha Canis Majoris" -142/100 27/10 'A1Vm)
                  (make-star "Alpha Carinae" -72/100 30 'F0Ib)
                  (make-star "Alpha Boötis" -5/100 112/10 'K2IIIp)
                  (make-star "Alpha-1 Centauri" -1/100 13/10 'G2V)
                  (make-star "Alpha Lyrae" 3/100 77/10 'A0Vvar))
    1 (star?)
    (match? star=? (make-star "Alpha-1 Centauri" -1/100 13/10 'G2V))))

;;; Test the procedures from the "Trees" section.

(suite tree? ()
  (test empty-tree
    (tree? (make-empty-tree))
    1 (boolean?)
    true?)
  (test non-empty-tree
    (tree? (make-non-empty-tree 'lreeh
             (make-non-empty-tree 'sfdpl
               (make-empty-tree)
               (make-empty-tree))
             (make-non-empty-tree 'osteiae
               (make-empty-tree)
               (make-empty-tree))))
    1 (boolean?)
    true?)
  (test random-value
    (tree? 'dhusils)
    1 (boolean?)
    not)
  (test too-few-components
    (tree? (list 'luen (make-non-empty-tree 'tnueoa
                         (make-empty-tree)
                         (make-empty-tree))))
    1 (boolean?)
    not)
  (test too-many-components
    (tree? (make-non-empty-tree
            'epfo
            (make-non-empty-tree 'ddame
              (make-empty-tree)
              (make-empty-tree))
            (list 'ipwho
                  (make-empty-tree)
                  (make-non-empty-tree 'uah
                    (make-empty-tree)
                    (make-empty-tree))
                  (make-empty-tree)
                  (make-non-empty-tree 'tlefidl
                    (make-empty-tree)
                    (make-empty-tree)))))
    1 (boolean?)
    not)
  (test bad-component-1
    (tree? (list 'dhlc 'smwsmg (make-non-empty-tree 'snfrn
                                 (make-empty-tree)
                                 (make-empty-tree))))
    1 (boolean?)
    not)
  (test bad-component-2
    (tree? (list 'rttet
                 (make-non-empty-tree 'wrkuea
                   (make-empty-tree)
                   (make-empty-tree))
                 (list (make-empty-tree) 'vrr 'hrhyss)))
    1 (boolean?)
    not))

(suite tree=? ()
  (test both-empty
    (tree=? (make-empty-tree) (make-empty-tree))
    1 (boolean?)
    true?)
  (test left-empty
    (tree=? (make-empty-tree) (make-non-empty-tree 'uiicg
                                (make-empty-tree)
                                (make-empty-tree)))
    1 (boolean?)
    not)
  (test right-empty
    (tree=? (make-non-empty-tree
             'twl
             (make-non-empty-tree 'aemfa
               (make-empty-tree)
               (make-empty-tree))
             (make-non-empty-tree 'dfs
               (make-empty-tree)
               (make-non-empty-tree 'fla
                 (make-empty-tree)
                 (make-empty-tree))))
            (make-empty-tree))
    1 (boolean?)
    not)
  (test equal-non-empty-trees
    (tree=? (make-non-empty-tree 'nait
              (make-empty-tree)
              (make-non-empty-tree 'ion
                (make-non-empty-tree 'ete
                  (make-empty-tree)
                  (make-empty-tree))
                (make-non-empty-tree 'hglt
                  (make-non-empty-tree 'iiaa
                    (make-empty-tree)
                    (make-empty-tree))
                  (make-empty-tree))))
            (make-non-empty-tree 'nait
              (make-empty-tree)
              (make-non-empty-tree 'ion
                (make-non-empty-tree 'ete
                  (make-empty-tree)
                  (make-empty-tree))
                (make-non-empty-tree 'hglt
                  (make-non-empty-tree 'iiaa
                    (make-empty-tree)
                    (make-empty-tree))
                  (make-empty-tree)))))
    1 (boolean?)
    true?)
  (test same-elements-in-same-inorder
    (tree=? (make-non-empty-tree 'oanisi
              (make-non-empty-tree 'mito
                (make-non-empty-tree 'ototll
                  (make-empty-tree)
                  (make-empty-tree))
                (make-non-empty-tree 'ire
                  (make-empty-tree)
                  (make-empty-tree)))
              (make-non-empty-tree 'cna
                (make-empty-tree)
                (make-empty-tree)))
            (make-non-empty-tree 'mito
              (make-non-empty-tree 'ototll
                (make-empty-tree)
                (make-empty-tree))
              (make-non-empty-tree 'oanisi
                (make-non-empty-tree 'ire
                  (make-empty-tree)
                  (make-empty-tree))
                (make-non-empty-tree 'cna
                  (make-empty-tree)
                  (make-empty-tree)))))
    1 (boolean?)
    not)
  (test different-elements-in-same-shape
    (tree=? (make-non-empty-tree 'pebe
              (make-non-empty-tree 'aadr
                (make-empty-tree)
                (make-non-empty-tree 'foiocat
                  (make-empty-tree)
                  (make-empty-tree)))
              (make-non-empty-tree 'lptnahs
                (make-empty-tree)
                (make-non-empty-tree 'riohn
                  (make-empty-tree)
                  (make-empty-tree))))
            (make-non-empty-tree 'pebe
              (make-non-empty-tree 'aadr
                (make-empty-tree)
                (make-non-empty-tree 'foiocat
                  (make-empty-tree)
                  (make-empty-tree)))
              (make-non-empty-tree 'tfnh
                (make-empty-tree)
                (make-non-empty-tree 'riohn
                  (make-empty-tree)
                  (make-empty-tree)))))
    1 (boolean?)
    not))

(suite singleton-tree ()
  (test satisfies-classification-predicate
    (tree? (singleton-tree 'gaea))
    1 (boolean?)
    true?)
  (test is-non-empty
    (non-empty-tree? (singleton-tree 'iadi))
    1 (boolean?)
    true?)
  (test has-correct-components
    (de-non-empty-tree (singleton-tree 'mii))
    3 (symbol? tree? tree?)
    (lambda (alpha beta gamma)
      (and (symbol=? alpha 'mii)
           (empty-tree? beta)
           (empty-tree? gamma)))))

(suite tree-of-numbers? ()
  (test empty-tree
    (tree-of-numbers? (make-empty-tree))
    1 (boolean?)
    true?)
  (test non-empty-tree
    (tree-of-numbers? (make-non-empty-tree 979
                        (singleton-tree 867)
                        (singleton-tree 63)))
    1 (boolean?)
    true?)
  (test random-value
    (tree-of-numbers? 508)
    1 (boolean?)
    not)
  (test too-few-components
    (tree-of-numbers? (list 596 (singleton-tree 770)))
    1 (boolean?)
    not)
  (test too-many-components
    (tree-of-numbers? (make-non-empty-tree 953
                        (singleton-tree 520)
                        (list 997
                              (make-empty-tree)
                              (singleton-tree 140)
                              (make-empty-tree)
                              (singleton-tree 972))))
    1 (boolean?)
    not)
  (test bad-component-1
    (tree-of-numbers? (list 922 325 (singleton-tree 565)))
    1 (boolean?)
    not)
  (test bad-component-2
    (tree-of-numbers? (list 973
                            (singleton-tree 448)
                            (list (make-empty-tree) 376 790)))
    1 (boolean?)
    not)
  (test non-number-at-root
    (tree-of-numbers? (make-non-empty-tree 'tioha
                        (make-empty-tree)
                        (make-non-empty-tree 359
                          (singleton-tree 262)
                          (make-empty-tree))))
    1 (boolean?)
    not)
  (test deeper-non-number
    (tree-of-numbers? (make-non-empty-tree 158
                        (singleton-tree 330)
                        (make-non-empty-tree 825
                          (singleton-tree 'bnmut)
                          (singleton-tree 175))))
    1 (boolean?)
    not))

(suite tree-of-numbers=? ((sample (make-non-empty-tree 493
                                    (make-non-empty-tree 167
                                      (singleton-tree 297)
                                      (make-empty-tree))
                                    (make-non-empty-tree 16
                                      (singleton-tree 399)
                                      (singleton-tree 708)))))
  (test both-empty
    (tree-of-numbers=? (make-empty-tree) (make-empty-tree))
    1 (boolean?)
    true?)
  (test left-empty
    (tree-of-numbers=? (make-empty-tree) (singleton-tree 32))
    1 (boolean?)
    not)
  (test right-empty
    (tree-of-numbers=? (make-non-empty-tree 24
                         (singleton-tree 89)
                         (make-non-empty-tree 134
                           (make-empty-tree)
                           (singleton-tree 547)))
                       (make-empty-tree))
    1 (boolean?)
    not)
  (test identical-non-empty-trees
    (tree-of-numbers=? sample sample)
    1 (boolean?)
    true?)
  (test duplicate-non-empty-trees
    (tree-of-numbers=? sample (make-non-empty-tree 493
                                    (make-non-empty-tree 167
                                      (singleton-tree 297)
                                      (make-empty-tree))
                                    (make-non-empty-tree 16
                                      (singleton-tree 399)
                                      (singleton-tree 708))))
    1 (boolean?)
    true?)
  (test same-elements-in-same-inorder
    (tree-of-numbers=? (make-non-empty-tree 163
                         (make-non-empty-tree 220
                           (singleton-tree 849)
                           (singleton-tree 754))
                         (singleton-tree 49))
                       (make-non-empty-tree 163
                         (make-non-empty-tree 849
                           (make-non-empty-tree 220
                             (make-empty-tree)
                             (singleton-tree 754))
                           (make-empty-tree))
                         (singleton-tree 49)))
    1 (boolean?)
    not)
  (test different-elements-in-same-shape
    (tree-of-numbers=? (make-non-empty-tree 931
                         (make-non-empty-tree 73
                           (make-empty-tree)
                           (singleton-tree 319))
                         (make-non-empty-tree 121
                           (make-empty-tree)
                           (singleton-tree 254)))
                       (make-non-empty-tree 314
                         (make-non-empty-tree 79
                           (make-empty-tree)
                           (singleton-tree 841))
                         (make-non-empty-tree 410
                           (make-empty-tree)
                           (singleton-tree 163))))
    1 (boolean?)
    not))

(suite tree-of ()
  (test symbols
    (tree-of symbol?)
    1 (procedure?)
    (lambda (tree-of-symbols?)
      (suite symbols-internal ()
        (test empty-tree
          (tree-of-symbols? (make-empty-tree))
          1 (boolean?)
          true?)
        (test non-empty-tree
          (tree-of-symbols? (make-non-empty-tree 'snhhd
                              (singleton-tree 'tnws)
                              (singleton-tree 'nfa)))
          1 (boolean?)
          true?)
        (test random-value
          (tree-of-symbols? 'mduur)
          1 (boolean?)
          not)
        (test too-few-components
          (tree-of-symbols? (list 'aohn (singleton-tree 'rrlud)))
          1 (boolean?)
          not)
        (test too-many-components
          (tree-of-symbols? (make-non-empty-tree 'idncse
                              (singleton-tree 'silda)
                              (list 'rsrnus
                                    (make-empty-tree)
                                    (singleton-tree 'sonlld)
                                    (make-empty-tree)
                                    (singleton-tree 'nra))))
          1 (boolean?)
          not)
        (test bad-component-1
          (tree-of-symbols? (list 'ethie 'eqrat (singleton-tree 'ueendp)))
          1 (boolean?)
          not)
        (test bad-component-2
          (tree-of-symbols? (list 'meor
                                  (singleton-tree 'lninar)
                                  (list (make-empty-tree) 'lyin 'ster)))
          1 (boolean?)
          not)
        (test non-symbol-at-root
          (tree-of-symbols? (make-non-empty-tree 143
                              (make-empty-tree)
                              (make-non-empty-tree 'hrldhsi
                                (singleton-tree 'dtewo)
                                (make-empty-tree))))
          1 (boolean?)
          not)
        (test deeper-non-symbol
          (tree-of-symbols? (make-non-empty-tree 'nhnafti
                              (singleton-tree 'iss)
                              (make-non-empty-tree 'srsf
                                (singleton-tree 482)
                                (singleton-tree 'dehens))))
          1 (boolean?)
          not)))))

(suite tree-of= ()
  (test symbol-equality
    (tree-of= symbol=?)
    1 (procedure?)
    (lambda (tree-of-symbols=?)
      (suite symbol-equality-internal
          ((sample (make-non-empty-tree 'ehu
                     (make-non-empty-tree 'cttso
                       (singleton-tree 'aee)
                       (make-empty-tree))
                     (make-non-empty-tree 'aphbd
                       (singleton-tree 'khnu)
                       (singleton-tree 'aslo)))))
        (test both-empty
          (tree-of-symbols=? (make-empty-tree) (make-empty-tree))
          1 (boolean?)
          true?)
        (test left-empty
          (tree-of-symbols=? (make-empty-tree) (singleton-tree 'ainndua))
          1 (boolean?)
          not)
        (test right-empty
          (tree-of-symbols=? (make-non-empty-tree 'dnknns
                               (singleton-tree 'sfeeit)
                               (make-non-empty-tree 'ere
                                 (make-empty-tree)
                                 (singleton-tree 'eaakn)))
                             (make-empty-tree))
          1 (boolean?)
          not)
        (test identical-non-empty-trees
          (tree-of-symbols=? sample sample)
          1 (boolean?)
          true?)
        (test duplicate-non-empty-trees
          (tree-of-symbols=? sample (make-non-empty-tree 'ehu
                                      (make-non-empty-tree 'cttso
                                        (singleton-tree 'aee)
                                        (make-empty-tree))
                                      (make-non-empty-tree 'aphbd
                                        (singleton-tree 'khnu)
                                        (singleton-tree 'aslo))))
          1 (boolean?)
          true?)
        (test same-elements-in-same-inorder
          (tree-of-symbols=? (make-non-empty-tree 'nnmrevo
                               (make-non-empty-tree 'nhhnoe
                                 (singleton-tree 'arec)
                                 (singleton-tree 'sob))
                               (singleton-tree 'hbni))
                             (make-non-empty-tree 'nnmrevo
                               (make-non-empty-tree 'arec
                                 (make-non-empty-tree 'nhhnoe
                                   (make-empty-tree)
                                   (singleton-tree 'sob))
                                 (make-empty-tree))
                               (singleton-tree 'hbni)))
          1 (boolean?)
          not)
        (test different-elements-in-same-shape
          (tree-of-symbols=? (make-non-empty-tree 'bak
                               (make-non-empty-tree 'huln
                                 (make-empty-tree)
                                 (singleton-tree 'tansvs))
                               (make-non-empty-tree 'nmhn
                                 (make-empty-tree)
                                 (singleton-tree 'eyrpett)))
                             (make-non-empty-tree 'srnoae
                               (make-non-empty-tree 'ebeo
                                 (make-empty-tree)
                                 (singleton-tree 'arta))
                               (make-non-empty-tree 'knt
                                 (make-empty-tree)
                                 (singleton-tree 'ennhhtw))))
          1 (boolean?)
          not)))))

(suite tree-size ()
  (test empty-tree
    (tree-size (make-empty-tree))
    1 (natural-number?)
    zero?)
  (test larger-tree
    (tree-size (make-non-empty-tree 'str
                 (make-non-empty-tree 'engssne
                   (make-empty-tree)
                   (make-non-empty-tree 'iteg
                     (make-empty-tree)
                     (singleton-tree 'cenqttw)))
                 (make-non-empty-tree 'nhuomaf
                   (make-non-empty-tree 'mnaie
                     (make-empty-tree)
                     (make-non-empty-tree 'ottuv
                       (make-empty-tree)
                       (singleton-tree 'rrmeeh)))
                   (make-non-empty-tree 'etin
                     (make-empty-tree)
                     (make-non-empty-tree 'utl
                       (make-empty-tree)
                       (singleton-tree 'tnn))))))
  1 (natural-number?)
  (match? = 11)))

(suite fold-tree ()
  (test minimal
    (fold-tree (create) black-hole)
    1 (procedure?)
    (lambda (idler)
      (test empty-tree
        (idler (make-empty-tree))
        0 ())
      (test larger-tree
        (idler (make-non-empty-tree 'loti
                 (singleton-tree 'rnmhpid)
                 (make-non-empty-tree 'ttef
                   (make-empty-tree)
                   (singleton-tree 'aaa))))
        0 ())))
  (test typical
    (fold-tree (create 0) +)
    1 (procedure?)
    (lambda (summer)
      (test empty-tree
        (summer (make-empty-tree))
        1 (exact-integer?)
        zero?)
      (test singleton-tree
        (summer (singleton-tree 770))
        1 (exact-integer?)
        (match? = 770))
      (test larger-tree
        (summer (make-non-empty-tree 679
                  (make-non-empty-tree 838
                    (make-non-empty-tree 214
                      (make-empty-tree)
                      (singleton-tree 10))
                    (make-non-empty-tree 536
                      (singleton-tree 41)
                      (make-empty-tree)))
                  (make-non-empty-tree 461
                    (singleton-tree 926)
                    (make-empty-tree))))
        1 (exact-integer?)
        (match? = (+ 679 838 214 10 536 41 461 926)))))
  (test multiple-results
    (fold-tree (create null null 0)
               (lambda (root l-left l-right l-depth r-left r-right r-depth)
                 (values (if (null? l-left) root l-left)
                         (if (null? r-right) root r-right)
                         (add1 (max l-depth r-depth)))))
    1 (procedure?)
    (lambda (multi)
      (suite multiple-internal ()
        (test empty-tree
          (multi (make-empty-tree))
          3 (null? null? natural-number?)
          (lambda (alpha beta gamma)
            (zero? gamma)))
        (test singleton
          (multi (singleton-tree 'anafh))
          3 (symbol? symbol? natural-number?)
          (lambda (alpha beta gamma)
            (and (symbol=? alpha 'anafh)
                 (symbol=? beta 'anafh)
                 (= gamma 1))))
        (test deeper-tree
          (multi (make-non-empty-tree 'ilohat
                   (make-non-empty-tree 'lota
                     (make-empty-tree)
                     (singleton-tree 'rortha))
                   (make-non-empty-tree 'bere
                     (singleton-tree 'fern)
                     (singleton-tree 'falasah))))
          3 (symbol? symbol? natural-number?)
          (lambda (alpha beta gamma)
            (and (symbol=? alpha 'lota)
                 (symbol=? beta 'falasah)
                 (= gamma 3))))))))

(suite tree-height ()
  (test empty-tree
    (tree-height (make-empty-tree))
    1 (natural-number?)
    zero?)
  (test singleton
    (tree-height (singleton-tree 'feodmr))
    1 (natural-number?)
    (match? = 1))
  (test balanced-tree
    (tree-height (make-non-empty-tree 'yhaiit
                   (make-non-empty-tree 'nrtcei
                     (make-non-empty-tree 'waemu
                       (make-non-empty-tree 'cigosr
                         (singleton-tree 'mer)
                         (singleton-tree 'feael))
                       (make-non-empty-tree 'mot
                         (singleton-tree 'daa)
                         (singleton-tree 'laaeyao)))
                     (make-non-empty-tree 'eeoh
                       (make-non-empty-tree 'eehwxea
                         (singleton-tree 'eryata)
                         (singleton-tree 'soosi))
                       (make-non-empty-tree 'wtne
                         (singleton-tree 'yod)
                         (singleton-tree 'aug))))
                   (make-non-empty-tree 'nuri
                     (make-non-empty-tree 'fen
                       (make-non-empty-tree 'hemp
                         (singleton-tree 'ori)
                         (singleton-tree 'hvwu))
                       (make-non-empty-tree 'npaowan
                         (singleton-tree 'fytw)
                         (singleton-tree 'toss)))
                     (make-non-empty-tree 'fyt
                       (make-non-empty-tree 'araso
                         (singleton-tree 'oluoh)
                         (singleton-tree 'klan))
                       (make-non-empty-tree 'ihiseir
                         (singleton-tree 'ceflpec)
                         (singleton-tree 'aedm))))))
    1 (natural-number?)
    (match? = 5))
  (test stringy-tree
    (tree-height (make-non-empty-tree 'shi
                   (singleton-tree 'ehsuhh)
                   (make-non-empty-tree 'ymiyoo
                     (make-non-empty-tree 'rrwmc
                       (make-non-empty-tree 'dpehghk
                         (make-empty-tree)
                         (make-non-empty-tree 'icaa
                           (make-non-empty-tree 'ititeon
                             (make-empty-tree)
                             (make-non-empty-tree 'rnifldd
                               (make-non-empty-tree 'ispig
                                 (singleton-tree 'rfnuusa)
                                 (make-non-empty-tree 'nmsh
                                   (make-empty-tree)
                                   (singleton-tree 'ehonp)))
                               (make-empty-tree)))
                           (make-non-empty-tree 'ecla
                             (make-non-empty-tree 'nraw
                               (make-empty-tree)
                               (make-non-empty-tree 'yshemrt
                                 (make-non-empty-tree 'bmp
                                   (make-empty-tree)
                                   (make-non-empty-tree 'ldbhoe
                                     (singleton-tree 'eefkc)
                                     (make-empty-tree)))
                                 (make-empty-tree)))
                             (make-empty-tree))))
                       (make-empty-tree))
                     (make-empty-tree))))
    1 (natural-number?)
    (match? = 11)))

(suite detree ()
  (test empty-tree
    (detree (make-empty-tree))
    0 ())
  (test singleton
    (detree (singleton-tree 'abutnda))
    1 (symbol?)
    (match? symbol=? 'abutnda))
  (test larger-tree
    (detree (make-non-empty-tree 'pep
              (make-non-empty-tree 'teanog
                (singleton-tree 'eeinn)
                (make-non-empty-tree 'feonan
                  (singleton-tree 'wouyohi)
                  (make-empty-tree)))
              (make-non-empty-tree 'idrt
                (singleton-tree 'ery)
                (singleton-tree 'esnhiw))))
    8 (symbol? symbol? symbol? symbol? symbol? symbol? symbol? symbol?)
    (match? symbol=? 'pep 'teanog 'eeinn 'feonan 'wouyohi 'idrt 'ery
                     'esnhiw)))

(suite for-all-in-tree? ()
  (test empty-tree
    (for-all-in-tree? (constant #f) (make-empty-tree))
    1 (boolean?)
    true?)
  (test good-singleton
    (for-all-in-tree? symbol? (singleton-tree 'otnfe))
    1 (boolean?)
    true?)
  (test bad-singleton
    (for-all-in-tree? string? (singleton-tree 'dosmni))
    1 (boolean?)
    not)
  (test bad-element-at-root
    (for-all-in-tree? odd? (make-non-empty-tree 168
                             (make-non-empty-tree 769
                               (singleton-tree 831)
                               (make-empty-tree))
                             (make-non-empty-tree 525
                               (singleton-tree 627)
                               (singleton-tree 237))))
    1 (boolean?)
    not)
  (test bad-element-in-subtree
    (for-all-in-tree? even? (make-non-empty-tree 92
                              (singleton-tree 190)
                              (make-non-empty-tree 188
                                (make-non-empty-tree 398
                                  (make-non-empty-tree 103
                                    (singleton-tree 760)
                                    (singleton-tree 40))
                                  (singleton-tree 470))
                                (singleton-tree 103))))
    1 (boolean?)
    not)
  (test good-larger-tree
    (for-all-in-tree? symbol? (make-non-empty-tree 'hae
                                (make-non-empty-tree 'ehadd
                                  (singleton-tree 'ote)
                                  (make-non-empty-tree 'uhewho
                                    (make-empty-tree)
                                    (singleton-tree 'aanohb)))
                                (make-non-empty-tree 'lrshow
                                  (make-non-empty-tree 'thgir
                                    (singleton-tree 'tomgrlm)
                                    (singleton-tree 'btri))
                                  (make-empty-tree))))
    1 (boolean?)
    true?))

(suite alternative-list->ordered-tree ()
  (test empty-list
    (alternative-list->ordered-tree (list))
    1 ((tree-of number?))
    empty-tree?)
  (test singleton
    (alternative-list->ordered-tree (list 645))
    1 ((tree-of number?))
    (match? tree=? (singleton-tree 645)))
  (test long-list
    (alternative-list->ordered-tree (list 235 80 132 432 565 881 726 138))
    1 ((tree-of number?))
    (match? tree=? (make-non-empty-tree 235
                     (make-non-empty-tree 80
                       (make-empty-tree)
                       (make-non-empty-tree 132
                         (make-empty-tree)
                         (singleton-tree 138)))
                     (make-non-empty-tree 432
                       (make-empty-tree)
                       (make-non-empty-tree 565
                         (make-empty-tree)
                         (make-non-empty-tree 881
                           (singleton-tree 726)
                           (make-empty-tree))))))))

(suite unfold-tree ()
  (test minimal
    (unfold-tree values? (create 'foo) black-hole black-hole)
    1 (procedure?)
    (lambda (nuller)
      (suite minimal-internal ()
        (test run-it
          (nuller)
          1 (tree?)
          empty-tree?))))
  (test multifoo
    (unfold-tree zero? (constant 'foo) sub1 sub1)
    1 (procedure?)
    (lambda (fooer)
      (suite multifoo-internal ()
        (test zero
          (fooer 0)
          1 (tree?)
          empty-tree?)
        (test nonzero
          (fooer 3)
          1 (tree?)
          (match? tree=? (make-non-empty-tree 'foo
                           (make-non-empty-tree 'foo
                             (singleton-tree 'foo)
                             (singleton-tree 'foo))
                           (make-non-empty-tree 'foo
                             (singleton-tree 'foo)
                             (singleton-tree 'foo)))))))))  

(suite list->ordered-tree ()
  (test empty-list
    (list->ordered-tree (list))
    1 ((tree-of number?))
    empty-tree?)
  (test singleton
    (list->ordered-tree (list 145))
    1 ((tree-of number?))
    (match? tree=? (singleton-tree 145)))
  (test long-list
    (list->ordered-tree (list 559 82 293 560 223 520 814))
    1 ((tree-of number?))
    (match? tree=? (make-non-empty-tree 559
                     (make-non-empty-tree 82
                       (make-empty-tree)
                       (make-non-empty-tree 293
                         (singleton-tree 223)
                         (singleton-tree 520)))
                     (make-non-empty-tree 560
                       (make-empty-tree)
                       (singleton-tree 814))))))

;;; Test the procedures in the "Bushes" section.

(suite bush ()
  (test no-arguments
    (bush)
    1 (list?)
    empty-list?)
  (test one-argument
    (bush 'mit)
    1 (list?)
    (match? (list-of= symbol=?) (list 'mit)))
  (test more-arguments
    (bush 'agsrt
          (bush 'bmfg (bush 'wfa) (bush 'eehus) (bush))
          (bush 'mnflad))
    1 (list?)
    (match? list=? (list 'agsrt
                         (list 'bmfg (list 'wfa) (list 'eehus) (list))
                         (list 'mnflad)))))

(suite bush-root ()
  (test singleton
    (bush-root (bush 'sronab))
    1 (symbol?)
    (match? symbol=? 'sronab))
  (test larger-bush
    (bush-root (bush 'asrnfhy
                     (bush 'hsiasl
                           (bush 'niena (bush 'tnhs))
                           (bush 'rwtshdh))
                     (bush 'shoc (bush) (bush))))
    1 (symbol?)
    (match? symbol=? 'asrnfhy)))

(suite bush-children ()
  (test singleton
    (bush-children (bush 'ualtio))
    1 (list?)
    empty-list?)
  (test larger-bush
    (bush-children (bush 'oohta
                         (bush 'hlseduo
                               (bush 'ist)
                               (bush)
                               (bush 'sfse (bush 'wluo)))
                         (bush)
                         (bush 'nfbwoyi (bush))))
    1 ((list-of list?))
    (match? (list-of= list=?)
            (list (list 'hlseduo
                        (list 'ist)
                        (list)
                        (list 'sfse (bush 'wluo)))
                  (list)
                  (list 'nfbwoyi (list))))))

(suite empty-bush? ()
  (test empty-bush
    (empty-bush? (bush))
    1 (boolean?)
    true?)
  (test singleton
    (empty-bush? (bush 'oioaabt))
    1 (boolean?)
    not)
  (test longer-bush
    (empty-bush? (bush 'oaedn
                       (bush 'hlyah) (bush 'shigst (bush 'toaeeo))))
    1 (boolean?)
    not))

(suite non-empty-bush? ()
  (test empty-bush
    (non-empty-bush? (bush))
    1 (boolean?)
    not)
  (test singleton
    (non-empty-bush? (bush 'ceree))
    1 (boolean?)
    true?)
  (test longer-bush
    (non-empty-bush? (bush 'eosan
                           (bush 'sioe (bush) (bush 'wkart
                                                    (bush 'aers)
                                                    (bush 'aihdin)))))
    1 (boolean?)
    true?))

(suite debush ()
  (test empty-bush
    (debush (bush))
    0 ())
  (test singleton
    (debush (bush 'maeea))
    1 (symbol?)
    (match? symbol=? 'maeea))
  (test larger-bush
    (debush (bush 'gsu
                  (bush)
                  (bush 'nhnimi
                        (bush 'erye)
                        (bush 'nothw (bush 'aai))
                        (bush))
                  (bush 'snocdta (bush))))
    4 (symbol? list? list? list?)
    (lambda (alpha beta gamma delta)
      (and (symbol=? alpha 'gsu)
           (list=? beta (list))
           (list=? gamma (list 'nhnimi
                               (list 'erye)
                               (list 'nothw (list 'aai))
                               (list)))
           (list=? delta (list 'snocdta (list)))))))

(suite bush? ()
  (test empty-bush
    (bush? (bush))
    1 (boolean?)
    true?)
  (test singleton-bush
    (bush? (bush 'rnnabel))
    1 (boolean?)
    true?)
  (test larger-bush
    (bush? (bush 'iuak
                 (bush 'xemfwl)
                 (bush 'ftuns (bush 'oeen (bush)) (bush))
                 (bush 'mttn (bush) (bush 'eao))))
    1 (boolean?)
    true?)
  (test unicorn
    (bush? 'oog)
    1 (boolean?)
    not)
  (test bad-element
    (bush? (list 'dsnaqep (bush 'itlc) 'hodnn))
    1 (boolean?)
    not)
  (test deeply-nested
    (bush? (list 'lya
                 (list 'shetn
                       (list 'asd
                             (list 'anrh (bush) (bush) 'fern)))))
    1 (boolean?)
    not)
  (test all-symbols
    (bush? (list 'epae 'sua 'fert 'out))
    1 (boolean?)
    not))

(suite bush=? ((sample (bush 'thieyb
                             (bush)
                             (bush 'euah (bush 'rltm))
                             (bush 'lebud)
                             (bush 'hrml (bush)))))
  (test empty-bushes
    (bush=? (bush) (bush))
    1 (boolean?)
    true?)
  (test left-bush-empty
    (bush=? (bush) (bush 'irlt
                         (bush 'cotrt (bush))
                         (bush 'nrrmtl
                               (bush 'cunmn)
                               (bush 'noc))))
    1 (boolean?)
    not)
  (test right-bush-empty
    (bush=? (bush 'ocwteyl (bush 'whn (bush 'esdmh)) (bush 'trt))
            (bush))
    1 (boolean?)
    not)
  (test identical
    (bush=? sample sample)
    1 (boolean?)
    true?)
  (test duplicate
    (bush=? sample (bush 'thieyb
                         (bush)
                         (bush 'euah (bush 'rltm))
                         (bush 'lebud)
                         (bush 'hrml (bush))))
    1 (boolean?)
    true?)
  (test different-root-element
    (bush=? (bush 'tdieutr (bush 'vwd) (bush 'oeeoa))
            (bush 'pdnh (bush 'vwd) (bush 'oeeoa)))
    1 (boolean?)
    not)
  (test different-numbers-of-children
    (bush=? (bush 'ndibbt 
                  (bush 'yefutn (bush 'khogim) (bush 'ceyaf))
                  (bush 'eutt))
            (bush 'ndibbt
                  (bush 'yefutn (bush 'khogim) (bush 'ceyaf))))
    1 (boolean?)
    not)
  (test different-children
    (bush=? (bush 'kltaoi
                  (bush 'aecin)
                  (bush 'ahodh (bush 'rnlsetn (bush 'doly))))
            (bush 'kltaoi
                  (bush 'aecin)
                  (bush 'ahodh (bush 'rnlsetn) (bush 'doly))))
    1 (boolean?)
    not))

(suite bush-of ()
  (test symbols
    (bush-of symbol?)
    1 (procedure?)
    (lambda (bush-of-symbols?)
      (suite bush-of-internal ()
        (test empty-bush
          (bush-of-symbols? (bush))
          1 (boolean?)
          true?)
        (test non-empty-bush-yes
          (bush-of-symbols? (bush 'ctet
                                  (bush)
                                  (bush 'eiido
                                        (bush 'ilh)
                                        (bush 'ssh)
                                        (bush)
                                        (bush 'tstobow))))
          1 (boolean?)
          true?)
        (test non-empty-bush-bad-root
          (bush-of-symbols? (bush 563
                                  (bush 'lveeaa
                                        (bush 'sokhr))
                                  (bush)
                                  (bush 'eonr)))
          1 (boolean?)
          not)
        (test non-empty-bush-bad-inner-element
          (bush-of-symbols? (bush 'aekthde
                                  (bush 'wnhhwle
                                        (bush)
                                        (bush 'yiiut)
                                        (bush 45)
                                        (bush 'cenw (bush 'ieldtde)))))
          1 (boolean?)
          not)
        (test unicorn
          (bush-of-symbols? 670)
          1 (boolean?)
          not)
        (test bad-element
          (bush-of-symbols? (list 'sth (bush 'vtmeird) 'nta (bush)))
          1 (boolean?)
          not)
        (test careless
          (bush-of-symbols? (list 'gerdetf 'unocc 'uaadao 'rtr))
          1 (boolean?)
          not)))))

(suite bush-of= ()
  (test numbers
    (bush-of= =)
    1 (procedure?)
    (lambda (bush-of-numbers=?)
      (suite bush-of=-internal ((sample (bush 80
                                              (bush)
                                              (bush 362 (bush 179))
                                              (bush 89))))
        (test empty-bushes
          (bush-of-numbers=? (bush) (bush))
          1 (boolean?)
          true?)
        (test left-bush-empty
          (bush-of-numbers=? (bush) sample)
          1 (boolean?)
          not)
        (test right-bush-empty
          (bush-of-numbers=? sample (bush))
          1 (boolean?)
          not)
        (test identical
          (bush-of-numbers=? sample sample)
          1 (boolean?)
          true?)
        (test duplicate
          (bush-of-numbers=? (bush 80
                                   (bush)
                                   (bush 362 (bush 179))
                                   (bush 89))
                             sample)
          1 (boolean?)
          true?)
        (test different-at-root
          (bush-of-numbers=? (bush 29
                                   (bush)
                                   (bush 362 (bush 179))
                                   (bush 89))
                             sample)
          1 (boolean?)
          not)
        (test different-shape
          (bush-of-numbers=? sample
                             (bush 80
                                   (bush 362 (bush))
                                   (bush 179 (bush 89))))
          1 (boolean?)
          not)
        (test different-children
          (bush-of-numbers=? sample
                             (bush 80
                                   (bush)
                                   (bush 272 (bush 499))
                                   (bush 212)))
          1 (boolean?)
          not)))))

(suite fold-bush ()
  (test bush-size
    (fold-bush (create 0) (run >all-but-initial + add1))
    1 (procedure?)
    (lambda (sizer)
      (suite bush-size-internal ()
        (test empty-bush-size
          (sizer (bush))
          1 (natural-number?)
          zero?)
        (test larger-bush-size
          (sizer (bush 'ugnons
                       (bush 'uoem
                             (bush 'aot
                                   (bush 'ocsgrsn))
                             (bush 'sabaoy
                                   (bush 'koy)
                                   (bush)
                                   (bush 'nhlr
                                         (bush 'ictoeod))))
                       (bush 'fhaleyx
                             (bush)
                             (bush 'oom))))
          1 (natural-number?)
          (match? = 10)))))
  (test bush->list
    (fold-bush (create (list))
               (pipe (dispatch >initial (pipe >all-but-initial append))
                     prepend))
    1 (procedure?)
    (lambda (lister)
      (suite bush->list-internal ()
        (test empty-bush->list
          (lister (bush))
          1 (list?)
          empty-list?)
        (test larger-bush->list
          (lister (bush 'nneo
                        (bush 'urfhal
                              (bush 'aisndo
                                    (bush)
                                    (bush 'ahrro)
                                    (bush 'eassip))
                              (bush 'agiswiv
                                    (bush 'dgefu)))
                        (bush)
                        (bush)))
          1 ((list-of symbol?))
          (match? (list-of= symbol=?)
                  (list 'nneo 'urfhal 'aisndo 'ahrro 'eassip 'agiswiv
                        'dgefu))))))
  (test evens-and-odds
    (fold-bush (create (list) (list))
               (lambda (root . others)
                 (receive (evens odds)
                          (let loop ((remaining others))
                            (if (empty-list? remaining)
                                (values (list) (list))
                                (receive (evens odds)
                                         (loop (rest (rest remaining)))
                                  (values (append (first remaining) evens)
                                          (append (first (rest remaining))
                                                  odds)))))
                   (if (even? root)
                       (values (cons root evens) odds)
                       (values evens (cons root odds))))))
    1 (procedure?)
    (lambda (separator)
      (suite evens-and-odds-internal ()
        (test empty-bush
          (separator (bush))
          2 ((list-of exact-integer?) (list-of exact-integer?))
          (match? (list-of= =) (list) (list)))
        (test larger-bush
          (separator (bush 700
                           (bush 655
                                 (bush 135 (bush))
                                 (bush 961 (bush 28) (bush 525))
                                 (bush))
                           (bush 205)
                           (bush 288 (bush 337) (bush 694))
                           (bush 414 (bush) (bush 162))))
          2 ((list-of exact-integer?) (list-of exact-integer?))
          (match? (list-of= =) (list 700 28 288 694 414 162)
                         (list 655 135 961 525 205 337)))))))

(suite bush->list ()
  (test empty-bush
    (bush->list (bush))
    1 (list?)
    empty-list?)
  (test singleton-bush
    (bush->list (bush 'tnb))
    1 ((list-of symbol?))
    (match? (list-of= symbol=?) (list 'tnb)))
  (test larger-bush
    (bush->list (bush 'rpdml
                      (bush 'bhsrsh
                            (bush)
                            (bush 'khrtlhe)
                            (bush 'tnr (bush) (bush 'stwapbr))
                            (bush 'nee (bush 'oalom) (bush)))))
    1 ((list-of symbol?))
    (match? (list-of= symbol=?)
            (list 'rpdml 'bhsrsh 'khrtlhe 'tnr 'stwapbr 'nee 'oalom))))

(suite for-all-in-bush? ()
  (test empty-bush
    (for-all-in-bush? (constant #f) (bush))
    1 (boolean?)
    true?)
  (test good-singleton
    (for-all-in-bush? symbol? (bush 'ispgd))
    1 (boolean?)
    true?)
  (test bad-singleton
    (for-all-in-bush? string? (bush 'slrpad))
    1 (boolean?)
    not)
  (test bad-root
    (for-all-in-bush? symbol? (bush 229
                                   (bush 'gdsmsgh
                                         (bush 'gbkto)
                                         (bush 'ionhir))
                                   (bush 'ften (bush 'waakrc))))
    1 (boolean?)
    not)
  (test bad-element-deeply-nested
    (for-all-in-bush? even? (bush 122
                                  (bush 540
                                        (bush 52
                                              (bush 787
                                                    (bush 540))
                                              (bush 424)
                                              (bush 766))
                                        (bush 538))
                                  (bush 552 (bush 464))))
    1 (boolean?)
    not)
  (test larger-bush
    (for-all-in-bush? symbol? (bush 'ieto
                                    (bush 'soohdt (bush 'cadsv))
                                    (bush 'waastca)
                                     (bush)
                                    (bush 'eea
                                          (bush 'nnyhr)
                                          (bush 'nrutt))))
    1 (boolean?)
    true?))

(suite unfold-bush ()
  (test minimal
    (unfold-bush values? (constant 'foo) black-hole)
    1 (procedure?)
    (lambda (nuller)
      (suite minimal-internal ()
        (test run-it
          (nuller 'bar)
          1 (bush?)
          empty-bush?))))
  (test increment
    (unfold-bush empty-bush?
                 (pipe bush-root add1)
                 (pipe bush-children delist))
    1 (procedure?)
    (lambda (incrementer)
      (suite increment-internal ()
        (test empty-bush
          (incrementer (bush))
          1 ((bush-of number?))
          empty-bush?)
        (test non-empty-bush
          (incrementer (bush 832
                             (bush 123)
                             (bush 255
                                   (bush 706 (bush))
                                   (bush 711))
                             (bush)
                             (bush 297 (bush 493))))
          1 ((bush-of number?))
          (match? (bush-of= =) (bush 833
                                     (bush 124)
                                     (bush 256
                                           (bush 707 (bush))
                                           (bush 712))
                                     (bush)
                                     (bush 298 (bush 494)))))))))

(suite iota-bush ((numbush? (bush-of natural-number?))
                  (numbush=? (bush-of= =)))
  (test zero
    (iota-bush 0)
    1 (numbush?)
    (match? numbush=? (bush 0)))
  (test one
    (iota-bush 1)
    1 (numbush?)
    (match? numbush=? (bush 1 (bush 0))))
  (test larger
    (iota-bush 5)
    1 (numbush?)
    (match? numbush=?
            (bush 5 (bush 4 (bush 3 (bush 2 (bush 1 (bush 0)) (bush 0))
                                    (bush 1 (bush 0))
                                    (bush 0))
                            (bush 2 (bush 1 (bush 0)) (bush 0))
                            (bush 1 (bush 0))
                            (bush 0))
                    (bush 3 (bush 2 (bush 1 (bush 0)) (bush 0))
                            (bush 1 (bush 0))
                            (bush 0))
                    (bush 2 (bush 1 (bush 0)) (bush 0))
                    (bush 1 (bush 0))
                    (bush 0)))))

;;; Test the procedures from the "Bags" section.

(suite bag? ()
  (test empty-bag
    (bag? (list))
    1 (boolean?)
    true?)
  (test fuller-bag
    (bag? (list 'ureild 'aytelec 'eesh 'leanio))
    1 (boolean?)
    true?)
  (test non-bag
    (bag? 'yctfhga)
    1 (boolean?)
    not))

(suite bag-of ()
  (test generic
    (bag-of values?)
    1 (procedure?)
    (lambda (alternative-bag?)
      (suite generic-internal ()
        (test empty-bag
          (alternative-bag? (list))
          1 (boolean?)
          true?)
        (test longer-bag
          (alternative-bag? (list 'lao 'oibo 'usntd 'hsny))
          1 (boolean?)
          true?)
        (test sneak
          (alternative-bag? (cons 'wntsidt
                                  (cons 'dtient (cons 'aiv 'hpv))))
          1 (boolean?)
          not)
        (test unicorn
          (alternative-bag? 415)
          1 (boolean?)
          not))))
  (test numbers
    (bag-of number?)
    1 (procedure?)
    (lambda (alternative-bag-of-numbers?)
      (suite numbers-internal ()
        (test empty
          (alternative-bag-of-numbers? (list))
          1 (boolean?)
          true?)
        (test good-singleton
          (alternative-bag-of-numbers? (list 445))
          1 (boolean?)
          true?)
        (test bad-singleton
          (alternative-bag-of-numbers? (list 'wegslt))
          1 (boolean?)
          not)
        (test good-long-bag
          (alternative-bag-of-numbers? (list 648 104 554 505 666))
          1 (boolean?)
          true?)
        (test one-wrong-un
          (alternative-bag-of-numbers? (list 395 940 'horhmgw 414 524))
          1 (boolean?)
          not)))))

(suite bag ()
  (test no-arguments
    (bag)
    1 (bag?)
    empty-list?)
  (test singleton
    (bag 'daee)
    1 (bag?)
    (match? (list-of= symbol=?) (list 'daee)))
  (test larger-bag
    (bag 'aeamdoe 'usypn 'ahiei 'fvu)
    1 (bag?)
    (match? (list-of= symbol=?) (list 'aeamdoe 'usypn 'ahiei 'fvu))))

(suite debag ()
  (test empty-bag
    (debag (bag))
    0 ())
  (test singleton
    (debag (bag 'map))
    1 (symbol?)
    (match? symbol=? 'map))
  (test larger-bag
    (debag (bag 'hirswca 'rwren 'norh 'eeyno 'eeney))
    5 (symbol? symbol? symbol? symbol? symbol?)
    (match? symbol=? 'hirswca 'rwren 'norh 'eeyno 'eeney)))

(suite alternative-put-into-bag ()
  (test empty-bag
    (alternative-put-into-bag 'mwoc (bag))
    1 (bag?)
    (pipe debag (^et (every symbol?) (match? symbol=? 'mwoc))))
  (test larger
    (alternative-put-into-bag 'eelt (bag 'tee 'mthin 'ttlsa 'uhte))
    1 (bag?)
    (pipe debag (^and (run list length (sect = <> 5))
                      (every symbol?)
                      (every (^or (sect symbol=? <> 'eelt)
                                  (sect symbol=? <> 'tee)
                                  (sect symbol=? <> 'mthin)
                                  (sect symbol=? <> 'ttlsa)
                                  (sect symbol=? <> 'uhte))))))
  (test repeats
    (alternative-put-into-bag 'ecuwam (bag 'pog 'ecuwam))
    1 (bag?)
    (pipe debag
          (^et (every symbol?)
               (pipe list
                     (^and (pipe length (sect = <> 3))
                           (pipe (conditionally-tally
                                   empty-list?
                                   (pipe car (sect symbol=? <> 'ecuwam))
                                   cdr)
                                 (sect = <> 2))
                           (pipe (conditionally-tally
                                   empty-list?
                                   (pipe car (sect symbol=? <> 'pog))
                                   cdr)
                                 (sect = <> 1))))))))

(suite put-into-bag ()
  (test empty-bag
    (put-into-bag 'drnehee (bag))
    1 (bag?)
    (pipe debag (^et (every symbol?) (match? symbol=? 'drnehee))))
  (test larger
    (put-into-bag 'loila (bag 'mcdoio 'nde 'dyhfir))
    1 (bag?)
    (pipe debag (^and (run list length (sect = <> 4))
                      (every symbol?)
                      (every (^or (sect symbol=? <> 'loila)
                                  (sect symbol=? <> 'mcdoio)
                                  (sect symbol=? <> 'nde)
                                  (sect symbol=? <> 'dyhfir))))))
  (test repeats
    (put-into-bag 'ayafyoy (bag 'ayafyoy 'oad 'ayafyoy))
    1 (bag?)
    (pipe debag
          (^et (every symbol?)
               (pipe list
                     (^and (pipe length (sect = <> 4))
                           (pipe (conditionally-tally
                                   empty-list?
                                   (pipe car (sect symbol=? <> 'ayafyoy))
                                   cdr)
                                 (sect = <> 3))
                           (pipe (conditionally-tally
                                   empty-list?
                                   (pipe car (sect symbol=? <> 'oad))
                                   cdr)
                                 (sect = <> 1))))))))

(suite alternative-take-from-bag ()
  (test singleton
    (alternative-take-from-bag (bag 'ilr))
    2 (symbol? bag?)
    (lambda (alpha beta)
      (symbol=? alpha 'ilr)
      (empty-list? ((pipe debag list) beta))))
  (test larger-bag
    (alternative-take-from-bag (bag 'vrdlma 'hsyarrd 'naomhoh))
    2 (symbol? bag?)
    (lambda (alpha beta)
      (pipe (cross-return-all identity debag)
            (^and (run list length (sect = <> 3))
                  (every symbol?)
                  (every (^or (sect symbol=? <> 'vrdlma)
                              (sect symbol=? <> 'hsyarrd)
                              (sect symbol=? <> 'naomhoh))))))))

(suite take-from-bag ()
  (test singleton
    (take-from-bag (bag 'eai))
    2 (symbol? bag?)
    (lambda (alpha beta)
      (symbol=? alpha 'eai)
      (empty-list? ((pipe debag list) beta))))
  (test larger-bag
    (take-from-bag (bag 'srass 'risro 'ahu 'lncdr))
    2 (symbol? bag?)
    (lambda (alpha beta)
      (pipe (cross-return-all identity debag)
            (^and (run list length (sect = <> 4))
                  (every symbol?)
                  (every (^or (sect symbol=? <> 'srass)
                              (sect symbol=? <> 'risro)
                              (sect symbol=? <> 'ahu)
                              (sect symbol=? <> 'lncdr))))))))

(suite empty-bag? ()
  (test empty-bag
    (empty-bag? (bag))
    1 (boolean?)
    true?)
  (test singleton
    (empty-bag? (bag 'lecaryg))
    1 (boolean?)
    not)
  (test larger-bag
    (empty-bag? (bag 'aoano 'ctaol 'nuf 'bnvrae 'ctaol 'aoano))
    1 (boolean?)
    not))

(suite bag=? ((sample (bag 'yneug 'desvoov 'rdh)))
  (test empty-bags
    (bag=? (bag) (bag))
    1 (boolean?)
    true?)
  (test left-bag-empty
    (bag=? (bag) (bag 'rsy 'vdmitr 'aiwo 'ofegyp 'teanm))
    1 (boolean?)
    not)
  (test right-bag-empty
    (bag=? (bag 'ytek 'gsain 'ehl 'abyt) (bag))
    1 (boolean?)
    not)
  (test identical-bag
    (bag=? sample sample)
    1 (boolean?)
    true?)
  (test duplicate-bag
    (bag=? sample (bag 'yneug 'desvoov 'rdh))
    1 (boolean?)
    true?)
  (test different-order
    (bag=? (bag 'rnvnn 'sni 'oojio 'pnygcr)
           (bag 'sni 'pnygcr 'rnvnn 'oojio))
    1 (boolean?)
    true?)
  (test different-sizes
    (bag=? (bag 'aiy 'aiy 'aiy 'aiy 'aiy)
           (bag 'aiy 'aiy 'aiy 'aiy)) 
    1 (boolean?)
    not)
  (test different-duplicate-elements
    (bag=? (bag 'trn 'yam 'yam) (bag 'trn 'trn 'yam))
    1 (boolean?)
    not)
  (test same-duplicate-elements
    (bag=? (bag 'mehahi 'stp 'sftpugo 'mehahi 'stp 'mehahi)
           (bag 'sftpugo 'stp 'stp 'mehahi 'mehahi 'mehahi))
    1 (boolean?)
    true?))

(suite bag-of= ()
  (test generic-equality
    (bag-of= equal?)
    1 (procedure?)
    (lambda (alternative-bag=?)
      (suite generic-equality-internal
          ((sample (bag 'oua 'orshs 'oua 'usea 'sislire)))
        (test empty-bags
          (alternative-bag=? (bag) (bag))
          1 (boolean?)
          true?)
        (test left-bag-empty
          (alternative-bag=? (bag) (bag 'saiivwr 'npinka))
          1 (boolean?)
          not)
        (test right-bag-empty
          (alternative-bag=? (bag 'ndeme 'otarly 'rlt 'houodad) (bag))
          1 (boolean?)
          not)
        (test identical-bag
          (alternative-bag=? sample sample)
          1 (boolean?)
          true?)
        (test duplicate-bag
          (alternative-bag=? sample (bag 'oua 'orshs 'oua 'usea 'sislire))
          1 (boolean?)
          true?)
        (test different-order
          (alternative-bag=? (bag 'mphahe 'tuoteaa 'tsrvsdt 'sth 'sodu)
                             (bag 'sth 'tsrvsdt 'mphahe 'sodu 'tuoteaa))
          1 (boolean?)
          true?)
        (test different-lengths
          (alternative-bag=? (bag 'pgoao 'pgoao 'pgoao)
                             (bag 'pgoao 'pgoao 'pgoao 'pgoao 'pgoao))
          1 (boolean?)
          not)
        (test different-duplicate-elements
          (alternative-bag=? (bag 'hra 'nue 'nue 'eohty)
                             (bag 'hra 'nue 'eohty 'hra))
          1 (boolean?)
          not))))
  (test number-equality
    (bag-of= =)
    1 (procedure?)
    (lambda (alternative-bag-of-numbers=?)
      (suite number-equality-internal ((sample (bag 562 653 262 653)))
        (test both-empty
          (alternative-bag-of-numbers=? (bag) (bag))
          1 (boolean?)
          true?)
        (test left-empty
          (alternative-bag-of-numbers=? (bag) (bag 448))
          1 (boolean?)
          not)
        (test right-empty
          (alternative-bag-of-numbers=? (bag 217 364 555 173 938 361 716)
                                        (bag))
          1 (boolean?)
          not)
        (test identical-bags
          (alternative-bag-of-numbers=? sample sample)
          1 (boolean?)
          true?)
        (test like-bags
          (alternative-bag-of-numbers=? sample (bag 262 562 653 653))
          1 (boolean?)
          true?)
        (test length-mismatch
          (alternative-bag-of-numbers=? (bag 972 903 893 250)
                                        (bag 972 903 893))
          1 (boolean?)
          not)
        (test value-mismatch 
          (alternative-bag-of-numbers=? (bag 172 516 350 742)
                                        (bag 172 516 278 742))
          1 (boolean?)
          not)))))

(suite extract-from-bag ()
  (test empty-bag
    (extract-from-bag values? (bag))
    2 ((^or box? boolean?) bag?)
    (lambda (alpha beta)
      (and (not alpha) (empty-bag? beta))))
  (test middle-of-larger-bag
    (extract-from-bag odd? (bag 832 346 844 385 268))
    2 ((^vel (box-of number?) boolean?) (bag-of number?))
    (lambda (alpha beta)
      (and (box? alpha)
           (= (debox alpha) 385)
           ((bag-of= =) beta (bag 832 346 844 268)))))
  (test multiple-matches
    (extract-from-bag odd? (bag 210 423 619 320 900))
    2 ((^vel (box-of number?) boolean?) (bag-of number?))
    (lambda (alpha beta)
      (and (box? alpha)
           (or (and (= (debox alpha) 423)
                    ((bag-of= =) beta (bag 210 619 320 900)))
               (and (= (debox alpha) 619)
                    ((bag-of= =) beta (bag 210 423 320 900)))))))
  (test no-matches
    (extract-from-bag (constant #f) (bag 'irupghg 'vatne 'snhsp 'teeots))
    2 ((^vel (box-of symbol?) boolean?) (bag-of symbol?))
    (lambda (alpha beta)
      (and (not alpha)
           ((bag-of= symbol=?) beta
                               (bag 'irupghg 'vatne 'snhsp 'teeots))))))

(suite bag-cardinality ()
  (test empty-bag
    (bag-cardinality (bag))
    1 (natural-number?)
    zero?)
  (test small-bag-without-duplicates
    (bag-cardinality (bag 'ntls 'pecte 'enwtd 'uoean 'josueu 'oatsno))
    1 (natural-number?)
    (match? = 6))
  (test small-bag-with-duplicates
    (bag-cardinality (bag 'reaeh 'ldnaa 'hsdl 'fvi 'ldnaa 'adnin 'emss))
    1 (natural-number?)
    (match? = 7))
  (test large-bag
    (bag-cardinality ((ply-natural bag put-into-bag) 292))
    1 (natural-number?)
    (match? = 292)))

(suite for-all-in-bag? ()
  (test empty-bag
    (for-all-in-bag? (constant #f) (bag))
    1 (boolean?)
    true?)
  (test singleton-yes
    (for-all-in-bag? even? (bag 980))
    1 (boolean?)
    true?)
  (test singleton-no
    (for-all-in-bag? even? (bag 933))
    1 (boolean?)
    not)
  (test large-bag-all
    (for-all-in-bag? even? (bag 746 482 466))
    1 (boolean?)
    true?)
  (test large-bag-some
    (for-all-in-bag? even? (bag 338 856 902 761 360))
    1 (boolean?)
    not)
  (test large-bag-none
    (for-all-in-bag? even? (bag 629 589 193 349 695 275))
    1 (boolean?)
    not))

(suite exists-in-bag? ()
  (test empty-bag
    (exists-in-bag? values? (bag))
    1 (boolean?)
    not)
  (test singleton-yes
    (exists-in-bag? even? (bag 330))
    1 (boolean?)
    true?)
  (test singleton-no
    (exists-in-bag? even? (bag 605))
    1 (boolean?)
    not)
  (test large-bag-all
    (exists-in-bag? even? (bag 690 896 222 138))
    1 (boolean?)
    true?)
  (test large-bag-some
    (exists-in-bag? even? (bag 663 813 83 293 227 924 707))
    1 (boolean?)
    true?)
  (test large-bag-none
    (exists-in-bag? even? (bag 411 19 473))
    1 (boolean?)
    not))

(suite fold-bag ()
  (test do-nothing
    (fold-bag (create) black-hole)
    1 (procedure?)
    (lambda (idler)
      (suite do-nothing-internal ()
        (test empty-bag
          (idler (bag))
          0 ())
        (test larger-bag
          (idler (bag 'uwt 'gsas 'teanb 'hehser 'piosrr))
          0 ()))))
  (test add-with-offset
    (fold-bag (create 482) +)
    1 (procedure?)
    (lambda (offset-adder)
      (suite add-with-offset-internal ()
        (test empty-bag
          (offset-adder (bag))
          1 (exact-integer?)
          (match? = 482))
        (test larger-bag
          (offset-adder (bag 765 7 638 592 527))
          1 (exact-integer?)
          (match? = (+ 482 765 7 638 592 527))))))
  (test tally-by-parity
    (fold-bag (create 0 0) (lambda (val evens odds)
                             (if (even? val)
                                 (values (add1 evens) odds)
                                 (values evens (add1 odds)))))
    1 (procedure?)
    (lambda (tallier)
      (suite tally-by-parity-internal ()
        (test empty-bag
          (tallier (bag))
          2 (natural-number? natural-number?)
          (match? = 0 0))
        (test larger-bag
          (tallier (bag 631 596 145 910 51 631 378))
          2 (natural-number? natural-number?)
          (match? = 3 4))))))

(suite bag-union ()
  (test both-empty
    (bag-union (bag) (bag))
    1 (bag?)
    empty-bag?)
  (test left-empty
    (bag-union (bag) (bag 'aeisesh 'hnihiar 'noshoa))
    1 ((bag-of symbol?))
    (match? (bag-of= symbol=?) (bag 'aeisesh 'hnihiar 'noshoa)))
  (test right-empty
    (bag-union (bag 'emgs 'hde) (bag))
    1 ((bag-of symbol?))
    (match? (bag-of= symbol=?) (bag 'emgs 'hde)))
  (test overlapping
    (bag-union (bag 'pntb 'smitset) (bag 'smitset 'dafh))
    1 ((bag-of symbol?))
    (match? (bag-of= symbol=?) (bag 'pntb 'smitset 'smitset 'dafh)))
  (test non-overlapping
    (bag-union (bag  'filwr 'dpary) (bag 'mis 'dmu))
    1 ((bag-of symbol?))
     (match? (bag-of= symbol=?) (bag 'filwr 'dpary 'mis 'dmu))))

(suite map-bag ()
  (test empty-bag
    (map-bag double (bag))
    1 (bag?)
    empty-bag?)
  (test larger-bag
    (map-bag square (bag 87 921 250 226 980 545))
    1 ((bag-of exact-integer?))
    (match? (bag-of= =) (bag (square 87) (square 921) (square 250)
                             (square 226) (square 980) (square 545))))
  (test duplicate-values
    (map-bag (sect mod <> 10) (bag 865 516 326 282 574 393 32 203 347 665))
    1 ((bag-of exact-integer?))
    (match? (bag-of= =) (bag 5 6 6 2 4 3 2 3 7 5)))
  (test duplicates-in-source
    (map-bag double (bag 690 23 690 47 23 943 690))
    1 ((bag-of exact-integer?))
    (match? (bag-of= =) (bag 1380 46 1380 94 46 1886 1380))))

(suite filter-bag ()
  (test empty-bag
    (filter-bag values? (bag))
    1 (bag?)
    empty-bag?)
  (test singleton-bag-satisfying
    (filter-bag even? (bag 228))
    1 ((bag-of number?))
    (match? (bag-of= =) (bag 228)))
  (test singleton-bag-unsatisfying
    (filter-bag even? (bag 885))
    1 ((bag-of number?))
    empty-bag?)
  (test all-elements-removed
    (filter-bag (constant #f) (bag 'iamn 'hsa 'ssr 'edowoi))
    1 ((bag-of symbol?))
    empty-bag?)
  (test some-elements-removed
    (filter-bag even? (bag 662 353 473 288 835 195 977 751 653))
    1 ((bag-of number?))
    (match? (bag-of= =) (bag 662 288)))
  (test no-elements-removed
    (filter-bag values? (bag 'sshe 'drh 'trnno 'bkddm 'hwso))
    1 ((bag-of symbol?))
    (match? (bag-of= symbol=?) (bag 'sshe 'drh 'trnno 'bkddm 'hwso))))

(suite remp-bag ()
  (test empty-bag
    (remp-bag (constant #f) (bag))
    1 (bag?)
    empty-bag?)
  (test singleton-bag-satisfying
    (remp-bag even? (bag 278))
    1 ((bag-of number?))
    empty-bag?)
  (test singleton-bag-unsatisfying
    (remp-bag even? (bag 539))
    1 ((bag-of number?))
    (match? (bag-of= =) (bag 539)))
  (test all-elements-removed
    (remp-bag values? (bag 'dnivaan 'niwt 'qpntt 'dsnm 'saoi))
    1 ((bag-of symbol?))
    empty-bag?)
  (test some-elements-removed
    (remp-bag even? (bag 427 289 824 291 691 9 82 559))
    1 ((bag-of number?))
    (match? (bag-of= =) (bag 427 289 291 691 9 559)))
  (test no-elements-removed
    (remp-bag (constant #f) (bag 'spytnt 'eadln 'eeutg 'othtsn 'museo))
    1 ((bag-of symbol?))
    (match? (bag-of= symbol=?)
            (bag 'spytnt 'eadln 'eeutg 'othtsn 'museo))))

(suite partition-bag ()
  (test empty-bag
    (partition-bag values? (bag))
    2 (bag? bag?)
    (match? bag=? (bag) (bag)))
  (test singleton-bag-satisfying
    (partition-bag even? (bag 664))
    2 ((bag-of number?) (bag-of number?))
    (match? (bag-of= =) (bag 664) (bag)))
  (test singleton-bag-unsatisfying
    (partition-bag even? (bag 479))
    2 ((bag-of number?) (bag-of number?))
    (match? (bag-of= =) (bag) (bag 479)))
  (test all-elements-out
    (partition-bag (constant #f) (bag 'gnhhuhi 'sbihe 'eaeeosi 'penaa))
    2 ((bag-of symbol?) (bag-of symbol?))
    (match? (bag-of= symbol=?)
            (bag)
            (bag 'gnhhuhi 'sbihe 'eaeeosi 'penaa)))
  (test distributed
    (partition-bag even? (bag 166 228 876 917 657 278 852 952 771))
    2 ((bag-of number?) (bag-of number?))
    (match? (bag-of= =)
            (bag 166 228 876 278 852 952)
            (bag 917 657 771)))
  (test all-elements-in
    (partition-bag values? (bag 'kenly 'httew 'tlaen 'yan 'dhetaw))
    2 ((bag-of symbol?) (bag-of symbol?))
    (match? (bag-of= symbol=?)
            (bag 'kenly 'httew 'tlaen 'yan 'dhetaw)
            (bag))))

(suite unfold-bag ()
  (test stops-immediately
    (unfold-bag values? (create 'foo) values)
    1 (procedure?)
    (lambda (nuller)
      (suite stops-immediately-internal ()
        (test run-it
          (nuller)
          1 (bag?)
          empty-bag?))))
  (test typical-case
    (unfold-bag odd? square halve)
    1 (procedure?)
    (lambda (bagger)
      (suite typical-case-internal ()
        (test no-steps
          (bagger 743)
          1 ((bag-of exact-integer?))
          empty-bag?)
        (test several-steps
         (bagger 56)
         1 ((bag-of exact-integer?))
         (match? (bag-of= =) (bag 3136 784 196))))))
  (test first-ten-squares
    ((unfold-bag (sect = <> 10) square add1) 0)
    1 ((bag-of exact-integer?))
    (match? (bag-of= =) (bag 0 1 4 9 16 25 36 49 64 81)))
  (test multiple-arguments
    ((unfold-bag (lambda (a b c) (< a 1))
                 *
                 (lambda (a b c)
                   (if (< 1 c)
                       (values a b (sub1 c))
                       (if (< 1 b)
                           (values a (sub1 b) 5)
                           (values (sub1 a) 4 5)))))
     3 4 5)
    1 ((bag-of exact-integer?))
    (match? (bag-of= =)
            (bag 60 48 36 24 12 45 36 27 18 9 30 24 18 12 6 15 12 9 6 3 40
                 32 24 16 8 30 24 18 12 6 20 16 12 8 4 10 8 6 4 2 20 16 12
                 8 4 15 12 9 6 3 10 8 6 4 2 5 4 3 2 1))))

;;; Test the procedures from the "Equivalence relations" section.

(suite pair-of ()
  (test pair-of-test
    (pair-of symbol? number?)
    1 (procedure?)
    (lambda (classer)
      (suite pair-of-internal ()
        (test yes
          (classer (cons 'optohhi 341))
          1 (boolean?)
          true?)
        (test bad-car
          (classer (cons 905 342))
          1 (boolean?)
          not)
        (test bad-cdr
          (classer (cons 'saihel 'ept))
          1 (boolean?)
          not)
        (test unicorn
          (classer 'innedih)
          1 (boolean?)
          not)))))

(suite pair-of= ()
  (test pair-of=-test
    (pair-of= symbol=? =)
    1 (procedure?)
    (lambda (equaler)
      (suite pair-of=-internal ()
        (test yes
          (equaler (cons 'vtaa 588) (cons 'vtaa 588))
          1 (boolean?)
          true?)
        (test bad-car
          (equaler (cons 'ael 961) (cons 'paohin 961))
          1 (boolean?)
          not)
        (test bad-cdr
          (equaler (cons 'thaohni 793) (cons 'thaohni 175))
          1 (boolean?)
          not)))))

(suite box-of ()
  (test box-of-test
    (box-of symbol?)
    1 (procedure?)
    (lambda (classer)
      (suite box-of-internal ()
        (test yes
          (classer (cons 'gtrhlis '()))
          1 (boolean?)
          true?)
        (test no
          (classer (cons 597 '()))
          1 (boolean?)
          not)
        (test unicorn
          (classer "rilpft")
          1 (boolean?)
          not)))))

(suite box-of= ()
  (test box-of=-test
    (box-of= =)
    1 (procedure?)
    (lambda (equaler)
      (suite box-of=-internal ()
        (test yes
          (equaler (cons 457 '()) (cons 457 '()))
          1 (boolean?)
          true?)
        (test no
          (equaler (cons 329 '()) (cons 397 '()))
          1 (boolean?)
          not)))))

(suite all-alike ()
  (test typical
    (all-alike symbol=?)
    1 (procedure?)
    (lambda (all-symbol=?)
      (suite typical-internal ()
        (test no-arguments
          (all-symbol=?)
          1 (boolean?)
          true?)
        (test one-argument
          (all-symbol=? 'tli)
          1 (boolean?)
          true?)
        (test two-differing
          (all-symbol=? 'vlgdc 'iusa)
          1 (boolean?)
          not)
        (test two-alike
          (all-symbol=? 'rvstca 'rvstca)
          1 (boolean?)
          true?)
        (test long-range-failure
          (all-symbol=? 'gnlyfw 'gnlyfw 'gnlyfw 'gnlyfw 'gnlyfw 'gnlyfw
                        'gnlyfw 'gnlyfw 'gnlyfw 'hneccmn 'gnlyfw 'gnlyfw)
          1 (boolean?)
          not)
        (test many-alike
          (all-symbol=? 'poavp 'poavp 'poavp 'poavp 'poavp 'poavp 'poavp
                        'poavp 'poavp 'poavp 'poavp 'poavp 'poavp
                        'poavp 'poavp 'poavp 'poavp)
          1 (boolean?)
          true?))))
  (test unusual-relation
    (all-alike (pipe + even?))
    1 (procedure?)
    (lambda (all-same-parity?)
      (suite unusual-relation-internal ()
        (test big-bag-of-duplicates
          (all-same-parity? 651 343 579 531 735 979)
          1 (boolean?)
          true?)))))

(suite all-different ()
  (test typical
    (all-different symbol=?)
    1 (procedure?)
    (lambda (none-symbol=?)
      (suite typical-internal ()
        (test no-arguments
          (none-symbol=?)
          1 (boolean?)
          true?)
        (test one-argument
          (none-symbol=? 'tey)
          1 (boolean?)
          true?)
        (test two-matching-arguments
          (none-symbol=? 'euat 'euat)
          1 (boolean?)
          not)
        (test two-differing-arguments
          (none-symbol=? 'esadyn 'sdoe)
          1 (boolean?)
          true?)
        (test long-range-match
          (none-symbol=? 'wiu 'fnfd 'rcmn 'cngavy 'arpdo 'oru 'deosakh
                         'wiu)
          1 (boolean?)
          not)
        (test many-arguments-without-duplicates
          (none-symbol=? 'erhdene 'detc 'ohs 'stotss 'tnt 'mea 'acoare
                         'isuo 'lngve)
          1 (boolean?)
          true?))))
  (test unusual-relation
    (all-different (pipe + odd?))
    1 (procedure?)
    (lambda (no-differing-parity?)
      (suite unusual-relation-internal ()
        (test many-arguments
          (no-differing-parity? 964 490 376 522 688 418 686 336 616)
          1 (boolean?)
          true?))))
  (test remainders-mod-ten
    (all-different (pipe (~each (sect mod <> 10)) =))  ;; example from text
    1 (procedure?)
    (lambda (last-digit-different?)
      (suite remainders-mod-10-internal ()
        (test two-different
          (last-digit-different? 872 656)
          1 (boolean?)
          true?)
        (test two-same
          (last-digit-different? 216 36)
          1 (boolean?)
          not)
        (test remote-duplicate
          (last-digit-different? 174 908 772 116 223 148 320)
          1 (boolean?)
          not)
        (test several-different
          (last-digit-different? 222 859 886 685 561 154 417)
          1 (boolean?)
          true?)))))

;;; Test the procedures from the "Sets" section.

(suite set-classification ()
  (test specific
    (set-classification (pipe (~each (sect mod <> 10)) =))
    1 (procedure?)
    (lambda (classifier)
      (suite specific-internal ()
        (test empty-bag
          (classifier (bag))
          1 (boolean?)
          true?)
        (test distinct-values
          (classifier (bag 62 571 393 558 594 266 530 317 995))
          1 (boolean?)
          true?)
        (test duplicate-values
          (classifier (bag 992 5 117 901 581))
          1 (boolean?)
          not)
        (test unicorn
          (classifier 'ahceo)
          1 (boolean?)
          not)))))

(suite set? ()
  (test empty-bag
    (set? (bag))
    1 (boolean?)
    true?)
  (test distinct-values
    (set? (bag 'trloee 'naadlt))
    1 (boolean?)
    true?)
  (test duplicate-values
    (set? (bag 'oea 'thv 'serhtss 'thv 'kwrauol))
    1 (boolean?)
    not)
  (test unicorn
    (set? 'dsteeer)
    1 (boolean?)
    not))

(suite set-of ()
  (test typical
    (set-of (bag-of integer?) bag=?)
    1 (procedure?)
    (lambda (classifier)
      (suite typical-internal ()
        (test empty-bag
          (classifier (bag))
          1 (boolean?)
          true?)
        (test distinct-values
          (classifier (bag (bag 155 615)
                           (bag)
                           (bag 308 28 498)
                           (bag 267 225)))
          1 (boolean?)
          true?)
        (test duplicate-values
          (classifier (bag (bag 574 808 512 800)
                           (bag 111 199 227 852)
                           (bag 574 808 512 800)
                           (bag 357)))
          1 (boolean?)
          not)
        (test subtle-duplicates
          (classifier (bag (bag 54 499)
                           (bag 374 194)
                           (bag 197 281)
                           (bag 499 54)))
          1 (boolean?)
          not)
        (test unicorn
          (classifier (bag (bag 280 721)
                           (bag)
                           (bag 47 'haltyp)
                           (bag 797 817)))
          1 (boolean?)
          not)))))

(suite set-adjoiner ()
  (test specific
    (set-adjoiner (pipe (~each (sect mod <> 10)) =))
    1 (procedure?)
    (lambda (adjoin)
      (suite specific-internal ()
        (test empty
          (adjoin 679 (bag))
          1 ((set-of exact-integer? =))
          (match? (set-of= =) (bag 679)))
        (test adjoin-new-value
          (adjoin 201 (bag 607 595 729))
          1 ((set-of exact-integer? =))
          (match? (set-of= =) (bag 201 607 595 729)))
        (test adjoin-matching-value
          (adjoin 769 (bag 59))
          1 ((set-of exact-integer? =))
          (match? (set-of= =) (bag 59)))))))

(suite put-into-set ()
  (test empty-set
    (put-into-set 'hmek (bag))
    1 (set?)
    (match? set=? (bag 'hmek)))
  (test adjoin-duplicate-value
    (put-into-set 'yiar (bag 'uis 'oiunt 'yiar 'memu))
    1 (set?)
    (match? set=? (bag 'uis 'oiunt 'yiar 'memu)))
  (test adjoin-new-value
    (put-into-set 'tsiueh (bag 'etwa 'shtei 'rcdwld))
    1 (set?)
    (match? set=? (bag 'tsiueh 'etwa 'shtei 'rcdwld))))

(suite fast-put-into-set ()
  (test empty-set
    (fast-put-into-set 841 (bag))
    1 (set?)
    (match? set=? (bag 841)))
  (test non-empty-set
    (fast-put-into-set 15 (bag 558 403 565 897 226))
    1 (set?)
    (match? set=? (bag 15 558 403 565 897 226))))

(suite set-maker ((same-last-digit? (pipe (~each (sect mod <> 10)) =)))
  (test special
    (set-maker same-last-digit?)
    1 (procedure?)
    (lambda (constructor)
      (suite special-internal ()
        (test no-arguments
          (constructor)
          1 ((set-of integer? same-last-digit?))
          (match? (set-of= same-last-digit?) (bag)))
        (test one-argument
          (constructor 977)
          1 ((set-of integer? same-last-digit?))
          (match? (set-of= same-last-digit?) (bag 977)))
        (test multiple-arguments
          (constructor 986 875 367 383 658 41 355 350 537 261)
          1 ((set-of integer? same-last-digit?))
          (match? (set-of= same-last-digit?) (bag 6 5 7 13 18 11 20)))))))

(suite set ()
  (test no-arguments
    (set)
    1 (set?)
    (match? set=? (bag)))
  (test one-argument
    (set 'watoest)
    1 (set?)
    (match? set=? (bag 'watoest)))
  (test no-duplicates
    (set 'ddodp 'iaahvr 'tttaf 'npuonae 'aiibap 'khf)
    1 (set?)
    (match? set=? (bag 'ddodp 'iaahvr 'tttaf 'npuonae 'aiibap 'khf)))
  (test duplicates
    (set 'etrrht 'tda 'nrryi 'tda 'ivm 'ivm)
    1 (set?)
    (match? set=? (bag 'tda 'nrryi 'ivm 'etrrht))))

(suite deset ()
  (test empty-set
    (deset (set))
    0 ())
  (test singleton
    (deset (set 395))
    1 (exact-integer?)
    (match? = 395))
  (test larger-set
    (deset (set 100 948 933 795 782))
    5 (exact-integer? exact-integer? exact-integer? exact-integer?
       exact-integer?)
    (match? = 100 948 933 795 782)))

(suite take-from-set ()
  (test singleton
    (take-from-set (set 'ntah))
    2 (symbol? set?)
    (lambda (alpha beta)
      (and (symbol=? alpha 'ntah)
           (empty-set? beta))))
  (test several-values
    (take-from-set (set 'hafn 'tecoga 'rmipy))
    2 (symbol? set?)
    (lambda (alpha beta)
      (or (and (symbol=? alpha 'hafn)
               (set=? beta (set 'tecoga 'rmipy)))
          (and (symbol=? alpha 'tecoga)
               (set=? beta (set 'hafn 'rmipy)))
          (and (symbol=? alpha 'rmipy)
               (set=? beta (set 'hafn 'tecoga))))))
  (test sets-contain-no-duplicates
    (take-from-set (set 'hiado 'hiado 'npotys 'npotys 'hiado 'hiado))
    2 (symbol? set?)
    (lambda (alpha beta)
      (or (and (symbol=? alpha 'hiado)
               (set=? beta (set 'npotys)))
          (and (symbol=? alpha 'npotys)
               (set=? beta (set 'hiado)))))))

(suite empty-set? ()
  (test yes
    (empty-set? (set))
    1 (boolean?)
    true?)
  (test no
    (empty-set? (set 'dhnnoof 'sayeei 'tteh 'eebsmcu 'ttcgeym))
    1 (boolean?)
    not))

(suite set=? ((sample (set 'hatna 'erhr 'eei 'cpotair)))
  (test empty-sets
    (set=? (set) (set))
    1 (boolean?)
    true?)
  (test left-set-empty
    (set=? (set) (set 'siuoend 'saseurw 'heeep 'smbdi 'osninnn))
    1 (boolean?)
    not)
  (test right-set-empty
    (set=? (set 'vewtarb 'hro 'hhatto 'rhota) (set))
    1 (boolean?)
    not)
  (test identical-set
    (set=? sample sample)
    1 (boolean?)
    true?)
  (test duplicate-set
    (set=? sample (set 'hatna 'erhr 'eei 'cpotair))
    1 (boolean?)
    true?)
  (test different-order
    (set=? (set 'rnrek 'hctnut 'anil 'ruyeimo 'omty)
           (set 'ruyeimo 'rnrek 'omty 'anil 'hctnut))
    1 (boolean?)
    true?)
  (test duplicating-one-member
    (set=? (set 'rtte 'rtte 'rtte 'rtte 'rtte) (set 'rtte 'rtte))
    1 (boolean?)
    true?)
  (test duplicating-several-nembers
    (set=? (set 'sim 'sim 'ymwidni) (set 'ymwidni 'sim 'ymwidni))
    1 (boolean?)
    true?))

(suite set-of= ((set-of-bags (set-maker bag=?)))
  (test special
    (set-of= bag=?)
    1 (procedure?)
    (lambda (bag-checker)
      (suite set-of=-internal ((sample (set-of-bags (bag 251)
                                                    (bag 283 133 937)
                                                    (bag 813 868))))
        (test empty-sets
          (bag-checker (set-of-bags) (set-of-bags))
          1 (boolean?)
          true?)
        (test left-set-empty
          (bag-checker (set-of-bags)
                       (set-of-bags (bag 342 330 211) (bag 705 280)))
          1 (boolean?)
          not)
        (test right-set-empty
          (bag-checker (set-of-bags (bag 717 125 764 354 404)
                                    (bag)
                                    (bag 935 463)
                                    (bag 783 144 402 70))
                       (set-of-bags))
          1 (boolean?)
          not)
        (test identical-sets
          (bag-checker sample sample)
          1 (boolean?)
          true?)
        (test duplicate-sets
          (bag-checker sample (set-of-bags (bag 251)
                                           (bag 283 133 937)
                                           (bag 813 868)))
          1 (boolean?)
          true?)
        (test rearrangement-of-sets
          (bag-checker (set-of-bags (bag 934 575)
                                    (bag 419 183 629)
                                    (bag 34 784))
                       (set-of-bags (bag 419 183 629)
                                    (bag 34 784)
                                    (bag 934 575)))
          1 (boolean?)
          true?)
        (test rearrangement-of-bags
          (bag-checker (set-of-bags (bag 646 731 525 264)
                                    (bag)
                                    (bag 141 862)
                                    (bag 113 404))
                       (set-of-bags (bag 525 264 646 731)
                                    (bag)
                                    (bag 862 141)
                                    (bag 404 113)))
          1 (boolean?)
          true?)
        (test subtly-different
          (bag-checker (set-of-bags (bag 38 820)
                                    (bag 10 596 944)
                                    (bag 219))
                       (set-of-bags (bag 38 820)
                                    (bag 10 594 944)
                                    (bag 219)))
          1 (boolean?)
          not)))))

(suite fold-set ()
  (test do-nothing
    (fold-set (create) black-hole)
    1 (procedure?)
    (lambda (idler)
      (suite do-nothing-internal ()
        (test do-nothing-briefly
          (idler (set))
          0 ())
        (test do-nothing-gracefully
          (idler (set 778 171 426 420 890 850))
          0 ()))))
  (test add-with-offset
    (fold-set (create 851) +)
    1 (procedure?)
    (lambda (adder)
      (suite add-with-offset-internal ()
        (test empty-set
          (adder (set))
          1 (exact-integer?)
          (match? = 851))
        (test larger-set
          (adder (set 734 834 796 174 461))
          1 (exact-integer?)
          (match? = (+ 851 734 834 796 174 461))))))
  (test tally-by-parity
    (fold-set (create 0 0) (lambda (member evens odds)
                             (if (even? member)
                                 (values (add1 evens) odds)
                                 (values evens (add1 odds)))))
    1 (procedure?)
    (lambda (tallier)
      (suite tally-by-parity-internal ()
        (test empty-set
          (tallier (set))
          2 (natural-number? natural-number?)
          (match? = 0 0))
        (test larger-set
          (tallier (set 882 113 228 581 579 68 977 905 935))
          2 (natural-number? natural-number?)
          (match? = 3 6))))))

(suite set-mapper ((same-last-digit? (pipe (~each (sect mod <> 10)) =)))
  (test specialized
    (set-mapper same-last-digit?)
    1 (procedure?)
    (lambda (mapper)
      (suite specialized-internal ()
        (test empty-set
          (mapper double (set))
          1 ((set-of number? same-last-digit?))
          empty-set?)
        (test larger-set
          (mapper square (set 489 738 303 835 242 849 743 947 7))
          1 ((set-of number? same-last-digit?))
          (match? (set-of= same-last-digit?) (set 1 4 9 5)))))))

(suite set-unfolder ((same-last-digit? (pipe (~each (sect mod <> 10)) =)))
  (test specialized
    (set-unfolder same-last-digit?)
    1 (procedure?)
    (lambda (meta)
      (suite mid-specialized ()
        (test minimal
          (meta values? (create 729) (create))
          1 (procedure?)
          (lambda (idler)
            (suite minimal-internal ()
              (test idler
                (idler)
                1 ((set-of exact-integer? same-last-digit?))
                empty-set?))))
        (test typical
          (meta empty-list? first rest)
          1 (procedure?)
          (lambda (unfolder)
            (suite typical-internal ()
              (test empty-list
                (unfolder (list))
                1 ((set-of exact-integer? same-last-digit?))
                empty-set?)
              (test big-list
                (unfolder (list 971 927 679 68 129 850 40 760 443 143 433))
                1 ((set-of exact-integer? same-last-digit?))
                (match? (set-of= same-last-digit?)
                        ((set-maker same-last-digit?) 1 7 9 8 0 3))))))
        (test multiple-arguments
          (meta (every empty-list?) (pipe (~each first) *) (~each rest))
          1 (procedure?)
          (lambda (unfolder)
            (suite multiple-arguments-internal ()
              (test empty-lists
                (unfolder (list) (list) (list))
                1 ((set-of exact-integer? same-last-digit?))
                empty-set?)
              (test big-lists
                (unfolder (list 623 82 128 204 707)
                          (list 695 123 424 876 68)
                          (list 27 161 321 17 532))
                1 ((set-of exact-integer? same-last-digit?))
                (match? (set-of= same-last-digit?)
                        ((set-maker same-last-digit?) 5 6 2 8))))))))))

(suite map-set ()
  (test empty-set
    (map-set double (set))
    1 (set?)
    empty-set?)
  (test longer-set
    (map-set square (set 526 4 317 115 432 15 191 428))
    1 (set?)
    (match? set=? (set (square 526) (square 4) (square 317) (square 115)
                       (square 432) (square 15) (square 191)
                       (square 428))))
  (test duplicate-values
    (map-set (sect mod <> 10)
             (set 171 584 804 73 624 401 28 742 630 678 554))
    1 (set?)
    (match? set=? (set 1 4 3 8 2 0))))

(suite unfold-set ()
  (test stops-immediately
    (unfold-set values? * values)
    1 (procedure?)
    (lambda (stopper)
      (suite stops-immediately-internal ()
        (test run-it
          (stopper 467 335 492 444)
          1 (set?)
          empty-set?))))
  (test typical
    (unfold-set odd? square halve)
    1 (procedure?)
    (lambda (squarer-of-evens)
      (suite typical-internal ()
        (test empty-set
          (squarer-of-evens 427)
          1 (set?)
          empty-set?)
        (test several-members
          (squarer-of-evens 616)
          1 (set?)
          (match? set=? (set (square 616) (square 308) (square 154)))))))
  (test first-ten-squares
    (unfold-set (sect = <> 10) square add1)
    1 (procedure?)
    (lambda (square-maker)
      (suite first-ten-squares-internal ()
        (test from-zero
          (square-maker 0)
          1 (set?)
          (match? set=? (set 0 1 4 9 16 25 36 49 64 81))))))
  (test compound-state
    (unfold-set (lambda (a b c) (< a 1))
                *
                (lambda (a b c)
                  (if (< 1 c)
                      (values a b (sub1 c))
                      (if (< 1 b)
                          (values a (sub1 b) 5)
                          (values (sub1 a) 4 5)))))
    1 (procedure?)
    (lambda (unfolder)
      (suite compound-state-internal ()
        (test try-it
          (unfolder 3 4 5)
          1 (set?)
          (match? set=? (set 60 48 36 24 12 45 27 18 9 30 6 15 3 40 32 16 8
                             12 20 4 10 2 5 1)))))))

(suite fast-map-set ()
  (test empty-set
    (fast-map-set (constant 642) (set))
    1 (set?)
    empty-set?)
  (test longer-set
    (fast-map-set square (set 755 486 9 378))
    1 (set?)
    (match? set=? (set (square 755) (square 486) (square 9)
                       (square 378)))))

(suite filter-set ()
  (test empty-set
    (filter-set values? (set))
    1 (set?)
    empty-set?)
  (test singleton-set-satisfying
    (filter-set even? (set 234))
    1 (set?)
    (match? set=? (set 234)))
  (test singleton-set-unsatisfying
    (filter-set even? (set 511))
    1 (set?)
    empty-set?)
  (test all-elements-removed
    (filter-set (constant #f)
                (set 'oouonee 'eewernt 'niettn 'hnwmon 'kirna))
    1 (set?)
    empty-set?)
  (test some-elements-removed
    (filter-set even? ((set-maker =) 620 425 653 42 584 745 593 875 903 720
                                     177 424))
    1 ((set-of exact-integer? =))
    (match? (set-of= =) ((set-maker =) 620 42 584 720 424)))
  (test no-elements-removed
    (filter-set values? ((set-maker symbol=?) 'see 'alahoe 'teylrsg
                                                  'cnbss 'toln 'rieeb))
    1 ((set-of symbol? symbol=?))
    (match? (set-of= symbol=?)
            ((set-maker symbol=?) 'see 'alahoe 'teylrsg 'cnbss 'toln
                                  'rieeb))))

(suite remp-set ()
  (test empty-set
    (remp-set (constant #f) (set))
    1 (set?)
    empty-set?)
  (test singleton-set-satisfying
    (remp-set even? (set 418))
    1 (set?)
    empty-set?)
  (test singleton-set-unsatisfying
    (remp-set even? (set 221))
    1 (set?)
    (match? set=? (set 221)))
  (test all-elements-removed
    (remp-set values? (set 'emuih 'crcgnyh 'yatid 'neeun 'fielio))
    1 (set?)
    empty-set?)
  (test some-elements-removed
    (remp-set even? ((set-maker =) 869 222 719 386 757 556 916 154 586 455
                                   580 528))
    1 ((set-of exact-integer? =))
    (match? (set-of= =) ((set-maker =) 869 719 757 455)))
  (test no-elements-removed
    (remp-set (constant #f)
              ((set-maker symbol=?) 'jwygv 'yto 'ittton 'aan 'gnntdp))
    1 ((set-of symbol? symbol=?))
    (match? (set-of= symbol=?)
            ((set-maker symbol=?) 'jwygv 'yto 'ittton 'aan 'gnntdp))))

(suite partition-set ()
  (test empty-set
    (partition-set values? (set))
    2 (set? set?)
    (every empty-set?))
  (test singleton-set-satisfying
    (partition-set even? (set 480))
    2 (set? set?)
    (match? set=? (set 480) (set)))
  (test singleton-set-unsatisfying
    (partition-set even? (set 919))
    2 (set? set?)
    (match? set=? (set) (set 919)))
  (test all-elements-out
    (partition-set (constant #f)
                   ((set-maker symbol=?) 'areeh 'trreet 'rqhc 'edpuo))
    2 ((set-of symbol? symbol=?) (set-of symbol? symbol=?))
    (match? (set-of= symbol=?)
            ((set-maker symbol=?))
            ((set-maker symbol=?) 'areeh 'trreet 'rqhc 'edpuo)))
  (test distributed
    (partition-set even? ((set-maker =) 545 888 203 735 219 211 813 846 537
                                        478 474 785))
    2 ((set-of exact-integer? =) (set-of exact-integer? =))
    (match? (set-of= =) ((set-maker =) 888 846 478 474)
                        ((set-maker =) 545 203 735 219 211 813 537 785)))
  (test all-elements-in
    (partition-set values?
                   ((set-maker symbol=?) 'aohns 'suf 'enienh 'hnee))
    2 ((set-of symbol? symbol=?) (set-of symbol? symbol=?))
    (match? (set-of= symbol=?)
            ((set-maker symbol=?) 'aohns 'suf 'enienh 'hnee)
            ((set-maker symbol=?)))))

(suite cardinality ()
  (test empty-set
    (cardinality (set))
    1 (natural-number?)
    zero?)
  (test small-set  ;; Note: one argument to the constructor is repeated.
    (cardinality (set 'tat 'rtt 'uoae 'rtt 'nsw 'ebihfn 'yttn))
    1 (natural-number?)
    (match? = 6))
  (test large-set
    (cardinality ((lower-ply-natural set fast-put-into-set) 292))
    1 (natural-number?)
    (match? = 292)))

(suite extract-from-set ()
  (test empty-set
    (extract-from-set values? (set))
    2 ((^or box? boolean?) set?)
    (lambda (alpha beta)
      (and (not alpha) (empty-set? beta))))
  (test middle-of-larger-set
    (extract-from-set odd? (set 982 900 841 856 384))
    2 ((^vel (box-of number?) boolean?) (set-of number? =))
    (lambda (alpha beta)
      (and (box? alpha)
           (= (debox alpha) 841)
           ((set-of= =) beta (set 982 900 856 384)))))
  (test multiple-matches
    (extract-from-set odd? (set 738 742 66 595 844 395 316))
    2 ((^vel (box-of number?) boolean?) (set-of number? =))
    (lambda (alpha beta)
      (and (box? alpha)
           (or (and (= (debox alpha) 595)
                    ((set-of= =) beta (set 738 742 66 844 395 316)))
               (and (= (debox alpha) 395)
                    ((set-of= =) beta (set 738 742 66 595 844 316)))))))
  (test no-matches
    (extract-from-set (constant #f) (set 'aottn 'rthhsd 'nfoli 'ninsin))
    2 ((^vel (box-of symbol?) boolean?) (set-of symbol? symbol=?))
    (lambda (alpha beta)
      (and (not alpha)
           ((set-of= symbol=?) beta
                               (set 'aottn 'rthhsd 'nfoli 'ninsin))))))

(suite remove-from-set ()
  (test singleton
    (remove-from-set 'oiwt (set 'oiwt))
    1 (set?)
    empty-set?)
  (test larger-set
    (remove-from-set 'eaee (set 'oiamws 'dlelfa 'eoin 'hirpya 'etedfvn
                                'nesn 'mrh 'eaee 'rotid 'tdatodh))
    1 (set?)
    (match? set=? (set 'oiamws 'dlelfa 'eoin 'hirpya 'etedfvn 'nesn 'mrh
                       'rotid 'tdatodh))))

(suite for-all-in-set?
    ((iota-set (lower-ply-natural set fast-put-into-set)))
  (test empty-set
    (for-all-in-set? (constant #f) (set))
    1 (boolean?)
    true?)
  (test satisfying-singleton
    (for-all-in-set? symbol? (set 'hoae))
    1 (boolean?)
    true?)
  (test unsatisfying-singleton
    (for-all-in-set? exact-integer? (set 'gseas))
    1 (boolean?)
    not)
  (test all-satisfying
    (for-all-in-set? natural-number? (iota-set 705))
    1 (boolean?)
    true?)
  (test none-satisfying
    (for-all-in-set? symbol? (iota-set 20))
    1 (boolean?)
    not)
  (test one-exception
    (for-all-in-set? (^not (sect = <> 662)) (iota-set 842))
    1 (boolean?)
    not))

(suite exists-in-set?
    ((iota-set (lower-ply-natural set fast-put-into-set))) 
  (test empty-set
    (exists-in-set? values? (set))
    1 (boolean?)
    not)
  (test satisfying-singleton
    (exists-in-set? symbol? (set 'oayp))
    1 (boolean?)
    true?)
  (test unsatisfying-singleton
    (exists-in-set? exact-integer? (set 'eow))
    1 (boolean?)
    not)
  (test all-satisfying
    (exists-in-set? natural-number? (iota-set 306))
    1 (boolean?)
    true?)
  (test none-satisfying
    (exists-in-set? symbol? (iota-set 935))
    1 (boolean?)
    not)
  (test one-exception
    (exists-in-set? (sect = <> 83) (iota-set 262))
    1 (boolean?)
    true?))

(suite member? ()
  (test empty-set
    (member? 'eiwtsf (set))
    1 (boolean?)
    not)
  (test first-of-non-empty-set
    (member? 'aalees (set 'aalees 'hdeeo 'etaeow 'aentnt))
    1 (boolean?)
    true?)
  (test last-of-non-empty-set
    (member? 'erhsiw (set 'hsanryo 'rro 'ocddaot 'sakuls 'erhsiw)) 
    1 (boolean?)
    true?)
  (test middle-of-non-empty-set
    (member? 'sewu (set 'utate 'saihu 'aoiyewn 'sewu 'tnttsai 'ersotc))
    1 (boolean?)
    true?)
  (test not-in-non-empty-set
    (member? 'reetio (set 'mnbd 'stil 'iay 'otfsi))
    1 (boolean?)
    not))

(suite set-membership ((same-last-digit? (pipe (~each (sect mod <> 10))
                                               =)))
  (test generic
    (set-membership equal?)
    1 (procedure?)
    (lambda (mem?)
      (suite generic-internal ()
        (test empty-set
          (mem? 443 (set))
          1 (boolean?)
          not)
        (test yes
          (mem? 205 (set 425 695 738 130 205 763 190))
          1 (boolean?)
          true?)
        (test no
          (mem? 690 (set 570 739 57 171))
          1 (boolean?)
          not))))
  (test special
    (set-membership same-last-digit?)
    1 (procedure?)
    (lambda (mem?)
      (suite special-internal ((arigilo (set-maker same-last-digit?)))
        (test empty-set
          (mem? 842 (arigilo))
          1 (boolean?)
          not)
        (test yes
          (mem? 506 (arigilo 370 883 926 300))
          1 (boolean?)
          true?)
        (test no
          (mem? 324 (arigilo 819 348))
          1 (boolean?)
          not)))))

(suite subset? ()
  (test both-empty
    (subset? (set) (set))
    1 (boolean?)
    true?)
  (test left-empty
    (subset? (set) (set 'ahdd 'eedeo 'touonp 'nieperd 'ruod))
    1 (boolean?)
    true?)
  (test right-empty
    (subset? (set 'ooeaah 'tnbieis 'distt) (set))
    1 (boolean?)
    not)
  (test disjoint-sets
    (subset? (set 'hwuux 'mreae 'orn 'irdw 'ngs)
             (set 'orntder 'ahn 'wheht 'mul 'pmseer 'uretat))
    1 (boolean?)
    not)
  (test overlapping
    (subset? (set 'anfa 'poer 'fasai 'ere 'iat)
             (set 'feyoth 'ere 'eikds 'fasai 'etedten))
    1 (boolean?)
    not)
  (test proper-superset
    (subset? (set 'hli 'shyeit 'ssahney 'vaga 'dhc 'suwcute)
             (set 'ssahney 'hli 'dhc))
    1 (boolean?)
    not)
  (test proper-subset
    (subset? (set 'eeey 'rosngad 'hcasnn 'etterd)
             (set 'hcasnn 'rosngad 'heooyuo 'eeey 'omditht 'etterd 'omwf))
    1 (boolean?)
    true?)
  (test equal-sets
    (subset? (set 'shapmil 'nmr 'ftfhaes 'staneor)
             (set 'ftfhaes 'shapmil 'staneor 'nmr))
    1 (boolean?)
    true?))

(suite set-subsethood ((same-last-digit? (pipe (~each (sect mod <> 10))
                                               =)))
  (test generic
    (set-subsethood equal?)
    1 (procedure?)
    (lambda (ss?)
      (suite generic-internal ()
        (test both-empty
          (ss? (set) (set))
          1 (boolean?)
          true?)
        (test left-empty
          (ss? (set) (set 554 163 317 808))
          1 (boolean?)
          true?)
        (test right-empty
          (ss? (set 908 876 636) (set))
          1 (boolean?)
          not)
        (test disjoint
          (ss? (set 381 492 160 127) (set 944 745 619))
          1 (boolean?)
          not)
        (test overlapping
          (ss? (set 695 406 354) (set 695 142 354 494))
          1 (boolean?)
          not)
        (test proper-superset
          (ss? (set 163 597 691 741 753) (set 741 163))
          1 (boolean?)
          not)
        (test proper-subset
          (ss? (set 90 108) (set 417 108 852 90))
          1 (boolean?)
          true?)
        (test equal-sets
          (ss? (set 971 290 801 628 241) (set 241 801 971 628 290))
          1 (boolean?)
          true?))))
  (test special
    (set-subsethood same-last-digit?)
    1 (procedure?)
    (lambda (ss?)
      (suite special-internal ((arigilo (set-maker same-last-digit?)))
        (test both-empty
          (ss? (arigilo) (arigilo))
          1 (boolean?)
          true?)
        (test left-empty
          (ss? (arigilo) (arigilo 99 214 413))
          1 (boolean?)
          true?)
        (test right-empty
          (ss? (arigilo 343 147 478) (arigilo))
          1 (boolean?)
          not)
        (test disjoint
          (ss? (arigilo 108 515) (arigilo 362 476))
          1 (boolean?)
          not)
        (test overlapping
          (ss? (arigilo 735 86) (arigilo 636 844))
          1 (boolean?)
          not)
        (test proper-superset
          (ss? (arigilo 684 359 438 983 635) (arigilo 553 88))
          1 (boolean?)
          not)
        (test proper-subset
          (ss? (arigilo 53 442 344) (arigilo 954 523 308 892))
          1 (boolean?)
          true?)
        (test equal-sets
          (ss? (arigilo 204 726) (arigilo 414 596))
          1 (boolean?)
          true?)))))

(suite set-unioner ((same-last-digit? (pipe (~each (sect mod <> 10)) =)))
  (test special
    (set-unioner same-last-digit?)
    1 (procedure?)
    (lambda (cup)
      (suite special-internal
          ((arigilo (set-maker same-last-digit?))
           (good-set? (set-of exact-integer? same-last-digit?))
           (matcher (set-of= same-last-digit?)))
        (test both-empty
          (cup (arigilo) (arigilo))
          1 (good-set?)
          empty-set?)
        (test left-empty
          (cup (arigilo) (arigilo 386 419 47 263))
          1 (good-set?)
          (match? matcher (arigilo 6 9 7 3)))
        (test right-empty
          (cup (arigilo 609 277 286 993 945) (arigilo))
          1 (good-set?)
          (match? matcher (arigilo 9 7 6 3 5)))
        (test overlapping
          (cup (arigilo 73 158 316 552 541 39 490) (arigilo 742 300 194))
          1 (good-set?)
          (match? matcher (arigilo 3 8 6 2 1 9 0 4)))
        (test disjoint
          (cup (arigilo 600 424) (arigilo 891 613 607))
          1 (good-set?)
          (match? matcher (arigilo 0 4 1 3 7)))))))

(suite union ()
  (test both-empty
    (union (set) (set))
    1 (set?)
    empty-set?)
  (test left-empty
    (union (set) (set 'ashti 'atdstu 'awcae 'hoessc 'imeriec))
    1 (set?)
    (match? set=? (set 'ashti 'atdstu 'awcae 'hoessc 'imeriec)))
  (test right-empty
    (union (set 'yrfs 'wepel 'sten 'say) (set))
    1 (set?)
    (match? set=? (set 'yrfs 'wepel 'sten 'say)))
  (test overlapping
    (union (set 'bdaahmi 'hhstost 'nts 'hieieae 'hdargar)
           (set 'igre 'hieieae 'hgat 'drewsre 'hhstost 'wmbwcct))
    1 (set?)
    (match? set=? (set 'bdaahmi 'hhstost 'nts 'hieieae 'hdargar 'igre 'hgat
                       'drewsre 'wmbwcct))))

(suite fast-union ()
  (test both-empty
    (fast-union (set) (set))
    1 (set?)
    empty-set?)
  (test left-empty
    (fast-union (set) (set 501 273 547))
    1 (set?)
    (match? set=? (set 501 273 547)))
  (test right-empty
    (fast-union (set 614 596 660 240 695) (set))
    1 (set?)
    (match? set=? (set 614 596 660 240 695)))
  (test non-overlapping
    (fast-union (set 369 57 258 782 794 780) (set 632 646 947))
    1 (set?)
    (match? set=? (set 369 57 258 782 794 780 632 646 947))))

(suite intersection ()
  (test left-empty
    (intersection (set) (set 'dsrcw 'snwaa 'rsars 'bohhoh))
    1 (set?)
    empty-set?)
  (test right-empty
    (intersection (set 'hmdt 'etg 'ntteagi 'esnn 'ufr) (set))
    1 (set?)
    empty-set?)
  (test overlapping
    (intersection (set 'seymeei 'potmei 'rrcpe 'apnoies 'sel 'iwd)
                  (set 'rrcpe 'pnmcrh 'iwd 'potmei 'hoha 'yshrr))
    1 (set?)
    (match? set=? (set 'potmei 'rrcpe 'iwd)))
  (test non-overlapping
    (intersection (set 'baoleos 'msneeoa 'nhfcoso) (set 'ttr 'dtomngl))
    1 (set?)
    empty-set?)
  (test equal-sets
    (intersection (set 'uhisn 'lhtiofi 'bta 'raof)
                  (set 'lhtiofi 'raof 'uhisn 'bta))
    1 (set?)
    (match? set=? (set 'lhtiofi 'raof 'uhisn 'bta))))

(suite set-intersectioner
    ((same-last-digit? (pipe (~each (sect mod <> 10)) =)))
  (test generic
    (set-intersectioner equal?)
    1 (procedure?)
    (lambda (cap)
      (suite generic-internal ()
        (test left-empty
          (cap (set) (set 621 369))
          1 (set?)
          empty-set?)
        (test right-empty
          (cap (set 892 332 291) (set))
          1 (set?)
          empty-set?)
        (test overlapping
          (cap (set 398 99 499 431) (set 713 351 431 99))
          1 (set?)
          (match? set=? (set 99 431)))
        (test non-overlapping
          (cap (set 680 322 125) (set 593 122 608 512 385 617))
          1 (set?)
          empty-set?)
        (test equal-sets
          (cap (set 232 880) (set 880 232))
          1 (set?)
          (match? set=? (set 232 880))))))
  (test special
    (set-intersectioner same-last-digit?)
    1 (procedure?)
    (lambda (cap)
      (suite special-internal
          ((arigilo (set-maker same-last-digit?))
           (good-set? (set-of exact-integer? same-last-digit?))
           (matcher (set-of= same-last-digit?)))
        (test left-empty
          (cap (arigilo) (arigilo 413 740))
          1 (good-set?)
          empty-set?)
        (test right-empty
          (cap (arigilo 530 57 918 924 495 388 911 573) (arigilo))
          1 (good-set?)
          empty-set?)
        (test overlapping
          (cap (arigilo 270 123 857 265 476 289 262)
               (arigilo 605 36 204 631 642 500 609))
          1 (good-set?)
          (match? matcher (arigilo 0 5 6 9 2)))
        (test non-overlapping
          (cap (arigilo 939 788 564 457) (arigilo 350 253))
          1 (good-set?)
          empty-set?)
        (test equal-sets
          (cap (arigilo 679 420 567 734 146)
               (arigilo 499 147 70 206 244))
          1 (good-set?)
          (match? matcher (arigilo 9 0 7 4 6)))))))

(suite disjoint? ()
  (test two-empty-sets
    (disjoint? (set) (set))
    1 (boolean?)
    true?)
  (test left-empty-set
    (disjoint? (set) (set 'linouew 'atw 'itta 'des 'iiutm))
    1 (boolean?)
    true?)
  (test right-empty-set
    (disjoint? (set 'bylea 'mlomd 'sedach 'snnieh) (set))
    1 (boolean?)
    true?)
  (test superset
    (disjoint? (set 'oldoie 'pigo 'aam)
               (set 'rnerute 'aeh 'oldoie 'ern 'pigo 'ihgwr 'aam))
    1 (boolean?)
    not)
  (test subset
    (disjoint? (set 'ohpd 'doar 'ihotea 'dafmb 'atbes 'diu 'rtoh 'puetue)
               (set 'ihotea 'dafmb 'atbes 'rtoh))
    1 (boolean?)
    not)
  (test overlapping
    (disjoint? (set 'ocetom 'erycrhr 'sac 'plh 'ddfpttt)
               (set 'psam 'sac 'fioly 'wotohn))
    1 (boolean?)
    not)
  (test non-overlapping
    (disjoint? (set 'aec 'hthioe 'innd 'ottgtn 'mhi)
               (set 'nemrdae 'ooe 'saaem 'ihy 'srrad 'eethe 'emne))
    1 (boolean?)
    true?))

(suite set-disjointness
    ((same-last-digit? (pipe (~each (sect mod <> 10)) =)))
  (test generic
    (set-disjointness equal?)
    1 (procedure?)
    (lambda (disj?)
      (suite generic-internal ()
        (test empty-sets
          (disj? (set) (set))
          1 (boolean?)
          true?)
        (test left-empty
          (disj? (set) (set 317 428 620))
          1 (boolean?)
          true?)
        (test right-empty
          (disj? (set 293) (set))
          1 (boolean?)
          true?)
        (test overlapping
          (disj? (set 196 515 552 977) (set 552 251 196))
          1 (boolean?)
          not)
        (test non-overlapping
          (disj? (set 811 491 70 499 93) (set 7 12 853))
          1 (boolean?)
          true?))))
  (test special
    (set-disjointness same-last-digit?)
    1 (procedure?)
    (lambda (disj?)
      (suite special-internal ((arigilo (set-maker same-last-digit?)))
        (test empty-sets
          (disj? (arigilo) (arigilo))
          1 (boolean?)
          true?)
        (test left-empty
          (disj? (arigilo) (arigilo 68 95 463))
          1 (boolean?)
          true?)
        (test right-empty
          (disj? (arigilo 86 942 998 313) (arigilo))
          1 (boolean?)
          true?)
        (test overlapping
          (disj? (arigilo 545 77 901 372)
                 (arigilo 179 921 707 655 378 164))
          1 (boolean?)
          not)
        (test non-overlapping
          (disj? (arigilo 855 477 279) (arigilo 336 568))
          1 (boolean?)
          true?)
        (test equal-as-sets
          (disj? (arigilo 155) (arigilo 225))
          1 (boolean?)
          not)))))

(suite set-difference ()
  (test left-empty
    (set-difference (set) (set 'snna 'afgrr 'tltophe 'luplr))
    1 (set?)
    empty-set?)
  (test right-empty
    (set-difference (set 'mtr 'ild 'ynnsrwl 'rads) (set))
    1 (set?)
    (match? set=? (set 'mtr 'ild 'ynnsrwl 'rads)))
  (test overlapping
    (set-difference (set 'anetg 'hthsao 'ldm 'niae 'nnrm 'ngoh)
                    (set 'evo 'niae 'anetg 'ryehede 'ient))
    1 (set?)
    (match? set=? (set 'hthsao 'ldm 'nnrm 'ngoh)))
  (test non-overlapping
    (set-difference (set 'nmtegt 'raa 'oneha 'inat)
                    (set 'ewuire 'akawwi 'rykeme 'ser 'eiyee))
    1 (set?)
    (match? set=? (set 'nmtegt 'raa 'oneha 'inat)))
  (test equal-sets
    (set-difference (set 'sota 'naawnch 'nnishg 'drcdat 'llay)
                    (set 'nnishg 'llay 'sota 'naawnch 'drcdat))
    1 (set?)
    empty-set?))

(suite set-differencer
    ((same-last-digit? (pipe (~each (sect mod <> 10)) =)))
  (test generic
    (set-differencer equal?)
    1 (procedure?)
    (lambda (differ)
      (suite generic-internal ()
        (test left-empty
          (differ (set) (set 576 267 515))
          1 (set?)
          empty-set?)
        (test right-empty
          (differ (set 773 288) (set))
          1 (set?)
          (match? set=? (set 288 773)))
        (test overlapping
          (differ (set 852 586 9 743 748) (set 2 586 81 989 9 852))
          1 (set?)
          (match? set=? (set 748 743)))
        (test non-overlapping
          (differ (set 296 93 908) (set 333 545 605 285))
          1 (set?)
          (match? set=? (set 296 93 908)))
        (test take-away-all
          (differ (set 558 10 525) (set 799 10 393 525 558 583))
          1 (set?)
          empty-set?))))
  (test special
    (set-differencer same-last-digit?)
    1 (procedure?)
    (lambda (differ)
      (suite special-internal
          ((arigilo (set-maker same-last-digit?))
           (good-set? (set-of exact-integer? same-last-digit?))
           (matcher (set-of= same-last-digit?)))
        (test left-empty
          (differ (arigilo) (arigilo 369 840 481 663))
          1 (good-set?)
          empty-set?)
        (test right-empty
          (differ (arigilo 980 952 793 929 156 911) (arigilo))
          1 (good-set?)
          (match? matcher (arigilo 0 2 3 9 6 1)))
        (test overlapping
          (differ (arigilo 182 538 19) (arigilo 409 493 86 128))
          1 (good-set?)
          (match? matcher (arigilo 2)))
        (test non-overlapping
          (differ (arigilo 356 229 951) (arigilo 367 915 274))
          1 (good-set?)
          (match? matcher (arigilo 6 9 1)))
        (test take-all
          (differ (arigilo 411 239) (arigilo 211 572 119))
          1 (good-set?)
          empty-set?)))))

;;; Test the procedures from the "Tables" section.

(suite table? ()
  (test empty-table
    (table? (bag))
    1 (boolean?)
    true?)
  (test singleton
    (table? (bag (cons 'ohlt 'vrrwges)))
    1 (boolean?)
    true?)
  (test larger-table
    (table? (bag (cons 'tnr 'wehsw)
                 (cons 'stmd 'ignx)
                 (cons 'loavbio 'yknaawi)
                 (cons 'wgt 'csrda)))
    1 (boolean?)
    true?)
  (test random-value
    (table? 'lob)
    1 (boolean?)
    not)
  (test non-pair-in-bag
    (table? (bag (cons 'resfy 'wadoh)
                 'khbte
                 (cons 'uorll 'otrla)))
    1 (boolean?)
    not)
  (test bag-with-duplicate-keys
    (table? (bag (cons 'ggedeoa 'ovhfve)
                 (cons 'froisov 'dmi)
                 (cons 'ggedeoa 'gkfnh)))
    1 (boolean?)
    not)
  (test bag-with-duplicate-associations
    (table? (bag (cons 'tvr 'diud)
                 (cons 'stavpi 'efyae)
                 (cons 'tvr 'diud)))
    1 (boolean?)
    not))

(suite table-of ((same-last-digit? (pipe (~each (sect mod <> 10)) =)))
  (test generic
    (table-of values? values? equal?)
    1 (procedure?)
    (lambda (tab?)
      (suite generic-internal ()
        (test empty-table
          (tab? (bag))
          1 (boolean?)
          true?)
        (test singleton
          (tab? (bag (cons 'soonegd 'vaofl)))
          1 (boolean?)
          true?)
        (test larger-table
          (tab? (bag (cons 'ltw 'blpnbe)
                     (cons 'rhtwnv 'sidhd)
                     (cons 'qamtt 'ogwooeh)
                     (cons 'etvrf 'fcsedr)
                     (cons 'rar 'rerb)))
          1 (boolean?)
          true?)
        (test random-value
          (tab? 303)
          1 (boolean?)
          not)
        (test non-pair-in-bag
          (tab? (bag (cons 'rnw 'seeddd)
                     113
                     (cons 'yst 'veaao)))
          1 (boolean?)
          not)
        (test bag-with-duplicate-keys
          (tab? (bag (cons 'fntt 'ornt)
                     (cons 'obsss 'ttdo)
                     (cons 'obsss 'heeonv)
                     (cons 'aooee 'lsrha)))
          1 (boolean?)
          not)
        (test bag-with-duplicate-associations
          (tab? (bag (cons 'aee 'eret)
                     (cons 'eueie 'smott)
                     (cons 'tet 'sda)
                     (cons 'eueie 'smott)))
          1 (boolean?)
          not))))
  (test special
    (table-of exact-integer? symbol? same-last-digit?)
    1 (procedure?)
    (lambda (tab?)
      (suite special-internal ()
        (test empty-table
          (tab? (bag))
          1 (boolean?)
          true?)
        (test singleton
          (tab? (bag (cons 684 'iheanas)))
          1 (boolean?)
          true?)
        (test larger-bag
          (tab? (bag (cons 797 'eouoe)
                     (cons 396 'rismian)
                     (cons 511 'anrrdt)
                     (cons 362 'eoogod)
                     (cons 70 'fil)))
          1 (boolean?)
          true?)
        (test random-value
          (tab? 947/955)
          1 (boolean?)
          not)
        (test non-pair-in-bag
          (tab? (bag (cons 380 'dyddo)
                     (cons 381 'ndrw)
                     791/155
                     (cons 990 'lhrt)))
          1 (boolean?)
          not)
        (test wrong-type-key
          (tab? (bag (cons 627 'hsrl)
                     (cons 'hmt 'nehrtae)
                     (cons 388 'henelo)))
          1 (boolean?)
          not)
        (test wrong-type-entry
          (tab? (bag (cons 494 'eghmj)
                     (cons 6 'taonont)
                     (cons 345 883)
                     (cons 391 'lshol)))
          1 (boolean?)
          not)
        (test duplicate-key
          (tab? (bag (cons 503 'eyuwf)
                     (cons 647 'ioyfor)
                     (cons 454 'mrnefno)
                     (cons 492 'inhe)
                     (cons 311 'nra)
                     (cons 663 'mll)))
          1 (boolean?)
          not)
        (test duplicate-association
          (tab? (bag (cons 835 'veenaui)
                     (cons 884 'asspu)
                     (cons 715 'veenaui)
                     (cons 9 'vtv)))
          1 (boolean?)
          not)))))

(suite fast-put-into-table ((table=? (bag-of= pair=?)))
  (test empty-table
    (fast-put-into-table 'admnr 'amest (bag))
    1 (table?)
    (match? table=? (bag (cons 'admnr 'amest))))
  (test larger-table
    (fast-put-into-table 'osah 'emh (bag (cons 'ueoab 'sedeu)
                                         (cons 'maf 'mwei)
                                         (cons 'nryeo 'dea)
                                         (cons 'klwber 'oeereny)))
    1 (table?)
    (match? table=? (bag (cons 'osah 'emh)
                         (cons 'ueoab 'sedeu)
                         (cons 'maf 'mwei)
                         (cons 'nryeo 'dea)
                         (cons 'klwber 'oeereny)))))

(suite put-into-table ((table=? (bag-of= pair=?)))
  (test empty-table
    (put-into-table 'stma 'onsk (bag))
    1 (table?)
    (match? table=? (bag (cons 'stma 'onsk))))
  (test new-key
    (put-into-table 'wkg 'eei (bag (cons 'idenhg 'hyo)
                                   (cons 'bbst 'yhsyug)
                                   (cons 'ahej 'irt)))
    1 (table?)
    (match? table=? (bag (cons 'wkg 'eei)
                         (cons 'idenhg 'hyo)
                         (cons 'bbst 'yhsyug)
                         (cons 'ahej 'irt))))
  (test duplicate-key
    (put-into-table 'crardl 'osnamiw (bag (cons 'gvrrwt 'aeuy)
                                          (cons 'tiaeni 'eyocrnr)
                                          (cons 'crardl 'ydtci)
                                          (cons 'reeiao 'otcroe)))
    1 (table?)
    (match? table=? (bag  (cons 'gvrrwt 'aeuy)
                          (cons 'tiaeni 'eyocrnr)
                          (cons 'crardl 'osnamiw)
                          (cons 'reeiao 'otcroe)))))

(suite table-adjoiner
    ((same-last-digit? (pipe (~each (sect mod <> 10)) =)))
  (test generic
    (table-adjoiner equal?)
    1 (procedure?)
    (lambda (adj)
      (suite generic-internal
          ((good-table? (table-of values? values? equal?))
           (matcher (bag-of= (pair-of= equal? equal?))))
        (test empty-table
          (adj 'asnef 'hage (bag))
          1 (good-table?)
          (match? matcher (bag (cons 'asnef 'hage))))
        (test larger-table
          (adj 'rsk 'dletsee (bag (cons 'ttmbs 'genw)
                                  (cons 'emamone 'oao)))
          1 (good-table?)
          (match? matcher (bag (cons 'rsk 'dletsee)
                               (cons 'ttmbs 'genw)
                               (cons 'emamone 'oao))))
        (test duplicate-key
          (adj 'ourks 'rre (bag (cons 'oehi 'alwp)
                                (cons 'ourks 'ahmth)
                                (cons 'swhlrdr 'erkano)))
          1 (good-table?)
          (match? matcher  (bag (cons 'oehi 'alwp)
                                (cons 'ourks 'rre)
                                (cons 'swhlrdr 'erkano)))))))
  (test specific
    (table-adjoiner same-last-digit?)
    1 (procedure?)
    (lambda (adj)
      (suite specific-internal
          ((good-table? (table-of exact-integer? symbol? same-last-digit?))
           (matcher (bag-of= (pair-of= same-last-digit? symbol=?))))
        (test empty-table
          (adj 907 'tsnnetr (bag))
          1 (good-table?)
          (match? matcher (bag (cons 7 'tsnnetr))))
        (test larger-table
          (adj 751 'rmpit (bag (cons 0 'isdtvh)
                               (cons 4 'btif)
                               (cons 9 'gnso)))
          1 (good-table?)
          (match? matcher (bag (cons 1 'rmpit)
                               (cons 0 'isdtvh)
                               (cons 4 'btif)
                               (cons 9 'gnso))))
        (test duplicate-key
          (adj 848 'reroete (bag (cons 712 'twiaaus)
                                 (cons 19 'dokrnvs)
                                 (cons 118 'hre)
                                 (cons 743 'gcshha)))
          1 (good-table?)
          (match? matcher (bag (cons 2 'twiaaus)
                               (cons 9 'dokrnvs)
                               (cons 8 'reroete)
                               (cons 3 'gcshha))))))))

(suite table ((table=? (bag-of= pair=?)))
  (test no-arguments
    (table)
    1 (table?)
    (match? table=? (bag)))
  (test several-arguments
    (table (cons 'mheatg 'drl)
           (cons 'cwuttf 'rhpwlc)
           (cons 'ytnnda 'aatrrs)
           (cons 'tdoe 'nha))
    1 (table?)
    (match? table=? (bag (cons 'mheatg 'drl)
                         (cons 'cwuttf 'rhpwlc)
                         (cons 'ytnnda 'aatrrs)
                         (cons 'tdoe 'nha)))))

(suite lookup ()
  (test minimal
    (lookup 405 (table))
    1 (boolean?)
    not)
  (test empty-table-with-default
    (lookup 'fhicsr (table) 'deeyi)
    1 (symbol?)
    (match? symbol=? 'deeyi))
  (test entry-from-singleton
    (lookup 'kisteo (table (cons 'kisteo 'tomdrp)) 'eshnt)
    1 (symbol?)
    (match? symbol=? 'tomdrp))
  (test entry-from-larger-table
    (lookup 'einl
            (table (cons 'int 'gddo)
                   (cons 'einl 'clm)
                   (cons 'aaodefa 'ofnjok))
            'byhc)
    1 (symbol?)
    (match? symbol=? 'clm))
  (test not-found
    (lookup 'nowd
            (table (cons 'wioeou 'eapiabo) 
                   (cons 'obo 'pdo)
                   (cons 'etadeah 'ohahn)
                   (cons 'wheuv 'aeoo)))
    1 (boolean?)
    not)
  (test not-found-with-default
    (lookup 'geh
            (table (cons 'ihn 'oslh)
                   (cons 'npkew 'aopt)
                   (cons 'dyhc 'tpoco))
            'mednto)
    1 (symbol?)
    (match? symbol=? 'mednto)))

(suite table-searcher ((same-last-digit? (pipe (~each (sect mod <> 10)) =)))
  (test generic
    (table-searcher equal?)
    1 (procedure?)
    (lambda (searcher)
      (suite generic-internal ()
        (test minimal
          (searcher 'tvhi (table))
          1 (boolean?)
          not)
        (test empty-table-with-default
          (searcher 345 (table) 560)
          1 (exact-integer?)
          (match? = 560))
        (test entry-from-singleton
          (searcher 945 (table (cons 945 62)) 632)
          1 (exact-integer?)
          (match? = 62))
        (test entry-from-larger-table
          (searcher 288
                    (table (cons 288 762)
                           (cons 747 290)
                           (cons 293 422)
                           (cons 486 678))
                    1)
          1 (exact-integer?)
          (match? = 762))
        (test not-found
          (searcher 75
                    (table (cons 301 917)
                           (cons 322 501)
                           (cons 709 483)))
          1 (boolean?)
          not)
        (test not-found-with-default
          (searcher 891
                    (table (cons 722 459)
                           (cons 469 474)
                           (cons 452 814)
                           (cons 814 140))
                    209)
          1 (exact-integer?)
          (match? = 209)))))
  (test specific
    (table-searcher same-last-digit?)
    1 (procedure?)
    (lambda (searcher)
      (suite specific-internal ()
        (test minimal
          (searcher 164 (table))
          1 (boolean?)
          not)
        (test empty-table-with-default
          (searcher 394 (table) 'moknnep)
          1 (symbol?)
          (match? symbol=? 'moknnep))
        (test entry-from-singleton
          (searcher 927 (table (cons 157 'rsh)) 'nup)
          1 (symbol?)
          (match? symbol=? 'rsh))
        (test entry-from-larger-table
          (searcher 918
                    (table (cons 306 'othdo)
                           (cons 475 'adag)
                           (cons 78 'dasgoio)
                           (cons 797 'ttmaeti))
                    'hohaao)
          1 (symbol?)
          (match? symbol=? 'dasgoio))
        (test not-found
          (searcher 390
                    (table (cons 155 'tnlne)
                           (cons 264 'ieneau)
                           (cons 499 'rasn)))
          1 (boolean?)
          not)
        (test not-found-with-default
          (searcher 188
                    (table (cons 980 'iof)
                           (cons 961 'tde)
                           (cons 447 'ifepaex)
                           (cons 405 'ertr))
                    'oianap)
          1 (symbol?)
          (match? symbol=? 'oianap))))))

(suite table-keys ()
  (test empty-table
    (table-keys (table))
    1 (set?)
    empty-set?)
  (test non-empty-table
    (table-keys (table (cons 'oodmr 'hteegec)
                       (cons 'lttvw 'rodptiy)
                       (cons 'ynsn 'vgf)))
    1 (set?)
    (match? set=? (set 'oodmr 'lttvw 'ynsn))))

(suite table-entries ()
  (test empty-table
    (table-entries (table))
    1 (bag?)
    empty-bag?)
  (test non-empty-table
    (table-entries (table (cons 'dir 'enurti)
                          (cons 'ancwd 'llsee)
                          (cons 'iolya 'lbt)
                          (cons 'hrlc 'mefbed)
                          (cons 'ncrr 'oanb)
                          (cons 'oittng 'thmv)))
    1 (bag?)
    (match? bag=? (bag 'enurti 'llsee 'lbt 'mefbed 'oanb 'thmv)))
  (test duplicate-entries
    (table-entries (table (cons 'lesdh 'mnalao)
                          (cons 'aawnwue 'tmseyes)
                          (cons 'soeaar 'reotoep)
                          (cons 'she 'tmseyes)))
    1 (bag?)
    (match? bag=? (bag 'mnalao 'tmseyes 'reotoep 'tmseyes))))

(suite delete-from-table ((table=? (bag-of= pair=?)))
  (test empty-table
    (delete-from-table 'rkoe (table))
    1 (table?)
    (match? table=? (table)))
  (test singleton-table
    (delete-from-table 'eetuso (table (cons 'eetuso 'eeat)))
    1 (table?)
    (match? table=? (table)))
  (test larger-table
    (delete-from-table 'hsri (table (cons 'snsmfcr 'erkemmi)
                                    (cons 'eeihn 'dtp)
                                    (cons 'hsri 'nhyowoe)
                                    (cons 'eldt 'iritoa)))
    1 (table?)
    (match? table=? (table (cons 'snsmfcr 'erkemmi)
                           (cons 'eeihn 'dtp)
                           (cons 'eldt 'iritoa))))
  (test unmatched-key
    (delete-from-table 'tbthoea (table (cons 'wealusr 'wrhch)
                                       (cons 'rerw 'ivo)
                                       (cons 'mitilm 'ces)
                                       (cons 'kmee 'aeht)))
    1 (table?)
    (match? table=? (table (cons 'wealusr 'wrhch)
                                       (cons 'rerw 'ivo)
                                       (cons 'mitilm 'ces)
                                       (cons 'kmee 'aeht)))))

(suite table-deleter ((same-last-digit? (pipe (~each (sect mod <> 10)) =)))
  (test generic
    (table-deleter equal?)
    1 (procedure?)
    (lambda (del)
      (suite generic-internal ((table=? (bag-of= pair=?)))
        (test empty-table
          (del 221 (table))
          1 (table?)
          (match? table=? (table)))
        (test singleton-table
          (del 100 (table (cons 100 295)))
          1 (table?)
          (match? table=? (table)))
        (test larger-table
          (del 495 (table (cons 508 824)
                          (cons 962 720)
                          (cons 495 727)))
          1 (table?)
          (match? table=? (table (cons 508 824) (cons 962 720))))
        (test unmatched-key
          (del 739 (table (cons 642 925)
                          (cons 190 104)
                          (cons 669 401)
                          (cons 826 672)
                          (cons 20 551)))
          1 (table?)
          (match? table=? (table (cons 642 925)
                                 (cons 190 104)
                                 (cons 669 401)
                                 (cons 826 672)
                                 (cons 20 551)))))))
  (test specific
    (table-deleter same-last-digit?)
    1 (procedure?)
    (lambda (del)
      (suite specific-internal
          ((good-table? (table-of exact-integer? symbol? same-last-digit?))
           (matcher (bag-of= (pair-of= same-last-digit? symbol=?))))
        (test empty-table
          (del 486 (table))
          1 (good-table?)
          (match? matcher (table)))
        (test singleton-table
          (del 997 (table (cons 357 'otud)))
          1 (good-table?)
          (match? matcher (table)))
        (test larger-table
          (del 611 (table (cons 768 'hntol)
                          (cons 710 'ritas)
                          (cons 41 'pteolw)
                          (cons 543 'aene)
                          (cons 842 'aaeoti)))
          1 (good-table?)
          (match? matcher (table (cons 8 'hntol)
                                 (cons 0 'ritas)
                                 (cons 3 'aene)
                                 (cons 2 'aaeoti))))
        (test unmatched-key
          (del 739 (table (cons 632 'nlg)
                          (cons 238 'tseamr)
                          (cons 670 'weeo)))
          1 (good-table?)
          (match? matcher (table (cons 2 'nlg)
                                 (cons 8 'tseamr)
                                 (cons 0 'weeo))))))))

(suite table-size ()
  (test zero
    (table-size (table))
    1 (natural-number?)
    zero?)
  (test larger
    (table-size (table (cons 'toneyw 'sohodiy)
                       (cons 'remnat 'hdxetod)
                       (cons 'trroey 'melrn)
                       (cons 'euong 'trmhnhy)
                       (cons 'adnhd 'rnho)
                       (cons 'aihetmr 'rseple)
                       (cons 'lrin 'etbu)))
    1 (natural-number?)
    (match? = 7)))

(suite table=? ((sample (table (cons 'raee 'sneye)
                               (cons 'ihca 'mset)
                               (cons 'cwphnee 'dea))))
  (test both-empty
    (table=? (table) (table))
    1 (boolean?)
    true?)
  (test empty-and-non-empty
    (table=? (table) (table (cons 'esl 'unmsi)))
    1 (boolean?)
    not)
  (test identical-tables
    (table=? sample sample)
    1 (boolean?)
    true?)
  (test duplicate-tables
    (table=? sample (table (cons 'raee 'sneye)
                           (cons 'ihca 'mset)
                           (cons 'cwphnee 'dea)))
    1 (boolean?)
    true?)
  (test permuted-tables
    (table=? (table (cons 'tmh 'mal)
                    (cons 'tlnhs 'liiir)
                    (cons 'ooatiei 'afbn))
             (table (cons 'ooatiei 'afbn)
                    (cons 'tmh 'mal)
                    (cons 'tlnhs 'liiir)))
    1 (boolean?)
    true?)
  (test swapped-entries
    (table=? (table (cons 'epd 'snthsto)
                    (cons 'aaanih 'inwdert)
                    (cons 'sddtl 'usyw))
             (table (cons 'epd 'inwdert)
                    (cons 'aaanih 'snthsto)
                    (cons 'sddtl 'usyw)))
    1 (boolean?)
    not)
  (test different-entry
    (table=? (table (cons 'nrtuo 'htdi)
                    (cons 'rgrhuar 'oid)
                    (cons 'ntcotr 'hdsh))
             (table (cons 'nrtuo 'htdi)
                    (cons 'rgrhuar 'hhesetd)
                    (cons 'ntcotr 'hdsh)))
    1 (boolean?)
    not)
  (test extra-key
    (table=? (table (cons 'iimaggn 'fhhpc)
                    (cons 'hbdrs 'iaihsr)
                    (cons 'uyaodoh 'nroa)
                    (cons 'iaem 'elh))
             (table (cons 'iimaggn 'fhhpc)
                    (cons 'hbdrs 'iaihsr)
                    (cons 'iaem 'elh)))
    1 (boolean?)
    not)
  (test missing-key
    (table=? (table (cons 'needaif 'idse)
                    (cons 'gfpdhsh 'ftllye)
                    (cons 'illlrse 'ddn))
             (table (cons 'needaif 'idse)
                    (cons 'ifist 'ndatcf)
                    (cons 'gfpdhsh 'ftllye)
                    (cons 'illlrse 'ddn)))
    1 (boolean?)
    not)
  (test unrelated-key
    (table=? (table (cons 'eiephon 'ixutnpw)
                    (cons 'ehpa 'esatehd)
                    (cons 'urdxi 'hron))
             (table (cons 'eiephon 'ixutnpw)
                    (cons 'fthtt 'esatehd)
                    (cons 'urdxi 'hron)))
    1 (boolean?)
    not))

(suite table-of= ((same-last-digit? (pipe (~each (sect mod <> 10)) =)))
  (test generic
    (table-of= equal? equal?)
    1 (procedure?)
    (lambda (tab=?)
      (suite generic-internal ((sample (table (cons 10 897)
                                              (cons 778 648)
                                              (cons 511 264)
                                              (cons 860 142))))
        (test both-empty
          (tab=? (table) (table))
          1 (boolean?)
          true?)
        (test empty-and-non-empty
          (tab=? (table) (table (cons 111 0)))
          1 (boolean?)
          not)
        (test identical-tables
          (tab=? sample sample)
          1 (boolean?)
          true?)
        (test duplicate-tables
          (tab=? sample (table (cons 10 897)
                               (cons 778 648)
                               (cons 511 264)
                               (cons 860 142)))
          1 (boolean?)
          true?)
        (test permuted-tables
          (tab=? (table (cons 276 728)
                        (cons 258 724)
                        (cons 866 251)
                        (cons 549 715))
                 (table (cons 258 724)
                        (cons 866 251)
                        (cons 276 728)
                        (cons 549 715)))
          1 (boolean?)
          true?)
        (test swapped-entries
          (tab=? (table (cons 995 457)
                        (cons 859 488))
                 (table (cons 995 488)
                        (cons 859 457)))
          1 (boolean?)
          not)
        (test different-entry
          (tab=? (table (cons 230 604)
                        (cons 577 844)
                        (cons 1 556))
                 (table (cons 230 604)
                        (cons 577 749)
                        (cons 1 556)))
          1 (boolean?)
          not)
        (test extra-key
          (tab=? (table (cons 151 491)
                        (cons 743 407)
                        (cons 332 606)
                        (cons 874 605))
                 (table (cons 151 491)
                        (cons 874 605)
                        (cons 332 606)))
          1 (boolean?)
          not)
        (test unrelated-key
          (tab=? (table (cons 522 508)
                        (cons 858 382))
                 (table (cons 31 508)
                        (cons 858 382)))
          1 (boolean?)
          not))))
  (test specific
    (table-of= same-last-digit? symbol=?)
    1 (procedure?)
    (lambda (tab=?)
      (suite specific-internal ((sample (table (cons 220 'etniah)
                                               (cons 573 'tti))))
        (test both-empty
          (tab=? (table) (table))
          1 (boolean?)
          true?)
        (test empty-and-non-empty
          (tab=? (table (cons 264 'ouoetfa) (cons 760 'eatbnoo))
                 (table))
          1 (boolean?)
          not)
        (test identical-tables
          (tab=? sample sample)
          1 (boolean?)
          true?)
        (test permuted-tables
          (tab=? sample (table (cons 743 'tti) (cons 560 'etniah)))
          1 (boolean?)
          true?)
        (test swapped-entries
          (tab=? (table (cons 712 'ahh)
                        (cons 848 'oapttrr)
                        (cons 19 'tlines))
                 (table (cons 62 'tlines)
                        (cons 118 'oapttrr)
                        (cons 69 'ahh)))
          1 (boolean?)
          not)
        (test different-entry
          (tab=? (table (cons 405 'mpe)
                        (cons 632 'tsi))
                 (table (cons 345 'oemu)
                        (cons 762 'tsi)))
          1 (boolean?)
          not)
        (test extra-key
          (tab=? (table (cons 945 'alt)
                        (cons 288 'utra)
                        (cons 747 'sos))
                 (table (cons 678 'utra)
                        (cons 147 'sos)))
          1 (boolean?)
          not)
        (test unrelated-key
          (tab=? (table (cons 290 'ytnrto)
                        (cons 293 'tsaireg)
                        (cons 422 'fcwt)
                        (cons 486 'etht))
                 (table (cons 333 'tsaireg)
                        (cons 246 'etht)
                        (cons 665 'ahnrt)
                        (cons 960 'ytnrto)))
          1 (boolean?)
          not)))))

(suite spectrum ()
  (test empty-bag
    (spectrum (bag))
    1 (table?)
    (match? table=? (table)))
  (test singleton-bag
    (spectrum (bag 'lds))
    1 (table?)
    (match? table=? (table (cons 'lds 1))))
  (test without-duplicates
    (spectrum (bag 'nsrr 'epnntek 'edehml 'wtaa 'sfektv 'cfdeiee 'neaim))
    1 (table?)
    (match? table=? (table (cons 'nsrr 1)
                           (cons 'epnntek 1)
                           (cons 'edehml 1)
                           (cons 'wtaa 1)
                           (cons 'sfektv 1)
                           (cons 'cfdeiee 1)
                           (cons 'neaim 1))))
  (test with-duplicates
    (spectrum (bag 'ilu 'hocdude 'ilu 'hocdude 'hocdude 'ilu 'hocdude 'ilu
                   'hocdude 'mnye 'hocdude 'mnye 'mnye 'ilu 'ilu 'hocdude
                   'ilu 'hocdude 'mnye 'ilu 'hocdude 'mnye 'ilu 'ilu 'mnye
                   'mnye 'mnye 'ilu 'hocdude))
    1 (table?)
    (match? table=? (table (cons 'ilu 11)
                           (cons 'hocdude 10)
                           (cons 'mnye 8)))))

(suite table-update ()
  (test empty-table
    (table-update 'rsa identity (table) 'hytdcey)
    1 (table?)
    (match? table=? (table (cons 'rsa 'hytdcey))))
  (test empty-no-default
    (table-update 'pbnee identity (table))
    1 (table?)
    (match? table=? (table (cons 'pbnee #f))))
  (test replacement
    (table-update 'eynm
                  (constant 'niee)
                  (table (cons 'eynm 'tspigfe)))
    1 (table?)
    (match? table=? (table (cons 'eynm 'niee))))
  (test subtler
    (table-update 'rndmeet square (table (cons 'rndmeet 693)) 1)
    1 (table?)
    (match? table=? (table (cons 'rndmeet (square 693)))))
  (test default-in-larger-table
    (table-update 'rlw
                  identity
                  (table (cons 'sat 'aelo)
                         (cons 'cdho 'ogac)
                         (cons 'ihcahe 'mhy))
                  'dslna)
    1 (table?)
    (match? table=? (table (cons 'sat 'aelo)
                           (cons 'cdho 'ogac)
                           (cons 'ihcahe 'mhy)
                           (cons 'rlw 'dslna))))
  (test replacement-in-larger-table
    (table-update 'tsi
                  halve
                  (table (cons 'mfee 816)
                         (cons 'tsi 244)
                         (cons 'hattsfv 342)
                         (cons 'oheo 122)))
    1 (table?)
    (match? table=? (table (cons 'mfee 816)
                           (cons 'tsi (halve 244))
                           (cons 'hattsfv 342)
                           (cons 'oheo 122)))))

(suite table-updater ((same-last-digit? (pipe (~each (sect mod <> 10)) =)))
  (test generic
    (table-updater equal?)
    1 (procedure?)
    (lambda (updater)
      (suite generic-internal ()
        (test empty-table
          (updater 624 identity (table) 21)
          1 (table?)
          (match? table=? (table (cons 624 21))))
        (test empty-no-default
          (updater 865 identity (table))
          1 (table?)
          (match? table=? (table (cons 865 #f))))
        (test replacement
          (updater 387 (constant 395) (table (cons 387 84)))
          1 (table?)
          (match? table=? (table (cons 387 395))))
        (test subtler
          (updater 123 square (table (cons 123 451)) 516)
          1 (table?)
          (match? table=? (table (cons 123 (square 451)))))
        (test default-in-larger-table
          (updater 723
                   identity
                   (table (cons 8 894)
                          (cons 619 920)
                          (cons 796 598)
                          (cons 426 417))
                   173)
          1 (table?)
          (match? table=? (table (cons 8 894)
                                 (cons 619 920)
                                 (cons 796 598)
                                 (cons 426 417)
                                 (cons 723 173))))
        (test replacement-in-larger-table
          (updater 296
                   double
                   (table (cons 961 522)
                          (cons 316 117)
                          (cons 296 616)
                          (cons 734 720)
                          (cons 254 672)))
          1 (table?)
          (match? table=? (table (cons 961 522)
                                 (cons 316 117)
                                 (cons 296 (double 616))
                                 (cons 734 720)
                                 (cons 254 672)))))))
  (test specific
    (table-updater same-last-digit?)
    1 (procedure?)
    (lambda (updater)
      (suite specific-internal ((good-table? (table-of exact-integer?
                                                       exact-integer?
                                                       same-last-digit?))
                                (matcher (table-of= same-last-digit? =)))
        (test empty-table
          (updater 146 add1 (table) 168)
          1 (good-table?)
          (match? matcher (table (cons 6 (add1 168)))))
        (test empty-no-default
          (updater 89
                   (^if exact-integer? add1 (constant 188))
                   (table))
          1 (good-table?)
          (match? matcher (table (cons 9 188))))
        (test replacement
          (updater 557 square (table (cons 557 367)) 19)
          1 (good-table?)
          (match? matcher (table (cons 557 (square 367)))))
        (test default-in-larger-table
          (updater 496
                   sub1
                   (table (cons 124 277)
                          (cons 538 117)
                          (cons 277 741)
                          (cons 743 486))
                   901)
          1 (good-table?)
          (match? matcher (table (cons 124 277)
                                 (cons 538 117)
                                 (cons 277 741)
                                 (cons 743 486)
                                 (cons 496 900))))
        (test replacement-in-larger-table
          (updater 793
                   double
                   (table (cons 793 447)
                          (cons 944 317)
                          (cons 230 362)
                          (cons 368 733)))
          1 (good-table?)
          (match? matcher (table (cons 793 (double 447))
                                 (cons 944 317)
                                 (cons 230 362)
                                 (cons 368 733))))))))

;;; Test the procedures from the "Buffers" section.

(suite buffer? ()
  (test non-pair
    (buffer? 'lntoreu)
    1 (boolean?)
    not)
  (test bad-fore
    (buffer? (cons 'iat (list 'crpahm 'heei 'nra)))
    1 (boolean?)
    not)
  (test bad-aft
    (buffer? (cons (list 'dssr 'ahrr 'shcre 'hamia) 214))
    1 (boolean?)
    not)
  (test empty-buffer
    (buffer? (cons (list) (list)))
    1 (boolean?)
    true?)
  (test typical-buffer
    (buffer? (cons (list 'teyhd 'rettms 'uatt 'gaed 'erangnx 'isaesro)
                   (list 'rdorhln 'rartt 'isol)))
    1 (boolean?)
    true?))

(suite buffer ((buffer=? (lambda (left right)
                           (list=? (append (car left) (reverse (cdr left)))
                                   (append (car right)
                                           (reverse (cdr right)))))))
  (test no-arguments
    (buffer)
    1 (buffer?)
    (match? buffer=? (cons (list) (list))))
  (test several-arguments
    (buffer 'osne 'oeue 'bew 'emnatwi 'tftst)
    1 (buffer?)
    (match? buffer=? (cons (list 'osne 'oeue 'bew 'emnatwi 'tftst)
                           (list)))))

(suite put-into-buffer ((buffer=?
                         (lambda (left right)
                           (list=? (append (car left) (reverse (cdr left)))
                                   (append (car right)
                                           (reverse (cdr right)))))))
  (test empty-buffer
    (put-into-buffer 'stvtoep (buffer))
    1 (buffer?)
    (match? buffer=? (buffer 'stvtoep)))
  (test larger-buffer
    (put-into-buffer 'lier (buffer 'oehnuur 'ierar 'aseh))
    1 (buffer?)
    (match? buffer=? (buffer 'oehnuur 'ierar 'aseh 'lier))))

(suite take-from-buffer ((buffer=?
                          (lambda (left right)
                            (list=? (append (car left) (reverse (cdr left)))
                                    (append (car right)
                                            (reverse (cdr right)))))))
  (test singleton-buffer
    (take-from-buffer (buffer 'omr))
    2 (symbol? buffer?)
    (lambda (alpha beta)
      (and (symbol=? alpha 'omr) (buffer=? beta (buffer)))))
  (test larger-buffer
    (take-from-buffer (put-into-buffer 'reacuh
                        (put-into-buffer 'fdtaar
                          (put-into-buffer 'aefrws (buffer)))))
    2 (symbol? buffer?)
    (lambda (alpha beta)
      (and (symbol=? alpha 'aefrws)
           (buffer=? beta (buffer 'fdtaar 'reacuh))))))

(suite empty-buffer? ()
  (test yes
    (empty-buffer? (buffer))
    1 (boolean?)
    true?)
  (test no
    (empty-buffer? (buffer 'ehooonl 'hetyd 'sot))
    1 (boolean?)
    not)
  (test fore-list-empty
    (empty-buffer? (put-into-buffer 'tte
                     (put-into-buffer 'tvy
                       (put-into-buffer 'astnaee (buffer)))))
    1 (boolean?)
    not))

(suite buffer-of= ((same-last-digit? (pipe (~each (sect mod <> 10)) =)))
  (test specific
    (buffer-of= same-last-digit?)
    1 (procedure?)
    (lambda (comparer)
      (suite buffer-of-internal
          ((sample (buffer 254 778 919 107 102 507 593)))
        (test empty-buffers
          (comparer (buffer) (buffer))
          1 (boolean?)
          true?)
        (test left-empty
          (comparer (buffer) (buffer 946 450 685))
          1 (boolean?)
          not)
        (test right-empty
          (comparer (buffer 891 196 17 833 361) (buffer))
          1 (boolean?)
          not)
        (test identical-buffers
          (comparer sample sample)
          1 (boolean?)
          true?)
        (test duplicate-buffers
          (comparer sample (buffer 4 608 589 197 332 687 63))
          1 (boolean?)
          true?)
        (test fore-and-aft
          (comparer (buffer 980 176 556 672)
                    (put-into-buffer 172
                      (put-into-buffer 516
                        (put-into-buffer 326 
                          (put-into-buffer 570 (buffer))))))
          1 (boolean?)
          true?)
        (test wrong-order
          (comparer (buffer 345 408 161 902 243)
                    (buffer 243 902 161 408 345))
          1 (boolean?)
          not)))))

(suite buffer=? ((sample (buffer 'hcaf 'ott 'tataes 'hiwmte)))
  (test empty-buffers
    (buffer=? (buffer) (buffer))
    1 (boolean?)
    true?)
  (test left-empty
    (buffer=? (buffer) (buffer 'rytirn 'onrgi 'hddm 'snrn))
    1 (boolean?)
    not)
  (test right-empty
    (buffer=? (buffer 'nsohw 'umrne 'tid) (buffer))
    1 (boolean?)
    not)
  (test identical-buffers
    (buffer=? sample sample)
    1 (boolean?)
    true?)
  (test duplicate-buffers
    (buffer=? sample (buffer 'hcaf 'ott 'tataes 'hiwmte))
    1 (boolean?)
    true?)
  (test fore-and-aft
    (buffer=? (buffer 'nyloios 'ttner 'kdtvtlr)
              (put-into-buffer 'kdtvtlr
                      (put-into-buffer 'ttner
                        (put-into-buffer 'nyloios (buffer)))))
    1 (boolean?)
    true?)
  (test wrong-order
    (buffer=? (buffer 'nouto 'hef 'uws 'ihuhhle)
              (buffer 'hef 'ihuhhle 'nouto 'uws))
    1 (boolean?)
    not))

(suite unfold-buffer ()
  (test minimal
    (unfold-buffer values? (create 'tsm) (create))
    1 (procedure?)
    (lambda (stopper)
      (suite minimal-internal ()
        (test run-it
          (stopper)
          1 (buffer?)
          empty-buffer?))))
  (test list->buffer
    (unfold-buffer empty-list? first rest)
    1 (procedure?)
    (lambda (list->buffer)
      (suite list->buffer-internal ()
        (test empty-list
          (list->buffer (list))
          1 (buffer?)
          empty-buffer?)
        (test fuller-list
          (list->buffer (list 'mrjernr 'dnnmfdh 'epahe 'dsowg 'remenn
                              'oygnr))
          1 (buffer?)
          (match? buffer=? (buffer 'oygnr 'remenn 'dsowg 'epahe 'dnnmfdh
                                   'mrjernr))))))
  (test multiargument
    (unfold-buffer (every empty-list?) (pipe (~each first) +) (~each rest))
    1 (procedure?)
    (lambda (adder)
      (suite multiargument-internal ()
        (test empties
          (adder (list) (list))
          1 (buffer?)
          empty-buffer?)
        (test pairwise-sums
          (adder (list 881 862 294 317) (list 693 585 973 480))
          1 (buffer?)
          (match? buffer=? (buffer (+ 317 480) (+ 294 973) (+ 862 585)
                                   (+ 881 693))))))))     

(suite process-buffer ((nanc (lambda (left right)
                               (* left (add1 right)))))
  (test do-nothing
    (process-buffer (create) (constant))
    1 (procedure?)
    (lambda (idler)
      (suite do-nothing-internal ()
        (test empty-buffer
          (idler (buffer))
          0 ())
        (test large-buffer
          (idler (buffer 'fdegalm 'tomcroo 'tydtst 'typneer 'teo 'idridla
                         'ihin))
          0 ()))))
  (test nanc
    (process-buffer (create 358) nanc)
    1 (procedure?)
    (lambda (nancer)
      (suite nanc-internal ()
        (test empty-buffer
          (nancer (buffer))
          1 (exact-integer?)
          (match? = 358))
        (test one-item-buffer
          (nancer (buffer 577))
          1 (exact-integer?)
          (match? = (nanc 577 358)))
        (test large-buffer
          (nancer (buffer 99 498 79 393 229))
          1 (exact-integer?)
          (match? = (nanc 229
                          (nanc 393
                                (nanc 79 (nanc 498 (nanc 99 358)))))))))))

;;; copyright (C) 2011, 2017 John David Stone

;;; This program is free software.  You may redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation -- either version 3 of the License, or (at
;;; your option) any later version.  A copy of the GNU General Public
;;; License is available on the World Wide Web at
;;;
;;;                http://www.gnu.org/licenses/gpl.html
;;;
;;; This program is distributed in the hope that it will be useful, but
;;; WITHOUT ANY WARRANTY -- without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; General Public License for more details.
