;;; This is the test program for the procedures introduced in the "Graphs"
;;; chapter.

;;; John David Stone
;;; Department of Computer Science
;;; Grinnell College

;;; created October 11, 2011
;;; last revised January 12, 2017

(import (afp primitives)
        (afp arithmetic)
        (afp extras)
        (only (afp constant-procedures) values? constant create)
        (only (afp couplers) pipe)
        (only (afp adapters) ~initial ~each)
        (only (afp predicate-operations) ^et ^vel)
        (only (afp pairs) pair-of pair-of=)
        (only (afp lists)
              first rest empty-list? prepend list-of list-of= fold-list
              ^and ^or run for-all? exists?)
        (only (afp bags) bag put-into-bag bag-of=)
        (only (afp sets)
              set? set-of set-maker set empty-set? set=? set-of= fold-set
              cardinality)
        (only (afp tables)
              table? table-of put-into-table table lookup table=?
              table-of=)
        (afp graphs)
        (afp depth-first-traversal)
        (afp paths)
        (afp breadth-first-traversal)
        (afp spanning-trees)
        (rename (afp shortest-paths Bellman-Ford)
                (shortest-paths-from Bellman-Ford)
                (safe-shortest-paths-from safe-Bellman-Ford))
        (rename (afp shortest-paths Dijkstra)
                (shortest-paths-from Dijkstra)
                (all-shortest-paths all-pairs-Dijkstra))
        (rename (afp shortest-paths Floyd-Warshall)
                (all-shortest-paths Floyd-Warshall))
        (afp flow-networks)
        (support testing)
        (support testing-procedures))

;;; Test the procedures from the "Implementing graphs" section.

(define arcset (set-maker (^et (pipe (~each arc-head) equal?)
                               (pipe (~each arc-tail) equal?))))

(suite make-graph
    ((graph? (pair-of set? (set-of arc?
                                   (^et (pipe (~each arc-head) equal?)
                                        (pipe (~each arc-tail) equal?)))))
     (graph=? (pair-of= set=? (set-of= (^et (pipe (~each arc-head) equal?)
                                            (pipe (~each arc-tail)
                                                  equal?))))))
  (test empty-graph
    (make-graph (set) (arcset))
    1 (graph?)
    (match? graph=? (cons (set) (arcset))))
  (test typical-graph
    (make-graph (set 'dyteht 'ess 'oylouha 'etoas)
                (arcset (make-arc 'dyteht 'oylouha 'nadya)
                        (make-arc 'dyteht 'ess 'isr)
                        (make-arc 'ess 'etoas 'vac)))
    1 (graph?)
    (match? graph=? (cons (set 'dyteht 'ess 'oylouha 'etoas)
                          (arcset (make-arc 'dyteht 'oylouha 'nadya)
                                  (make-arc 'dyteht 'ess 'isr)
                                  (make-arc 'ess 'etoas 'vac))))))

(suite vertices ()
  (test empty-graph
    (vertices (make-graph (set) (arcset)))
    1 (set?)
    empty-set?)
  (test larger-graph
    (vertices (make-graph (set 'ertt 'ikoec 'rioodr 'rfft 'yubhca)
                          (arcset (make-arc 'ertt 'rioodr 'emhnnl)
                                  (make-arc 'ertt 'rfft 'tsord)
                                  (make-arc 'ertt 'yubhca 'ndot)
                                  (make-arc 'ikoec 'rioodr 'oat)
                                  (make-arc 'ikoec 'yubhca 'ahdtlan)
                                  (make-arc 'rioodr 'rfft 'toeils)
                                  (make-arc 'rioodr 'yubhca 'teyea)
                                  (make-arc 'rfft 'yubhca 'dtttmsy))))
    1 ((set-of symbol? symbol=?))
    (match? (set-of= symbol=?) (set 'ertt 'ikoec 'rioodr 'rfft 'yubhca))))

(suite arcs ((good-set? (set-of arc? (^et (pipe (~each arc-head) equal?)
                                          (pipe (~each arc-tail) equal?))))
             (same-arcset? (set-of= (^et (pipe (~each arc-head) equal?)
                                          (pipe (~each arc-tail) equal?)))))
  (test empty-graph
    (arcs (make-graph (set) (arcset)))
    1 (set?)
    empty-set?)
  (test larger-graph
    (arcs (make-graph (set 'hae 'lgiaihr 'ethd 'sonok)
                      (arcset (make-arc 'hae 'hae 'beocnae)
                              (make-arc 'hae 'lgiaihr 'diaodae)
                              (make-arc 'hae 'ethd 'eviie)
                              (make-arc 'lgiaihr 'hae 'nktl)
                              (make-arc 'lgiaihr 'ethd 'ulu)
                              (make-arc 'ethd 'hae 'ddfue)
                              (make-arc 'sonok 'ethd 'reoaots)
                              (make-arc 'sonok 'sonok 'ani))))
    1 (good-set?)
    (match? same-arcset? (arcset (make-arc 'hae 'hae 'beocnae)
                                 (make-arc 'hae 'lgiaihr 'diaodae)
                                 (make-arc 'hae 'ethd 'eviie)
                                 (make-arc 'lgiaihr 'hae 'nktl)
                                 (make-arc 'lgiaihr 'ethd 'ulu)
                                 (make-arc 'ethd 'hae 'ddfue)
                                 (make-arc 'sonok 'ethd 'reoaots)
                                 (make-arc 'sonok 'sonok 'ani)))))

(suite arc=? ((sample (make-arc 'aeewnro 'ineenvo 'eeha)))
  (test identical-arcs
    (arc=? sample sample)
    1 (boolean?)
    true?)
  (test duplicate-arcs
    (arc=? sample (make-arc 'aeewnro 'ineenvo 'eeha))
    1 (boolean?)
    true?)
  (test label-mismatch
    (arc=? (make-arc 'oree 'dhe 'ist) (make-arc 'oree 'dhe 'orr))
    1 (boolean?)
    not)
  (test one-label-missing
    (arc=? (make-arc 'altia 'euufy 'mgi) (make-arc 'altia 'euufy null))
    1 (boolean?)
    not)
  (test complete-failure
    (arc=? (make-arc 'idt 'enth 'nrgenu) (make-arc 'tttul 'nnit 'ifrc))
    1 (boolean?)
    not)
  (test half-failure
    (arc=? (make-arc 'eyhn 'eeihoe 'srrn) (make-arc 'drim 'eeihoe 'srrn))
    1 (boolean?)
    not)
  (test out-of-order
    (arc=? (make-arc 'nrgdau 'ire 'edrhogi)
           (make-arc 'ire 'nrgdau 'edrhogi))
    1 (boolean?)
    not))

(suite same-endpoints? ((sample (make-arc 'aphphh 'woo 'oweany)))
  (test identical-arcs
    (same-endpoints? sample sample)
    1 (boolean?)
    true?)
  (test duplicate-arcs
    (same-endpoints? sample (make-arc 'aphphh 'woo 'oweany))
    1 (boolean?)
    true?)
  (test label-mismatch
    (same-endpoints? (make-arc 'elatpg 'phdhb 'iofih)
                     (make-arc 'elatpg 'phdhb 'crri))
    1 (boolean?)
    true?)
  (test one-label-missing
    (same-endpoints? (make-arc 'tvea 'rpmipr 'snb)
                     (make-arc 'tvea 'rpmipr null))
    1 (boolean?)
    true?)
  (test complete-failure
    (same-endpoints? (make-arc 'iesuwh 'rotsnrd 'ieios)
                     (make-arc 'cdctas 'lau 'etwgotl))
    1 (boolean?)
    not)
  (test half-failure
    (same-endpoints? (make-arc 'teaaisg 'yuthhfn 'sror)
                     (make-arc 'teaaisg 'sehuib 'sror))
    1 (boolean?)
    not)
  (test out-of-order
    (same-endpoints? (make-arc 'eaet 'ohmedt 'iwnaoe)
                     (make-arc 'ohmedt 'eaet 'iwnaoe))
    1 (boolean?)
    not))

(suite reverse-arc ()
  (test labeled-arc
    (reverse-arc (make-arc 'tch 'dtsrhe 'akitk))
    1 (arc?)
    (match? arc=? (make-arc 'dtsrhe 'tch 'akitk)))
  (test unlabeled-arc
    (reverse-arc (make-arc 'tnd 'orisat null))
    1 (arc?)
    (match? arc=? (make-arc 'orisat 'tnd null))))

(suite graph? ()
  (test non-pair
    (graph? 'bttdhr)
    1 (boolean?)
    not)
  (test car-not-set
    (graph? (cons 'onehda (arcset (make-arc 'onehda 'onehda 'odiraa))))
    1 (boolean?)
    not)
  (test car-bag-with-repeated-elements
    (graph? (cons (bag 'nrebhta 'hlut 'nrebhta)
                  (arcset (make-arc 'nrebhta 'hlut 'hltt))))
    1 (boolean?)
    not)
  (test cdr-not-set
    (graph? (cons (set 'lfaa 'nnadner 'uenso) 'eertr))
    1 (boolean?)
    not)
  (test non-arc-in-cdr
    (graph? (cons (set 'etdh 'hre 'nypod)
                  (set (make-arc 'etdh 'nypod 'enooru)
                       'nnadner 
                       (make-arc 'hre 'etdh 'rets))))
    1 (boolean?)
    not)
  (test tail-not-a-vertex
    (graph? (cons (set 'lsi 'bos 'llae 'otrio)
                  (arcset (make-arc 'lsi 'bos 'hri)
                          (make-arc 'ddao 'llae 'hupitte)
                          (make-arc 'otrio 'lsi 'tly))))
    1 (boolean?))
  (test head-not-a-vertex
    (graph? (cons (set 'ndtw 'ydin 'ogfut 'usea 'svothr)
                  (arcset (make-arc 'ogfut 'usea 'etet)
                          (make-arc 'svothr 'ydin 'eoy)
                          (make-arc 'ndtw 'sefeoei 'lhhoa)
                          (make-arc 'ydin 'ogfut 'rpot))))
    1 (boolean?)
    not)
  (test duplicated-endpoints
    (graph? (cons (set 'ioeoshg 'rnttey 'hnte)
                  (set (make-arc 'rnttey 'ioeoshg 'inrmpwe)
                       (make-arc 'ioeoshg 'ioeoshg 'uaorotb)
                       (make-arc 'rnttey 'ioeoshg 'mcu))))
    1 (boolean?)
    not)
  (test empty-graph
    (graph? (make-graph (set) (arcset)))
    1 (boolean?)
    true?)
  (test larger-graph
    (graph? (make-graph (set 'iahlai 'htrnaa 'yee 'cuev 'hor)
                        (arcset (make-arc 'iahlai 'htrnaa 'epmsi)
                                (make-arc 'iahlai 'cuev 'edow)
                                (make-arc 'iahlai 'hor 'prto)
                                (make-arc 'htrnaa 'yee 'ilodhee)
                                (make-arc 'htrnaa 'cuev 'doofuo)
                                (make-arc 'htrnaa 'hor 'bsnda)
                                (make-arc 'yee 'cuev 'itopd)
                                (make-arc 'yee 'hor 'hens)
                                (make-arc 'cuev 'hor 'aouup))))
    1 (boolean?)
    true?)
  (test unlabeled-graph
    (graph? (make-graph (set 'ltc 'eir 'riplecg 'dath)
                        (arcset (make-arc 'ltc 'riplecg null)
                                (make-arc 'ltc 'dath null)
                                (make-arc 'eir 'ltc null)
                                (make-arc 'eir 'dath null)
                                (make-arc 'riplecg 'eir null)
                                (make-arc 'riplecg 'riplecg null)
                                (make-arc 'dath 'eir null))))
    1 (boolean?)
    true?)
  (test undirected-graph
    (graph? (make-graph (set 'adrprc 'tlia 'eosdrot 'argtei)
                        (arcset (make-arc 'adrprc 'tlia 'lfawes)
                                (make-arc 'tlia 'adrprc 'lfawes)
                                (make-arc 'adrprc 'argtei 'eelt)
                                (make-arc 'argtei 'adrprc 'eelt)
                                (make-arc 'tlia 'eosdrot 'thvh)
                                (make-arc 'eosdrot 'tlia 'thvh))))
    1 (boolean?)
    true?))

(suite graph=?
    ((sample (make-graph (set 'pthls 'ghaliat 'yrkrela)
                         (arcset (make-arc 'pthls 'pthls 'von)
                                 (make-arc 'ghaliat 'ghaliat 'toahwt)
                                 (make-arc 'ghaliat 'yrkrela 'tedtebt)
                                 (make-arc 'yrkrela 'pthls 'sdulaon)
                                 (make-arc 'yrkrela 'ghaliat 'nendl)
                                 (make-arc 'yrkrela 'yrkrela 'ltodr)))))
  (test empty-graphs
    (graph=? (make-graph (set) (arcset)) (make-graph (set) (arcset)))
    1 (boolean?)
    true?)
  (test only-one-empty
    (graph=? (make-graph (set) (arcset))
             (make-graph (set 'miuiiwe 'pdow 'tgrnasg)
                         (arcset (make-arc 'miuiiwe 'pdow 'eett)
                                 (make-arc 'pdow 'miuiiwe 'nil))))
    1 (boolean?)
    not)
  (test identical-graphs
    (graph=? sample sample)
    1 (boolean?)
    true?)
  (test duplicate-graphs
    (graph=? sample
             (make-graph (set 'pthls 'ghaliat 'yrkrela)
                         (arcset (make-arc 'pthls 'pthls 'von)
                                 (make-arc 'ghaliat 'ghaliat 'toahwt)
                                 (make-arc 'ghaliat 'yrkrela 'tedtebt)
                                 (make-arc 'yrkrela 'pthls 'sdulaon)
                                 (make-arc 'yrkrela 'ghaliat 'nendl)
                                 (make-arc 'yrkrela 'yrkrela 'ltodr))))
    1 (boolean?)
    true?)
  (test different-vertices
    (graph=? (make-graph (set 'acemte 'elarvn 'sieyt)
                         (arcset (make-arc 'acemte 'acemte 'dlfdp)
                                 (make-arc 'elarvn 'acemte 'insmma)
                                 (make-arc 'elarvn 'elarvn 'ineein)))
             (make-graph (set 'aaoc 'acemte 'elarvn)
                         (arcset (make-arc 'acemte 'acemte 'dlfdp)
                                 (make-arc 'elarvn 'acemte 'insmma)
                                 (make-arc 'elarvn 'elarvn 'ineein))))
    1 (boolean?)
    not)
  (test different-arcs
    (graph=? (make-graph (set 'reuwvt 'shengr 'sppbrw)
                         (arcset (make-arc 'reuwvt 'reuwvt 'eew)
                                 (make-arc 'reuwvt 'sppbrw 'ieyu)
                                 (make-arc 'shengr 'reuwvt 'nti)
                                 (make-arc 'shengr 'shengr 'mrsh)
                                 (make-arc 'sppbrw 'reuwvt 'gtiony)
                                 (make-arc 'sppbrw 'sppbrw 'wth)))
             (make-graph (set 'reuwvt 'shengr 'sppbrw)
                         (arcset (make-arc 'reuwvt 'shengr 'elee)
                                 (make-arc 'shengr 'reuwvt 'nti)
                                 (make-arc 'shengr 'shengr 'mrsh))))
    1 (boolean?)
    not)
  (test different-arc-label
    (graph=? (make-graph (set 'eaterow 'esii 'ods)
                         (arcset (make-arc 'eaterow 'eaterow 'eovobwt)
                                 (make-arc 'eaterow 'esii 'idun)
                                 (make-arc 'esii 'esii 'rsrr)
                                 (make-arc 'ods 'esii 'hnsibbi)
                                 (make-arc 'ods 'ods 'uuin)))
             (make-graph (set 'eaterow 'esii 'ods)
                         (arcset (make-arc 'eaterow 'eaterow 'eovobwt)
                                 (make-arc 'eaterow 'esii 'idun)
                                 (make-arc 'esii 'esii 'rsrr)
                                 (make-arc 'ods 'esii 'goanyi)
                                 (make-arc 'ods 'ods 'uuin))))
    1 (boolean?)
    not))

(suite empty-graph ()
  (test does-it-work?
    (empty-graph)
    1 (graph?)
    (match? graph=? (make-graph (set) (arcset)))))

(suite add-vertex ()
  (test empty-graph
    (add-vertex 'mrrndto (empty-graph))
    1 (graph?)
    (match? graph=? (make-graph (set 'mrrndto) (arcset))))
  (test larger-graph
    (add-vertex 'ignf
                (make-graph (set 'peedr 'feemsp 'vrryuee 'dwe)
                            (arcset (make-arc 'peedr 'vrryuee 'wahe)
                                    (make-arc 'peedr 'dwe 'ereuavh)
                                    (make-arc 'feemsp 'vrryuee 'hnsw)
                                    (make-arc 'feemsp 'dwe 'aelas))))
    1 (graph?)
    (match? graph=? (make-graph (set 'peedr 'feemsp 'vrryuee 'dwe 'ignf)
                                (arcset (make-arc 'peedr 'vrryuee 'wahe)
                                        (make-arc 'peedr 'dwe 'ereuavh)
                                        (make-arc 'feemsp 'vrryuee 'hnsw)
                                        (make-arc 'feemsp 'dwe 'aelas))))))

(suite arcless-graph ()
  (test no-vertices
    (arcless-graph (set))
    1 (graph?)
    (match? graph=? (empty-graph)))
  (test typical-case
    (arcless-graph (set 'weohdhu 'lswr 'lrottan 'uamgsl))
    1 (graph?)
    (match? graph=? (make-graph (set 'weohdhu 'lswr 'lrottan 'uamgsl)
                                (arcset)))))

(suite add-labeled-arc ()
  (test minimal-case
    (add-labeled-arc 'dlg 'caiobaa 'ititn
                     (make-graph (set 'dlg 'caiobaa) (arcset)))
    1 (graph?)
    (match? graph=? (make-graph (set 'dlg 'caiobaa)
                                (arcset (make-arc 'dlg 'caiobaa 'ititn)))))
  (test larger-case
    (add-labeled-arc 'edeo 'frrrrh 'iepdato
                     (make-graph (set 'saaia 'frrrrh 'isohddo 'edeo 'ays)
                                 (arcset (make-arc 'saaia 'frrrrh 'rce)
                                         (make-arc 'saaia 'ays 'atnovaf)
                                         (make-arc 'frrrrh 'isohddo 'daahm)
                                         (make-arc 'isohddo 'edeo 'syttf)
                                         (make-arc 'edeo 'ays 'eoacwh))))
    1 (graph?)
    (match? graph=? (make-graph (set 'saaia 'frrrrh 'isohddo 'edeo 'ays)
                                (arcset (make-arc 'saaia 'frrrrh 'rce)
                                        (make-arc 'saaia 'ays 'atnovaf)
                                        (make-arc 'frrrrh 'isohddo 'daahm)
                                        (make-arc 'isohddo 'edeo 'syttf)
                                        (make-arc 'edeo 'frrrrh 'iepdato)
                                        (make-arc 'edeo 'ays 'eoacwh))))))

(suite add-arc ()
  (test minimal-case
    (add-arc 'iakert 'tte (make-graph (set 'iakert 'tte) (arcset)))
    1 (graph?)
    (match? graph=? (make-graph (set 'iakert 'tte)
                                (arcset (make-arc 'iakert 'tte null)))))
  (test larger-case
    (add-arc 'hafhcfd 'eddg
             (make-graph (set 'uutse 'eddg 'reeoi 'hafhcfd 'etm)
                         (arcset (make-arc 'uutse 'eddg null)
                                 (make-arc 'uutse 'etm null)
                                 (make-arc 'eddg 'reeoi null)
                                 (make-arc 'reeoi 'hafhcfd null)
                                 (make-arc 'hafhcfd 'etm null))))
    1 (graph?)
    (match? graph=? (make-graph (set 'uutse 'eddg 'reeoi 'hafhcfd 'etm)
                                (arcset (make-arc 'uutse 'eddg null)
                                        (make-arc 'uutse 'etm null)
                                        (make-arc 'eddg 'reeoi null)
                                        (make-arc 'reeoi 'hafhcfd null)
                                        (make-arc 'hafhcfd 'eddg null)
                                        (make-arc 'hafhcfd 'etm null))))))

(suite delete-arc ()
  (test minimal-case
    (delete-arc 'een 'mieeiau
                (make-graph (set 'een 'mieeiau)
                            (arcset (make-arc 'een 'mieeiau 'ulnwre))))
    1 (graph?)
    (match? graph=? (make-graph (set 'een 'mieeiau) (arcset))))
  (test larger-case
    (delete-arc 'talnn 'ooti
                (make-graph (set 'talnn 'ltty 'nadent 'vetntaa 'ooti)
                            (arcset (make-arc 'talnn 'ltty 'lef)
                                    (make-arc 'talnn 'ooti 'brlai)
                                    (make-arc 'ltty 'nadent 'eiost)
                                    (make-arc 'nadent 'vetntaa 'lbasle)
                                    (make-arc 'vetntaa 'ooti 'iairsbu))))
    1 (graph?)
    (match? graph=?
            (make-graph (set 'talnn 'ltty 'nadent 'vetntaa 'ooti)
                        (arcset (make-arc 'talnn 'ltty 'lef)
                                (make-arc 'ltty 'nadent 'eiost) 
                                (make-arc 'nadent 'vetntaa 'lbasle)
                                (make-arc 'vetntaa 'ooti 'iairsbu)))))
  (test arc-not-present
    (delete-arc 'ewoov 'slekoe
                (make-graph (set 'slekoe 'rmaces 'ewoov)
                            (arcset (make-arc 'slekoe 'rmaces 'enoi)
                                    (make-arc 'rmaces 'ewoov 'aton))))
    1 (graph?)
    (match? graph=?
            (make-graph (set 'slekoe 'rmaces 'ewoov)
                        (arcset (make-arc 'slekoe 'rmaces 'enoi)
                                (make-arc 'rmaces 'ewoov 'aton))))))

(suite replace-arc ()
  (test minimal-case
    (replace-arc 'etwrn 'hehnatc 'itf
                 (make-graph (set 'etwrn 'hehnatc)
                             (arcset (make-arc 'etwrn 'hehnatc 'ureeaf))))
    1 (graph?)
    (match? graph=? (make-graph (set 'etwrn 'hehnatc)
                                (arcset (make-arc 'etwrn 'hehnatc 'itf)))))
  (test larger-case
    (replace-arc 'ccern 'rcrh 'tedeueh
                 (make-graph (set 'usrienm 'rcrh 'euxn 'ccern 'tew)
                             (arcset (make-arc 'usrienm 'rcrh 'ris)
                                     (make-arc 'usrienm 'tew 'weeae)
                                     (make-arc 'rcrh 'euxn 'lindg)
                                     (make-arc 'euxn 'ccern 'lat)
                                     (make-arc 'ccern 'rcrh 'rssce)
                                     (make-arc 'ccern 'tew 'jrns))))
  1 (graph?)
  (match? graph=? (make-graph (set 'usrienm 'rcrh 'euxn 'ccern 'tew)
                              (arcset (make-arc 'usrienm 'rcrh 'ris)
                                      (make-arc 'usrienm 'tew 'weeae)
                                      (make-arc 'rcrh 'euxn 'lindg)
                                      (make-arc 'euxn 'ccern 'lat)
                                      (make-arc 'ccern 'rcrh 'tedeueh)
                                      (make-arc 'ccern 'tew 'jrns))))))

(suite delete-vertex ()
  (test single-vertex-graph
    (delete-vertex 'daah
                   (make-graph (set 'daah)
                               (arcset (make-arc 'daah 'daah 'ncnnc))))
    1 (graph?)
    (match? graph=? (empty-graph)))
  (test multiple-vertex-graph
    (delete-vertex 'uwtd
                   (make-graph (set 'hoi 'aehooy 'uwtd 'tiihfns)
                               (arcset (make-arc 'hoi 'hoi null)
                                       (make-arc 'hoi 'aehooy null)
                                       (make-arc 'hoi 'uwtd null)
                                       (make-arc 'hoi 'tiihfns null)
                                       (make-arc 'aehooy 'hoi null)
                                       (make-arc 'aehooy 'aehooy null)
                                       (make-arc 'aehooy 'uwtd null)
                                       (make-arc 'aehooy 'tiihfns null)
                                       (make-arc 'uwtd 'hoi null)
                                       (make-arc 'uwtd 'aehooy null)
                                       (make-arc 'uwtd 'uwtd null)
                                       (make-arc 'uwtd 'tiihfns null)
                                       (make-arc 'tiihfns 'hoi null)
                                       (make-arc 'tiihfns 'aehooy null)
                                       (make-arc 'tiihfns 'uwtd null)
                                       (make-arc 'tiihfns 'tiihfns null))))
    1 (graph?)
    (match? graph=?
            (make-graph (set 'hoi 'aehooy 'tiihfns)
                        (arcset (make-arc 'hoi 'hoi null)
                                (make-arc 'hoi 'aehooy null)
                                (make-arc 'hoi 'tiihfns null)
                                (make-arc 'aehooy 'hoi null)
                                (make-arc 'aehooy 'aehooy null)
                                (make-arc 'aehooy 'tiihfns null)
                                (make-arc 'tiihfns 'hoi null)
                                (make-arc 'tiihfns 'aehooy null)
                                (make-arc 'tiihfns 'tiihfns null)))))
  (test delend-not-present
    (delete-vertex 'wnmc
                   (make-graph (set 'otlt 'nlse 'noneoho)
                               (arcset (make-arc 'otlt 'otlt 'tse)
                                       (make-arc 'nlse 'otlt 'umions)
                                       (make-arc 'noneoho 'otlt 'syg))))
    1 (graph?)
    (match? graph=? (make-graph (set 'otlt 'nlse 'noneoho)
                                (set (make-arc 'otlt 'otlt 'tse)
                                     (make-arc 'nlse 'otlt 'umions)
                                     (make-arc 'noneoho 'otlt 'syg))))))

(suite restriction ()
  (test empty-graph
    (restriction (set) (empty-graph))
    1 (graph?)
    (match? graph=? (empty-graph)))
  (test single-vertex-unrestricted
    (restriction (set 'aok) (complete-graph (set 'aok)))
    1 (graph?)
    (match? graph=? (complete-graph (set 'aok))))
  (test single-vertex-restricted
    (restriction (set) (complete-graph (set 'meta)))
    1 (graph?)
    (match? graph=? (empty-graph)))
  (test two-vertex
    (restriction (set 'boect) (complete-graph (set 'vbwtnr 'boect)))
    1 (graph?)
    (match? graph=? (complete-graph (set 'boect))))
  (test contrived
    (restriction (set 'iouigho 'eaoi 'hmrste 'medecao 'eyo 'reort 'ieedem
                      'welc)
                 (make-graph (set 'iouigho 'eaoi 'hmrste 'medecao 'eyo
                                  'reort 'ieedem 'welc 'tidnde 'rcgah 'rwae
                                  'rhns 'rmhmuh 'ookeen 'ero 'myfwuut)
                             (arcset (make-arc 'eaoi 'hmrste 'ilwvthp)
                                     (make-arc 'medecao 'rhns 'ifa)
                                     (make-arc 'reort 'ieedem 'aasne)
                                     (make-arc 'reort 'ero 'otuhg)
                                     (make-arc 'ieedem 'reort 'adyree)
                                     (make-arc 'ieedem 'ero 'wineees)
                                     (make-arc 'welc 'welc 'lomh)
                                     (make-arc 'rcgah 'rwae 'awt)
                                     (make-arc 'rmhmuh 'eyo 'rueo)
                                     (make-arc 'ookeen 'reort 'bro)
                                     (make-arc 'ookeen 'ieedem 'tset)
                                     (make-arc 'ookeen 'ero 'een)
                                     (make-arc 'ero 'ookeen 'iha)
                                     (make-arc 'myfwuut 'myfwuut 'stiw))))
    1 (graph?)
    (match? graph=? (make-graph (set 'iouigho 'eaoi 'hmrste 'medecao 'eyo
                                     'reort 'ieedem 'welc)
                                (arcset (make-arc 'eaoi 'hmrste 'ilwvthp)
                                        (make-arc 'reort 'ieedem 'aasne)
                                        (make-arc 'ieedem 'reort 'adyree)
                                        (make-arc 'welc 'welc 'lomh)))))
  (test random-graph
    (restriction (set 'leeato 'moda 'lao 'incn)
                 (make-graph (set 'osn 'leeato 'res 'moda 'aeo 'lao 'ehsot
                                  'incn)
                             (arcset (make-arc 'osn 'leeato 'neapdmr)
                                     (make-arc 'osn 'ehsot 'asniv)
                                     (make-arc 'osn 'incn 'stdteo)
                                     (make-arc 'leeato 'leeato 'efleenn)
                                     (make-arc 'leeato 'res 'snhn)
                                     (make-arc 'leeato 'aeo 'pkg)
                                     (make-arc 'leeato 'ehsot 'teks)
                                     (make-arc 'res 'osn 'supigrn)
                                     (make-arc 'res 'leeato 'nest)
                                     (make-arc 'res 'moda 'rheooe)
                                     (make-arc 'res 'aeo 'atoawmt)
                                     (make-arc 'res 'ehsot 'sssa)
                                     (make-arc 'res 'incn 'sehoshh)
                                     (make-arc 'moda 'res 'tsrohta)
                                     (make-arc 'moda 'moda 'uoaehei)
                                     (make-arc 'moda 'aeo 'ssxtoeh)
                                     (make-arc 'moda 'lao 'oeeac)
                                     (make-arc 'moda 'ehsot 'mcdaeti)
                                     (make-arc 'aeo 'osn 'nynh)
                                     (make-arc 'aeo 'ehsot 'oeu)
                                     (make-arc 'aeo 'incn 'eeedau)
                                     (make-arc 'lao 'osn 'ogi)
                                     (make-arc 'lao 'leeato 'threoao)
                                     (make-arc 'lao 'res 'hakl)
                                     (make-arc 'lao 'lao 'wnhofo)
                                     (make-arc 'lao 'ehsot 'eieutoc)
                                     (make-arc 'lao 'incn 'eoerp)
                                     (make-arc 'ehsot 'osn 'eipee)
                                     (make-arc 'ehsot 'leeato 'hrh)
                                     (make-arc 'ehsot 'res 'ailo)
                                     (make-arc 'ehsot 'aeo 'wruypd)
                                     (make-arc 'ehsot 'lao 'doeah)
                                     (make-arc 'incn 'leeato 'ote)
                                     (make-arc 'incn 'aeo 'lldsaca))))
    1 (graph?)
    (match? graph=? (make-graph (set 'leeato 'moda 'lao 'incn)
                                (arcset (make-arc 'leeato 'leeato 'efleenn)
                                        (make-arc 'moda 'moda 'uoaehei)
                                        (make-arc 'moda 'lao 'oeeac)
                                        (make-arc 'lao 'leeato 'threoao)
                                        (make-arc 'lao 'lao 'wnhofo)
                                        (make-arc 'lao 'incn 'eoerp)
                                        (make-arc 'incn 'leeato 'ote))))))

(suite relation-graph ()
  (test empty-domain
    (relation-graph (set) values?)
    1 (graph?)
    (match? graph=? (empty-graph)))
  (test singleton-domain-no-arcs
    (relation-graph (set 'imivuol) (constant #f))
    1 (graph?)
    (match? graph=? (make-graph (set 'imivuol) (arcset))))
  (test singleton-domain-complete
    (relation-graph (set 'ngtd) values?)
    1 (graph?)
    (match? graph=? (make-graph (set 'ngtd)
                                (arcset (make-arc 'ngtd 'ngtd null)))))
  (test larger-domain
    (relation-graph (set 768 501 317 682 229) <=)
    1 (graph?)
    (match? graph=? (make-graph (set 768 501 317 682 229)
                                (arcset (make-arc 768 768 null)
                                        (make-arc 501 768 null)
                                        (make-arc 501 501 null)
                                        (make-arc 501 682 null)
                                        (make-arc 317 768 null)
                                        (make-arc 317 501 null)
                                        (make-arc 317 317 null)
                                        (make-arc 317 682 null)
                                        (make-arc 682 768 null)
                                        (make-arc 682 682 null)
                                        (make-arc 229 768 null)
                                        (make-arc 229 501 null)
                                        (make-arc 229 317 null)
                                        (make-arc 229 682 null)
                                        (make-arc 229 229 null))))))

(suite alternative-related-by ()
  (test no-arcs
    (alternative-related-by
      (make-graph (set 'hcen 'biaehe 'fdoaen 'dnonoif 'pwlenno 'rneu)
                  (arcset)))
    1 (procedure?)
    (lambda (arc-exists?)
      (suite no-arcs-internal ()
        (test any-source-and-target
          (arc-exists? 'euo 'onuuoh)
          1 (boolean?)
          not))))
  (test more-arcs
    (alternative-related-by
      (make-graph (set 'usgh 'mlr 'hfcv 'eegnno)
                  (arcset (make-arc 'usgh 'usgh 'elotb)
                          (make-arc 'usgh 'hfcv 'neehe)
                          (make-arc 'usgh 'eegnno 'hetlna)
                          (make-arc 'mlr 'usgh 'ordtdr)
                          (make-arc 'mlr 'hfcv 'wpo)
                          (make-arc 'mlr 'eegnno 'lod)
                          (make-arc 'hfcv 'usgh 'entsan)
                          (make-arc 'hfcv 'mlr 'gnbhl)
                          (make-arc 'hfcv 'eegnno 'vhlh)
                          (make-arc 'eegnno 'mlr 'etle)
                          (make-arc 'eegnno 'eegnno 'uop))))
    1 (procedure?)
    (lambda (arc-exists?)
      (suite more-arcs-internal ()
        (test yes
          (arc-exists? 'hfcv 'eegnno)
          1 (boolean?)
          true?)
        (test endpoints-reversed
          (arc-exists? 'eegnno 'hfcv)
          1 (boolean?)
          not)
        (test no
          (arc-exists? 'usgh 'mlr)
          1 (boolean?)
          not)))))

(suite adjacency-representation ()
  (test empty-graph
    (adjacency-representation (empty-graph))
    1 (table?)
    (match? table=? (table)))
  (test general-case
    (adjacency-representation
      (make-graph (set 'ooaekmf 'ndsant 'lremt 'lpksma 'eue 'gbrdfng 'mrl)
                  (arcset (make-arc 'ooaekmf 'ooaekmf null)
                          (make-arc 'ooaekmf 'ndsant null)
                          (make-arc 'ooaekmf 'gbrdfng null)
                          (make-arc 'ndsant 'ooaekmf null)
                          (make-arc 'ndsant 'lremt null)
                          (make-arc 'ndsant 'eue null)
                          (make-arc 'ndsant 'mrl null)
                          (make-arc 'lremt 'lremt null)
                          (make-arc 'lremt 'eue null)
                          (make-arc 'lpksma 'ooaekmf null)
                          (make-arc 'lpksma 'ndsant null)
                          (make-arc 'lpksma 'lremt null)
                          (make-arc 'lpksma 'lpksma null)
                          (make-arc 'lpksma 'mrl null)
                          (make-arc 'eue 'ooaekmf null)
                          (make-arc 'eue 'lremt null)
                          (make-arc 'eue 'gbrdfng null)
                          (make-arc 'gbrdfng 'ooaekmf null)
                          (make-arc 'gbrdfng 'ndsant null)
                          (make-arc 'gbrdfng 'lpksma null)
                          (make-arc 'gbrdfng 'gbrdfng null)
                          (make-arc 'mrl 'ooaekmf null)
                          (make-arc 'mrl 'ndsant null)
                          (make-arc 'mrl 'lremt null)
                          (make-arc 'mrl 'lpksma null))))
    1 ((table-of symbol? (list-of symbol?) symbol=?))
    (match? (table-of= symbol=? (list-of= symbol=?))
            (table (cons 'ooaekmf (list 'ooaekmf 'ndsant 'gbrdfng))
                   (cons 'ndsant (list 'ooaekmf 'lremt 'eue 'mrl))
                   (cons 'lremt (list 'lremt 'eue))
                   (cons 'lpksma (list 'ooaekmf 'ndsant 'lremt 'lpksma
                                       'mrl))
                   (cons 'eue (list 'ooaekmf 'lremt 'gbrdfng))
                   (cons 'gbrdfng (list 'ooaekmf 'ndsant 'lpksma 'gbrdfng))
                   (cons 'mrl (list 'ooaekmf 'ndsant 'lremt 'lpksma))))))

(suite related-by ()
  (test no-arcs
    (related-by (make-graph (set 'tamnv 'onuuoh 'alm 'euo 'tsl) (arcset)))
    1 (procedure?)
    (lambda (arc-exists?)
      (suite no-arcs-internal ()
        (test any-source-and-target
          (arc-exists? 'euo 'onuuoh)
          1 (boolean?)
          not))))
  (test more-arcs
    (related-by (make-graph (set 'sgdh 'faestw 'ronah 'esc)
                            (arcset (make-arc 'sgdh 'sgdh 'heme)
                                    (make-arc 'sgdh 'esc 'toiyt)
                                    (make-arc 'faestw 'sgdh 'ilhse)
                                    (make-arc 'faestw 'esc 'tme)
                                    (make-arc 'ronah 'faestw 'ydtar)
                                    (make-arc 'esc 'sgdh 'gnyu)
                                    (make-arc 'esc 'ronah 'sihwpar)
                                    (make-arc 'esc 'esc 'oddymvg))))
    1 (procedure?)
    (lambda (arc-exists?)
      (suite more-arcs-internal ()
        (test yes
          (arc-exists? 'ronah 'faestw)
          1 (boolean?)
          true?)
        (test endpoints-reversed
          (arc-exists? 'ronah 'esc)
          1 (boolean?)
          not)
        (test no
          (arc-exists? 'sgdh 'ronah)
          1 (boolean?)
          not)))))

(suite complete-graph ()
  (test empty-set
    (complete-graph (set))
    1 (graph?)
    (match? graph=? (empty-graph)))
  (test singleton-set
    (complete-graph (set 'ntwdas))
    1 (graph?)
    (match? graph=? (make-graph (set 'ntwdas)
                                (arcset (make-arc 'ntwdas 'ntwdas null)))))
  (test larger-set
    (complete-graph (set 'oelrd 'afaodo 'euem 'esor))
    1 (graph?)
    (match? graph=? (make-graph (set 'oelrd 'afaodo 'euem 'esor)
                                (arcset (make-arc 'oelrd 'oelrd null)
                                        (make-arc 'oelrd 'afaodo null)
                                        (make-arc 'oelrd 'euem null)
                                        (make-arc 'oelrd 'esor null)
                                        (make-arc 'afaodo 'oelrd null)
                                        (make-arc 'afaodo 'afaodo null)
                                        (make-arc 'afaodo 'euem null)
                                        (make-arc 'afaodo 'esor null)
                                        (make-arc 'euem 'oelrd null)
                                        (make-arc 'euem 'afaodo null)
                                        (make-arc 'euem 'euem null)
                                        (make-arc 'euem 'esor null)
                                        (make-arc 'esor 'oelrd null)
                                        (make-arc 'esor 'afaodo null)
                                        (make-arc 'esor 'euem null)
                                        (make-arc 'esor 'esor null))))))

(suite graph-converse ()
  (test empty-graph
    (graph-converse (empty-graph))
    1 (graph?)
    (match? graph=? (empty-graph)))
  (test small-graph
    (graph-converse (make-graph (set 'svyt 'dredsap)
                                (arcset (make-arc 'svyt 'dredsap 'tcae))))
    1 (graph?)
    (match? graph=? (make-graph (set 'svyt 'dredsap)
                                (arcset (make-arc 'dredsap 'svyt 'tcae)))))
  (test larger-graph
    (graph-converse (make-graph (set 'srbes 'leonatc 'medtrh 'aopl 'ffrh)
                                (arcset (make-arc 'srbes 'leonatc null)
                                        (make-arc 'srbes 'aopl null)
                                        (make-arc 'leonatc 'srbes null)
                                        (make-arc 'leonatc 'medtrh null)
                                        (make-arc 'leonatc 'ffrh null)
                                        (make-arc 'medtrh 'leonatc null)
                                        (make-arc 'medtrh 'ffrh null)
                                        (make-arc 'aopl 'srbes null)
                                        (make-arc 'aopl 'leonatc null)
                                        (make-arc 'ffrh 'srbes null)
                                        (make-arc 'ffrh 'leonatc null)
                                        (make-arc 'ffrh 'medtrh null)
                                        (make-arc 'ffrh 'ffrh null))))
    1 (graph?)
    (match? graph=? (make-graph (set 'srbes 'leonatc 'medtrh 'aopl 'ffrh)
                                (arcset (make-arc 'srbes 'leonatc null)
                                        (make-arc 'srbes 'aopl null)
                                        (make-arc 'srbes 'ffrh null)
                                        (make-arc 'leonatc 'srbes null)
                                        (make-arc 'leonatc 'medtrh null)
                                        (make-arc 'leonatc 'aopl null)
                                        (make-arc 'leonatc 'ffrh null)
                                        (make-arc 'medtrh 'leonatc null)
                                        (make-arc 'medtrh 'ffrh null)
                                        (make-arc 'aopl 'srbes null)
                                        (make-arc 'ffrh 'leonatc null)
                                        (make-arc 'ffrh 'medtrh null)
                                        (make-arc 'ffrh 'ffrh null))))))

(suite alternative-graph-converse ()
  (test empty-graph
    (alternative-graph-converse (empty-graph))
    1 (graph?)
    (match? graph=? (empty-graph)))
  (test small-graph
    (alternative-graph-converse
      (make-graph (set 'bhioue 'wwshg)
                  (arcset (make-arc 'bhioue 'wwshg 'ymydoso))))
    1 (graph?)
    (match? graph=? (make-graph (set 'bhioue 'wwshg)
                                (arcset (make-arc 'wwshg 'bhioue null)))))
  (test larger-graph
    (alternative-graph-converse
      (make-graph (set 'nhr 'lylwsh 'tia 'hmo)
                  (arcset (make-arc 'nhr 'nhr null)
                          (make-arc 'lylwsh 'nhr null)
                          (make-arc 'lylwsh 'hmo null)
                          (make-arc 'tia 'nhr null)
                          (make-arc 'tia 'tia null)
                          (make-arc 'hmo 'nhr null))))
    1 (graph?)
    (match? graph=? (make-graph (set 'nhr 'lylwsh 'tia 'hmo)
                                (arcset (make-arc 'nhr 'nhr null)
                                        (make-arc 'nhr 'lylwsh null)
                                        (make-arc 'nhr 'tia null)
                                        (make-arc 'nhr 'hmo null)
                                        (make-arc 'tia 'tia null)
                                        (make-arc 'hmo 'lylwsh null))))))

(suite graph-complement ()
  (test empty-graph
    (graph-complement (empty-graph))
    1 (graph?)
    (match? graph=? (empty-graph)))
  (test singleton-graph
    (graph-complement (make-graph (set 'ennl) (arcset)))
    1 (graph?)
    (match? graph=?
            (make-graph (set 'ennl) (arcset (make-arc 'ennl 'ennl null)))))
  (test larger-graph
    (graph-complement
      (make-graph (set 'dcneetm 'lahtim 'mae 'tskimi)
                  (arcset (make-arc 'dcneetm 'lahtim 'nlas)
                          (make-arc 'lahtim 'dcneetm 'taorko)
                          (make-arc 'lahtim 'tskimi 'toerknt)
                          (make-arc 'mae 'dcneetm 'nrlrre)
                          (make-arc 'mae 'mae 'iwe)
                          (make-arc 'tskimi 'lahtim 'nauscu)
                          (make-arc 'tskimi 'mae 'tlm))))
    1 (graph?)
    (match? graph=?
            (make-graph (set 'dcneetm 'lahtim 'mae 'tskimi)
                        (arcset (make-arc 'dcneetm 'dcneetm null)
                                (make-arc 'dcneetm 'mae null)
                                (make-arc 'dcneetm 'tskimi null)
                                (make-arc 'lahtim 'lahtim null)
                                (make-arc 'lahtim 'mae null)
                                (make-arc 'mae 'lahtim null)
                                (make-arc 'mae 'tskimi null)
                                (make-arc 'tskimi 'dcneetm null)
                                (make-arc 'tskimi 'tskimi null))))))

(suite graph-product ()
  (test empty-graphs
    (graph-product (empty-graph) (empty-graph))
    1 (graph?)
    (match? graph=? (empty-graph)))
  (test no-left-arcs
    (graph-product (make-graph (set 'dnblgoi 'ndhe 'reolse) (arcset))
                   (complete-graph (set 'dnblgoi 'ndhe 'reolse)))
    1 (graph?)
    (match? graph=? (make-graph (set 'dnblgoi 'ndhe 'reolse) (arcset))))
  (test no-right-arcs
    (graph-product (complete-graph (set 'eatne 'nahngra 'tnd 'omcnlww))
                   (make-graph (set 'eatne 'nahngra 'tnd 'omcnlww)
                               (arcset)))
    1 (graph?)
    (match? graph=? (make-graph (set 'eatne 'nahngra 'tnd 'omcnlww)
                                (arcset))))
  (test no-arcs-in-result
    (graph-product (make-graph (set 'rmhsfm 'ero 'mtae)
                               (arcset (make-arc 'rmhsfm 'ero null)
                                       (make-arc 'ero 'ero null)
                                       (make-arc 'mtae 'ero null)))
                   (make-graph (set 'rmhsfm 'ero 'mtae)
                               (arcset (make-arc 'rmhsrm 'rmhsrm null)
                                       (make-arc 'rmhsrm 'ero null)
                                       (make-arc 'rmhsrm 'mtae null)
                                       (make-arc 'mtae 'rmhsrm null)
                                       (make-arc 'mtae 'ero null)
                                       (make-arc 'mtae 'mtae null))))
    1 (graph?)
    (match? graph=? (make-graph (set 'rmhsfm 'ero 'mtae) (arcset))))
  (test general-case
    (graph-product (make-graph (set 'olit 'ufevsyi 'haonscs 'eddcd 'reiha)
                               (arcset (make-arc 'olit 'olit null)
                                       (make-arc 'olit 'reiha null)
                                       (make-arc 'ufevsyi 'haonscs null)
                                       (make-arc 'ufevsyi 'eddcd null)
                                       (make-arc 'ufevsyi 'reiha null)
                                       (make-arc 'haonscs 'ufevsyi null)
                                       (make-arc 'haonscs 'eddcd null)
                                       (make-arc 'eddcd 'reiha null)))
                   (make-graph (set 'olit 'ufevsyi 'haonscs 'eddcd 'reiha)
                               (arcset (make-arc 'olit 'olit null)
                                       (make-arc 'olit 'ufevsyi null)
                                       (make-arc 'olit 'haonscs null)
                                       (make-arc 'olit 'eddcd null)
                                       (make-arc 'ufevsyi 'olit null)
                                       (make-arc 'ufevsyi 'ufevsyi null)
                                       (make-arc 'ufevsyi 'haonscs null)
                                       (make-arc 'ufevsyi 'reiha null)
                                       (make-arc 'haonscs 'olit null)
                                       (make-arc 'haonscs 'ufevsyi null)
                                       (make-arc 'haonscs 'haonscs null)
                                       (make-arc 'haonscs 'reiha null)
                                       (make-arc 'eddcd 'ufevsyi null)
                                       (make-arc 'eddcd 'haonscs null)
                                       (make-arc 'reiha 'olit null)
                                       (make-arc 'reiha 'ufevsyi null))))
    1 (graph?)
    (match? graph=? (make-graph (set 'olit 'ufevsyi 'haonscs 'eddcd 'reiha)
                                (arcset (make-arc 'olit 'olit null)
                                        (make-arc 'olit 'ufevsyi null)
                                        (make-arc 'olit 'haonscs null)
                                        (make-arc 'olit 'eddcd null)
                                        (make-arc 'ufevsyi 'olit null)
                                        (make-arc 'ufevsyi 'ufevsyi null)
                                        (make-arc 'ufevsyi 'haonscs null)
                                        (make-arc 'ufevsyi 'reiha null)
                                        (make-arc 'haonscs 'olit null)
                                        (make-arc 'haonscs 'ufevsyi null)
                                        (make-arc 'haonscs 'haonscs null)
                                        (make-arc 'haonscs 'reiha null)
                                        (make-arc 'eddcd 'olit null)
                                        (make-arc 'eddcd 'ufevsyi null))))))

(suite reflexive? ()
  (test empty-graph
    (reflexive? (empty-graph))
    1 (boolean?)
    true?)
  (test one-vertex-no-arcs
    (reflexive? (make-graph (set 'einy) (arcset)))
    1 (boolean?)
    not)
  (test one-vertex-one-arc
    (reflexive? (make-graph (set 'syds)
                            (arcset (make-arc 'syds 'syds 'iials))))
    1 (boolean?)
    true?)
  (test many-vertices-yes
    (reflexive? (make-graph (set 'hewdin 'ryun 'cei 'wiienn)
                            (arcset (make-arc 'hewdin 'hewdin 'itme)
                                    (make-arc 'hewdin 'ryun 'genhhoe)
                                    (make-arc 'hewdin 'cei 'eoieih)
                                    (make-arc 'ryun 'ryun 'mntr)
                                    (make-arc 'cei 'hewdin 'aadasas)
                                    (make-arc 'cei 'ryun 'hassae)
                                    (make-arc 'cei 'cei 'deu)
                                    (make-arc 'cei 'wiienn 'ffea)
                                    (make-arc 'wiienn 'cei 'gco)
                                    (make-arc 'wiienn 'wiienn 'ahegt))))
    1 (boolean?)
    true?)
  (test many-vertices-no
    (reflexive? (make-graph (set 'dlu 'egaerra 'aaarnw)
                            (arcset (make-arc 'dlu 'dlu 'riude)
                                    (make-arc 'dlu 'egaerra 'gwrine)
                                    (make-arc 'dlu 'aaarnw 'iervss)
                                    (make-arc 'egaerra 'egaerra 'eosu)
                                    (make-arc 'egaerra 'aaarnw 'eeldu))))
    1 (boolean?)
    not)
  (test many-vertices-irreflexive
    (reflexive? (make-graph (set 'tta 'bfulc 'aoinm)
                            (arcset (make-arc 'tta 'bfulc 'trein)
                                    (make-arc 'tta 'aoinm 'ioinnet)
                                    (make-arc 'aoinm 'tta 'eoeseit))))
    1 (boolean?)
    not))

(suite irreflexive? ()
  (test empty-graph
    (irreflexive? (empty-graph))
    1 (boolean?)
    true?)
  (test one-vertex-no-arcs
    (irreflexive? (make-graph (set 'atget) (arcset)))
    1 (boolean?)
    true?)
  (test one-vertex-one-arc
    (irreflexive? (make-graph (set 'kopmro)
                              (arcset (make-arc 'kopmro 'kopmro 'roo))))
    1 (boolean?)
    not)
  (test many-vertices-some-arcs-irreflexive
    (irreflexive? (make-graph (set 'gtwu 'naosuw 'lsuos 'nsfrir)
                              (arcset (make-arc 'gtwu 'naosuw 'hobeet)
                                      (make-arc 'gtwu 'nsfrir 'rsrm)
                                      (make-arc 'naosuw 'lsuos 'ywsaaf)
                                      (make-arc 'lsuos 'nsfrir 'ifehaio))))
    1 (boolean?)
    true?)
  (test many-vertices-some-arcs-not-irreflexive
    (irreflexive? (make-graph (set 'eihhps 'ilns 'hrtoein)
                              (arcset (make-arc 'eihhps 'ilns 'dhiad)
                                      (make-arc 'ilns 'eihhps 'dhsin)
                                      (make-arc 'ilns 'ilns 'thaeiwn)
                                      (make-arc 'hrtoein 'eihhps 'ear))))
    1 (boolean?)
    not))

(suite symmetric? ()
  (test empty-graph
    (symmetric? (empty-graph))
    1 (boolean?)
    true?)
  (test single-vertex
    (symmetric? (make-graph (set 'sred)
                            (arcset (make-arc 'sred 'sred 'swlum))))
    1 (boolean?)
    true?)
  (test no-arcs
    (symmetric? (make-graph (set 'osvfie 'netbh 'ariti) (arcset)))
    1 (boolean?)
    true?)
  (test minimal-non-symmetric
    (symmetric? (make-graph (set 'bth 'rdph)
                            (arcset (make-arc 'bth 'rdph 'nadhhen))))
    1 (boolean?)
    not)
  (test label-mismatch
    (symmetric? (make-graph (set 'aes 'iua)
                            (arcset (make-arc 'aes 'iua 'naearny)
                                    (make-arc 'iua 'aes  'asln))))
    1 (boolean?)
    not)
  (test larger-symmetric
    (symmetric? (make-graph (set 'uygdov 'dtuihh 'datoyy 'nvhnutp)
                            (arcset (make-arc 'uygdov 'uygdov 'ton)
                                    (make-arc 'uygdov 'dtuihh 'ddyor)
                                    (make-arc 'dtuihh 'uygdov 'ddyor)
                                    (make-arc 'nvhnutp 'nvhnutp
                                              'tulrtdn))))
    1 (boolean?)
    true?)
  (test larger-non-symmetric
    (symmetric? (make-graph (set 'sartonh 'wscltw 'itah 'scrlnja 'didnm)
                            (arcset (make-arc 'sartonh 'itah 'pcs)
                                    (make-arc 'itah 'sartonh 'pcs)
                                    (make-arc 'wscltw 'didnm 'smef)
                                    (make-arc 'didnm 'scrlnja 'srdemsr)
                                    (make-arc 'scrlnja 'wscltw 'frteaea)
                                    (make-arc 'itah 'didnm 'unshh)
                                    (make-arc 'didnm 'itah 'unshh))))
    1 (boolean?)
    not))

(suite asymmetric? ()
  (test empty-graph
    (asymmetric? (empty-graph))
    1 (boolean?)
    true?)
  (test single-vertex
    (asymmetric? (make-graph (set 'wrbgr)
                             (arcset (make-arc 'wrbgr 'wrbgr 'ato))))
    1 (boolean?)
    not)
  (test no-arcs
    (symmetric? (make-graph (set 'fyoad 'feet 'sdy 'dbtpd 'eodhc)
                            (arcset)))
    1 (boolean?)
    true?)
  (test minimal-non-asymmetric
    (asymmetric? (make-graph (set 'gaoya 'neloa)
                             (arcset (make-arc 'gaoya 'neloa 'osese)
                                     (make-arc 'neloa 'gaoya 'osese))))
    1 (boolean?)
    not)
  (test label-mismatch
    (asymmetric? (make-graph (set 'gptooot 'ltla)
                             (arcset (make-arc 'gptooot 'ltla 'dssnte)
                                     (make-arc 'ltla 'gptooot 'wtersn))))
    1 (boolean?)
    not)
  (test larger-asymmetric
    (asymmetric? (make-graph (set 'bnd 'nok 'ttfeer 'butaod)
                             (arcset (make-arc 'bnd 'ttfeer null)
                                     (make-arc 'bnd 'butaod null)
                                     (make-arc 'nok 'butaod null)
                                     (make-arc 'ttfeer 'nok null)
                                     (make-arc 'butaod 'ttfeer null))))
    1 (boolean?)
    true?)
  (test larger-non-asymmetric
    (asymmetric? (make-graph (set 'aondo 'autol 'ouetd)
                             (arcset (make-arc 'aondo 'aondo 'nnhe)
                                     (make-arc 'aondo 'ouetd 'itt)
                                     (make-arc 'autol 'ouetd 'rteovae)
                                     (make-arc 'ouetd 'aondo 'itt)
                                     (make-arc 'ouetd 'ouetd 'parueio))))
    1 (boolean?)
    not))

(suite connected? ()
  (test empty-graph
    (connected? (empty-graph))
    1 (boolean?)
    true?)
  (test singleton-yes
    (connected? (make-graph (set 'rhieuhs)
                            (arcset (make-arc 'rhieuhs 'rhieuhs 'rtslli))))
    1 (boolean?)
    true?)
  (test singleton-no
    (connected? (make-graph (set 'ahdnt) (arcset)))
    1 (boolean?)
    not)
  (test minimal-nontrivial
    (connected? (make-graph (set 'icpaf 'epi)
                            (arcset (make-arc 'icpaf 'icpaf 'kab)
                                    (make-arc 'icpaf 'epi 'wratd)
                                    (make-arc 'epi 'epi 'yiinnit))))
    1 (boolean?)
    true?)
  (test larger-yes
    (connected? (make-graph (set 'apsatn 'hnrai 'enarm 'rnsbdeo)
                            (arcset (make-arc 'apsatn 'apsatn null)
                                    (make-arc 'apsatn 'hnrai null)
                                    (make-arc 'apsatn 'rnsbdeo null)
                                    (make-arc 'hnrai 'apsatn null)
                                    (make-arc 'hnrai 'hnrai null)
                                    (make-arc 'hnrai 'rnsbdeo null)
                                    (make-arc 'enarm 'apsatn null)
                                    (make-arc 'enarm 'hnrai null)
                                    (make-arc 'enarm 'enarm null)
                                    (make-arc 'rnsbdeo 'enarm null)
                                    (make-arc 'rnsbdeo 'rnsbdeo null))))
    1 (boolean?)
    true?)
  (test larger-no
    (connected? (make-graph (set 'tscud 'tao 'nslscl 'ioa)
                            (arcset (make-arc 'tscud 'tscud 'pnoot)
                                    (make-arc 'tao 'tscud 'dfnal)
                                    (make-arc 'tao 'tao 'ntaege)
                                    (make-arc 'tao 'ioa 'eeoo)
                                    (make-arc 'nslscl 'nslscl 'aser)
                                    (make-arc 'nslscl 'ioa 'eruetn)
                                    (make-arc 'ioa 'tao 'eeoo)
                                    (make-arc 'ioa 'ioa 'hswrh))))
    1 (boolean?)
    not))    

(suite transitive? ()
  (test empty-graph
    (transitive? (empty-graph))
    1 (boolean?)
    true?)
  (test minimal-nontrivial-yes
    (transitive? (make-graph (set 'oleeset 'oforel 'agttnol)
                             (arcset (make-arc 'oleeset 'oforel 'saoeieg)
                                     (make-arc 'oleeset 'agttnol 'iat)
                                     (make-arc 'oforel 'agttnol
                                               'tidnadt))))
    1 (boolean?)
    true?)
  (test minimal-non-trivial-no
    (transitive? (make-graph (set 'ayb 'eac 'uesic)
                             (arcset (make-arc 'ayb 'eac null)
                                     (make-arc 'eac 'uesic null))))
    1 (boolean?)
    not)
  (test larger-yes
    (transitive? (relation-graph (set 529 946 600 84 679 55 612) <))
    1 (boolean?)
    true?)
  (test larger-no
    (transitive? (relation-graph (set 180 891 440 821 677 435)
                                 (pipe + (sect < <> 1000))))
    1 (boolean?)
    not))

(suite intransitive? ()
  (test empty-graph
    (intransitive? (empty-graph))
    1 (boolean?)
    true?)
  (test minimal-nontrivial-yes
    (intransitive? (make-graph (set 'stact 'rmtf 'avid)
                               (arcset (make-arc 'stact 'rmtf 'nlssrtr)
                                       (make-arc 'rmtf 'stact 'etfdili))))
    1 (boolean?)
    true?)
  (test minimal-nontrivial-no
    (intransitive? (make-graph (set 'titeo 'osaa 'see)
                               (arcset (make-arc 'titeo 'osaa 'itweeca)
                                       (make-arc 'titeo 'see 'nbstd)
                                       (make-arc 'osaa 'see 'obae))))
    1 (boolean?)
    not)
  (test larger-yes
    (intransitive? (make-graph (set 'ilehl 'huta 'nri 'grk 'ogioaom)
                               (arcset (make-arc 'huta 'nri null)
                                       (make-arc 'nri 'grk null)
                                       (make-arc 'grk 'huta null)
                                       (make-arc 'grk 'ogioaom null)
                                       (make-arc 'ogioaom 'grk null))))
    1 (boolean?)
    true?)
  (test larger-no
    (intransitive? (make-graph (set 'etast 'omdmwh 'nyyamnb 'hedt 'otoh)
                               (arcset (make-arc 'etast 'omdmwh null)
                                       (make-arc 'etast 'hedt null)
                                       (make-arc 'omdmwh 'etast null)
                                       (make-arc 'omdmwh 'hedt null)
                                       (make-arc 'omdmwh 'otoh null)
                                       (make-arc 'nyyamnb 'omdmwh null)
                                       (make-arc 'nyyamnb 'otoh null)
                                       (make-arc 'hedt 'etast null)
                                       (make-arc 'hedt 'nyyamnb null)
                                       (make-arc 'hedt 'otoh null)
                                       (make-arc 'otoh 'omdmwh null)
                                       (make-arc 'otoh 'hedt null))))
    1 (boolean?)
    not))

(suite equivalence-relation? ()
  (test empty-graph
    (equivalence-relation? (empty-graph))
    1 (boolean?)
    true?)
  (test nontrivial-yes
    (equivalence-relation?
      (make-graph (set 'entrm 'rfdenhe 'hio)
                  (arcset (make-arc 'entrm 'entrm 'eesu)
                          (make-arc 'rfdenhe 'rfdenhe 'eevhcha)
                          (make-arc 'rfdenhe 'hio 'rtl)
                          (make-arc 'hio 'rfdenhe 'rtl)
                          (make-arc 'hio 'hio 'eped))))
    1 (boolean?)
    true?)
  (test minimal-non-reflexive
    (equivalence-relation? (make-graph (set 'utih) (arcset)))
    1 (boolean?)
    not)
  (test minimal-non-symmetric
    (equivalence-relation?
      (make-graph (set 'bah 'hieao)
                  (arcset (make-arc 'bah 'bah null)
                          (make-arc 'bah 'hieao null)
                          (make-arc 'hieao 'hiaeo null))))
    1 (boolean?)
    not)
  (test minimal-non-transitive
    (equivalence-relation?
      (make-graph (set 'eeedsc 'ahntru 'twewga)
                  (arcset (make-arc 'eeedsc 'eeedsc 'rul)
                          (make-arc 'eeedsc 'ahntru 'mpanyfn)
                          (make-arc 'ahntru 'eeedsc 'mpanyfn)
                          (make-arc 'ahntru 'ahntru 'iysos)
                          (make-arc 'ahntru 'twewga 'yisei)
                          (make-arc 'twewga 'ahntru 'yisei)
                          (make-arc 'twewga 'twewga 'etpvsrn))))
    1 (boolean?)
    not)
  (test larger-yes
    (equivalence-relation?
      (make-graph (set 'tnnnp 'iwohsot 'rjde 'tboi 'yset 'ieb)
                  (arcset (make-arc 'tnnnp 'tnnnp null)
                          (make-arc 'tnnnp 'tboi null)
                          (make-arc 'tnnnp 'ieb null)
                          (make-arc 'iwohsot 'iwohsot null)
                          (make-arc 'iwohsot 'yset null)
                          (make-arc 'rjde 'rjde null)
                          (make-arc 'tboi 'tnnnp null)
                          (make-arc 'tboi 'tboi null)
                          (make-arc 'tboi 'ieb null)
                          (make-arc 'yset 'iwohsot null)
                          (make-arc 'yset 'yset null)
                          (make-arc 'ieb 'tnnnp null)
                          (make-arc 'ieb 'tboi null)
                          (make-arc 'ieb 'ieb null))))
    1 (boolean?)
    true?)
  (test larger-no
    (equivalence-relation?
      (make-graph (set 'ilc 'arshier 'inlma 'sio 'leifaa)
                  (arcset (make-arc 'ilc 'sio 'neht)
                          (make-arc 'ilc 'leifaa 'ciaayp)
                          (make-arc 'arshier 'ilc 'ermudyi)
                          (make-arc 'arshier 'arshier 'ahmayo)
                          (make-arc 'arshier 'inlma 'asrg)
                          (make-arc 'arshier 'leifaa 'hotrsr)
                          (make-arc 'inlma 'arshier 'asrg)
                          (make-arc 'inlma 'sio 'iawuio)
                          (make-arc 'inlma 'leifaa 'syee)
                          (make-arc 'sio 'arshier 'eatatdn)
                          (make-arc 'leifaa 'ilc 'ciaayp)
                          (make-arc 'leifaa 'arshier 'hotrsr)
                          (make-arc 'leifaa 'inlma 'syee)
                          (make-arc 'leifaa 'leifaa 'onnh))))
    1 (boolean?)
    not))

(suite ordering-relation? ()
  (test empty-graph
    (ordering-relation? (empty-graph))
    1 (boolean?)
    true?)
  (test complete-graph
    (ordering-relation?
      (complete-graph (set 'uihnt 'rdeeao 'rbf 'elsdsn 'eaisrof)))
    1 (boolean?)
    true?)
  (test minimal-nontrivial
    (ordering-relation?
      (make-graph (set 'eeatfr 'syiun)
                  (arcset (make-arc 'eeatfr 'eeatfr null)
                          (make-arc 'syiun 'eeatfr null)
                          (make-arc 'syiun 'syiun null))))
    1 (boolean?)
    true?)
  (test unconnected
    (ordering-relation?
      (make-graph (set 'ewpi 'tode 'bymnvne 'hhegl)
                  (arcset (make-arc 'ewpi 'ewpi 'imtaolm)
                          (make-arc 'ewpi 'tode 'emstis)
                          (make-arc 'ewpi 'bymnvne 'tafluha)
                          (make-arc 'ewpi 'hhegl 'etbrmt)
                          (make-arc 'tode 'tode 'tsyh)
                          (make-arc 'tode 'hhegl 'whir)
                          (make-arc 'bymnvne 'bymnvne 'ooeo)
                          (make-arc 'bymnvne 'hhegl 'nmaro)
                          (make-arc 'hhegl 'hhegl 'eaela))))
    1 (boolean?)
    not)
  (test nontransitive
    (ordering-relation?
      (make-graph (set 'hold 'oupi 'dhnu)
                  (arcset (make-arc 'hold 'hold null)
                          (make-arc 'hold 'oupi null)
                          (make-arc 'oupi 'oupi null)
                          (make-arc 'oupi 'dhnu null)
                          (make-arc 'dhnu 'hold null)
                          (make-arc 'dhnu 'dhnu null))))
    1 (boolean?)
    not)
  (test larger-yes
    (ordering-relation? (relation-graph (set 200 248 155 692 872 157 25)
                                        >=))
    1 (boolean?)
    true?)
  (test larger-no
    (ordering-relation? (relation-graph (set 446 780 930 859 324 115 296)
                                        (pipe + (sect <= <> 1000))))
    1 (boolean?)
    not))

(suite arcs-leaving ()
  (test no-arcs
    (arcs-leaving 'leeln
      (make-graph (set 'hssfc 'leeln 'ohloel 'isiacc 'walpyio 'atfoh
                       'fylhhos 'losetuy 'aomioh 'ebnwrr)
                  (arcset (make-arc 'hssfc 'hssfc 'nhpast)
                          (make-arc 'hssfc 'leeln 'hlifn)
                          (make-arc 'hssfc 'isiacc 'foat)
                          (make-arc 'hssfc 'atfoh 'eoysnr)
                          (make-arc 'hssfc 'fylhhos 'eer)
                          (make-arc 'hssfc 'losetuy 'eniiu)
                          (make-arc 'hssfc 'aomioh 'vhfnt)
                          (make-arc 'ohloel 'ohloel 'hlhhalp)
                          (make-arc 'ohloel 'walpyio 'aghrite)
                          (make-arc 'ohloel 'atfoh 'orseoc)
                          (make-arc 'ohloel 'losetuy 'ghlerlp)
                          (make-arc 'isiacc 'hssfc 'stynit)
                          (make-arc 'isiacc 'ohloel 'rdsei)
                          (make-arc 'isiacc 'isiacc 'lduo)
                          (make-arc 'isiacc 'walpyio 'onnse)
                          (make-arc 'isiacc 'atfoh 'qet)
                          (make-arc 'isiacc 'fylhhos 'stlu)
                          (make-arc 'isiacc 'aomioh 'gemnsah)
                          (make-arc 'isiacc 'ebnwrr 'gljh)
                          (make-arc 'walpyio 'walpyio 'gda)
                          (make-arc 'walpyio 'atfoh 'nee)
                          (make-arc 'walpyio 'fylhhos 'indy)
                          (make-arc 'walpyio 'losetuy 'nigb)
                          (make-arc 'walpyio 'aomioh 'teeo)
                          (make-arc 'walpyio 'ebnwrr 'aoac)
                          (make-arc 'atfoh 'hssfc 'karekd)
                          (make-arc 'atfoh 'ohloel 'eopm)
                          (make-arc 'atfoh 'losetuy 'ego)
                          (make-arc 'atfoh 'aomioh 'shwdo)
                          (make-arc 'fylhhos 'hssfc 'lln)
                          (make-arc 'fylhhos 'leeln 'cgdoo)
                          (make-arc 'fylhhos 'atfoh 'dvcs)
                          (make-arc 'fylhhos 'fylhhos 'oag)
                          (make-arc 'losetuy 'hssfc 'ipcyi)
                          (make-arc 'losetuy 'leeln 'wra)
                          (make-arc 'losetuy 'ohloel 'agohsul)
                          (make-arc 'losetuy 'walpyio 'sdv)
                          (make-arc 'losetuy 'atfoh 'hnyrhhf)
                          (make-arc 'losetuy 'losetuy 'tujdf)
                          (make-arc 'losetuy 'ebnwrr 'wbdeep)
                          (make-arc 'aomioh 'isiacc 'stdiefa)
                          (make-arc 'aomioh 'walpyio 'ruandbe)
                          (make-arc 'aomioh 'atfoh 'knoetod)
                          (make-arc 'aomioh 'fylhhos 'hyn)
                          (make-arc 'aomioh 'aomioh 'rorlgrh)
                          (make-arc 'ebnwrr 'leeln 'idtuid)
                          (make-arc 'ebnwrr 'walpyio 'esrerh)
                          (make-arc 'ebnwrr 'fylhhos 'elateee))))
    1 ((set-of arc? same-endpoints?))
    empty-set?)
  (test some-arcs
    (arcs-leaving 'dkavash
      (make-graph (set 'sud 'sdabpnr 'dkavash 'ehssera)
                  (arcset (make-arc 'sud 'sud 'eelw)
                          (make-arc 'sud 'sdabpnr 'fhnh)
                          (make-arc 'sud 'dkavash 'aen)
                          (make-arc 'sdabpnr 'sud 'tft)
                          (make-arc 'sdabpnr 'ehssera 'aeattdi)
                          (make-arc 'dkavash 'sdabpnr 'oaytlrg)
                          (make-arc 'dkavash 'dkavash 'idsss)
                          (make-arc 'dkavash 'ehssera 'iedwt)
                          (make-arc 'ehssera 'ehssera 'brade))))
    1 ((set-of arc? same-endpoints?))
    (match? (set-of= arc=?) (arcset (make-arc 'dkavash 'sdabpnr 'oaytlrg)
                                    (make-arc 'dkavash 'dkavash 'idsss)
                                    (make-arc 'dkavash 'ehssera 'iedwt)))))

(suite arcs-arriving ()
  (test no-arcs
    (arcs-arriving 'caeo
                   (make-graph (set 'eeeeras 'lmimogi 'oews 'caeo 'srs
                                    'arluuv) 
                               (arcset (make-arc 'eeeeras 'eeeeras null)
                                       (make-arc 'eeeeras 'lmimogi null)
                                       (make-arc 'eeeeras 'srs null)
                                       (make-arc 'eeeras 'arluuv null)
                                       (make-arc 'lmimogi 'eeeeras null)
                                       (make-arc 'lmimogi 'lmimogi null)
                                       (make-arc 'lmimogi 'oews null)
                                       (make-arc 'lmimogi 'srs null)
                                       (make-arc 'oews 'lmimogi null)
                                       (make-arc 'oews 'srs null)
                                       (make-arc 'caeo 'eeeeras null)
                                       (make-arc 'caeo 'lmimogi null)
                                       (make-arc 'caeo 'oews null)
                                       (make-arc 'caeo 'arluuv null)
                                       (make-arc 'srs 'lmimogi null)
                                       (make-arc 'srs 'oews null)
                                       (make-arc 'srs 'srs null)
                                       (make-arc 'srs 'arluuv null)
                                       (make-arc 'arluuv 'eeeeras null)
                                       (make-arc 'arluuv 'oews null))))
    1 ((set-of arc? same-endpoints?))
    empty-set?)
  (test some-arcs
    (arcs-arriving 'trhg (make-graph (set 'wtnt 'anuu 'elieai 'trhg)
                                     (arcset (make-arc 'wtnt 'anuu 584)
                                             (make-arc 'wtnt 'elieai 300)
                                             (make-arc 'anuu 'trhg 140)
                                             (make-arc 'elieai 'anuu 817)
                                             (make-arc 'elieai 'elieai 263)
                                             (make-arc 'elieai 'trhg 998)
                                             (make-arc 'trhg 'wtnt 826))))
    1 ((set-of arc? same-endpoints?))
    (match? (set-of= arc=?) (arcset (make-arc 'anuu 'trhg 140)
                                    (make-arc 'elieai 'trhg 998)))))

(suite slow-neighbors
    ((sample (make-graph (set 'tevhnh 'ngermo 'mne 'naotwa 'iesnt)
                         (arcset (make-arc 'ngermo 'naotwa null)
                                 (make-arc 'mne 'ngermo null)
                                 (make-arc 'mne 'mne null)
                                 (make-arc 'mne 'naotwa null)
                                 (make-arc 'naotwa 'naotwa null)
                                 (make-arc 'iesnt 'iesnt null)))))
  (test no-arcs
    (slow-neighbors 'eeuci
                    (make-graph (set 'glaens 'odse 'mnmelni 'eeuci
                                     'lecanst)
                                (arcset)))
    1 (set?)
    empty-set?)
  (test self-sufficient
    (slow-neighbors 'iesnt sample)
    1 (set?)
    (match? set=? (set 'iesnt)))
  (test curmudgeon
    (slow-neighbors 'tevhnh sample)
    1 (set?)
    empty-set?)
  (test edge-of-town
    (slow-neighbors 'ngermo sample)
    1 (set?)
    (match? set=? (set 'naotwa)))
  (test more-neighbors
    (slow-neighbors 'mne sample)
    1 (set?)
    (match? set=? (set 'ngermo 'mne 'naotwa))))

(suite neighbors ()
  (test no-arcs
    (neighbors (make-graph (set 'sew 'evegd 'omoa 'aaaosor) (arcset)))
    1 (procedure?)
    (lambda (neighbor-provider)
      (suite no-arcs-internal ()
        (test any-vertex
          (neighbor-provider 'evegd)
          1 (set?)
          empty-set?))))
  (test variety-of-cases
    (neighbors (make-graph (set 'redatts 'saoo 'ede 'znteup 'ohe 'airu)
                           (arcset (make-arc 'saoo 'ede 'aet)
                                   (make-arc 'ede 'znteup 'hioa)
                                   (make-arc 'ede 'ohe 'dau)
                                   (make-arc 'ede 'airu 'ante)
                                   (make-arc 'znteup 'ohe 'wecan)
                                   (make-arc 'ohe 'airu 'iunmond))))
    1 (procedure?)
    (lambda (neighbor-provider)
      (suite variety-internal ()
        (test hermit
          (neighbor-provider 'redatts)
          1 (set?)
          empty-set?)
        (test edge-of-town
          (neighbor-provider 'saoo)
          1 (set?)
          (match? set=? (set 'ede)))
        (test more-neighbors
          (neighbor-provider 'ede)
          1 (set?)
          (match? set=? (set 'znteup 'ohe 'airu)))))))

(suite label-structure
    ((right-type? (table-of symbol? (table-of symbol? symbol? symbol=?)
                            symbol=?))
     (equiv? (table-of= symbol=? (table-of= symbol=? symbol=?))))
  (test empty-graph
    (label-structure (empty-graph))
    1 (table?)
    (match? table=? (table)))
  (test singleton-graph
    (label-structure
      (make-graph (set 'iachhhc)
                  (arcset (make-arc 'iachhhc 'iachhhc 'hsruchf))))
    1 (right-type?)
    (match? equiv? (table (cons 'iachhhc
                                (table (cons 'iachhhc 'hsruchf))))))
  (test larger-graph
    (label-structure
      (make-graph (set 'nsore 'regefs 'soaoog 'ttyxw)
                  (arcset (make-arc 'nsore 'soaoog 'ovetg)
                          (make-arc 'regefs 'regefs 'niaa)
                          (make-arc 'regefs 'ttyxw 'eoi)
                          (make-arc 'soaoog 'nsore 'meayian)
                          (make-arc 'soaoog 'regefs 'hnlae)
                          (make-arc 'soaoog 'ttyxw 'uoraal)
                          (make-arc 'ttyxw 'nsore 'sotttha))))
    1 (right-type?)
    (match? equiv? (table (cons 'nsore (table (cons 'soaoog 'ovetg)))
                          (cons 'regefs (table (cons 'regefs 'niaa)
                                               (cons 'ttyxw 'eoi)))
                          (cons 'soaoog (table (cons 'nsore 'meayian)
                                               (cons 'regefs 'hnlae)
                                               (cons 'ttyxw 'uoraal)))
                          (cons 'ttyxw (table (cons 'nsore 'sotttha)))))))

(suite label-lookup ()
  (test minimal-case
    (label-lookup (make-graph (set 'hhti 'solht)
                              (arcset (make-arc 'hhti 'solht 'fdawgoe))))
    1 (procedure?)
    (lambda (get-label)
      (suite minimal-internal ()
        (test succeed
          (get-label 'hhti 'solht)
          1 ((^vel symbol? not))
          (^et symbol? (match? symbol=? 'fdawgoe)))
        (test reversed-endpoints
          (get-label 'solht 'hhti)
          1 ((^vel symbol? not))
          not))))
  (test more-complicated
    (label-lookup (make-graph (set 'dthod 'teaatt 'hfbhee 'mnlethi 'urhdt)
                              (arcset (make-arc 'dthod 'dthod 'ltducrt)
                                      (make-arc 'dthod 'hfbhee 'ligh)
                                      (make-arc 'dthod 'urhdt 'gpew)
                                      (make-arc 'teaatt 'hfbhee 'eheow)
                                      (make-arc 'hfbhee 'mnlethi 'sontt)
                                      (make-arc 'hfbhee 'urhdt 'poeaaal)
                                      (make-arc 'mnlethi 'teaatt 'wese)
                                      (make-arc 'mnlethi 'urhdt 'aero)
                                      (make-arc 'urhdt 'mnlethi 'eatuco)))
                  'failed)
    1 (procedure?)
    (lambda (get-label)
      (suite more-complicated-internal ()
        (test succeed
          (get-label 'mnlethi 'urhdt)
          1 (symbol?)
          (match? symbol=? 'aero))
        (test back-arc
          (get-label 'urhdt 'mnlethi)
          1 (symbol?)
          (match? symbol=? 'eatuco))
        (test no-arc
          (get-label 'dthod 'mnlethi)
          1 (symbol?)
          (match? symbol=? 'failed))
        (test endpoints-match
          (get-label 'teaatt 'teaatt)
          1 (symbol?)
          (match? symbol=? 'failed))))))

(suite out-degree ()
  (test minimal-case
    (out-degree (make-graph (set 'easo) (arcset)))
    1 (procedure?)
    (lambda (degree-finder)
      (suite minimal-internal ()
        (test only-vertex
          (degree-finder 'easo)
          1 (natural-number?)
          zero?))))
  (test larger-graph
    (out-degree (make-graph (set 'eayvpae 'rlllltr 'leltihp 'teg 'pheaab
                                 'eah 'mrttwa)
                            (arcset (make-arc 'eayvpae 'eayvpae 'eop)
                                    (make-arc 'eayvpae 'rlllltr 'reto)
                                    (make-arc 'eayvpae 'leltihp 'lnf)
                                    (make-arc 'eayvpae 'teg 'ocitso)
                                    (make-arc 'rlllltr 'rlllltr 'tebl)
                                    (make-arc 'rlllltr 'leltihp 'evseek)
                                    (make-arc 'rlllltr 'mrttwa 'ileoso)
                                    (make-arc 'teg 'eayvpae 'syrwrll)
                                    (make-arc 'teg 'leltihp 'ldiblt)
                                    (make-arc 'teg 'teg 'iaraede)
                                    (make-arc 'teg 'pheaab 'tsoul)
                                    (make-arc 'teg 'eah 'hrndtge)
                                    (make-arc 'teg 'mrttwa 'emdante)
                                    (make-arc 'pheaab 'eayvpae 'ovib)
                                    (make-arc 'pheaab 'rlllltr 'pislnhb)
                                    (make-arc 'pheaab 'leltihp 'ohm)
                                    (make-arc 'pheaab 'eah 'iraoto)
                                    (make-arc 'eah 'eayvpae 'hur)
                                    (make-arc 'eah 'leltihp 'ievtia)
                                    (make-arc 'eah 'teg 'rdflrro)
                                    (make-arc 'eah 'pheaab 'rtyeoaa)
                                    (make-arc 'eah 'eah 'ihf)
                                    (make-arc 'mrttwa 'eayvpae 'nrmo)
                                    (make-arc 'mrttwa 'eah 'fktabn)
                                    (make-arc 'mrttwa 'mrttwa 'horesta))))
    1 (procedure?)
    (lambda (degree-finder)
      (suite larger-case-internal ()
        (test out-degree-zero
          (degree-finder 'leltihp)
          1 (natural-number?)
          zero?)
        (test out-degree-non-zero
          (degree-finder 'teg)
          1 (natural-number?)
          (match? = 6))))))

(suite in-degree ()
  (test minimal-case
    (in-degree (make-graph (set 'aalss) (arcset)))
    1 (procedure?)
    (lambda (degree-finder)
      (suite minimal-internal ()
        (test only-vertex
          (degree-finder 'aalss)
          1 (natural-number?)
          zero?))))
  (test larger-graph
    (in-degree (make-graph (set 'edgo 'rttyss 'igreln 'ibr 'fkecst 'htenm)
                           (arcset (make-arc 'edgo 'edgo 'ydo)
                                   (make-arc 'edgo 'rttyss 'eoi)
                                   (make-arc 'edgo 'ibr 'lhltf)
                                   (make-arc 'edgo 'htenm 'anu)
                                   (make-arc 'rttyss 'rttyss 'tedrai)
                                   (make-arc 'rttyss 'igreln 'ethga)
                                   (make-arc 'rttyss 'ibr 'lnoug)
                                   (make-arc 'igreln 'edgo 'eart)
                                   (make-arc 'igreln 'igreln 'pkt)
                                   (make-arc 'igreln 'ibr 'eia)
                                   (make-arc 'ibr 'edgo 'ahirot)
                                   (make-arc 'ibr 'rttyss 'eot)
                                   (make-arc 'fkecst 'edgo 'etoti)
                                   (make-arc 'fkecst 'igreln 'deemyi)
                                   (make-arc 'fkecst 'htenm 'sad)
                                   (make-arc 'htenm 'edgo 'otieayf)
                                   (make-arc 'htenm 'rttyss 'rst)
                                   (make-arc 'htenm 'igreln 'tnont)
                                   (make-arc 'htenm 'htenm 'bnrwroc))))
    1 (procedure?)
    (lambda (degree-finder)
      (suite larger-case-internal ()
        (test in-degree-zero
          (degree-finder 'fkecst)
          1 (natural-number?)
          zero?)
        (test in-degree-non-zero
          (degree-finder 'edgo)
          1 (natural-number?)
          (match? = 5))))))

(suite undirected? ()
  (test empty-graph
    (undirected? (empty-graph))
    1 (boolean?)
    true?)
  (test one-vertex-no-arcs
    (undirected? (make-graph (set 'adsd) (arcset)))
    1 (boolean?)
    true?)
  (test one-vertex-one-arc
    (undirected? (make-graph (set 'hso)
                             (arcset (make-arc 'hso 'hso 'uio))))
    1 (boolean?)
    not)
  (test no-arcs
    (undirected? (make-graph (set 'aeevi 'ien 'uata 'eyoewtn 'ilhegn
                                  'rropwe 'ggogdr)
                             (arcset)))
    1 (boolean?)
    true?)
  (test larger-graph-undirected
    (undirected? (make-graph (set 'hle 'aihhemt 'tseodpo 'ieot)
                             (arcset (make-arc 'hle 'aihhemt 'irewo)
                                     (make-arc 'aihhemt 'hle 'irewo)
                                     (make-arc 'hle 'tseodpo 'emci)
                                     (make-arc 'tseodpo 'hle 'emci)
                                     (make-arc 'hle 'ieot 'athch)
                                     (make-arc 'ieot 'hle 'athch)
                                     (make-arc 'aihhemt 'ieot 'deus)
                                     (make-arc 'ieot 'aihhemt 'deus))))
    1 (boolean?)
    true?)
  (test larger-graph-not-irreflexive
    (undirected? (make-graph (set 'drtt 'idu 'ntoshor 'aeu)
                             (arcset (make-arc 'drtt 'idu 'eam)
                                     (make-arc 'idu 'drtt 'eam)
                                     (make-arc 'idu 'ntoshor 'foiadd)
                                     (make-arc 'ntoshor 'idu 'foiadd)
                                     (make-arc 'idu 'idu 'ogrdfho)
                                     (make-arc 'ntoshor 'aeu 'ttsi)
                                     (make-arc 'aeu 'ntoshor 'ttsi))))
    1 (boolean?)
    not)
  (test larger-graph-not-symmetric
    (undirected? (make-graph (set 'iydehe 'iad 'itiola)
                             (arcset (make-arc 'iydehe 'iad 'spte)
                                     (make-arc 'itiola 'iydehe 'gryer)
                                     (make-arc 'iad 'itiola 'hrktnt))))
    1 (boolean?)
    not))

(suite add-labeled-edge ()
  (test minimal-case
    (add-labeled-edge 'ict 'fghole 'sedgs
      (make-graph (set 'ict 'fghole) (arcset)))
    1 (graph?)
    (match? graph=? (make-graph (set 'ict 'fghole)
                                (arcset (make-arc 'ict 'fghole 'sedgs)
                                        (make-arc 'fghole 'ict 'sedgs)))))
  (test larger-case
    (add-labeled-edge 'etk 'varneto 'erelrh
      (make-graph (set 'ueip 'varneto 'baloeiu 'etk 'odeec)
                  (arcset (make-arc 'ueip 'varneto 'tnth)
                          (make-arc 'ueip 'odeec 'nhhedt)
                          (make-arc 'varneto 'ueip 'tnth)
                          (make-arc 'varneto 'baloeiu 'gooenrh)
                          (make-arc 'baloeiu 'varneto 'gooenrh)
                          (make-arc 'baloeiu 'etk 'extsmi)
                          (make-arc 'etk 'baloeiu 'extsmi)
                          (make-arc 'etk 'odeec 'dey)
                          (make-arc 'odeec 'ueip 'nhhedt)
                          (make-arc 'odeec 'etk 'dey))))
    1 (graph?)
    (match? graph=?
            (make-graph (set 'ueip 'varneto 'baloeiu 'etk 'odeec)
                        (arcset (make-arc 'ueip 'varneto 'tnth)
                                (make-arc 'ueip 'odeec 'nhhedt)
                                (make-arc 'varneto 'ueip 'tnth)
                                (make-arc 'varneto 'baloeiu 'gooenrh)
                                (make-arc 'varneto 'etk 'erelrh)
                                (make-arc 'baloeiu 'varneto 'gooenrh)
                                (make-arc 'baloeiu 'etk 'extsmi)
                                (make-arc 'etk 'varneto 'erelrh)
                                (make-arc 'etk 'baloeiu 'extsmi)
                                (make-arc 'etk 'odeec 'dey)
                                (make-arc 'odeec 'ueip 'nhhedt)
                                (make-arc 'odeec 'etk 'dey))))))

(suite add-edge ()
  (test minimal-case
    (add-edge 'iauurfa 'ene (make-graph (set 'iauurfa 'ene) (arcset)))
    1 (graph?)
    (match? graph=? (make-graph (set 'iauurfa 'ene)
                                (arcset (make-arc 'iauurfa 'ene null)
                                        (make-arc 'ene 'iauurfa null)))))
  (test larger-case
    (add-edge 'neaaa 'atdhth
      (make-graph (set 'fsn 'atdhth 'tieow 'neaaa 'tsnnfa)
                  (arcset (make-arc 'fsn 'atdhth null)
                          (make-arc 'fsn 'tsnnfa null)
                          (make-arc 'atdhth 'fsn null)
                          (make-arc 'atdhth 'tieow null)
                          (make-arc 'tieow 'atdhth null)
                          (make-arc 'tieow 'neaaa null)
                          (make-arc 'neaaa 'tieow null)
                          (make-arc 'neaaa 'tsnnfa null)
                          (make-arc 'tsnnfa 'fsn null)
                          (make-arc 'tsnnfa 'neaaa null))))
    1 (graph?)
    (match? graph=? (make-graph (set 'fsn 'atdhth 'tieow 'neaaa 'tsnnfa)
                                (arcset (make-arc 'fsn 'atdhth null)
                                        (make-arc 'fsn 'tsnnfa null)
                                        (make-arc 'atdhth 'fsn null)
                                        (make-arc 'atdhth 'tieow null)
                                        (make-arc 'atdhth 'neaaa null)
                                        (make-arc 'tieow 'atdhth null)
                                        (make-arc 'tieow 'neaaa null)
                                        (make-arc 'neaaa 'atdhth null)
                                        (make-arc 'neaaa 'tieow null)
                                        (make-arc 'neaaa 'tsnnfa null)
                                        (make-arc 'tsnnfa 'fsn null)
                                        (make-arc 'tsnnfa 'neaaa null))))))

(suite delete-edge ()
  (test minimal-case
    (delete-edge 'shrunon 'haici
      (make-graph (set 'shrunon 'haici)
                  (arcset (make-arc 'shrunon 'haici 'phsaaef)
                          (make-arc 'haici 'shrunon 'phsaaef))))
    1 (graph?)
    (match? graph=? (make-graph (set 'shrunon 'haici) (arcset))))
  (test larger-case
    (delete-edge 'aeic 'lms
      (make-graph (set 'aeic 'trila 'nnoaeai 'oligldo 'lms)
                  (arcset (make-arc 'aeic 'trila 'neelk)
                          (make-arc 'aeic 'lms 'tlte)
                          (make-arc 'trila 'aeic 'neelk)
                          (make-arc 'trila 'nnoaeai 'trihrg)
                          (make-arc 'nnoaeai 'trila 'trihrg)
                          (make-arc 'nnoaeai 'oligldo 'hcmra)
                          (make-arc 'oligldo 'nnoaeai 'hcmra)
                          (make-arc 'oligldo 'lms 'hdoaewo)
                          (make-arc 'lms 'aeic 'tlte)
                          (make-arc 'lms 'oligldo 'hdoaewo))))
    1 (graph?)
    (match? graph=?
            (make-graph (set 'aeic 'trila 'nnoaeai 'oligldo 'lms)
                        (arcset (make-arc 'aeic 'trila 'neelk)
                                (make-arc 'trila 'aeic 'neelk)
                                (make-arc 'trila 'nnoaeai 'trihrg)
                                (make-arc 'nnoaeai 'trila 'trihrg)
                                (make-arc 'nnoaeai 'oligldo 'hcmra)
                                (make-arc 'oligldo 'nnoaeai 'hcmra)
                                (make-arc 'oligldo 'lms 'hdoaewo)
                                (make-arc 'lms 'oligldo 'hdoaewo))))))

(suite edge? ()
  (test random-value
    (edge? 'usr)
    1 (boolean?)
    not)
  (test empty-list
    (edge? (list))
    1 (boolean?)
    not)
  (test one-element
    (edge? (list (set 'sonotrh 'agoet)))
    1 (boolean?)
    not)
  (test too-many-elements
    (edge? (list (set 'ftudte 'atoes) 'edwtner 'ninhg))
    1 (boolean?)
    not)
  (test first-element-not-set
    (edge? (list 'sftrb 'sitreea))
    1 (boolean?)
    not)
  (test first-element-null-set
    (edge? (list (set) 'tutah))
    1 (boolean?)
    not)
  (test first-element-one-member-set
    (edge? (list (set 'toae) 'ogter))
    1 (boolean?)
    not)
  (test first-element-too-large
    (edge? (list (set 'eshvay 'tdroet 'rnrego) 'mhis))
    1 (boolean?)
    not)
  (test make-edge-result
    (edge? (make-edge (set 'redtedc 'atanoea) 'uth))
    1 (boolean?)
    true?))

(suite edge=? ((sample-edge (make-edge (set 'ngnewrs 'aepsuen) 'ley)))
  (test identical
    (edge=? sample-edge sample-edge)
    1 (boolean?)
    true?)
  (test duplicate
    (edge=? (make-edge (set 'sweohts 'dondlr) 'hidros)
            (make-edge (set 'sweohts 'dondlr) 'hidros))
    1 (boolean?)
    true?)
  (test reversed-components
    (edge=? (make-edge (set 'etro 'uefoisu) 'seyd)
            (make-edge (set 'uefoisu 'etro) 'seyd))
    1 (boolean?)
    true?)
  (test different-end
    (edge=? (make-edge (set 'ccdia 'nrwk) 'eeeoe)
            (make-edge (set 'nnkyy 'nrwk) 'eeeoe))
    1 (boolean?)
    not)
  (test different-label
    (edge=? (make-edge (set 'wuti 'nto) 'layvs)
            (make-edge (set 'wuti 'nto) 'edruam))
    1 (boolean?)
    not)
  (test no-labels
    (edge=? (make-edge (set 'eolh 'srhi) null)
            (make-edge (set 'eolh 'srhi) null))
    1 (boolean?)
    true?)
  (test one-label
    (edge=? (make-edge (set 'eik 'uenr) null)
            (make-edge (set 'eik 'uenr) 'bmts))
    1 (boolean?)
    not))

(suite ends ()
  (test unlabeled-edge
    (ends (make-edge (set 'uscu 'tts) null))
    2 (symbol? symbol?)
    (^vel (match? symbol=? 'uscu 'tts)
          (match? symbol=? 'tts 'uscu)))
  (test labeled-edge
    (ends (make-edge (set 'auhsl 'dofd) 'mforh))
    2 (symbol? symbol?)
    (^vel (match? symbol=? 'auhsl 'dofd)
          (match? symbol=? 'dofd 'auhsl))))

(suite edges ((same-ends? (pipe (~each edge-endpoints) set=?)))
  (test empty-graph
    (edges (empty-graph))
    1 ((set-of edge? same-ends?))
    empty-set?)
  (test larger-graph
    (edges (make-graph (set 'thnhneg 'nia 'steyabe 'tisp)
                       (arcset (make-arc 'thnhneg 'nia 'mapvpth)
                               (make-arc 'thnhneg 'steyabe 'eviie)
                               (make-arc 'nia 'thnhneg 'mapvpth)
                               (make-arc 'nia 'steyabe 'ertte)
                               (make-arc 'steyabe 'thnhneg 'eviie)
                               (make-arc 'steyabe 'nia 'ertte)
                               (make-arc 'steyabe 'tisp 'swotimm)
                               (make-arc 'tisp 'steyabe 'swotimm))))
    1 ((set-of edge? same-ends?))
    (match? (set-of= edge=?)
            ((set-maker same-ends?)
             (make-edge (set 'thnhneg 'nia) 'mapvpth)
             (make-edge (set 'thnhneg 'steyabe) 'eviie)
             (make-edge (set 'nia 'steyabe) 'ertte)
             (make-edge (set 'tisp 'steyabe) 'swotimm)))))

(suite degree ()
  (test no-edges
    (degree (make-graph (set 'awwdn 'teegn 'giehh) (arcset)))
    1 (procedure?)
    (lambda (degree-finder)
      (suite no-edges-internal ()
        (test any-vertex
          (degree-finder 'teegn)
          1 (natural-number?)
          zero?))))
  (test larger-graph
    (degree (make-graph (set 'lleess 'agohrsr 'opgo 'eemeoto 'dhtbily
                             'dthrdtr)
                        (arcset (make-arc 'lleess 'dhtbily 'aeeal)
                                (make-arc 'dhtbily 'lleess 'aeeal)
                                (make-arc 'lleess 'dthrdtr 'wteoe)
                                (make-arc 'dthrdtr 'lleess 'wteoe)
                                (make-arc 'agohrsr 'eemeoto 'efdel)
                                (make-arc 'eemeoto 'agohrsr 'efdel)
                                (make-arc 'agohrsr 'dthrdtr 'aybl)
                                (make-arc 'dthrdtr 'agohrsr 'aybl)
                                (make-arc 'eemeoto 'dhtbily 'cmogai)
                                (make-arc 'dhtbily 'eemeoto 'cmogai)
                                (make-arc 'eemeoto 'dthrdtr 'uakdtcd)
                                (make-arc 'dthrdtr 'eemeoto 'uakdtcd)
                                (make-arc 'dhtbily 'dthrdtr 'ismh)
                                (make-arc 'dthrdtr 'dhtbily 'ismh))))
    1 (procedure?)
    (lambda (degree-finder)
      (suite larger-graph-internal ()
        (test hermit
          (degree-finder 'opgo)
          1 (natural-number?)
          zero?)
        (test small
          (degree-finder 'lleess)
          1 (natural-number?)
          (match? = 2))
        (test larger
          (degree-finder 'dthrdtr)
          1 (natural-number?)
          (match? = 4))))))

;;; Test the procedures from the section on depth-first traversal.

(suite depth-first-visit ()
  (test no-advance
    (depth-first-visit (sect lookup <> (table (cons 'lfiydo (set))
                                              (cons 'ttif (set))
                                              (cons 'hfi (set))
                                              (cons 'etieins (set))))
                       prepend
                       prepend)
    1 (procedure?)
    (lambda (visitor)
      (suite no-advance ()
        (test no-advance-skip
          (visitor 'ttif (set 'lfiydo 'ttif 'etieins) (list))
          2 (set? list?)
          (lambda (alpha beta)
            (and (set=? alpha (set 'lfiydo 'ttif 'etieins))
                 (empty-list? beta))))
        (test no-advance-operate
          (visitor 'hfi (set) (list))
          2 (set? (list-of symbol?))
          (lambda (alpha beta)
            (and (set=? alpha (set 'hfi))
                 ((list-of= symbol=?) beta (list 'hfi 'hfi))))))))
  (test self-loop
    (depth-first-visit (sect lookup <> (table (cons 'omot (set 'omot))
                                              (cons 'nicwe
                                                    (set 'nicwe 'racha))
                                              (cons 'racha (set 'racha))))
                       prepend
                       prepend)
    1 (procedure?)
    (lambda (visitor)
      (suite self-loop-internal ()
        (test self-loop-operate
          (visitor 'omot (set) (list))
          2 (set? (list-of symbol?))
          (lambda (alpha beta)
            (and (set=? alpha (set 'omot))
                 ((list-of= symbol=?) beta (list 'omot 'omot))))))))
  (test one-step
    (depth-first-visit (sect lookup <> (table (cons 'sogef (set 'oapseee))
                                              (cons 'oapseee (set))
                                              (cons 'dhroie
                                                    (set 'dhroie))))
                       prepend
                       prepend)
    1 (procedure?)
    (lambda (visitor)
      (suite one-step-internal ()
        (test one-step-operate
          (visitor 'sogef (set) (list))
          2 (set? (list-of symbol?))
          (lambda (alpha beta)
            (and (set=? alpha (set 'sogef 'oapseee))
                 ((list-of= symbol=?) beta (list 'sogef 'oapseee 'oapseee
                                                 'sogef))))))))
  (test three-step
    (depth-first-visit
      (sect lookup <> (table (cons 'syhwiee (set 'ase 'pecs 'sdyat))
                             (cons 'gso (set 'erleir))
                             (cons 'ase (set 'argurre 'amen))
                             (cons 'argurre (set 'gso 'syhwiee 'eni))
                             (cons 'ipd (set 'ipd 'ase))
                             (cons 'pecs (set 'hweruc))
                             (cons 'amen (set 'ipd 'ase 'gso 'eni))
                             (cons 'eni (set 'essaa 'hweruc 'gso 'ipd))
                             (cons 'sdyat (set 'amen 'essaa 'erlier))
                             (cons 'erleir (set))
                             (cons 'hweruc (set 'gso))
                             (cons 'essaa (set 'hweruc 'essaa 'pecs))))
      prepend
      prepend)
    1 (procedure?)
    (lambda (visitor)
      (suite three-step-internal ()
        (test three-step-operate
          (visitor 'pecs (set 'argurre 'syhwiee 'eni 'essaa 'ipd) (list))
          2 (set? (list-of symbol?))
          (lambda (alpha beta)
            (and (set=? alpha (set 'pecs 'hweruc 'gso 'erleir 'argurre
                                   'syhwiee 'eni 'essaa 'ipd))
                 ((list-of= symbol=?) beta (list 'pecs 'hweruc 'gso 'erleir
                                                 'erleir 'gso 'hweruc
                                                 'pecs))))))))
  (test cycle
    (depth-first-visit (sect lookup <> (table (cons 'heasso (set 'tphhhib))
                                              (cons 'tphhhib
                                                    (set 'heasso))))
                       prepend
                       prepend)
    1 (procedure?)
    (lambda (visitor)
      (suite cycle-internal ()
        (test cycle-operate
          (visitor 'heasso (set) (list))
          2 (set? (list-of symbol?))
          (lambda (alpha beta)
            (and (set=? alpha (set 'heasso 'tphhhib))
                 ((list-of= symbol=?) beta (list 'heasso 'tphhhib 'tphhhib
                                                 'heasso))))))))
  (test backtrack
    (depth-first-visit (sect lookup <> (table (cons 'artnyo
                                                    (set 'sgrtn 'citihg))
                                              (cons 'sgrtn (set))
                                              (cons 'citihg (set))))
                       prepend
                       prepend)
    1 (procedure?)
    (lambda (visitor)
      (suite backtrack-internal ()
        (test backtrack-operate
          (visitor 'artnyo (set) (list))
          2 (set? (list-of symbol?))
          (lambda (alpha beta)
            (and (set=? alpha (set 'artnyo 'sgrtn 'citihg))
                 (or ((list-of= symbol=?) beta
                                          (list 'artnyo 'sgrtn 'sgrtn
                                                'citihg 'citihg 'artnyo))
                     ((list-of= symbol=?) beta
                                          (list 'artnyo 'citihg 'citihg
                                                'sgrtn 'sgrtn
                                                'artnyo)))))))))
  (test depth-first
    (depth-first-visit (sect lookup <> (table (cons 'coos
                                                    (set 'nen 'icotnro))
                                              (cons 'nen (set 'edt))
                                              (cons 'icotnro (set 'selr))
                                              (cons 'edt (set))
                                              (cons 'selr (set))))
                       prepend
                       prepend)
    1 (procedure?)
    (lambda (visitor)
      (suite depth-first-internal ()
        (test depth-first-operate
          (visitor 'coos (set) (list))
          2 (set? (list-of symbol?))
          (lambda (alpha beta)
            (and (set=? alpha (set 'coos 'icotnro 'nen 'edt 'selr))
                 (or ((list-of= symbol=?) beta
                                          (list 'coos 'nen 'edt 'edt 'nen
                                                'icotnro 'selr 'selr
                                                'icotnro 'coos))
                     ((list-of= symbol=?) beta
                                          (list 'coos 'icotnro 'selr 'selr
                                                'icotnro 'nen 'edt 'edt 'nen 
                                                'coos)))))))))
  (test back-links
    (depth-first-visit
      (sect lookup <> (table (cons 'cifgsgo (set 'kasotds 'uts))
                             (cons 'kasotds (set 'tohnk))
                             (cons 'uts (set 'edaa))
                             (cons 'tohnk (set 'cifgsgo 'uts 'kasotds))
                             (cons 'edaa (set 'cifgsgo 'kasotds 'uts))
                             (cons 'alicra (set 'lil))
                             (cons 'lil (set 'alicra 'lil))))
      prepend
      prepend)
    1 (procedure?)
    (lambda (visitor)
      (suite back-links-internal ()
        (test back-links-operate
          (visitor 'cifgsgo (set 'alicra 'lil) (list))
          2 (set? (list-of symbol?))
          (lambda (alpha beta)
            (and (set=? alpha (set 'alicra 'lil 'cifgsgo 'kasotds 'uts
                                   'tohnk 'edaa))
                 (or ((list-of= symbol=?) beta
                                          (list 'cifgsgo 'kasotds 'tohnk
                                                'uts 'edaa 'edaa 'uts
                                                'tohnk 'kasotds 'cifgsgo))
                     ((list-of= symbol=?) beta
                                          (list 'cifgsgo 'uts 'edaa
                                                'kasotds 'tohnk 'tohnk
                                                'kasotds 'edaa 'uts
                                                'cifgsgo)))))))))
  (test multiple-paths
    (depth-first-visit (sect lookup <> (table (cons 'lsrtsm
                                                    (set 'roalco 'uude))
                                              (cons 'roalco (set 'ttc))
                                              (cons 'uude (set 'ttc))
                                              (cons 'ttc (set))))
                       prepend
                       prepend)
    1 (procedure?)
    (lambda (visitor)
      (suite multiple-paths-internal ()
        (test multiple-paths-operate
          (visitor 'lsrtsm (set) (list))
          2 (set? (list-of symbol?))
          (lambda (alpha beta)
            (and (set=? alpha (set 'lsrtsm 'roalco 'uude 'ttc))
                 (or ((list-of= symbol=?) beta
                                          (list 'lsrtsm 'uude 'uude 'roalco
                                                'ttc 'ttc 'roalco 'lsrtsm))
                     ((list-of= symbol=?) beta
                                          (list 'lsrtsm 'roalco 'roalco
                                                'uude 'ttc 'ttc 'uude
                                                'lsrtsm))))))))))

(suite depth-first-traversal ((symlist? (list-of symbol?))
                              (symlist=? (list-of= symbol=?)))
  (test tracer
    (depth-first-traversal list prepend prepend)
    1 (procedure?)
    (lambda (tracer)
      (suite depth-first-traversal-internal ()
        (test empty-graph
          (tracer (empty-graph))
          1 (list?)
          empty-list?)
        (test one-vertex-no-arcs
          (tracer (make-graph (set 'nntu) (arcset)))
          1 (symlist?)
          (match? symlist=? (list 'nntu 'nntu)))
        (test one-vertex-one-arc
          (tracer (make-graph (set 'gps)
                              (arcset (make-arc 'gps 'gps 'nysiecf))))
          1 (symlist?)
          (match? symlist=? (list 'gps 'gps)))
        (test two-vertices-one-arc
          (tracer (make-graph (set 'slrt 'surd)
                              (arcset (make-arc 'slrt 'surd 'ysdmae))))
          1 (symlist?)
          (^vel (match? symlist=? (list 'slrt 'surd 'surd 'slrt))
                (match? symlist=? (list 'slrt 'slrt 'surd 'surd))))
        (test line-graph
          (tracer (make-graph (set 'ceue 'ehnhrt 'cgbahhr 'iwwee)
                              (arcset (make-arc 'ceue 'ehnhrt 'nnwhat)
                                      (make-arc 'ehnhrt 'cgbahhr 'wsdc)
                                      (make-arc 'cgbahhr 'iwwee 'fri))))
          1 (symlist?)
          (^or (match? symlist=? (list 'ceue 'ehnhrt 'cgbahhr 'iwwee 'iwwee
                                       'cgbahhr 'ehnhrt 'ceue))
               (match? symlist=? (list 'ceue 'ceue 'ehnhrt 'cgbahhr 'iwwee
                                       'iwwee 'cgbahhr 'ehnhrt))
               (match? symlist=? (list 'ceue 'ehnhrt 'ehnhrt 'ceue 'cgbahhr
                                       'iwwee 'iwwee 'cgbahhr))
               (match? symlist=? (list 'ceue 'ceue 'ehnhrt 'ehnhrt 'cgbahhr
                                       'iwwee 'iwwee 'cgbahhr))
               (match? symlist=? (list 'ceue 'ehnhrt 'cgbahhr 'cgbahhr
                                       'ehnhrt 'ceue 'iwwee 'iwwee))
               (match? symlist=? (list 'ceue 'ceue 'ehnhrt 'cgbahhr
                                       'cgbahhr 'ehnhrt 'iwwee 'iwwee))
               (match? symlist=? (list 'ceue 'ehnhrt 'ehnhrt 'ceue 'cgbahhr
                                       'cgbahhr 'iwwee 'iwwee))
               (match? symlist=? (list 'ceue 'ceue 'ehnhrt 'ehnhrt 'cgbahhr
                                       'cgbahhr 'iwwee 'iwwee))))
        (test cycle
          (tracer (make-graph (set 'hhot 'tsn 'ciiu)
                              (arcset (make-arc 'hhot 'tsn 'hthi)
                                      (make-arc 'tsn 'ciiu 'raeeeey)
                                      (make-arc 'ciiu 'hhot 'nlude))))
          1 (symlist?)
          (^or (match? symlist=? (list 'hhot 'tsn 'ciiu 'ciiu 'tsn 'hhot))
               (match? symlist=? (list 'tsn 'ciiu 'hhot 'hhot 'ciiu 'tsn))
               (match? symlist=? (list 'ciiu 'hhot 'tsn 'tsn 'hhot
                                       'ciiu))))
        (test harder-case
          (tracer (make-graph (set 'erioa 'nehdi 'detaott 'htle 'swis)
                              (arcset (make-arc 'erioa 'nehdi 'iism)
                                      (make-arc 'erioa 'detaott 'dil)
                                      (make-arc 'nehdi 'swis 'ofeuait)
                                      (make-arc 'detaott 'htle 'vne)
                                      (make-arc 'htle 'swis 'nms)
                                      (make-arc 'swis 'erioa 'iaiglh))))
          1 (symlist?)
          (^or (match? symlist=? (list 'erioa 'detaott 'htle 'htle 'detaott
                                       'nehdi 'swis 'swis 'nehdi 'erioa))
               (match? symlist=? (list 'erioa 'nehdi 'nehdi 'detaott 'htle
                                       'swis 'swis 'htle 'detaott 'erioa))
               (match? symlist=? (list 'nehdi 'swis 'erioa 'detaott 'htle
                                       'htle 'detaott 'erioa 'swis 'nehdi))
               (match? symlist=? (list 'detaott 'htle 'swis 'erioa 'nehdi
                                       'nehdi 'erioa 'swis 'htle 'detaott))
               (match? symlist=? (list 'htle 'swis 'erioa 'detaott 'detaott
                                       'nehdi 'nehdi 'erioa 'swis 'htle))
               (match? symlist=? (list 'htle 'swis 'erioa 'nehdi 'nehdi
                                       'detaott 'detaott 'erioa 'swis
                                       'htle))
               (match? symlist=? (list 'swis 'erioa 'detaott 'htle 'htle
                                       'detaott 'nehdi 'nehdi 'erioa
                                       'swis))
               (match? symlist=? (list 'swis 'erioa 'nehdi 'nehdi 'detaott
                                       'htle 'htle 'detaott 'erioa 'swis))))
      (test disconnected
        (tracer (make-graph (set 'scrwh 'sflit 'eycde 'aweht)
                            (arcset (make-arc 'scrwh 'sflit 'nir)
                                    (make-arc 'sflit 'scrwh 'tuw)
                                    (make-arc 'eycde 'aweht 'rdpweep)
                                    (make-arc 'aweht 'eycde 'rrprior))))
        1 (symlist?)
        (^or (match? symlist=? (list 'scrwh 'sflit 'sflit 'scrwh 'eycde
                                     'aweht 'aweht 'eycde))
             (match? symlist=? (list 'sflit 'scrwh 'scrwh 'sflit 'eycde
                                     'aweht 'aweht 'eycde))
             (match? symlist=? (list 'scrwh 'sflit 'sflit 'scrwh 'aweht
                                     'eycde 'eycde 'aweht))
             (match? symlist=? (list 'sflit 'scrwh 'scrwh 'sflit 'aweht
                                     'eycde 'eycde 'aweht))
             (match? symlist=? (list 'eycde 'aweht 'aweht 'eycde 'scrwh
                                     'sflit 'sflit 'scrwh))
             (match? symlist=? (list 'aweht 'eycde 'eycde 'aweht 'scrwh
                                     'sflit 'sflit 'scrwh))
             (match? symlist=? (list 'eycde 'aweht 'aweht 'eycde 'sflit
                                     'scrwh 'scrwh 'sflit))
             (match? symlist=? (list 'aweht 'eycde 'eycde 'aweht 'sflit
                                     'scrwh 'scrwh 'sflit))))
      (test square
        (tracer (make-graph (set 'and 'nnrrrit 'testio 'ssn)
                            (arcset (make-arc 'and 'nnrrrit 'oivh)
                                    (make-arc 'and 'testio 'uerobd)
                                    (make-arc 'nnrrrit 'ssn 'sif)
                                    (make-arc 'testio 'ssn 'fhc))))
        1 (symlist?)
        (^or (match? symlist=? (list 'and 'testio 'testio 'nnrrrit 'ssn
                                     'ssn 'nnrrrit 'and))
             (match? symlist=? (list 'and 'nnrrrit 'nnrrrit 'testio 'ssn
                                     'ssn 'testio 'and))
             (match? symlist=? (list 'and 'testio 'testio 'and 'nnrrrit
                                     'ssn 'ssn 'nnrrrit))
             (match? symlist=? (list 'and 'and 'testio 'testio 'nnrrrit
                                     'ssn 'ssn 'nnrrrit))
             (match? symlist=? (list 'and 'nnrrrit 'nnrrrit 'and 'testio
                                     'ssn 'ssn 'testio))
             (match? symlist=? (list 'and 'and 'nnrrrit 'nnrrrit 'testio
                                     'ssn 'ssn 'testio))
             (match? symlist=? (list 'and 'testio 'testio 'nnrrrit 'nnrrrit
                                     'and 'ssn 'ssn)) 
             (match? symlist=? (list 'and 'nnrrrit 'nnrrrit 'testio 'testio
                                     'and 'ssn 'ssn))
             (match? symlist=? (list 'and 'testio 'testio 'and 'nnrrrit
                                     'nnrrrit 'ssn 'ssn))
             (match? symlist=? (list 'and 'and 'testio 'testio 'nnrrrit
                                     'nnrrrit 'ssn 'ssn))
             (match? symlist=? (list 'and 'nnrrrit 'nnrrrit 'and 'testio
                                     'testio 'ssn 'ssn))
             (match? symlist=? (list 'and 'and 'nnrrrit 'nnrrrit 'testio
                                     'testio 'ssn 'ssn)))))
    (test double-triangle
      (tracer (make-graph (set 'nir 'tuw 'rdpweep 'rrprior 'dnnlal)
                          (arcset (make-arc 'nir 'tuw 'dtsn)
                                  (make-arc 'nir 'rrprior 'otm)
                                  (make-arc 'tuw 'rdpweep 'nell)
                                  (make-arc 'rdpweep 'nir 'etssy)
                                  (make-arc 'rrprior 'dnnlal 'sia)
                                  (make-arc 'dnnlal 'nir 'fhmms))))
      1 (symlist?)
      (^or (match? symlist=? (list 'nir 'rrprior 'dnnlal 'dnnlal 'rrprior
                                   'tuw 'rdpweep 'rdpweep 'tuw 'nir))
           (match? symlist=? (list 'nir 'tuw 'rdpweep 'rdpweep 'tuw
                                   'rrprior 'dnnlal 'dnnlal 'rrprior 'nir))
           (match? symlist=? (list 'tuw 'rdpweep 'nir 'rrprior 'dnnlal
                                   'dnnlal 'rrprior 'nir 'rdpweep 'tuw))
           (match? symlist=? (list 'rdpweep 'nir 'rrprior 'dnnlal 'dnnlal
                                   'rrprior 'tuw 'tuw 'nir 'rdpweep))
           (match? symlist=? (list 'rdpweep 'nir 'tuw 'tuw 'rrprior 'dnnlal
                                   'dnnlal 'rrprior 'nir 'rdpweep))
           (match? symlist=? (list 'rrprior 'dnnlal 'nir 'tuw 'rdpweep
                                   'rdpweep 'tuw 'nir 'dnnlal 'rrprior))
           (match? symlist=? (list 'dnnlal 'nir 'rrprior 'rrprior 'tuw
                                   'rdpweep 'rdpweep 'tuw 'nir 'dnnlal))
           (match? symlist=? (list 'dnnlal 'nir 'tuw 'rdpweep 'rdpweep 'tuw
                                   'rrprior 'rrprior 'nir 'dnnlal)))))))

(suite topological-sort
    ((symlist? (list-of symbol?))
     (symlist=? (list-of= symbol=?))
     (list->bag (fold-list bag put-into-bag))
     (topologically-sorted?
      (lambda (ls graph)
        (for-all? (lambda (arc)
                    (let ((tail (arc-tail arc))
                          (head (arc-head arc)))
                      ((rec (precedes? sublist)
                         (or (equal? (first sublist) tail)
                             (and (not (equal? (first sublist) head))
                                  (precedes? (rest sublist)))))
                       ls)))
                  (arcs graph))))

     ;; From _Introduction to algorithms_, third edition, by Thomas
     ;; H. Cormen, Charles E. Leiserson, Ronald L. Rivest, and Clifford
     ;; Stein, p. 615:

     (Cormen-graph (make-graph (set 'm 'n 'o 'p 'q 'r 's 't 'u 'v 'w 'x 'y
                                    'z)
                               (arcset (make-arc 'm 'q null)
                                       (make-arc 'm 'r null)
                                       (make-arc 'm 'x null)
                                       (make-arc 'n 'o null)
                                       (make-arc 'n 'q null)
                                       (make-arc 'n 'u null)
                                       (make-arc 'o 'r null)
                                       (make-arc 'o 's null)
                                       (make-arc 'o 'v null)
                                       (make-arc 'p 'o null)
                                       (make-arc 'p 's null)
                                       (make-arc 'p 'z null)
                                       (make-arc 'q 't null)
                                       (make-arc 'r 'u null)
                                       (make-arc 'r 'y null)
                                       (make-arc 's 'r null)
                                       (make-arc 'u 't null)
                                       (make-arc 'v 'w null)
                                       (make-arc 'v 'x null)
                                       (make-arc 'w 'z null)
                                       (make-arc 'y 'v null))))

     ;; From Robert Sedgewick's _Algorithms_ (second edition, p. 479):

     (Sedgewick-graph (make-graph (set 'a 'b 'c 'd 'e 'f 'g 'h 'i 'j 'k 'l
                                       'm)
                                  (arcset (make-arc 'a 'b null)
                                          (make-arc 'a 'c null)
                                          (make-arc 'a 'f null)
                                          (make-arc 'a 'g null)
                                          (make-arc 'e 'd null)
                                          (make-arc 'f 'd null)
                                          (make-arc 'f 'e null)
                                          (make-arc 'g 'c null)
                                          (make-arc 'g 'e null)
                                          (make-arc 'g 'h null)
                                          (make-arc 'h 'i null)
                                          (make-arc 'j 'g null)
                                          (make-arc 'j 'k null)
                                          (make-arc 'j 'l null)
                                          (make-arc 'j 'm null)
                                          (make-arc 'l 'g null)
                                          (make-arc 'l 'm null)))))
  (test empty-graph
    (topological-sort (empty-graph))
    1 (list?)
    empty-list?)
  (test one-vertex-graph
    (topological-sort (make-graph (set 'uhenir) (arcset)))
    1 (symlist?)
    (match? symlist=? (list 'uhenir)))
  (test no-arcs
    (topological-sort (make-graph (set 'eyhlymn 'edeaesh 'iexaltm 'gheeg
                                       'lrewes 'wibegf 'aai 'eynq 'thgcni)
                                  (arcset)))
    1 (symlist?)
    (pipe list->bag (match? (bag-of= symbol=?)
                            (bag 'eyhlymn 'edeaesh 'iexaltm 'gheeg 'lrewes 
                                 'wibegf 'aai 'eynq 'thgcni))))
  (test several-vertices
    (topological-sort
      (make-graph (set 'tlolu 'oao 'tbuy 'hetkrnl 'ehsh 'vtvlcod)
                  (arcset (make-arc 'tbuy 'vtvlcod 'mied)
                          (make-arc 'vtvlcod 'tlolu 'src)
                          (make-arc 'oao 'ehsh 'caeg)
                          (make-arc 'hetkrnl 'tbuy 'yidan)
                          (make-arc 'tlolu 'oao 'etutsig))))
    1 (symlist?)
    (match? symlist=? (list 'hetkrnl 'tbuy 'vtvlcod 'tlolu 'oao 'ehsh)))
  (test Cormen
    (topological-sort Cormen-graph)
    1 (symlist?)
    (^et (pipe list->bag (match? (bag-of= symbol=?)
                                 (bag 'm 'n 'o 'p 'q 'r 's 't 'u 'v 'w 'x
                                      'y 'z)))
         (sect topologically-sorted? <> Cormen-graph)))
  (test Sedgewick
    (topological-sort Sedgewick-graph)
    1 (symlist?)
    (^et (pipe list->bag (match? (bag-of= symbol=?)
                                 (bag 'a 'b 'c 'd 'e 'f 'g 'h 'i 'j 'k 'l
                                      'm)))
         (sect topologically-sorted? <> Sedgewick-graph))))

;;; Test the procedures from the section on paths.

(suite reachables ((symset? (set-of symbol? symbol=?))
                   (symset=? (set-of= symbol=?)))
  (test one-vertex
    (reachables (make-graph (set 'rad) (arcset)))
    1 (procedure?)
    (lambda (collector)
      (suite one-vertex-internal ()
        (test start-only
          (collector 'rad)
          1 (symset?)
          (match? symset=? (set 'rad))))))
  (test two-vertex
    (reachables (make-graph (set 'rthtn 'tytrrer)
                            (arcset (make-arc 'rthtn 'tytrrer null))))
    1 (procedure?)
    (lambda (collector)
      (suite two-vertex-internal ()
        (test blocked
          (collector 'tytrrer)
          1 (symset?)
          (match? symset=? (set 'tytrrer)))
        (test open
          (collector 'rthtn)
          1 (symset?)
          (match? symset=? (set 'rthtn 'tytrrer))))))
  (test random-graph
    (reachables (make-graph (set 'acle 'ehhtcy 'thhua 'wetrah 'soal 'npa)
                            (arcset (make-arc 'acle 'ehhtcy 'ioeam)
                                    (make-arc 'acle 'thhua 'caier)
                                    (make-arc 'acle 'wetrah 'nwueomn)
                                    (make-arc 'ehhtcy 'ehhtcy 'unr)
                                    (make-arc 'ehhtcy 'thhua 'fit)
                                    (make-arc 'ehhtcy 'wetrah 'ooaf)
                                    (make-arc 'ehhtcy 'soal 'sed)
                                    (make-arc 'thhua 'acle 'wlptmgt)
                                    (make-arc 'thhua 'ehhtcy 'ylow)
                                    (make-arc 'thhua 'thhua 'nscoene)
                                    (make-arc 'thhua 'soal 'aokshi)
                                    (make-arc 'soal 'ehhtcy 'bdsledl)
                                    (make-arc 'soal 'wetrah 'rlacem)
                                    (make-arc 'npa 'acle 'einir)
                                    (make-arc 'npa 'ehhtcy 'ehtlnef))))
    1 (procedure?)
    (lambda (collector)
      (suite random-graph-internal ()
        (test out-degree-zero
          (collector 'wetrah)
          1 (symset?)
          (match? symset=? (set 'wetrah)))
        (test pickups
          (collector 'soal)
          1 (symset?)
          (match? symset=? (set 'acle 'ehhtcy 'thhua 'wetrah 'soal)))
        (test universe
          (collector 'npa)
          1 (symset?)
          (match? symset=?
                  (set 'acle 'ehhtcy 'thhua 'wetrah 'soal 'npa)))))))

(suite path-finder ((right-type? (^vel not (list-of symbol?)))
                    (symlist? (list-of symbol?))
                    (symlist=? (list-of= symbol=?)))
  (test one-vertex-graph
    (path-finder (make-graph (set 'nnecvet) (arcset)))
    1 (procedure?)
    (lambda (finder)
      (suite one-vertex-graph-internal ()
        (test self-path
          (finder 'nnecvet 'nnecvet)
          1 (right-type?)
          (^et symlist? (match? symlist=? (list 'nnecvet)))))))
  (test two-vertex-graph
    (path-finder (make-graph (set 'onesa 'msayoui)
                             (arcset (make-arc 'onesa 'msayoui 'net))))
    1 (procedure?)
    (lambda (finder)
      (suite two-vertex-graph-internal ()
        (test self-path
          (finder 'onesa 'onesa)
          1 (symlist?)
          (^et symlist? (match? symlist=? (list 'onesa))))
        (test along-arc
          (finder 'onesa 'msayoui)
          1 (right-type?)
          (^et symlist? (match? symlist=? (list 'onesa 'msayoui))))
        (test no-path
          (finder 'msayoui 'onesa)
          1 (right-type?)
          not))))
  (test larger-graph
    (path-finder (make-graph (set 'vandel 'tni 'eedrsa 'nlstrpa 'alef
                                  'lssmu 'edfofee 'eaot 'antwis)
                             (arcset (make-arc 'vandel 'tni 'lri)
                                     (make-arc 'vandel 'eedrsa 'bsie)
                                     (make-arc 'vandel 'nlstrpa 'iitatt)
                                     (make-arc 'tni 'vandel 'arkse)
                                     (make-arc 'eedrsa 'tni 'oeeiar)
                                     (make-arc 'eedrsa 'alef 'ehrh)
                                     (make-arc 'eedrsa 'nlstrpa 'wtguen)
                                     (make-arc 'alef 'nlstrpa 'hscai)
                                     (make-arc 'alef 'lssmu 'adatit)
                                     (make-arc 'alef 'edfofee 'amjewo)
                                     (make-arc 'alef 'eaot 'ttieme)
                                     (make-arc 'alef 'tni 'ises)
                                     (make-arc 'edfofee 'antwis 'tirpmu))))
    1 (procedure?)
    (lambda (finder)
      (suite larger-graph-internal ()
        (test difficult-path
          (finder 'vandel 'antwis)
          1 (right-type?)
          (^et symlist?
               (match? symlist=?
                       (list 'vandel 'eedrsa 'alef 'edfofee 'antwis))))
        (test impossible-path
          (finder 'edfofee 'eedrsa)
          1 (right-type?)
          not)))))

(suite path-connected? ()
  (test empty-graph
    (path-connected? (empty-graph))
    1 (boolean?)
    true?)
  (test one-vertex
    (path-connected? (make-graph (set 'enesnae) (arcset)))
    1 (boolean?)
    true?)
  (test two-vertex-without-edge
    (path-connected? (make-graph (set 'liwgr 'oha) (set)))
    1 (boolean?)
    not)
  (test two-vertex-with-edge
    (path-connected? (make-graph (set 'ahadf 'haaou)
                                 (arcset (make-arc 'ahadf 'haaou null)
                                         (make-arc 'haaou 'ahadf null))))
    1 (boolean?)
    true?)
  (test larger-case
    (path-connected?
      (make-graph (set 'uoi 'hnrasdi 'neiss 'hotge 'hneuhy
                       'tnfeonl 'tea 'pieuk 'fipd 'eonaw)
                  (arcset (make-arc 'uoi 'hnrasdi 'leton)
                          (make-arc 'hnrasdi 'uoi 'leton)
                          (make-arc 'uoi 'neiss 'iotrndt)
                          (make-arc 'neiss 'uoi 'iotrndt)
                          (make-arc 'uoi 'hotge 'lolaoe)
                          (make-arc 'hotge 'uoi 'lolaoe)
                          (make-arc 'hnrasdi 'hneuhy 'bsrweea)
                          (make-arc 'hneuhy 'hnrasdi 'bsrweea)
                          (make-arc 'hnrasdi 'neiss 'tkhe)
                          (make-arc 'neiss 'hnrasdi 'tkhe)
                          (make-arc 'hnrasdi 'hotge 'mslseol)
                          (make-arc 'hotge 'hnrasdi 'mslseol)
                          (make-arc 'hnrasdi 'eonaw 'adoat)
                          (make-arc 'eonaw 'hnrasdi 'adoat)
                          (make-arc 'neiss 'tnfeonl 'teo)
                          (make-arc 'tnfeonl 'neiss 'teo)
                          (make-arc 'neiss 'eonaw 'amstwob)
                          (make-arc 'eonaw 'neiss 'amstwob)
                          (make-arc 'hotge 'tnfeonl 'eebwoa)
                          (make-arc 'tnfeonl 'hotge 'eebwoa)
                          (make-arc 'hotge 'tea 'ipeyd)
                          (make-arc 'tea 'hotge 'ipeyd)
                          (make-arc 'hneuhy 'tnfeonl 'tsn)
                          (make-arc 'tnfeonl 'hneuhy 'tsn)
                          (make-arc 'hneuhy 'tea 'ehhohd)
                          (make-arc 'tea 'hneuhy 'ehhohd)
                          (make-arc 'pieuk 'fipd 'ebeter)
                          (make-arc 'fipd 'pieuk 'ebeter))))
    1 (boolean?)
    not)
  (test linear
    (path-connected? (make-graph (set 'sibc 'ofhwnkc 'dyndmnn 'numhw
                                      'dmooiue 'kenufh)
                                 (arcset (make-arc 'sibc 'ofhwnkc null)
                                         (make-arc 'ofhwnkc 'sibc null)
                                         (make-arc 'dyndmnn 'numhw null)
                                         (make-arc 'numhw 'dyndmnn null)
                                         (make-arc 'dmooiue 'kenufh null)
                                         (make-arc 'kenufh 'dmooiue null)
                                         (make-arc 'ofhwnkc 'dyndmnn null)
                                         (make-arc 'dyndmnn 'ofhwnkc null)
                                         (make-arc 'numhw 'dmooiue null)
                                         (make-arc 'dmooiue 'numhw null))))
    1 (boolean?)
    true?))

(suite connected-components ((graphset? (set-of graph? graph=?))
                             (graphset=? (set-of= graph=?)))
  (test empty-graph
    (connected-components (empty-graph))
    1 (set?)
    empty-set?)
  (test one-vertex
    (connected-components (make-graph (set 'estwe) (arcset)))
    1 (graphset?)
    (match? graphset=? (set (make-graph (set 'estwe) (arcset)))))
  (test two-vertex
    (connected-components (make-graph (set 'hyn 'rdotsi) (arcset)))
    1 (graphset?)
    (match? graphset=? (set (make-graph (set 'hyn) (arcset))
                            (make-graph (set 'rdotsi) (arcset)))))
  (test contrived
    (connected-components
      (make-graph (set 'ves 'wulosef 'cennher 'rre 'nmcschi 'nwnd 'seenh
                       'tgtmesn 'enn 'com 'abaai 'hhiiayl 'tcai 'ttohot
                       'ghlot 'ndp 'oeoc 'slcchun)
                  (arcset (make-arc 'wulosef 'cennher 'smhg)
                          (make-arc 'cennher 'wulosef 'smhg)
                          (make-arc 'rre 'nmcschi 'outot)
                          (make-arc 'nmcschi 'rre 'outot)
                          (make-arc 'rre 'nwnd 'rnidw)
                          (make-arc 'nwnd 'rre 'rnidw)
                          (make-arc 'rre 'seenh 'ouardn)
                          (make-arc 'seenh 'rre 'ouardn)
                          (make-arc 'nwnd 'seenh 'gotaoc)
                          (make-arc 'seenh 'nwnd 'gotaoc)
                          (make-arc 'tgtmesn 'enn 'padbr)
                          (make-arc 'enn 'tgtmesn 'padbr)
                          (make-arc 'tgtmesn 'com 'vral)
                          (make-arc 'com 'tgtmesn 'vral)
                          (make-arc 'tgtmesn 'abaai 'mne)
                          (make-arc 'abaai 'tgtmesn 'mne)
                          (make-arc 'enn 'com 'euuaeht)
                          (make-arc 'com 'enn 'euuaeht)
                          (make-arc 'enn 'abaai 'drayo)
                          (make-arc 'abaai 'enn 'drayo)
                          (make-arc 'com 'abaai 'idhud)
                          (make-arc 'abaai 'com 'idhud)
                          (make-arc 'hhiiayl 'ttohot 'eosbfe)
                          (make-arc 'ttohot 'hhiiayl 'eosbfe)
                          (make-arc 'tcai 'ttohot 'kfcnyom)
                          (make-arc 'ttohot 'tcai 'kfcnyom)
                          (make-arc 'ttohot 'ghlot 'oaeotn)
                          (make-arc 'ghlot 'ttohot 'oaeotn)
                          (make-arc 'ghlot 'ndp 'arrkswe)
                          (make-arc 'ndp 'ghlot 'arrkswe)
                          (make-arc 'ghlot 'oeoc 'oweebmr)
                          (make-arc 'oeoc 'ghlot 'oweebmr)
                          (make-arc 'oeoc 'slcchun 'arho)
                          (make-arc 'slcchun 'oeoc 'arho))))
    1 (graphset?)
    (match? graphset=?
            (set (make-graph (set 'ves) (arcset))
                 (make-graph (set 'wulosef 'cennher)
                             (arcset (make-arc 'wulosef 'cennher 'smhg)
                                     (make-arc 'cennher 'wulosef 'smhg)))
                 (make-graph (set 'rre 'nmcschi 'nwnd 'seenh)
                             (arcset (make-arc 'rre 'nmcschi 'outot)
                                     (make-arc 'nmcschi 'rre 'outot)
                                     (make-arc 'rre 'nwnd 'rnidw)
                                     (make-arc 'nwnd 'rre 'rnidw)
                                     (make-arc 'rre 'seenh 'ouardn)
                                     (make-arc 'seenh 'rre 'ouardn)
                                     (make-arc 'nwnd 'seenh 'gotaoc)
                                     (make-arc 'seenh 'nwnd 'gotaoc)))
                 (make-graph (set 'tgtmesn 'enn 'com 'abaai)
                             (arcset (make-arc 'tgtmesn 'enn 'padbr)
                                     (make-arc 'enn 'tgtmesn 'padbr)
                                     (make-arc 'tgtmesn 'com 'vral)
                                     (make-arc 'com 'tgtmesn 'vral)
                                     (make-arc 'tgtmesn 'abaai 'mne)
                                     (make-arc 'abaai 'tgtmesn 'mne)
                                     (make-arc 'enn 'com 'euuaeht)
                                     (make-arc 'com 'enn 'euuaeht)
                                     (make-arc 'enn 'abaai 'drayo)
                                     (make-arc 'abaai 'enn 'drayo)
                                     (make-arc 'com 'abaai 'idhud)
                                     (make-arc 'abaai 'com 'idhud)))
                 (make-graph (set 'hhiiayl 'tcai 'ttohot 'ghlot 'ndp 'oeoc
                                  'slcchun)
                             (arcset (make-arc 'hhiiayl 'ttohot 'eosbfe)
                                     (make-arc 'ttohot 'hhiiayl 'eosbfe)
                                     (make-arc 'tcai 'ttohot 'kfcnyom)
                                     (make-arc 'ttohot 'tcai 'kfcnyom)
                                     (make-arc 'ttohot 'ghlot 'oaeotn)
                                     (make-arc 'ghlot 'ttohot 'oaeotn)
                                     (make-arc 'ghlot 'ndp 'arrkswe)
                                     (make-arc 'ndp 'ghlot 'arrkswe)
                                     (make-arc 'ghlot 'oeoc 'oweebmr)
                                     (make-arc 'oeoc 'ghlot 'oweebmr)
                                     (make-arc 'oeoc 'slcchun 'arho)
                                     (make-arc 'slcchun 'oeoc 'arho))))))
  (test single-component
    (connected-components
      (make-graph (set 'nipd 'eut 'egar 'hdtes 'alh 'lpb 'waii 'heegn)
                  (arcset (make-arc 'nipd 'eut 'aucea)
                          (make-arc 'eut 'nipd 'aucea)
                          (make-arc 'nipd 'egar 'rsimes)
                          (make-arc 'egar 'nipd 'rsimes)
                          (make-arc 'nipd 'hdtes 'pae)
                          (make-arc 'hdtes 'nipd 'pae)
                          (make-arc 'eut 'alh 'eai)
                          (make-arc 'alh 'eut 'eai)
                          (make-arc 'eut 'heegn 'dteic)
                          (make-arc 'heegn 'eut 'dteic)
                          (make-arc 'eut 'egar 'etaesv)
                          (make-arc 'egar 'eut 'etaesv)
                          (make-arc 'eut 'lpb 'lat)
                          (make-arc 'lpb 'eut 'lat)
                          (make-arc 'eut 'waii 'llwrr)
                          (make-arc 'waii 'eut 'llwrr)
                          (make-arc 'egar 'lpb 'gtoen)
                          (make-arc 'lpb 'egar 'gtoen)
                          (make-arc 'egar 'hdtes 'seooir)
                          (make-arc 'hdtes 'egar 'seooir)
                          (make-arc 'hdtes 'alh 'tnt)
                          (make-arc 'alh 'hdtes 'tnt)
                          (make-arc 'hdtes 'lpb 'swctfs)
                          (make-arc 'lpb 'hdtes 'swctfs)
                          (make-arc 'alh 'lpb 'sae)
                          (make-arc 'lpb 'alh 'sae)
                          (make-arc 'alh 'waii 'fuei)
                          (make-arc 'waii 'alh 'fuei)
                          (make-arc 'lpb 'heegn 'een)
                          (make-arc 'heegn 'lpb 'een)
                          (make-arc 'waii 'heegn 'nrhbto)
                          (make-arc 'heegn 'waii 'nrhbto))))
    1 (graphset?)
    (match? graphset=?
            (set (make-graph (set 'nipd 'eut 'egar 'hdtes 'alh 'lpb 'waii
                                  'heegn)
                             (arcset (make-arc 'nipd 'eut 'aucea)
                                     (make-arc 'eut 'nipd 'aucea)
                                     (make-arc 'nipd 'egar 'rsimes)
                                     (make-arc 'egar 'nipd 'rsimes)
                                     (make-arc 'nipd 'hdtes 'pae)
                                     (make-arc 'hdtes 'nipd 'pae)
                                     (make-arc 'eut 'alh 'eai)
                                     (make-arc 'alh 'eut 'eai)
                                     (make-arc 'eut 'heegn 'dteic)
                                     (make-arc 'heegn 'eut 'dteic)
                                     (make-arc 'eut 'egar 'etaesv)
                                     (make-arc 'egar 'eut 'etaesv)
                                     (make-arc 'eut 'lpb 'lat)
                                     (make-arc 'lpb 'eut 'lat)
                                     (make-arc 'eut 'waii 'llwrr)
                                     (make-arc 'waii 'eut 'llwrr)
                                     (make-arc 'egar 'lpb 'gtoen)
                                     (make-arc 'lpb 'egar 'gtoen)
                                     (make-arc 'egar 'hdtes 'seooir)
                                     (make-arc 'hdtes 'egar 'seooir)
                                     (make-arc 'hdtes 'alh 'tnt)
                                     (make-arc 'alh 'hdtes 'tnt)
                                     (make-arc 'hdtes 'lpb 'swctfs)
                                     (make-arc 'lpb 'hdtes 'swctfs)
                                     (make-arc 'alh 'lpb 'sae)
                                     (make-arc 'lpb 'alh 'sae)
                                     (make-arc 'alh 'waii 'fuei)
                                     (make-arc 'waii 'alh 'fuei)
                                     (make-arc 'lpb 'heegn 'een)
                                     (make-arc 'heegn 'lpb 'een)
                                     (make-arc 'waii 'heegn 'nrhbto)
                                     (make-arc 'heegn 'waii 'nrhbto)))))))

;;; Test the procedures from the section on breadth-first traversal.

(suite breadth-first-visit ((symset? (set-of symbol? symbol=?))
                            (symset=? (set-of= symbol=?))
                            (symlist? (list-of symbol?))
                            (symlist=? (list-of= symbol=?)))
  (test no-advance
    (breadth-first-visit (sect lookup <> (table (cons 'cih (set))
                                                (cons 'oeldog (set))
                                                (cons 'nssr (set))
                                                (cons 'fsdo (set))))
                         prepend)
    1 (procedure?)
    (lambda (visitor)
      (suite no-advance-internal ()
        (test no-advance-skip
          (visitor 'oeldog (set 'cih 'oeldog 'fsdo) (list))
          2 (symset? symlist?)
          (lambda (alpha beta)
            (and (symset=? alpha (set 'cih 'oeldog 'fsdo))
                 (empty-list? beta))))
        (test no-advance-operate
          (visitor 'nssr (set) (list))
          2 (symset? symlist?)
          (lambda (alpha beta)
            (and (symset=? alpha (set 'nssr))
                 (symlist=? beta (list 'nssr))))))))
  (test self-loop
    (breadth-first-visit (sect lookup <> (table (cons 'detl (set 'detl))))
                         postpend)
    1 (procedure?)
    (lambda (visitor)
      (suite self-loop-internal ()
        (test self-loop-operate
          (visitor 'detl (set) (list))
          2 (symset? symlist?)
          (lambda (alpha beta)
            (and (symset=? alpha (set 'detl))
                 (symlist=? beta (list 'detl))))))))
  (test one-step
    (breadth-first-visit (sect lookup <> (table (cons 'qteoi (set 'opiy))
                                                (cons 'opiy (set))))
                         prepend)
    1 (procedure?)
    (lambda (visitor)
      (suite one-step-internal ()
        (test one-step-operate
          (visitor 'qteoi (set) (list))
          2 (symset? symlist?)
          (lambda (alpha beta)
            (and (symset=? alpha (set 'qteoi 'opiy))
                 (symlist=? beta (list 'opiy 'qteoi))))))))
  (test three-step
    (breadth-first-visit
      (sect lookup <> (table (cons 'tycca (set 'brletoa))
                             (cons 'etdretl (set 'uttto))
                             (cons 'uttto (set 'tycca))
                             (cons 'brletoa (set))))
      prepend)
    1 (procedure?)
    (lambda (visitor)
      (suite three-step-internal ()
        (test three-step-operate
          (visitor 'etdretl
                   (set 'dyiigge 'laio 'eboio 'aufsd 'lirsoi 'hetiam)
                   (list))
          2 (symset? symlist?)
          (lambda (alpha beta)
            (and (symset=? alpha
                           (set 'tycca 'brletoa 'etdretl 'uttto 'dyiigge
                                'laio 'eboio 'aufsd 'lirsoi 'hetiam))
                 (symlist=? beta
                            (list 'brletoa 'tycca 'uttto 'etdretl))))))))
  (test cycle
    (breadth-first-visit (sect lookup <> (table (cons 'htts (set 'ofi))
                                                (cons 'ofi (set 'htts))))
                         prepend)
    1 (procedure?)
    (lambda (visitor)
      (suite cycle-internal ()
        (test cycle-operate
          (visitor 'ofi (set) (list))
          2 (symset? symlist?)
          (lambda (alpha beta)
            (and (symset=? alpha (set 'htts 'ofi))
                 (symlist=? beta (list 'htts 'ofi))))))))
  (test alternatives
    (breadth-first-visit
      (sect lookup <> (table (cons 'aae (set 'ctasku 'aaoo))
                             (cons 'ctasku (set))
                             (cons 'aaoo (set))))
      prepend)
    1 (procedure?)
    (lambda (visitor)
      (suite alternatives-internal ()
        (test alternatives-operate
          (visitor 'aae (set) (list))
          2 (symset? symlist?)
          (lambda (alpha beta)
            (and (symset=? alpha (set 'aae 'ctasku 'aaoo))
                 (or (symlist=? beta (list 'ctasku 'aaoo 'aae))
                     (symlist=? beta (list 'aaoo 'ctasku 'aae)))))))))
  (test breadth-first
    (breadth-first-visit
      (sect lookup <> (table (cons 'utetao (set 'hkt 'tsroo))
                             (cons 'hkt (set 'oskt))
                             (cons 'oskt (set))
                             (cons 'tsroo (set 'mttct))
                             (cons 'mttct (set))))
      prepend)
    1 (procedure?)
    (lambda (visitor)
      (suite breadth-first-internal ()
        (test breadth-first-operate
          (visitor 'utetao (set) (list))
          2 (symset? symlist?)
          (lambda (alpha beta)
            (and (symset=? alpha (set 'utetao 'hkt 'oskt 'tsroo 'mttct))
                 (or (symlist=? beta
                                (list 'mttct 'oskt 'tsroo 'hkt 'utetao))
                     (symlist=? beta
                                (list 'oskt 'mttct 'hkt 'tsroo
                                      'utetao)))))))))
  (test back-links
    (breadth-first-visit
      (sect lookup <> (table (cons 'eaiesl (set 'tiatvas 'gcsraso))
                             (cons 'gcsraso (set 'oicoaf))
                             (cons 'tiatvas (set 'hha))
                             (cons 'hha (set 'eaiesl 'tiatvas 'gcsraso))
                             (cons 'oicoaf
                                   (set 'gcsraso 'eaiesl 'tiatvas))))
      prepend)
    1 (procedure?)
    (lambda (visitor)
      (suite back-links-internal ()
        (test back-links-operate
          (visitor 'eaiesl (set 'idegeds 'hhi 'whhets) (list))
          2 (symset? symlist?)
          (lambda (alpha beta)
            (and (symset=? alpha (set 'eaiesl 'gcsraso 'tiatvas 'hha
                                      'oicoaf 'idegeds 'hhi 'whhets))
                 (or (symlist=? beta (list 'oicoaf 'hha 'gcsraso 'tiatvas
                                           'eiaesl))
                     (symlist=? beta (list 'hha 'oicoaf 'tiatvas 'gcsraso
                                           'eaiesl)))))))))
  (test multiple-paths
    (breadth-first-visit
      (sect lookup <> (table (cons 'iodee (set 'eatve 'hun))
                             (cons 'hyfnepd (set))
                             (cons 'eatve (set 'hyfnepd))
                             (cons 'hun (set 'hyfnepd))))
      prepend)
    1 (procedure?)
    (lambda (visitor)
      (suite multiple-paths-internal ()
        (test multiple-paths-operate
          (visitor 'iodee (set) (list))
          2 (symset? symlist?)
          (lambda (alpha beta)
            (and (symset=? alpha (set 'iodee 'hyfnepd 'eatve 'hun))
                 (or (symlist=? beta (list 'hyfnepd 'hun 'eatve 'iodee))
                     (symlist=? beta (list 'hyfnepd 'eatve 'hun
                                           'iodee))))))))))

(suite breadth-first-traversal
    ((step (lambda (vertex counter tabula)
             (values (add1 counter)
                     (put-into-table vertex counter tabula))))
     (tab? (table-of symbol? natural-number? symbol=?))
     (tab=? (table-of= symbol=? =)))
  (test vertex-labeling
    (breadth-first-traversal (create 0 (table)) step)
    1 (procedure?)
    (lambda (labeler)
      (suite breadth-first-traversal-internal ()
        (test empty-graph
          (labeler (empty-graph))
          2 (natural-number? tab?)
          (lambda (alpha beta)
            (and (zero? alpha)
                 (tab=? beta (table)))))
        (test one-vertex-no-arcs
          (labeler (make-graph (set 'yestl) (arcset)))
          2 (natural-number? tab?)
          (lambda (alpha beta)
            (and (= alpha 1)
                 (tab=? beta (table (cons 'yestl 0))))))
        (test one-vertex-one-arc
          (labeler (make-graph (set 'ressht)
                               (arcset (make-arc 'ressht 'ressht 'knu))))
          2 (natural-number? tab?)
          (lambda (alpha beta)
            (and (= alpha 1)
                 (tab=? beta (table (cons 'ressht 0))))))
        (test two-vertices-one-arc
          (labeler (make-graph (set 'dyrepmi 'eibtni)
                               (arcset (make-arc 'dyrepmi 'eibtni
                                                 'yrnni))))
          2 (natural-number? tab?)
          (lambda (alpha beta)
            (and (= alpha 2)
                 (or (tab=? beta (table (cons 'dyrepmi 0)
                                        (cons 'eibtni 1)))
                     (tab=? beta (table (cons 'dyrepmi 1)
                                        (cons 'eibtni 0)))))))
        (test line-graph
          (labeler (make-graph (set 'nlifh 'onug 'enabge 'ieho 'hiohuev)
                               (arcset (make-arc 'nlifh 'onug 'nesg)
                                       (make-arc 'onug 'enabge 'lowoit)
                                       (make-arc 'enabge 'ieho 'hyeesa)
                                       (make-arc 'ieho 'hiohuev 'etue))))
          2 (natural-number? tab?)
          (lambda (alpha beta)
            (and (= alpha 5)
                 (exists? (sect tab=? beta <>)
                          (list (table (cons 'nlifh 0)
                                       (cons 'onug 1)
                                       (cons 'enabge 2)
                                       (cons 'ieho 3)
                                       (cons 'hiohuev 4))
                                (table (cons 'nlifh 4)
                                       (cons 'onug 0)
                                       (cons 'enabge 1)
                                       (cons 'ieho 2)
                                       (cons 'hiohuev 3))
                                (table (cons 'nlifh 3)
                                       (cons 'onug 4)
                                       (cons 'enabge 0)
                                       (cons 'ieho 1)
                                       (cons 'hiohuev 2))
                                (table (cons 'nlifh 4)
                                       (cons 'onug 3)
                                       (cons 'enabge 0)
                                       (cons 'ieho 1)
                                       (cons 'hiohuev 2))
                                (table (cons 'nlifh 2)
                                       (cons 'onug 3)
                                       (cons 'enabge 4)
                                       (cons 'ieho 0)
                                       (cons 'hiohuev 1))
                                (table (cons 'nlifh 4)
                                       (cons 'onug 2)
                                       (cons 'enabge 3)
                                       (cons 'ieho 0)
                                       (cons 'hiohuev 1))
                                (table (cons 'nlifh 3)
                                       (cons 'onug 4)
                                       (cons 'enabge 2)
                                       (cons 'ieho 0)
                                       (cons 'hiohuev 1))
                                (table (cons 'nlifh 4)
                                       (cons 'onug 3)
                                       (cons 'enabge 2)
                                       (cons 'ieho 0)
                                       (cons 'hiohuev 1))
                                (table (cons 'nlifh 1)
                                       (cons 'onug 2)
                                       (cons 'enabge 3)
                                       (cons 'ieho 4)
                                       (cons 'hiohuev 0))
                                (table (cons 'nlifh 4)
                                       (cons 'onug 1)
                                       (cons 'enabge 2)
                                       (cons 'ieho 3)
                                       (cons 'hiohuev 0))
                                (table (cons 'nlifh 3)
                                       (cons 'onug 4)
                                       (cons 'enabge 1)
                                       (cons 'ieho 2)
                                       (cons 'hiohuev 0))
                                (table (cons 'nlifh 4)
                                       (cons 'onug 3)
                                       (cons 'enabge 1)
                                       (cons 'ieho 2)
                                       (cons 'hiohuev 0))
                                (table (cons 'nlifh 2)
                                       (cons 'onug 3)
                                       (cons 'enabge 4)
                                       (cons 'ieho 1)
                                       (cons 'hiohuev 0))
                                (table (cons 'nlifh 4)
                                       (cons 'onug 2)
                                       (cons 'enabge 3)
                                       (cons 'ieho 1)
                                       (cons 'hiohuev 0))
                                (table (cons 'nlifh 3)
                                       (cons 'onug 4)
                                       (cons 'enabge 2)
                                       (cons 'ieho 1)
                                       (cons 'hiohuev 0))
                                (table (cons 'nlifh 4)
                                       (cons 'onug 3)
                                       (cons 'enabge 2)
                                       (cons 'ieho 1)
                                       (cons 'hiohuev 0)))))))
        (test cycle
          (labeler (make-graph (set 'slny 'apsni 'etht 'aaa)
                               (arcset (make-arc 'slny 'apsni 'edefui)
                                       (make-arc 'apsni 'etht 'selhbth)
                                       (make-arc 'etht 'aaa 'eieenib)
                                       (make-arc 'aaa 'slny 'coegots))))
          2 (natural-number? tab?)
          (lambda (alpha beta)
            (and (= alpha 4)
                 (exists? (sect tab=? beta <>)
                          (list (table (cons 'etht 0)
                                       (cons 'aaa 1)
                                       (cons 'slny 2)
                                       (cons 'apsni 3))
                                (table (cons 'etht 3)
                                       (cons 'aaa 0)
                                       (cons 'slny 1)
                                       (cons 'apsni 2))
                                (table (cons 'etht 2)
                                       (cons 'aaa 3)
                                       (cons 'slny 0)
                                       (cons 'apsni 1))
                                (table (cons 'etht 1)
                                       (cons 'aaa 2)
                                       (cons 'slny 3)
                                       (cons 'apsni 0)))))))
        (test breadth-first-check
          (labeler (make-graph (set 'nneoueh 'negh 'dne 'cwdher 'egheaeu
                                    'hnergo)
                               (arcset (make-arc 'nneoueh 'dne 'uici)
                                       (make-arc 'nneoueh 'cwdher 'lstl)
                                       (make-arc 'negh 'dne 'uhaptst)
                                       (make-arc 'negh 'cwdher 'yad)
                                       (make-arc 'dne 'egheaeu 'nwlit)
                                       (make-arc 'cwdher 'hnergo 'koogote)
                                       (make-arc 'egheaeu 'nneoueh 'harnt)
                                       (make-arc 'hnergo 'negh 'aes))))
          2 (natural-number? tab?)
          (lambda (alpha beta)
            (and (= alpha 6)
                 (exists? (sect tab=? beta <>)
                          (list (table (cons 'nneoueh 0)
                                       (cons 'negh 5)
                                       (cons 'dne 1)
                                       (cons 'cwdher 2)
                                       (cons 'egheaeu 3)
                                       (cons 'hnergo 4))
                                (table (cons 'nneoueh 0)
                                       (cons 'negh 5)
                                       (cons 'dne 2)
                                       (cons 'cwdher 1)
                                       (cons 'egheaeu 4)
                                       (cons 'hnergo 3))
                                (table (cons 'nneoueh 5)
                                       (cons 'negh 0)
                                       (cons 'dne 1)
                                       (cons 'cwdher 2)
                                       (cons 'egheaeu 3)
                                       (cons 'hnergo 4))
                                (table (cons 'nneoueh 5)
                                       (cons 'negh 0)
                                       (cons 'dne 2)
                                       (cons 'cwdher 1)
                                       (cons 'egheaeu 4)
                                       (cons 'hnergo 3))
                                (table (cons 'nneoueh 2)
                                       (cons 'negh 5)
                                       (cons 'dne 0)
                                       (cons 'cwdher 3)
                                       (cons 'egheaeu 1)
                                       (cons 'hnergo 4))
                                (table (cons 'nneoueh 5)
                                       (cons 'negh 2)
                                       (cons 'dne 3)
                                       (cons 'cwdher 0)
                                       (cons 'egheaeu 4)
                                       (cons 'hnergo 1))
                                (table (cons 'nneoueh 1)
                                       (cons 'negh 5)
                                       (cons 'dne 2)
                                       (cons 'cwdher 3)
                                       (cons 'egheaeu 0)
                                       (cons 'hnergo 4))
                                (table (cons 'nneoueh 1)
                                       (cons 'negh 5)
                                       (cons 'dne 3)
                                       (cons 'cwdher 2)
                                       (cons 'egheaeu 0)
                                       (cons 'hnergo 4))
                                (table (cons 'nneoueh 5)
                                       (cons 'negh 1)
                                       (cons 'dne 2)
                                       (cons 'cwdher 3)
                                       (cons 'egheaeu 4)
                                       (cons 'hnergo 0))
                                (table (cons 'nneoueh 5)
                                       (cons 'negh 1)
                                       (cons 'dne 3)
                                       (cons 'cwdher 2)
                                       (cons 'egheaeu 4)
                                       (cons 'hnergo 0)))))))))))

;;; Test the procedures from the section on spanning trees.

(suite slow-spanning-tree
    ((span (lambda (verts size)
             (^and undirected?
                   path-connected?
                   (pipe vertices (match? set=? verts))
                   (run edges cardinality (match? = size))))))
  (test null-graph
    (slow-spanning-tree (empty-graph))
    1 (graph?)
    (match? graph=? (empty-graph)))
  (test one-vertex
    (slow-spanning-tree (make-graph (set 'ainwws) (arcset)))
    1 (graph?)
    (match? graph=? (make-graph (set 'ainwws) (arcset))))
  (test two-vertex
    (slow-spanning-tree (make-graph (set 'ryn 'caeis)
                                    (arcset (make-arc 'ryn 'caeis null)
                                            (make-arc 'caeis 'ryn null))))
    1 (graph?)
    (match? graph=? (make-graph (set 'ryn 'caeis)
                                (arcset (make-arc 'ryn 'caeis null)
                                        (make-arc 'caeis 'ryn null)))))
  (test three-vertex
    (slow-spanning-tree (make-graph (set 'wif 'rhoeo 'rldh)
                                    (arcset (make-arc 'wif 'rhoeo null)
                                            (make-arc 'rhoeo 'wif null)
                                            (make-arc 'wif 'rldh null)
                                            (make-arc 'rldh 'wif null)
                                            (make-arc 'rhoeo 'rldh null)
                                            (make-arc 'rldh 'rhoeo null))))
    1 (graph?)
    (span (set 'wif 'rhoeo 'rldh) 2))
  (test random-graph
    (slow-spanning-tree
     (make-graph (set 'radn 'nae 'tio 'esscaes 'ein 'noibh 'sdaoadu)
                 (arcset (make-arc 'radn 'nae 'onhorl)
                         (make-arc 'nae 'radn 'onhorl)
                         (make-arc 'radn 'tio 'miy)
                         (make-arc 'tio 'radn 'miy)
                         (make-arc 'radn 'esscaes 'tsgor)
                         (make-arc 'esscaes 'radn 'tsgor)
                         (make-arc 'radn 'ein 'cmhyih)
                         (make-arc 'ein 'radn 'cmhyih)
                         (make-arc 'radn 'noibh 'rnodd)
                         (make-arc 'noibh 'radn 'rnodd)
                         (make-arc 'nae 'tio 'feensnp)
                         (make-arc 'tio 'nae 'feensnp)
                         (make-arc 'nae 'ein 'feherh)
                         (make-arc 'ein 'nae 'feherh)
                         (make-arc 'nae 'noibh 'nyhei)
                         (make-arc 'noibh 'nae 'nyhei)
                         (make-arc 'nae 'sdaoadu 'rdta)
                         (make-arc 'sdaoadu 'nae 'rdta)
                         (make-arc 'tio 'ein 'trohrrs)
                         (make-arc 'ein 'tio 'trohrrs)
                         (make-arc 'tio 'noibh 'ribshkr)
                         (make-arc 'noibh 'tio 'ribshkr)
                         (make-arc 'tio 'sdaoadu 'seatnlt)
                         (make-arc 'sdaoadu 'tio 'seatnlt)
                         (make-arc 'esscaes 'ein 'aolt)
                         (make-arc 'ein 'esscaes 'aolt)
                         (make-arc 'esscaes 'noibh 'iluye)
                         (make-arc 'noibh 'esscaes 'iluye)
                         (make-arc 'esscaes 'sdaoadu 'oiiwa)
                         (make-arc 'sdaoadu 'esscaes 'oiiwa)
                         (make-arc 'ein 'sdaoadu 'thnhfva)
                         (make-arc 'sdaoadu 'ein 'thnhfva)
                         (make-arc 'noibh 'sdaoadu 'dts)
                         (make-arc 'sdaoadu 'noibh 'dts))))
    1 (graph?)
    (span (set 'radn 'nae 'tio 'esscaes 'ein 'noibh 'sdaoadu) 6)))

(suite spanning-tree
    ((span (lambda (verts size)
             (^and undirected?
                   path-connected?
                   (pipe vertices (match? set=? verts))
                   (run edges cardinality (match? = size))))))
  (test empty-graph
    (spanning-tree (empty-graph))
    1 (graph?)
    (match? graph=? (empty-graph)))
  (test one-vertex
    (spanning-tree (make-graph (set 'otulih) (arcset)))
    1 (graph?)
    (match? graph=? (make-graph (set 'otulih) (arcset))))
  (test two-vertex
    (spanning-tree (make-graph (set 'easit 'tito)
                               (arcset (make-arc 'easit 'tito null)
                                       (make-arc 'tito 'easit null))))
    1 (graph?)
    (match? graph=? (make-graph (set 'easit 'tito)
                                (arcset (make-arc 'easit 'tito null)
                                        (make-arc 'tito 'easit null)))))
  (test three-vertex
    (spanning-tree
      (make-graph (set 'faa 'eaodtsy 'hwwes)
                  (arcset (make-arc 'faa 'eaodtsy 'gddv)
                          (make-arc 'eaodtsy 'faa 'gddv)
                          (make-arc 'faa 'hwwes 'oeiph)
                          (make-arc 'hwwes 'faa 'oeiph)
                          (make-arc 'eaodtsy 'hwwes 'rslwdlf)
                          (make-arc 'hwwes 'eaodtsy 'rslwdlf))))
    1 (graph?)
    (span (set 'faa 'eaodtsy 'hwwes) 2))
  (test random-graph
    (spanning-tree
      (make-graph (set 'fiyl 'bfythum 'iaaei 'ryov 'osr 'oegw 'ihetdhe)
                  (arcset (make-arc 'fiyl 'iaaei null)
                          (make-arc 'fiyl 'ryov null)
                          (make-arc 'fiyl 'osr null)
                          (make-arc 'fiyl 'ihetdhe null)
                          (make-arc 'bfythum 'iaaei null)
                          (make-arc 'bfythum 'osr null)
                          (make-arc 'bfythum 'oegw null)
                          (make-arc 'bfythum 'ihetdhe null)
                          (make-arc 'iaaei 'fiyl null)
                          (make-arc 'iaaei 'bfythum null)
                          (make-arc 'iaaei 'osr null)
                          (make-arc 'iaaei 'oegw null)
                          (make-arc 'iaaei 'ihetdhe null)
                          (make-arc 'ryov 'fiyl null)
                          (make-arc 'ryov 'oegw null)
                          (make-arc 'ryov 'ihetdhe null)
                          (make-arc 'osr 'fiyl null)
                          (make-arc 'osr 'bfythum null)
                          (make-arc 'osr 'iaaei null)
                          (make-arc 'osr 'oegw null)
                          (make-arc 'osr 'ihetdhe null)
                          (make-arc 'oegw 'bfythum null)
                          (make-arc 'oegw 'iaaei null)
                          (make-arc 'oegw 'ryov null)
                          (make-arc 'oegw 'osr null)
                          (make-arc 'oegw 'ihetdhe null)
                          (make-arc 'ihetdhe 'fiyl null)
                          (make-arc 'ihetdhe 'bfythum null)
                          (make-arc 'ihetdhe 'iaaei null)
                          (make-arc 'ihetdhe 'ryov null)
                          (make-arc 'ihetdhe 'osr null)
                          (make-arc 'ihetdhe 'oegw null))))
    1 (graph?)
    (span (set 'fiyl 'bfythum 'iaaei 'ryov 'osr 'oegw 'ihetdhe) 6)))

(suite minimum-spanning-tree
    ((minspan (lambda (verts size weight-sum)
                (^and undirected?
                      path-connected?
                      (pipe vertices (match? set=? verts))
                      (run edges cardinality (match? = size))
                      (run edges
                           (fold-set (create 0)
                                     (pipe (~initial edge-label) +))
                           (match? = weight-sum))))))
  (test empty-graph
    (minimum-spanning-tree (empty-graph))
    1 (graph?)
    (match? graph=? (empty-graph)))
  (test one-vertex
    (minimum-spanning-tree (make-graph (set 'orne) (arcset)))
    1 (graph?)
    (match? graph=? (make-graph (set 'orne) (arcset))))
  (test two-vertices
    (minimum-spanning-tree
      (make-graph (set 'trysda 'mhoaal)
                  (arcset (make-arc 'trysda 'mhoaal 331)
                          (make-arc 'mhoaal 'trysda 331))))
    1 (graph?)
    (match? graph=? (make-graph (set 'trysda 'mhoaal)
                                (arcset (make-arc 'trysda 'mhoaal 331)
                                        (make-arc 'mhoaal 'trysda 331)))))
  (test simplest-non-trivial
    (minimum-spanning-tree
      (make-graph (set 'iyi 'ettl 'reeema)
                  (arcset (make-arc 'iyi 'ettl 133)
                          (make-arc 'ettl 'iyi 133)
                          (make-arc 'iyi 'reeema 862)
                          (make-arc 'reeema 'iyi 862)
                          (make-arc 'ettl 'reeema 845)
                          (make-arc 'reeema 'ettl 845))))
    1 (graph?)
    (match? graph=? (make-graph (set 'iyi 'ettl 'reeema)
                    (arcset (make-arc 'iyi 'ettl 133)
                            (make-arc 'ettl 'iyi 133)
                            (make-arc 'ettl 'reeema 845)
                            (make-arc 'reeema 'ettl 845)))))
  (test easy-choices
    (minimum-spanning-tree (make-graph (set 'a 'b 'c 'd 'e 'f 'g 'h 'i)
                                       (arcset (make-arc 'a 'b 2)
                                               (make-arc 'b 'a 2)
                                               (make-arc 'a 'c 2)
                                               (make-arc 'c 'a 2)
                                               (make-arc 'a 'd 3)
                                               (make-arc 'd 'a 3)
                                               (make-arc 'a 'g 4)
                                               (make-arc 'g 'a 4)
                                               (make-arc 'b 'c 3)
                                               (make-arc 'c 'b 3)
                                               (make-arc 'b 'i 4)
                                               (make-arc 'i 'b 4)
                                               (make-arc 'c 'e 4)
                                               (make-arc 'e 'c 4)
                                               (make-arc 'd 'e 3)
                                               (make-arc 'e 'd 3)
                                               (make-arc 'd 'f 2)
                                               (make-arc 'f 'd 2)
                                               (make-arc 'd 'g 4)
                                               (make-arc 'g 'd 4)
                                               (make-arc 'e 'f 2)
                                               (make-arc 'f 'e 2)
                                               (make-arc 'f 'h 3)
                                               (make-arc 'h 'f 3)
                                               (make-arc 'g 'h 2)
                                               (make-arc 'h 'g 2)
                                               (make-arc 'g 'i 3)
                                               (make-arc 'i 'g 3)
                                               (make-arc 'h 'i 2)
                                               (make-arc 'i 'h 2))))
    1 (graph?)
    (match? graph=? (make-graph (set 'a 'b 'c 'd 'e 'f 'g 'h 'i)
                    (arcset (make-arc 'a 'b 2)
                            (make-arc 'b 'a 2)
                            (make-arc 'a 'c 2)
                            (make-arc 'c 'a 2)
                            (make-arc 'a 'd 3)
                            (make-arc 'd 'a 3)
                            (make-arc 'd 'f 2)
                            (make-arc 'f 'd 2)
                            (make-arc 'e 'f 2)
                            (make-arc 'f 'e 2)
                            (make-arc 'f 'h 3)
                            (make-arc 'h 'f 3)
                            (make-arc 'g 'h 2)
                            (make-arc 'h 'g 2)
                            (make-arc 'h 'i 2)
                            (make-arc 'i 'h 2)))))

  ;; An example from Cormen _et al._ (_Introduction to algorithms_, third
  ;; edition, p. 625):

  (test Cormen
    (minimum-spanning-tree (make-graph (set 'a 'b 'c 'd 'e 'f 'g 'h 'i)
                                       (arcset (make-arc 'a 'b 4)
                                               (make-arc 'b 'a 4)
                                               (make-arc 'a 'h 8)
                                               (make-arc 'h 'a 8)
                                               (make-arc 'b 'c 8)
                                               (make-arc 'c 'b 8)
                                               (make-arc 'b 'h 11)
                                               (make-arc 'h 'b 11)
                                               (make-arc 'c 'd 7)
                                               (make-arc 'd 'c 7)
                                               (make-arc 'c 'f 4)
                                               (make-arc 'f 'c 4)
                                               (make-arc 'c 'i 2)
                                               (make-arc 'i 'c 2)
                                               (make-arc 'd 'e 9)
                                               (make-arc 'e 'd 9)
                                               (make-arc 'd 'f 14)
                                               (make-arc 'f 'd 14)
                                               (make-arc 'e 'f 10)
                                               (make-arc 'f 'e 10)
                                               (make-arc 'f 'g 2)
                                               (make-arc 'g 'f 2)
                                               (make-arc 'g 'h 1)
                                               (make-arc 'h 'g 1)
                                               (make-arc 'g 'i 6)
                                               (make-arc 'i 'g 6)
                                               (make-arc 'h 'i 7)
                                               (make-arc 'i 'h 7))))
    1 (graph?)
    (minspan (set 'a 'b 'c 'd 'e 'f 'g 'h 'i) 8 37))

  ;; An example from Rabhi and Lapalme (_Algorithms: a functional
  ;; programming approach_, p. 148).

  (test Rabhi
    (minimum-spanning-tree (make-graph (set 1 2 3 4 5)
                                       (arcset (make-arc 1 2 12)
                                               (make-arc 2 1 12)
                                               (make-arc 1 3 34)
                                               (make-arc 3 1 34)
                                               (make-arc 1 5 78)
                                               (make-arc 5 1 78)
                                               (make-arc 2 4 55)
                                               (make-arc 4 2 55)
                                               (make-arc 2 5 32)
                                               (make-arc 5 2 32)
                                               (make-arc 3 4 61)
                                               (make-arc 4 3 61)
                                               (make-arc 3 5 44)
                                               (make-arc 5 3 44)
                                               (make-arc 4 5 93)
                                               (make-arc 5 4 93))))
    1 (graph?)
    (match? graph=? (make-graph (set 1 2 3 4 5)
                                (arcset (make-arc 1 2 12)
                                        (make-arc 2 1 12)
                                        (make-arc 1 3 34)
                                        (make-arc 3 1 34)
                                        (make-arc 2 4 55)
                                        (make-arc 4 2 55)
                                        (make-arc 2 5 32)
                                        (make-arc 5 2 32)))))

  ;; An example from Sedgewick (_Algorithms_, second edition, p. 452).

  (test Sedgewick
    (minimum-spanning-tree (make-graph (set 'a 'b 'c 'd 'e 'f 'g 'h 'i 'j
                                            'k 'l 'm)
                                       (arcset (make-arc 'a 'b 1)
                                               (make-arc 'b 'a 1)
                                               (make-arc 'a 'f 2)
                                               (make-arc 'f 'a 2)
                                               (make-arc 'a 'g 6)
                                               (make-arc 'g 'a 6)
                                               (make-arc 'b 'c 1)
                                               (make-arc 'c 'b 1)
                                               (make-arc 'b 'd 2)
                                               (make-arc 'd 'b 2)
                                               (make-arc 'b 'e 4)
                                               (make-arc 'e 'b 4)
                                               (make-arc 'c 'e 4)
                                               (make-arc 'e 'c 4)
                                               (make-arc 'd 'e 2)
                                               (make-arc 'e 'd 2)
                                               (make-arc 'd 'f 1)
                                               (make-arc 'f 'd 1)
                                               (make-arc 'e 'f 2)
                                               (make-arc 'f 'e 2)
                                               (make-arc 'e 'g 1)
                                               (make-arc 'g 'e 1)
                                               (make-arc 'e 'l 4)
                                               (make-arc 'l 'e 4)
                                               (make-arc 'f 'l 2)
                                               (make-arc 'l 'f 2)
                                               (make-arc 'g 'h 3)
                                               (make-arc 'h 'g 3)
                                               (make-arc 'g 'j 1)
                                               (make-arc 'j 'g 1)
                                               (make-arc 'g 'l 5)
                                               (make-arc 'l 'g 5)
                                               (make-arc 'h 'i 2)
                                               (make-arc 'i 'h 2)
                                               (make-arc 'i 'k 1)
                                               (make-arc 'k 'i 1)
                                               (make-arc 'j 'k 1)
                                               (make-arc 'k 'j 1)
                                               (make-arc 'j 'l 3)
                                               (make-arc 'l 'j 3)
                                               (make-arc 'j 'm 2)
                                               (make-arc 'm 'j 2)
                                               (make-arc 'l 'm 1)
                                               (make-arc 'm 'l 1))))
    1 (graph?)
    (minspan (set 'a 'b 'c 'd 'e 'f 'g 'h 'i 'j 'k 'l 'm) 12 16))

  ;; Randomly constructed examples.

  (test random-0
    (minimum-spanning-tree
      (make-graph (set 'rueurs 'rraaea 'rmeotob 'isu 'enrln 'wehnh 'gsnmeta
                       'raprldh)
                  (arcset (make-arc 'rueurs 'rraaea 3)
                          (make-arc 'rraaea 'rueurs 3)
                          (make-arc 'rueurs 'rmeotob 1)
                          (make-arc 'rmeotob 'rueurs 1)
                          (make-arc 'rueurs 'isu 8)
                          (make-arc 'isu 'rueurs 8)
                          (make-arc 'rueurs 'enrln 3)
                          (make-arc 'enrln 'rueurs 3)
                          (make-arc 'rraaea 'wehnh 6)
                          (make-arc 'wehnh 'rraaea 6)
                          (make-arc 'rraaea 'rmeotob 9)
                          (make-arc 'rmeotob 'rraaea 9)
                          (make-arc 'rraaea 'isu 2)
                          (make-arc 'isu 'rraaea 2)
                          (make-arc 'rraaea 'enrln 5)
                          (make-arc 'enrln 'rraaea 5)
                          (make-arc 'rraaea 'gsnmeta 2)
                          (make-arc 'gsnmeta 'rraaea 2)
                          (make-arc 'rmeotob 'wehnh 6)
                          (make-arc 'wehnh 'rmeotob 6)
                          (make-arc 'rmeotob 'raprldh 7)
                          (make-arc 'raprldh 'rmeotob 7)
                          (make-arc 'rmeotob 'gsnmeta 7)
                          (make-arc 'gsnmeta 'rmeotob 7)
                          (make-arc 'isu 'wehnh 1)
                          (make-arc 'wehnh 'isu 1)
                          (make-arc 'isu 'enrln 6)
                          (make-arc 'enrln 'isu 6)
                          (make-arc 'isu 'raprldh 8)
                          (make-arc 'raprldh 'isu 8)
                          (make-arc 'isu 'gsnmeta 6)
                          (make-arc 'gsnmeta 'isu 6)
                          (make-arc 'enrln 'wehnh 2)
                          (make-arc 'wehnh 'enrln 2)
                          (make-arc 'enrln 'raprldh 4)
                          (make-arc 'raprldh 'enrln 4)
                          (make-arc 'enrln 'gsnmeta 2)
                          (make-arc 'gsnmeta 'enrln 2)
                          (make-arc 'wehnh 'raprldh 4)
                          (make-arc 'raprldh 'wehnh 4))))
    1 (graph?)
    (minspan (set 'rueurs 'rraaea 'rmeotob 'isu 'enrln 'wehnh 'gsnmeta
                  'raprldh)
             7
             15))

  (test random-1
    (minimum-spanning-tree
      (make-graph (set 'crweee 'gsiigoi 'efesfrc 'enetl 'dfviy 'eetkneo
                       'aaa 'rtwhtie 'yda)
                  (arcset (make-arc 'crweee 'gsiigoi 1)
                          (make-arc 'gsiigoi 'crweee 1)
                          (make-arc 'crweee 'efesfrc 6)
                          (make-arc 'efesfrc 'crweee 6)
                          (make-arc 'crweee 'enetl 1)
                          (make-arc 'enetl 'crweee 1)
                          (make-arc 'gsiigoi 'rtwhtie 7)
                          (make-arc 'rtwhtie 'gsiigoi 7)
                          (make-arc 'gsiigoi 'efesfrc 6)
                          (make-arc 'efesfrc 'gsiigoi 6)
                          (make-arc 'gsiigoi 'enetl 9)
                          (make-arc 'enetl 'gsiigoi 9)
                          (make-arc 'gsiigoi 'yda 5)
                          (make-arc 'yda 'gsiigoi 5)
                          (make-arc 'efesfrc 'dfviy 9)
                          (make-arc 'dfviy 'efesfrc 9)
                          (make-arc 'efesfrc 'aaa 6)
                          (make-arc 'aaa 'efesfrc 6)
                          (make-arc 'efesfrc 'enetl 2)
                          (make-arc 'enetl 'efesfrc 2)
                          (make-arc 'enetl 'eetkneo 5)
                          (make-arc 'eetkneo 'enetl 5)
                          (make-arc 'enetl 'aaa 7)
                          (make-arc 'aaa 'enetl 7)
                          (make-arc 'enetl 'rtwhtie 4)
                          (make-arc 'rtwhtie 'enetl 4)
                          (make-arc 'enetl 'yda 6)
                          (make-arc 'yda 'enetl 6)
                          (make-arc 'dfviy 'eetkneo 1)
                          (make-arc 'eetkneo 'dfviy 1)
                          (make-arc 'dfviy 'aaa 4)
                          (make-arc 'aaa 'dfviy 4)
                          (make-arc 'dfviy 'rtwhtie 3)
                          (make-arc 'rtwhtie 'dfviy 3)
                          (make-arc 'dfviy 'yda 3)
                          (make-arc 'yda 'dfviy 3)
                          (make-arc 'eetkneo 'rtwhtie 1)
                          (make-arc 'rtwhtie 'eetkneo 1)
                          (make-arc 'aaa 'rtwhtie 5)
                          (make-arc 'rtwhtie 'aaa 5))))
    1 (graph?)
    (minspan (set 'crweee 'gsiigoi 'efesfrc 'enetl 'dfviy 'eetkneo
                  'aaa 'rtwhtie 'yda)
             8
             17)))

(suite crossings ((edgeset? (set-of edge? edge=?))
                  (edgeset=? (set-of= edge=?)))
  (test minimal
    (crossings (make-graph (set 'cdtrem 'yio)
                           (arcset (make-arc 'cdtrem 'yio null)
                                   (make-arc 'yio 'cdtrem null))))
    1 (procedure?)
    (lambda (crossings-builder)
      (suite minimal-internal ()
        (test minimal-cut
          (crossings-builder (set 'cdtrem))
          1 (edgeset?)
          (match? edgeset=?
                 (set (make-edge (set 'cdtrem 'yio) null)))))))
  (test typical
    (crossings (make-graph (set 'ueooy 'oaoy 'toi 'wfnrpn 'tastvfk 'orfcoyo
                                'jto 'leno 'vtofhoe)
                           (arcset (make-arc 'ueooy 'oaoy 'iouliro)
                                   (make-arc 'oaoy 'ueooy 'iouliro)
                                   (make-arc 'ueooy 'toi 'sena)
                                   (make-arc 'toi 'ueooy 'sena)
                                   (make-arc 'ueooy 'wfnrpn 'srawo)
                                   (make-arc 'wfnrpn 'ueooy 'srawo)
                                   (make-arc 'ueooy 'tastvfk 'oowyhu)
                                   (make-arc 'tastvfk 'ueooy 'oowyhu)
                                   (make-arc 'ueooy 'orfcoyo 'mrsa)
                                   (make-arc 'orfcoyo 'ueooy 'mrsa)
                                   (make-arc 'ueooy 'jto 'raeai)
                                   (make-arc 'jto 'ueooy 'raeai)
                                   (make-arc 'oaoy 'toi 'edrl)
                                   (make-arc 'toi 'oaoy 'edrl)
                                   (make-arc 'oaoy 'orfcoyo 'hnrhmrs)
                                   (make-arc 'orfcoyo 'oaoy 'hnrhmrs)
                                   (make-arc 'oaoy 'leno 'nusruh)
                                   (make-arc 'leno 'oaoy 'nusruh)
                                   (make-arc 'oaoy 'vtofhoe 'hyioa)
                                   (make-arc 'vtofhoe 'oaoy 'hyioa)
                                   (make-arc 'toi 'tastvfk 'emswur)
                                   (make-arc 'tastvfk 'toi 'emswur)
                                   (make-arc 'toi 'orfcoyo 'tll)
                                   (make-arc 'orfcoyo 'toi 'tll)
                                   (make-arc 'toi 'vtofhoe 'tfvohnn)
                                   (make-arc 'vtofhoe 'toi 'tfvohnn)
                                   (make-arc 'wfnrpn 'tastvfk 'bteoie)
                                   (make-arc 'tastvfk 'wfnrpn 'bteoie)
                                   (make-arc 'wfnrpn 'orfcoyo 'seta)
                                   (make-arc 'orfcoyo 'wfnrpn 'seta)
                                   (make-arc 'wfnrpn 'leno 'saaa)
                                   (make-arc 'leno 'wfnrpn 'saaa)
                                   (make-arc 'wfnrpn 'vtofhoe 'vfynw)
                                   (make-arc 'vtofhoe 'wfnrpn 'vfynw)
                                   (make-arc 'tastvfk 'vtofhoe 'nni)
                                   (make-arc 'vtofhoe 'tastvfk 'nni)
                                   (make-arc 'tastvfk 'jto 'ahleont)
                                   (make-arc 'jto 'tastvfk 'ahleont)
                                   (make-arc 'orfcoyo 'leno 'hgh)
                                   (make-arc 'leno 'orfcoyo 'hgh)
                                   (make-arc 'orfcoyo 'vtofhoe 'eesgoee)
                                   (make-arc 'vtofhoe 'orfcoyo 'eesgoee)
                                   (make-arc 'jto 'leno 'bhnh)
                                   (make-arc 'leno 'jto 'bhnh))))
    1 (procedure?)
    (lambda (crossings-builder)
      (suite typical-internal ()
        (test low-degree
          (crossings-builder (set 'jto))
          1 (edgeset?)
          (match? edgeset=?
                  (set (make-edge (set 'jto 'leno) 'bhnh)
                       (make-edge (set 'ueooy 'jto) 'raeai)
                       (make-edge (set 'tastvfk 'jto) 'ahleont))))
        (test inverted-low-degree
          (crossings-builder (set 'ueooy 'oaoy 'toi 'wfnrpn 'tastvfk
                                  'orfcoyo 'leno 'vtofhoe))
          1 (edgeset?)
          (match? edgeset=?
                  (set (make-edge (set 'leno 'jto) 'bhnh)
                       (make-edge (set 'jto 'ueooy) 'raeai)
                       (make-edge (set 'tastvfk 'jto) 'ahleont))))
        (test high-degree
          (crossings-builder (set 'ueooy 'oaoy 'leno 'vtofhoe))
          1 (edgeset?)
          (match? edgeset=?
                  (set (make-edge (set 'ueooy 'toi) 'sena)
                       (make-edge (set 'ueooy 'wfnrpn) 'srawo)
                       (make-edge (set 'ueooy 'tastvfk) 'oowyhu)
                       (make-edge (set 'ueooy 'orfcoyo) 'mrsa)
                       (make-edge (set 'ueooy 'jto) 'raeai)
                       (make-edge (set 'oaoy 'toi) 'edrl)
                       (make-edge (set 'oaoy 'orfcoyo) 'hnrhmrs)
                       (make-edge (set 'toi 'vtofhoe) 'tfvohnn)
                       (make-edge (set 'wfnrpn 'leno) 'saaa)
                       (make-edge (set 'wfnrpn 'vtofhoe) 'vfynw)
                       (make-edge (set 'tastvfk 'vtofhoe) 'nni)
                       (make-edge (set 'orfcoyo 'leno) 'hgh)
                       (make-edge (set 'orfcoyo 'vtofhoe) 'eesgoee)
                       (make-edge (set 'jto 'leno) 'bhnh))))))))

(suite light-edge ()
  (test minimal
    (light-edge (make-graph (set 'soo 'gihmpdl)
                            (arcset (make-arc 'soo 'gihmpdl 85)
                                    (make-arc 'gihmpdl 'soo 85))))
    1 (procedure?)
    (lambda (light-edge-finder)
      (suite minimal-internal ()
        (test unique-edge
          (light-edge-finder (set 'soo))
          1 (edge?)
          (match? edge=? (make-edge (set 'soo 'gihmpdl) 85))))))
  (test random-graph
    (light-edge (make-graph (set 'ftll 'dbu 'eawd 'hvrleah 'nyo 'nah)
                            (arcset (make-arc 'ftll 'dbu 136)
                                    (make-arc 'dbu 'ftll 136)
                                    (make-arc 'ftll 'eawd 9)
                                    (make-arc 'eawd 'ftll 9)
                                    (make-arc 'dbu 'hvrleah 698)
                                    (make-arc 'hvrleah 'dbu 698)
                                    (make-arc 'dbu 'nyo 683)
                                    (make-arc 'nyo 'dbu 683)
                                    (make-arc 'eawd 'nah 82)
                                    (make-arc 'nah 'eawd 82)
                                    (make-arc 'eawd 'hvrleah 280)
                                    (make-arc 'hvrleah 'eawd 280)
                                    (make-arc 'eawd 'nyo 626)
                                    (make-arc 'nyo 'eawd 626)
                                    (make-arc 'hvrleah 'nah 429)
                                    (make-arc 'nah 'hvrleah 429)
                                    (make-arc 'hvrleah 'nyo 480)
                                    (make-arc 'nyo 'hvrleah 480)
                                    (make-arc 'nyo 'nah 478)
                                    (make-arc 'nah 'nyo 478))))
    1 (procedure?)
    (lambda (light-edge-finder)
      (suite random-internal ()
        (test low-degree
          (light-edge-finder (set 'nyo))
          1 (edge?)
          (match? edge=? (make-edge (set 'nyo 'nah) 478)))
        (test inverted-low-degree
          (light-edge-finder (set 'ftll 'dbu 'eawd 'hvrleah 'nah))
          1 (edge?)
          (match? edge=? (make-edge (set 'nyo 'nah) 478)))
        (test high-degree
          (light-edge-finder (set 'ftll 'dbu 'eawd))
          1 (edge?)
          (match? edge=? (make-edge (set 'eawd 'nah) 82))))))
  (test cormen-graph
    (light-edge (make-graph (set 'a 'b 'c 'd 'e 'f 'g 'h 'i)
                            (arcset (make-arc 'a 'b 4)
                                    (make-arc 'b 'a 4)
                                    (make-arc 'a 'h 8)
                                    (make-arc 'h 'a 8)
                                    (make-arc 'b 'h 11)
                                    (make-arc 'h 'b 11)
                                    (make-arc 'b 'c 8)
                                    (make-arc 'c 'b 8)
                                    (make-arc 'h 'i 7)
                                    (make-arc 'i 'h 7)
                                    (make-arc 'i 'c 2)
                                    (make-arc 'c 'i 2)
                                    (make-arc 'h 'g 1)
                                    (make-arc 'g 'h 1)
                                    (make-arc 'i 'g 6)
                                    (make-arc 'g 'i 6)
                                    (make-arc 'c 'f 4)
                                    (make-arc 'f 'c 4)
                                    (make-arc 'g 'f 2)
                                    (make-arc 'f 'g 2)
                                    (make-arc 'c 'd 7)
                                    (make-arc 'd 'c 7)
                                    (make-arc 'd 'f 14)
                                    (make-arc 'f 'd 14)
                                    (make-arc 'd 'e 9)
                                    (make-arc 'e 'd 9)
                                    (make-arc 'f 'e 10)
                                    (make-arc 'e 'f 10))))
    1 (procedure?)
    (lambda (light-edge-finder)
      (suite cormen-internal ()
        (test cormen
          (light-edge-finder (set 'a 'b 'd 'e))
          1 (edge?)
          (match? edge=? (make-edge (set 'c 'd) 7)))
        (test prim-1
          (light-edge-finder (set 'a))
          1 (edge?)
          (match? edge=? (make-edge (set 'a 'b) 4)))
        (test prim-2
          (light-edge-finder (set 'a 'b))
          1 (edge?)
          (^vel (match? edge=? (make-edge (set 'b 'c) 8))
                (match? edge=? (make-edge (set 'a 'h) 8))))
        (test prim-3
          (light-edge-finder (set 'a 'b 'c))
          1 (edge?)
          (match? edge=? (make-edge (set 'i 'c) 2)))
        (test prim-4
          (light-edge-finder (set 'a 'b 'c 'i))
          1 (edge?)
          (match? edge=? (make-edge (set 'c 'f) 4)))
        (test prim-5
          (light-edge-finder (set 'a 'b 'c 'i 'f))
          1 (edge?)
          (match? edge=? (make-edge (set 'g 'f) 2)))
        (test prim-6
          (light-edge-finder (set 'a 'b 'c 'i 'f 'g))
          1 (edge?)
          (match? edge=? (make-edge (set 'h 'g) 1)))
        (test prim-7
          (light-edge-finder (set 'a 'b 'c 'i 'f 'g 'h))
          1 (edge?)
          (match? edge=? (make-edge (set 'c 'd) 7)))
        (test prim-8
          (light-edge-finder (set 'a 'b 'c 'i 'f 'g 'h 'd))
          1 (edge?)
          (match? edge=? (make-edge (set 'd 'e) 9)))))))

(suite alternative-minimum-spanning-tree
    ((minspan (lambda (verts size weight-sum)
                (^and undirected?
                      path-connected?
                      (pipe vertices (match? set=? verts))
                      (run edges cardinality (match? = size))
                      (run edges
                           (fold-set (create 0)
                                     (pipe (~initial edge-label) +))
                           (match? = weight-sum))))))
  (test empty-graph
    (alternative-minimum-spanning-tree (empty-graph))
    1 (graph?)
    (match? graph=? (empty-graph)))
  (test one-vertex
    (alternative-minimum-spanning-tree (make-graph (set 'hhrnm) (arcset)))
    1 (graph?)
    (match? graph=? (make-graph (set 'hhrnm) (arcset))))
  (test two-vertices
    (alternative-minimum-spanning-tree
      (make-graph (set 'liwdene 'anhvose)
                  (arcset (make-arc 'liwdene 'anhvose 529)
                          (make-arc 'anhvose 'liwdene 529))))
    1 (graph?)
    (match? graph=?
            (make-graph (set 'liwdene 'anhvose)
                        (arcset (make-arc 'liwdene 'anhvose 529)
                                (make-arc 'anhvose 'liwdene 529)))))
  (test simplest-non-trivial
    (alternative-minimum-spanning-tree
      (make-graph (set 'oaashgy 'hiheai 'aafnsgc)
                  (arcset (make-arc 'oaashgy 'hiheai 29)
                          (make-arc 'hiheai 'oaashgy 29)
                          (make-arc 'oaashgy 'aafnsgc 470)
                          (make-arc 'aafnsgc 'oaashgy 470)
                          (make-arc 'hiheai 'aafnsgc 38)
                          (make-arc 'aafnsgc 'hiheai 38))))
    1 (graph?)
    (match? graph=? (make-graph (set 'oaashgy 'hiheai 'aafnsgc)
                                (arcset (make-arc 'oaashgy 'hiheai 29)
                                        (make-arc 'hiheai 'oaashgy 29)
                                        (make-arc 'hiheai 'aafnsgc 38)
                                        (make-arc 'aafnsgc 'hiheai 38)))))
  (test easy-choices
    (alternative-minimum-spanning-tree
      (make-graph (set 'a 'b 'c 'd 'e 'f 'g 'h 'i)
             (arcset (make-arc 'a 'b 2)
                     (make-arc 'b 'a 2)
                     (make-arc 'a 'c 2)
                     (make-arc 'c 'a 2)
                     (make-arc 'a 'd 3)
                     (make-arc 'd 'a 3)
                     (make-arc 'a 'g 4)
                     (make-arc 'g 'a 4)
                     (make-arc 'b 'c 3)
                     (make-arc 'c 'b 3)
                     (make-arc 'b 'i 4)
                     (make-arc 'i 'b 4)
                     (make-arc 'c 'e 4)
                     (make-arc 'e 'c 4)
                     (make-arc 'd 'e 3)
                     (make-arc 'e 'd 3)
                     (make-arc 'd 'f 2)
                     (make-arc 'f 'd 2)
                     (make-arc 'd 'g 4)
                     (make-arc 'g 'd 4)
                     (make-arc 'e 'f 2)
                     (make-arc 'f 'e 2)
                     (make-arc 'f 'h 3)
                     (make-arc 'h 'f 3)
                     (make-arc 'g 'h 2)
                     (make-arc 'h 'g 2)
                     (make-arc 'g 'i 3)
                     (make-arc 'i 'g 3)
                     (make-arc 'h 'i 2)
                     (make-arc 'i 'h 2))))
    1 (graph?)
    (match? graph=? (make-graph (set 'a 'b 'c 'd 'e 'f 'g 'h 'i)
                                (arcset (make-arc 'a 'b 2)
                                        (make-arc 'b 'a 2)
                                        (make-arc 'a 'c 2)
                                        (make-arc 'c 'a 2)
                                        (make-arc 'a 'd 3)
                                        (make-arc 'd 'a 3)
                                        (make-arc 'd 'f 2)
                                        (make-arc 'f 'd 2)
                                        (make-arc 'e 'f 2)
                                        (make-arc 'f 'e 2)
                                        (make-arc 'f 'h 3)
                                        (make-arc 'h 'f 3)
                                        (make-arc 'g 'h 2)
                                        (make-arc 'h 'g 2)
                                        (make-arc 'h 'i 2)
                                        (make-arc 'i 'h 2)))))
  (test Cormen
    (alternative-minimum-spanning-tree
      (make-graph (set 'a 'b 'c 'd 'e 'f 'g 'h 'i)
                  (arcset (make-arc 'a 'b 4)
                          (make-arc 'b 'a 4)
                          (make-arc 'a 'h 8)
                          (make-arc 'h 'a 8)
                          (make-arc 'b 'c 8)
                          (make-arc 'c 'b 8)
                          (make-arc 'b 'h 11)
                          (make-arc 'h 'b 11)
                          (make-arc 'c 'd 7)
                          (make-arc 'd 'c 7)
                          (make-arc 'c 'f 4)
                          (make-arc 'f 'c 4)
                          (make-arc 'c 'i 2)
                          (make-arc 'i 'c 2)
                          (make-arc 'd 'e 9)
                          (make-arc 'e 'd 9)
                          (make-arc 'd 'f 14)
                          (make-arc 'f 'd 14)
                          (make-arc 'e 'f 10)
                          (make-arc 'f 'e 10)
                          (make-arc 'f 'g 2)
                          (make-arc 'g 'f 2)
                          (make-arc 'g 'h 1)
                          (make-arc 'h 'g 1)
                          (make-arc 'g 'i 6)
                          (make-arc 'i 'g 6)
                          (make-arc 'h 'i 7)
                          (make-arc 'i 'h 7))))
    1 (graph?)
    (minspan (set 'a 'b 'c 'd 'e 'f 'g 'h 'i) 8 37))
  (test Rabhi
    (alternative-minimum-spanning-tree
      (make-graph (set 1 2 3 4 5)
                  (arcset (make-arc 1 2 12)
                          (make-arc 2 1 12)
                          (make-arc 1 3 34)
                          (make-arc 3 1 34)
                          (make-arc 1 5 78)
                          (make-arc 5 1 78)
                          (make-arc 2 4 55)
                          (make-arc 4 2 55)
                          (make-arc 2 5 32)
                          (make-arc 5 2 32)
                          (make-arc 3 4 61)
                          (make-arc 4 3 61)
                          (make-arc 3 5 44)
                          (make-arc 5 3 44)
                          (make-arc 4 5 93)
                          (make-arc 5 4 93))))
    1 (graph?)
    (match? graph=? (make-graph (set 1 2 3 4 5)
                                (arcset (make-arc 1 2 12)
                                        (make-arc 2 1 12)
                                        (make-arc 1 3 34)
                                        (make-arc 3 1 34)
                                        (make-arc 2 4 55)
                                        (make-arc 4 2 55)
                                        (make-arc 2 5 32)
                                        (make-arc 5 2 32)))))
  (test Sedgewick
    (alternative-minimum-spanning-tree
      (make-graph (set 'a 'b 'c 'd 'e 'f 'g 'h 'i 'j 'k 'l 'm)
                  (arcset (make-arc 'a 'b 1)
                          (make-arc 'b 'a 1)
                          (make-arc 'a 'f 2)
                          (make-arc 'f 'a 2)
                          (make-arc 'a 'g 6)
                          (make-arc 'g 'a 6)
                          (make-arc 'b 'c 1)
                          (make-arc 'c 'b 1)
                          (make-arc 'b 'd 2)
                          (make-arc 'd 'b 2)
                          (make-arc 'b 'e 4)
                          (make-arc 'e 'b 4)
                          (make-arc 'c 'e 4)
                          (make-arc 'e 'c 4)
                          (make-arc 'd 'e 2)
                          (make-arc 'e 'd 2)
                          (make-arc 'd 'f 1)
                          (make-arc 'f 'd 1)
                          (make-arc 'e 'f 2)
                          (make-arc 'f 'e 2)
                          (make-arc 'e 'g 1)
                          (make-arc 'g 'e 1)
                          (make-arc 'e 'l 4)
                          (make-arc 'l 'e 4)
                          (make-arc 'f 'l 2)
                          (make-arc 'l 'f 2)
                          (make-arc 'g 'h 3)
                          (make-arc 'h 'g 3)
                          (make-arc 'g 'j 1)
                          (make-arc 'j 'g 1)
                          (make-arc 'g 'l 5)
                          (make-arc 'l 'g 5)
                          (make-arc 'h 'i 2)
                          (make-arc 'i 'h 2)
                          (make-arc 'i 'k 1)
                          (make-arc 'k 'i 1)
                          (make-arc 'j 'k 1)
                          (make-arc 'k 'j 1)
                          (make-arc 'j 'l 3)
                          (make-arc 'l 'j 3)
                          (make-arc 'j 'm 2)
                          (make-arc 'm 'j 2)
                          (make-arc 'l 'm 1)
                          (make-arc 'm 'l 1))))
    1 (graph?)
    (minspan (set 'a 'b 'c 'd 'e 'f 'g 'h 'i 'j 'k 'l 'm) 12 16))
  (test random-0
    (alternative-minimum-spanning-tree
      (make-graph (set 'rtioyin 'eaocdfh 'elaa 'mmtnaw)
                  (arcset (make-arc 'rtioyin 'eaocdfh 677)
                          (make-arc 'eaocdfh 'rtioyin 677)
                          (make-arc 'rtioyin 'elaa 332)
                          (make-arc 'elaa 'rtioyin 332)
                          (make-arc 'eaocdfh 'mmtnaw 562)
                          (make-arc 'mmtnaw 'eaocdfh 562)
                          (make-arc 'elaa 'mmtnaw 692)
                          (make-arc 'elaa 'mmtnaw 692))))
    1 (graph?)
    (match? graph=?
            (make-graph (set 'rtioyin 'eaocdfh 'elaa 'mmtnaw)
                        (arcset (make-arc 'rtioyin 'eaocdfh 677)
                                (make-arc 'eaocdfh 'rtioyin 677)
                                (make-arc 'rtioyin 'elaa 332)
                                (make-arc 'elaa 'rtioyin 332)
                                (make-arc 'eaocdfh 'mmtnaw 562)
                                (make-arc 'mmtnaw 'eaocdfh 562)))))
  (test random-1
    (alternative-minimum-spanning-tree
      (make-graph (set 'ceois 'aarsee 'iecuism 'thl 'rswi 'rbi 'orl 'loo
                       'tyotiga 'eiehpha)
                  (arcset (make-arc 'ceois 'aarsee 2)
                          (make-arc 'aarsee 'ceois 2)
                          (make-arc 'ceois 'iecuism 9)
                          (make-arc 'iecuism 'ceois 9)
                          (make-arc 'ceois 'thl 4)
                          (make-arc 'thl 'ceois 4)
                          (make-arc 'ceois 'rswi 10)
                          (make-arc 'rswi 'ceois 10)
                          (make-arc 'ceois 'rbi 3)
                          (make-arc 'rbi 'ceois 3)
                          (make-arc 'aarsee 'iecuism 16)
                          (make-arc 'iecuism 'aarsee 16)
                          (make-arc 'aarsee 'thl 15)
                          (make-arc 'thl 'aarsee 15)
                          (make-arc 'aarsee 'orl 17)
                          (make-arc 'orl 'aarsee 17)
                          (make-arc 'aarsee 'rbi 12)
                          (make-arc 'rbi 'aarsee 12)
                          (make-arc 'aarsee 'loo 2)
                          (make-arc 'loo 'aarsee 2)
                          (make-arc 'iecuism 'thl 6)
                          (make-arc 'thl 'iecuism 6)
                          (make-arc 'iecuism 'orl 9)
                          (make-arc 'orl 'iecuism 9)
                          (make-arc 'iecuism 'rswi 15)
                          (make-arc 'rswi 'iecuism 15)
                          (make-arc 'iecuism 'rbi 15)
                          (make-arc 'rbi 'iecuism 15)
                          (make-arc 'thl 'tyotiga 14)
                          (make-arc 'tyotiga 'thl 14)
                          (make-arc 'thl 'orl 5)
                          (make-arc 'orl 'thl 5)
                          (make-arc 'thl 'rswi 1)
                          (make-arc 'rswi 'thl 1)
                          (make-arc 'thl 'rbi 2)
                          (make-arc 'rbi 'thl 2)
                          (make-arc 'rswi 'rbi 1)
                          (make-arc 'rbi 'rswi 1)
                          (make-arc 'rswi 'loo 11)
                          (make-arc 'loo 'rswi 11)
                          (make-arc 'rbi 'tyotiga 4)
                          (make-arc 'tyotiga 'rbi 4)
                          (make-arc 'rbi 'eiehpha 5)
                          (make-arc 'eiehpha 'rbi 5)
                          (make-arc 'orl 'tyotiga 14)
                          (make-arc 'tyotiga 'orl 14)
                          (make-arc 'orl 'eiehpha 18)
                          (make-arc 'eiehpha 'orl 18)
                          (make-arc 'loo 'tyotiga 10)
                          (make-arc 'tyotiga 'loo 10)
                          (make-arc 'loo 'eiehpha 10)
                          (make-arc 'eiehpha 'loo 10)
                          (make-arc 'tyotiga 'eiehpha 19)
                          (make-arc 'eiehpha 'tyotiga 19))))
    1 (graph?)
    (match? graph=?
            (make-graph (set 'ceois 'aarsee 'iecuism 'thl 'rswi 'rbi 'orl
                             'loo 'tyotiga 'eiehpha)
                        (arcset (make-arc 'ceois 'aarsee 2)
                                (make-arc 'aarsee 'ceois 2)
                                (make-arc 'ceois 'rbi 3)
                                (make-arc 'rbi 'ceois 3)
                                (make-arc 'aarsee 'loo 2)
                                (make-arc 'loo 'aarsee 2)
                                (make-arc 'iecuism 'thl 6)
                                (make-arc 'thl 'iecuism 6)
                                (make-arc 'thl 'orl 5)
                                (make-arc 'orl 'thl 5)
                                (make-arc 'thl 'rswi 1)
                                (make-arc 'rswi 'thl 1)
                                (make-arc 'rswi 'rbi 1)
                                (make-arc 'rbi 'rswi 1)
                                (make-arc 'rbi 'tyotiga 4)
                                (make-arc 'tyotiga 'rbi 4)
                                (make-arc 'rbi 'eiehpha 5)
                                (make-arc 'eiehpha 'rbi 5))))))

;;; Test the procedures from the "Shortest paths" section.

(suite relaxable? ()
  (test no-tail
    (relaxable? (make-arc 'ctiehe 'sanah 788) (table))
    1 (boolean?)
    not)
  (test no-head
    (relaxable? (make-arc 'fdote 'aao 332)
                (table (cons 'fdote
                             (cons 562 (list 'fdote 'aenmt 'iytshd
                                             'heheced)))))
    1 (boolean?)
    true?)
  (test no-improvement
    (relaxable? (make-arc 'ntgt 'pcrtnda 758)
                (table (cons 'ntgt (cons 782 (list 'ntgt 'femhr 'adaeeua)))
                       (cons 'pcrtnda (cons 647 (list 'pcrtnda 'ayn 'nni 
                                                      'adaeeua)))))
    1 (boolean?)
    not)
  (test improvement
    (relaxable? (make-arc 'ited 'aguemt 289)
                (table (cons 'ited (cons 506 (list 'ited 'eecahd)))
                       (cons 'aguemt (cons 898 (list 'aguemt 'tdisdt
                                                     'eecahd)))))
    1 (boolean?)
    true?))

(suite relax
    ((path-table? (table-of symbol?
                            (pair-of number? (list-of symbol?))
                            symbol=?))
     (path-table=? (table-of= symbol=? (pair-of= = (list-of= symbol=?)))))
  (test no-tail
    (relax (make-arc 'padbr 'vral 677) (table))
    1 (path-table?)
    (match? path-table=? (table)))
  (test no-head
    (relax (make-arc 'mne 'euuaeht 332)
           (table (cons 'mne (cons 562 (list 'mne 'otfnnes 'aii)))))
    1 (path-table?)
    (match? path-table=?
            (table (cons 'mne (cons 562 (list 'mne 'otfnnes 'aii)))
                   (cons 'euuaeht
                         (cons 894 (list 'euuaeht 'mne 'otfnnes 'aii))))))
  (test no-improvement
    (relax (make-arc 'idhud 'eosbfe 399)
           (table (cons 'idhud (cons 999 (list 'idhud 'khcnyon)))
                  (cons 'eosbfe (cons 685 (list 'eosbfe 'oaeotn)))))
    1 (path-table?)
    (match? path-table=?
            (table (cons 'idhud (cons 999 (list 'idhud 'khcnyon)))
                   (cons 'eosbfe (cons 685 (list 'eosbfe 'oaeotn))))))
  (test improvement
    (relax (make-arc 'arrkswe 'oweebmr 712)
           (table (cons 'arrkswe
                        (cons 215 (list 'arrkswe 'arho 'nheol 'hso)))
                  (cons 'oweebmr (cons 997 (list 'oweebmr 'itfeu 'mde)))))
    1 (path-table?)
    (match? path-table=?
            (table (cons 'arrkswe
                         (cons 215 (list 'arrkswe 'arho 'nheol 'hso)))
                   (cons 'oweebmr
                         (cons 927 (list 'oweebmr 'arrkswe 'arho 'nheol
                                         'hso)))))))

(suite Bellman-Ford
    ((path-table? (table-of symbol?
                            (pair-of number? (list-of symbol?))
                            symbol=?))
     (path-table=? (table-of= symbol=? (pair-of= = (list-of= symbol=?)))))
  (test one-vertex-no-arcs
    (Bellman-Ford (make-graph (set 'afcrth) (arcset)) 'afcrth)
    1 (path-table?)
    (match? path-table=? (table (cons 'afcrth (cons 0 (list 'afcrth))))))
  (test two-vertices-connected
    (Bellman-Ford (make-graph (set 'toid 'fersa)
                              (arcset (make-arc 'toid 'fersa 148)))
                  'toid)
    1 (path-table?)
    (match? path-table=?
            (table (cons 'fersa (cons 148 (list 'fersa 'toid)))
                   (cons 'toid (cons 0 (list 'toid))))))
  (test unreachable-vertex
    (Bellman-Ford (make-graph (set 'acnioso 'iiet 'amhpom)
                              (arcset (make-arc 'iiet 'acnioso 217)))
                  'iiet)
    1 (path-table?)
    (match? path-table=?
            (table (cons 'acnioso (cons 217 (list 'acnioso 'iiet)))
                   (cons 'iiet (cons 0 (list 'iiet))))))
  (test alternate-paths
    (Bellman-Ford (make-graph (set 'btetbdf 'rttf 'mwk 'tnoh)
                              (arcset (make-arc 'btetbdf 'rttf 445)
                                      (make-arc 'btetbdf 'mwk 192)
                                      (make-arc 'rttf 'tnoh 541)
                                      (make-arc 'mwk 'tnoh 759)))
                  'btetbdf)
    1 (path-table?)
    (match? path-table=?
            (table (cons 'tnoh (cons 951 (list 'tnoh 'mwk 'btetbdf)))
                   (cons 'mwk (cons 192 (list 'mwk 'btetbdf)))
                   (cons 'rttf (cons 445 (list 'rttf 'btetbdf)))
                   (cons 'btetbdf (cons 0 (list 'btetbdf))))))
  (test negative-weights
    (Bellman-Ford (make-graph (set 'ctptu 'shnntta 'ortwetd 'qled 'udaache
                                   'rtmol)
                              (arcset (make-arc 'ctptu 'shnntta 759)
                                      (make-arc 'ctptu 'ortwetd -541)
                                      (make-arc 'shnntta 'qled 314)
                                      (make-arc 'shnntta 'udaache -876)
                                      (make-arc 'ortwetd 'qled 795)
                                      (make-arc 'qled 'rtmol -161)
                                      (make-arc 'udaache 'rtmol 6)))
                  'ctptu)
    1 (path-table?)
    (match? path-table=?
            (table (cons 'ctptu (cons 0 (list 'ctptu)))
                   (cons 'shnntta (cons 759 (list 'shnntta 'ctptu)))
                   (cons 'ortwetd (cons -541 (list 'ortwetd 'ctptu)))
                   (cons 'qled (cons 254 (list 'qled 'ortwetd 'ctptu)))
                   (cons 'udaache
                         (cons -117 (list 'udaache 'shnntta 'ctptu)))
                   (cons 'rtmol
                         (cons -111 (list 'rtmol 'udaache 'shnntta
                                          'ctptu))))))
  (test with-cycle
    (Bellman-Ford (make-graph (set 'aniase 'csrigdh 'wicge 'rnvamh 'ofd)
                              (arcset (make-arc 'aniase 'csrigdh 558)
                                      (make-arc 'csrigdh 'wicge 611)
                                      (make-arc 'csrigdh 'ofd 638)
                                      (make-arc 'wicge 'rnvamh 209)
                                      (make-arc 'rnvamh 'csrigdh 494)))
                  'aniase)
    1 (path-table?)
    (match? path-table=?
            (table (cons 'aniase (cons 0 (list 'aniase)))
                   (cons 'csrigdh (cons 558 (list 'csrigdh 'aniase)))
                   (cons 'wicge (cons 1169 (list 'wicge 'csrigdh 'aniase)))
                   (cons 'rnvamh (cons 1378 (list 'rnvamh 'wicge 'csrigdh
                                                  'aniase)))
                   (cons 'ofd (cons 1196 (list 'ofd 'csrigdh 'aniase))))))
  (test cormen-24.2
    (Bellman-Ford (make-graph (set 's 't 'x 'y 'z)
                              (arcset (make-arc 's 't 3)
                                      (make-arc 's 'y 5)
                                      (make-arc 't 'x 6)
                                      (make-arc 't 'y 2)
                                      (make-arc 'x 'z 2)
                                      (make-arc 'y 't 1)
                                      (make-arc 'y 'x 4)
                                      (make-arc 'y 'z 6)
                                      (make-arc 'z 's 3)
                                      (make-arc 'z 'x 7)))
                  's)
    1 (path-table?)
    (lambda (alpha)
      (exists? (sect path-table=? alpha <>)
              (list (table (cons 's (cons 0 (list 's)))
                           (cons 't (cons 3 (list 't 's)))
                           (cons 'x (cons 9 (list 'x 't 's)))
                           (cons 'y (cons 5 (list 'y 's)))
                           (cons 'z (cons 11 (list 'z 'y 's))))
                    (table (cons 's (cons 0 (list 's)))
                           (cons 't (cons 3 (list 't 's)))
                           (cons 'x (cons 9 (list 'x 'y 's)))
                           (cons 'y (cons 5 (list 'y 's)))
                           (cons 'z (cons 11 (list 'z 'y 's))))
                    (table (cons 's (cons 0 (list 's)))
                           (cons 't (cons 3 (list 't 's)))
                           (cons 'x (cons 9 (list 'x 'y 't 's)))
                           (cons 'y (cons 5 (list 'y 's)))
                           (cons 'z (cons 11 (list 'z 'y 's))))
                    (table (cons 's (cons 0 (list 's)))
                           (cons 't (cons 3 (list 't 's)))
                           (cons 'x (cons 9 (list 'x 't 's)))
                           (cons 'y (cons 5 (list 'y 't 's)))
                           (cons 'z (cons 11 (list 'z 'y 't 's))))
                    (table (cons 's (cons 0 (list 's)))
                           (cons 't (cons 3 (list 't 's)))
                           (cons 'x (cons 9 (list 'x 'y 's)))
                           (cons 'y (cons 5 (list 'y 't 's)))
                           (cons 'z (cons 11 (list 'z 'y 't 's))))
                    (table (cons 's (cons 0 (list 's)))
                           (cons 't (cons 3 (list 't 's)))
                           (cons 'x (cons 9 (list 'x 'y 't 's)))
                           (cons 'y (cons 5 (list 'y 't 's)))
                           (cons 'z (cons 11 (list 'z 'y 't 's))))
                    (table (cons 's (cons 0 (list 's)))
                           (cons 't (cons 3 (list 't 's)))
                           (cons 'x (cons 9 (list 'x 'y 't 's)))
                           (cons 'y (cons 5 (list 'y 't 's)))
                           (cons 'z (cons 11 (list 'z 'x 'y 't 's))))
                    (table (cons 's (cons 0 (list 's)))
                           (cons 't (cons 3 (list 't 's)))
                           (cons 'x (cons 9 (list 'x 't 's)))
                           (cons 'y (cons 5 (list 'y 's)))
                           (cons 'z (cons 11 (list 'z 'x 't 's))))
                    (table (cons 's (cons 0 (list 's)))
                           (cons 't (cons 3 (list 't 's)))
                           (cons 'x (cons 9 (list 'x 't 's)))
                           (cons 'y (cons 5 (list 'y 't 's)))
                           (cons 'z (cons 11 (list 'z 'x 't 's))))
                    (table (cons 's (cons 0 (list 's)))
                           (cons 't (cons 3 (list 't 's)))
                           (cons 'x (cons 9 (list 'x 'y 's)))
                           (cons 'y (cons 5 (list 'y 's)))
                           (cons 'z (cons 11 (list 'z 'x 'y 's))))))))
  (test cormen-24.4
    (Bellman-Ford (make-graph (set 's 't 'x 'y 'z)
                              (arcset (make-arc 's 't 6)
                                      (make-arc 's 'y 7)
                                      (make-arc 't 'x 5)
                                      (make-arc 't 'y 8)
                                      (make-arc 't 'z -4)
                                      (make-arc 'x 't -2)
                                      (make-arc 'y 'x -3)
                                      (make-arc 'y 'z 9)
                                      (make-arc 'z 's 2)
                                      (make-arc 'z 'x 7)))
                  's)
    1 (path-table?)
    (match? path-table=?
            (table (cons 's (cons 0 (list 's)))
                   (cons 't (cons 2 (list 't 'x 'y 's)))
                   (cons 'x (cons 4 (list 'x 'y 's)))
                   (cons 'y (cons 7 (list 'y 's)))
                   (cons 'z (cons -2 (list 'z 't 'x 'y 's))))))
  (test cormen-24.6
    (Bellman-Ford (make-graph (set 's 't 'x 'y 'z)
                              (arcset (make-arc 's 't 10)
                                      (make-arc 's 'y 5)
                                      (make-arc 't 'x 1)
                                      (make-arc 't 'y 2)
                                      (make-arc 'x 'z 4)
                                      (make-arc 'y 't 3)
                                      (make-arc 'y 'x 9)
                                      (make-arc 'y 'z 2)
                                      (make-arc 'z 's 7)
                                      (make-arc 'z 'x 6)))
                  's)
    1 (path-table?)
    (match? path-table=?
            (table (cons 's (cons 0 (list 's)))
                   (cons 't (cons 8 (list 't 'y 's)))
                   (cons 'x (cons 9 (list 'x 't 'y 's)))
                   (cons 'y (cons 5 (list 'y 's)))
                   (cons 'z (cons 7 (list 'z 'y 's)))))))

(suite safe-Bellman-Ford
    ((right-type? (^vel (table-of symbol?
                                  (pair-of number? (list-of symbol?))
                                  symbol=?)
                        not))
     (path-table=? (table-of= symbol=? (pair-of= = (list-of= symbol=?)))))
  (test single-vertex
    (safe-Bellman-Ford (make-graph (set 'ynaau) (arcset)) 'ynaau)
    1 (right-type?)
    (^et table?
         (match? path-table=?
                 (table (cons 'ynaau (cons 0 (list 'ynaau)))))))
  (test minimal-irreflexive
    (safe-Bellman-Ford
      (make-graph (set 'aldec 'amuf)
                  (arcset (make-arc 'aldec 'amuf -737)
                          (make-arc 'amuf 'aldec 156)))
      'aldec)
    1 (right-type?)
    not)
  (test small-instance
    (safe-Bellman-Ford
      (make-graph (set 'oweeor 'ttlsth 'ago 'swghe 'emlooa)
                  (arcset (make-arc 'oweeor 'ttlsth 558)
                          (make-arc 'ttlsth 'ago -343)
                          (make-arc 'ago 'swghe -678)
                          (make-arc 'swghe 'ttlsth 817)
                          (make-arc 'ttlsth 'emlooa 743)))
      'oweeor)
    1 (right-type?)
    not)
  (test zero-cycle
    (safe-Bellman-Ford
      (make-graph (set 'oued 'tndantd 'sanoine 'iefhmpm 'snihha)
                  (arcset (make-arc 'oued 'tndantd 13)
                          (make-arc 'tndantd 'sanoine -796)
                          (make-arc 'sanoine 'iefhmpm 169)
                          (make-arc 'iefhmpm 'tndantd 627)
                          (make-arc 'tndantd 'snihha 871)))
      'oued)
    1 (right-type?)
    (^et table?
         (match? path-table=?
                 (table (cons 'oued (cons 0 (list 'oued)))
                        (cons 'tndantd (cons 13 (list 'tndantd 'oued)))
                        (cons 'sanoine
                              (cons -783 (list 'sanoine 'tndantd 'oued)))
                        (cons 'iefhmpm
                              (cons -614 (list 'iefhmpm 'sanoine 'tndantd
                                               'oued)))
                        (cons 'snihha
                              (cons 884 (list 'snihha 'tndantd 'oued)))))))
  (test inaccessible-negative-cycle
    (safe-Bellman-Ford
      (make-graph (set 'oiacc 'lepis 'snucdi 'cfi 'ahge)
                  (arcset (make-arc 'oiacc 'lepis 453)
                          (make-arc 'snucdi 'cfi -592)
                          (make-arc 'cfi 'ahge 79)
                          (make-arc 'ahge 'snucdi 251)
                          (make-arc 'cfi 'lepis 92)))
      'oiacc)
    1 (right-type?)
    (^et table?
         (match? path-table=?
                 (table (cons 'oiacc (cons 0 (list 'oiacc)))
                        (cons 'lepis (cons 453 (list 'lepis 'oiacc))))))))

(suite nearest ()
  (test minimal
    (nearest (set 'eoe) (table (cons 'eoe (cons 0 (list 'eoe)))))
    1 (symbol?)
    (match? symbol=? 'eoe))
  (test singleton-set
    (nearest (set 'oulntm)
             (table (cons 'sitrh
                          (cons 816 (list 'sitrh 'yhieetf 'daup 'hctt)))
                    (cons 'srahef
                          (cons 97 (list 'srahef 'oeh 'hyfnon 'hctt)))
                    (cons 'oulntm (cons 597 (list 'oulntm 'agrm 'hctt)))))
    1 (symbol?)
    (match? symbol=? 'oulntm))
  (test typical
    (nearest (set 'nnipicr 'wkergdl 'auidnei 'peh)
             (table (cons 'nnipicr
                          (cons 942 (list 'nnipicr 'eoaw 'acnwo 'epmirge)))
                    (cons 'litr (cons 394 (list 'litr 'cae 'epmirge)))
                    (cons 'wkergdl
                          (cons 274 (list 'wkergdl 'aoeepn 'ted 'aapepf
                                          'epmirge)))
                    (cons 'atoleh
                          (cons 46 (list 'atoleh 'dtm 'eiatay 'epmirge)))
                    (cons 'auidnei
                          (cons 777 (list 'auidnei 'sdfwhsm 'epmirge)))
                    (cons 'peh
                          (cons 679 (list 'peh 'aiei 'hurtto 'hbh
                                          'epmirge)))))
    1 (symbol?)
    (match? symbol=? 'wkergdl)))

(suite Dijkstra
    ((path-table? (table-of symbol?
                            (pair-of number? (list-of symbol?))
                            symbol=?))
     (path-table=? (table-of= symbol=? (pair-of= = (list-of= symbol=?)))))
  (test one-vertex-no-arcs
    (Dijkstra (make-graph (set 'nnnhfeb) (arcset)) 'nnnhfeb)
    1 (path-table?)
    (match? path-table=? (table (cons 'nnnhfeb (cons 0 (list 'nnnhfeb))))))
  (test two-vertices-connected
    (Dijkstra (make-graph (set 'loinu 'leiem)
                          (arcset (make-arc 'loinu 'leiem 743)))
              'loinu)
    1 (path-table?)
    (match? path-table=?
            (table (cons 'leiem (cons 743 (list 'leiem 'loinu)))
                   (cons 'loinu (cons 0 (list 'loinu))))))
  (test unreachable-vertex
    (Dijkstra (make-graph (set 'ihetadi 'mtse 'eeanvee)
                          (arcset (make-arc 'mtse 'ihetadi 973)))
              'mtse)
    1 (path-table?)
    (match? path-table=?
            (table (cons 'ihetadi (cons 973 (list 'ihetadi 'mtse))) 
                   (cons 'mtse (cons 0 (list 'mtse))))))
  (test alternate-paths
    (Dijkstra (make-graph (set 'ayaaoer 'rosm 'elrds 'rolphw)
                          (arcset (make-arc 'ayaaoer 'rosm 735)
                                  (make-arc 'ayaaoer 'elrds 437)
                                  (make-arc 'rosm 'rolphw 574)
                                  (make-arc 'elrds 'rolphw 544)))
              'ayaaoer)
    1 (path-table?)
    (match? path-table=?
            (table (cons 'rolphw (cons 981 (list 'rolphw 'elrds 'ayaaoer)))
                   (cons 'elrds (cons 437 (list 'elrds 'ayaaoer)))
                   (cons 'rosm (cons 735 (list 'rosm 'ayaaoer)))
                   (cons 'ayaaoer (cons 0 (list 'ayaaoer))))))
  (test with-cycle
    (Dijkstra (make-graph (set 'oyih 'rewiy 'hrnea 'eota 'eutat)
                          (arcset (make-arc 'oyih 'rewiy 192)
                                  (make-arc 'rewiy 'hrnea 781)
                                  (make-arc 'rewiy 'eutat 316)
                                  (make-arc 'hrnea 'eota 817)
                                  (make-arc 'eota 'rewiy 673)))
              'oyih)
    1 (path-table?)
    (match? path-table=?
            (table (cons 'oyih (cons 0 (list 'oyih)))
                   (cons 'rewiy (cons 192 (list 'rewiy 'oyih)))
                   (cons 'hrnea (cons 973 (list 'hrnea 'rewiy 'oyih)))
                   (cons 'eota
                         (cons 1790 (list 'eota 'hrnea 'rewiy 'oyih)))
                   (cons 'eutat
                         (cons 508 (list 'eutat 'rewiy 'oyih))))))
  (test cormen-24.2
    (Dijkstra (make-graph (set 's 't 'x 'y 'z)
                          (arcset (make-arc 's 't 3)
                                  (make-arc 's 'y 5)
                                  (make-arc 't 'x 6)
                                  (make-arc 't 'y 2)
                                  (make-arc 'x 'z 2)
                                  (make-arc 'y 't 1)
                                  (make-arc 'y 'x 4)
                                  (make-arc 'y 'z 6)
                                  (make-arc 'z 's 3)
                                  (make-arc 'z 'x 7)))
                  's)
    1 (path-table?)
    (lambda (alpha)
      (exists? (sect path-table=? alpha <>)
              (list (table (cons 's (cons 0 (list 's)))
                           (cons 't (cons 3 (list 't 's)))
                           (cons 'x (cons 9 (list 'x 't 's)))
                           (cons 'y (cons 5 (list 'y 's)))
                           (cons 'z (cons 11 (list 'z 'y 's))))
                    (table (cons 's (cons 0 (list 's)))
                           (cons 't (cons 3 (list 't 's)))
                           (cons 'x (cons 9 (list 'x 'y 's)))
                           (cons 'y (cons 5 (list 'y 's)))
                           (cons 'z (cons 11 (list 'z 'y 's))))
                    (table (cons 's (cons 0 (list 's)))
                           (cons 't (cons 3 (list 't 's)))
                           (cons 'x (cons 9 (list 'x 'y 't 's)))
                           (cons 'y (cons 5 (list 'y 's)))
                           (cons 'z (cons 11 (list 'z 'y 's))))
                    (table (cons 's (cons 0 (list 's)))
                           (cons 't (cons 3 (list 't 's)))
                           (cons 'x (cons 9 (list 'x 't 's)))
                           (cons 'y (cons 5 (list 'y 't 's)))
                           (cons 'z (cons 11 (list 'z 'y 't 's))))
                    (table (cons 's (cons 0 (list 's)))
                           (cons 't (cons 3 (list 't 's)))
                           (cons 'x (cons 9 (list 'x 'y 's)))
                           (cons 'y (cons 5 (list 'y 't 's)))
                           (cons 'z (cons 11 (list 'z 'y 't 's))))
                    (table (cons 's (cons 0 (list 's)))
                           (cons 't (cons 3 (list 't 's)))
                           (cons 'x (cons 9 (list 'x 'y 't 's)))
                           (cons 'y (cons 5 (list 'y 't 's)))
                           (cons 'z (cons 11 (list 'z 'y 't 's))))
                    (table (cons 's (cons 0 (list 's)))
                           (cons 't (cons 3 (list 't 's)))
                           (cons 'x (cons 9 (list 'x 'y 't 's)))
                           (cons 'y (cons 5 (list 'y 't 's)))
                           (cons 'z (cons 11 (list 'z 'x 'y 't 's))))
                    (table (cons 's (cons 0 (list 's)))
                           (cons 't (cons 3 (list 't 's)))
                           (cons 'x (cons 9 (list 'x 't 's)))
                           (cons 'y (cons 5 (list 'y 's)))
                           (cons 'z (cons 11 (list 'z 'x 't 's))))
                    (table (cons 's (cons 0 (list 's)))
                           (cons 't (cons 3 (list 't 's)))
                           (cons 'x (cons 9 (list 'x 't 's)))
                           (cons 'y (cons 5 (list 'y 't 's)))
                           (cons 'z (cons 11 (list 'z 'x 't 's))))
                    (table (cons 's (cons 0 (list 's)))
                           (cons 't (cons 3 (list 't 's)))
                           (cons 'x (cons 9 (list 'x 'y 's)))
                           (cons 'y (cons 5 (list 'y 's)))
                           (cons 'z (cons 11 (list 'z 'x 'y 's))))))))

  (test cormen-24.6
    (Dijkstra (make-graph (set 's 't 'x 'y 'z)
                          (arcset (make-arc 's 't 10)
                                  (make-arc 's 'y 5)
                                  (make-arc 't 'x 1)
                                  (make-arc 't 'y 2)
                                  (make-arc 'x 'z 4)
                                  (make-arc 'y 't 3)
                                  (make-arc 'y 'x 9)
                                  (make-arc 'y 'z 2)
                                  (make-arc 'z 's 7)
                                  (make-arc 'z 'x 6)))
              's)
    1 (path-table?)
    (match? path-table=?
            (table (cons 's (cons 0 (list 's)))
                   (cons 't (cons 8 (list 't 'y 's)))
                   (cons 'x (cons 9 (list 'x 't 'y 's)))
                   (cons 'y (cons 5 (list 'y 's)))
                   (cons 'z (cons 7 (list 'z 'y 's))))))
  (test manber-7.18
    (Dijkstra (make-graph (set 'v 'a 'b 'c 'd 'e 'f 'g 'h)
                          (arcset (make-arc 'v 'a 1)
                                  (make-arc 'v 'b 5)
                                  (make-arc 'v 'd 9)
                                  (make-arc 'a 'c 2)
                                  (make-arc 'b 'e 3)
                                  (make-arc 'c 'd 4)
                                  (make-arc 'c 'f 9)
                                  (make-arc 'd 'e 2)
                                  (make-arc 'd 'g 4)
                                  (make-arc 'e 'h 1)
                                  (make-arc 'g 'f 2)
                                  (make-arc 'h 'g 3)))
              'v)
    1 (path-table?)
    (match? path-table=?
            (table (cons 'v (cons 0 (list 'v)))
                   (cons 'a (cons 1 (list 'a 'v)))
                   (cons 'b (cons 5 (list 'b 'v)))
                   (cons 'c (cons 3 (list 'c 'a 'v)))
                   (cons 'd (cons 7 (list 'd 'c 'a 'v)))
                   (cons 'e (cons 8 (list 'e 'b 'v)))
                   (cons 'f (cons 12 (list 'f 'c 'a 'v)))
                   (cons 'g (cons 11 (list 'g 'd 'c 'a 'v)))
                   (cons 'h (cons 9 (list 'h 'e 'b 'v)))))))

(suite all-pairs-Dijkstra
    ((two-level-path-table?
      (table-of symbol?
                (table-of symbol?
                          (pair-of number? (list-of symbol?))
                          symbol=?)
                symbol=?))
     (two-level-path-table=?
      (table-of= symbol=?
                 (table-of= symbol=? (pair-of= = (list-of= symbol=?))))))
  (test empty-graph
    (all-pairs-Dijkstra (empty-graph))
    1 (two-level-path-table?)
    (match? two-level-path-table=? (table)))
  (test one-vertex-no-arcs
    (all-pairs-Dijkstra (make-graph (set 'ciiertd) (arcset)))
    1 (two-level-path-table?)
    (match? two-level-path-table=?
            (table (cons 'ciiertd
                         (table (cons 'ciiertd
                                      (cons 0 (list 'ciiertd))))))))
  (test one-vertex-with-loop
    (all-pairs-Dijkstra (make-graph (set 'wawh)
                                    (arcset (make-arc 'wawh 'wawh 715))))
    1 (two-level-path-table?)
    (match? two-level-path-table=?
            (table (cons 'wawh
                         (table (cons 'wawh (cons 0 (list 'wawh))))))))
  (test small-graph
    (all-pairs-Dijkstra
      (make-graph (set 'tua 'dgoopoi 'ile 'mixseir)
                  (arcset (make-arc 'tua 'dgoopoi 935)
                          (make-arc 'tua 'ile 492)
                          (make-arc 'dgoopoi 'tua 234)
                          (make-arc 'dgoopoi 'mixseir 233)
                          (make-arc 'ile 'tua 801)
                          (make-arc 'ile 'mixseir 817)
                          (make-arc 'mixseir 'dgoopoi 852)
                          (make-arc 'mixseir 'ile 557))))
    1 (two-level-path-table?)
    (match? two-level-path-table=?
            (table (cons 'tua
                         (table (cons 'tua (cons 0 (list 'tua)))
                                (cons 'dgoopoi
                                      (cons 935 (list 'dgoopoi 'tua)))
                                (cons 'ile (cons 492 (list 'ile 'tua)))
                                (cons 'mixseir
                                      (cons 1168 (list 'mixseir 'dgoopoi
                                                       'tua)))))
                   (cons 'dgoopoi
                         (table (cons 'tua (cons 234 (list 'tua 'dgoopoi)))
                                (cons 'dgoopoi (cons 0 (list 'dgoopoi)))
                                (cons 'ile
                                      (cons 726 (list 'ile 'tua 'dgoopoi)))
                                (cons 'mixseir
                                      (cons 233 (list 'mixseir
                                                      'dgoopoi)))))
                   (cons 'ile
                         (table (cons 'tua (cons 801 (list 'tua 'ile)))
                                (cons 'dgoopoi
                                      (cons 1669 (list 'dgoopoi 'mixseir
                                                       'ile)))
                                (cons 'ile (cons 0 (list 'ile)))
                                (cons 'mixseir
                                      (cons 817 (list 'mixseir 'ile)))))
                   (cons 'mixseir
                         (table (cons 'tua
                                      (cons 1086 (list 'tua 'dgoopoi
                                                       'mixseir)))
                                (cons 'dgoopoi
                                      (cons 852 (list 'dgoopoi 'mixseir)))
                                (cons 'ile (cons 557 (list 'ile 'mixseir)))
                                (cons 'mixseir
                                      (cons 0 (list 'mixseir))))))))
  (test smith-2.33
    (all-pairs-Dijkstra (make-graph (set 1 2 3 4)
                                    (arcset (make-arc 1 2 7)
                                            (make-arc 1 3 2)
                                            (make-arc 2 4 5)
                                            (make-arc 3 2 3)
                                            (make-arc 3 4 9)
                                            (make-arc 4 1 1)
                                            (make-arc 4 3 6))))
    1 ((table-of natural-number?
                 (table-of natural-number?
                           (pair-of number? (list-of natural-number?))
                           =)
                  =))
    (match? (table-of= = (table-of= = (pair-of= = (list-of= =))))
            (table (cons 1 (table (cons 1 (cons 0 (list 1)))
                                  (cons 2 (cons 5 (list 2 3 1)))
                                  (cons 3 (cons 2 (list 3 1)))
                                  (cons 4 (cons 10 (list 4 2 3 1)))))
                   (cons 2 (table (cons 1 (cons 6 (list 1 4 2)))
                                  (cons 2 (cons 0 (list 2)))
                                  (cons 3 (cons 8 (list 3 1 4 2)))
                                  (cons 4 (cons 5 (list 4 2)))))
                   (cons 3 (table (cons 1 (cons 9 (list 1 4 2 3)))
                                  (cons 2 (cons 3 (list 2 3)))
                                  (cons 3 (cons 0 (list 3)))
                                  (cons 4 (cons 8 (list 4 2 3)))))
                   (cons 4 (table (cons 1 (cons 1 (list 1 4)))
                                  (cons 2 (cons 6 (list 2 3 1 4)))
                                  (cons 3 (cons 3 (list 3 1 4)))
                                  (cons 4 (cons 0 (list 4)))))))))

(suite trivial-table
    ((two-level-path-table?
      (table-of symbol?
                (table-of symbol?
                          (pair-of number? (list-of symbol?))
                          symbol=?)
                symbol=?))
     (two-level-path-table=?
      (table-of= symbol=?
                 (table-of= symbol=? (pair-of= = (list-of= symbol=?))))))
  (test empty-graph
    (trivial-table (empty-graph))
    1 (two-level-path-table?)
    (match? two-level-path-table=? (table)))
  (test single-vertex-no-arc
    (trivial-table (make-graph (set 'lade) (arcset)))
    1 (two-level-path-table?)
    (match? two-level-path-table=?
            (table (cons 'lade
                         (table (cons 'lade (cons 0 (list 'lade))))))))
  (test single-vertex-with-arc
    (trivial-table (make-graph (set 'nischa)
                               (arcset (make-arc 'nischa 'nischa 813))))
    1 (two-level-path-table?)
    (match? two-level-path-table=?
            (table (cons 'nischa
                         (table (cons 'nischa (cons 0 (list 'nischa))))))))
  (test random-graph
    (trivial-table
      (make-graph (set 'tthho 'rern 'itaahtn 'ftagsca 'eesc 'ansm)
                  (arcset (make-arc 'tthho 'rern 454)
                          (make-arc 'tthho 'ftagsca 129)
                          (make-arc 'tthho 'ansm 815)
                          (make-arc 'rern 'ftagsca 947)
                          (make-arc 'itaahtn 'tthho 323)
                          (make-arc 'itaahtn 'rern 86)
                          (make-arc 'itaahtn 'ftagsca 362)
                          (make-arc 'itaahtn 'ansm 451)
                          (make-arc 'ftagsca 'rern 425)
                          (make-arc 'ftagsca 'ftagsca 243)
                          (make-arc 'ftagsca 'ansm 26)
                          (make-arc 'eesc 'rern 238)
                          (make-arc 'eesc 'itaahtn 896)
                          (make-arc 'eesc 'ftagsca 758)
                          (make-arc 'ansm 'rern 438)
                          (make-arc 'ansm 'ftagsca 909)
                          (make-arc 'ansm 'eesc 118))))
    1 (two-level-path-table?)
    (match? two-level-path-table=?
            (table (cons 'tthho
                         (table (cons 'tthho (cons 0 (list 'tthho)))
                                (cons 'rern (cons 454 (list 'rern 'tthho)))
                                (cons 'ftagsca
                                      (cons 129 (list 'ftagsca 'tthho)))
                                (cons 'ansm
                                      (cons 815 (list 'ansm 'tthho)))))
                   (cons 'rern
                         (table (cons 'rern (cons 0 (list 'rern)))
                                (cons 'ftagsca
                                      (cons 947 (list 'ftagsca 'rern)))))
                   (cons 'itaahtn
                         (table (cons 'tthho
                                      (cons 323 (list 'tthho 'itaahtn)))
                                (cons 'rern
                                      (cons 86 (list 'rern 'itaahtn)))
                                (cons 'itaahtn (cons 0 (list 'itaahtn)))
                                (cons 'ftagsca
                                      (cons 362 (list 'ftagsca 'itaahtn)))
                                (cons 'ansm
                                      (cons 451 (list 'ansm 'itaahtn)))))
                   (cons 'ftagsca
                         (table (cons 'rern
                                      (cons 425 (list 'rern 'ftagsca)))
                                (cons 'ftagsca (cons 0 (list 'ftagsca)))
                                (cons 'ansm
                                      (cons 26 (list 'ansm 'ftagsca)))))
                   (cons 'eesc
                         (table (cons 'rern
                                      (cons 238 (list 'rern 'eesc)))
                                (cons 'itaahtn
                                      (cons 896 (list 'itaahtn 'eesc)))
                                (cons 'ftagsca
                                      (cons 758 (list 'ftagsca 'eesc)))
                                (cons 'eesc (cons 0 (list 'eesc)))))
                   (cons 'ansm
                         (table (cons 'rern
                                      (cons 438 (list 'rern 'ansm)))
                                (cons 'ftagsca
                                      (cons 909 (list 'ftagsca 'ansm)))
                                (cons 'eesc
                                      (cons 118 (list 'eesc 'ansm)))
                                (cons 'ansm (cons 0 (list 'ansm)))))))))

(suite splice-paths ((path-pair? (pair-of number? (list-of symbol?)))
                     (path-pair=? (pair-of= = (list-of= symbol=?))))
  (test minimal-case
    (splice-paths (cons 0 (list 'ifew)) (cons 0 (list 'ifew)))
    1 (path-pair?)
    (match? path-pair=? (cons 0 (list 'ifew))))
  (test typical-case
    (splice-paths (cons 505 (list 'odvsnho 'annr 'tlliem 'rings 'pue 'oaoe
                                  'tulas))
                  (cons 748 (list 'tuhstt 'hdahrah 'etawrra 'ete 'amew
                                  'odvsnho)))
    1 (path-pair?)
    (match? path-pair=?
            (cons 1253 (list 'tuhstt 'hdahrah 'etawrra 'ete 'amew 'odvsnho
                             'annr 'tlliem 'rings 'pue 'oaoe 'tulas)))))

(suite test-connection
    ((path-table? (table-of symbol?
                            (pair-of number? (list-of symbol?))
                            symbol=?))
     (path-table=? (table-of= symbol=? (pair-of= = (list-of= symbol=?)))))
  (test direct-is-shorter
    (test-connection (cons 482 (list 'phidne 'uqplo 'titstn))
                     (cons 281 (list 'arefo 'tred 'tare 'phidne))
                     (cons 104 (list 'arefo 'iei 'emctii 'titstn))
                     'arefo
                     (table (cons 'tred (cons 403 (list 'tred 'selr 'aaergh
                                                        'titstn)))
                            (cons 'titstn (cons 0 (list 'titstn)))
                            (cons 'ootrtco (cons 707 (list 'ootrtco
                                                           'titstn)))))
    1 (path-table?)
    (match? path-table=?
            (table (cons 'tred (cons 403 (list 'tred 'selr 'aaergh
                                               'titstn)))
                   (cons 'titstn (cons 0 (list 'titstn)))
                   (cons 'ootrtco (cons 707 (list 'ootrtco 'titstn)))
                   (cons 'arefo (cons 104 (list 'arefo 'iei 'emctii
                                                'titstn))))))
  (test spliced-is-shorter
    (test-connection (cons 241 (list 'dilltmi 'nloiuan 'nsto 'rhe))
                     (cons 80 (list 'aohisod 'hwa 'dilltmi))
                     (cons 642 (list 'aohisod 'tstsdop 'dhp 'hmwidrh 'rhe))
                     'aohisod
                     (table (cons 'hwa (cons 805 (list 'hwa 'dilltmi 'rhe)))
                            (cons 'rhe (cons 0 (list 'rhe)))
                            (cons 'nsto (cons 813 (list 'nsto 'rhe)))))
    1 (path-table?)
    (match? path-table=?
            (table (cons 'hwa (cons 805 (list 'hwa 'dilltmi 'rhe)))
                   (cons 'rhe (cons 0 (list 'rhe)))
                   (cons 'nsto (cons 813 (list 'nsto 'rhe)))
                   (cons 'aohisod (cons 321 (list 'aohisod 'hwa 'dilltmi
                                                  'nloiuan 'nsto 'rhe))))))
  (test no-direct-path
    (test-connection (cons 523 (list 'tlnoel 'tlfn 'moer))
                     (cons 755 (list 'saoi 'cnmn 'mne 'ltdnnrg 'tlnoel))
                     #f
                     'saoi
                     (table (cons 'moer (cons 0 (list 'moer)))
                            (cons 'cnmn (cons 336 (list 'cnmn 'mne 'moer)))
                            (cons 'tlfn (cons 433 (list 'tlfn 'moer)))))
    1 (path-table?)
    (match? path-table=?
            (table (cons 'moer (cons 0 (list 'moer)))
                   (cons 'cnmn (cons 336 (list 'cnmn 'mne 'moer)))
                   (cons 'tlfn (cons 433 (list 'tlfn 'moer)))
                   (cons 'saoi (cons 1278 (list 'saoi 'cnmn 'mne 'ltdnnrg
                                                'tlnoel 'tlfn 'moer))))))
  (test no-fore-path
    (test-connection #f
                     (cons 717 (list 'hwuago 'taesu 'tyaayfu 'tnhgwf))
                     (cons 926 (list 'hwuago 'audhlst 'ondtycr 'nxaa))
                     'hwuago
                     (table (cons 'taesu (cons 717 (list 'taesu 'tyaayfu
                                                         'tnhgwf 'nxaa)))
                            (cons 'ondtycr (cons 439 (list 'ondtycr
                                                           'nxaa)))
                            (cons 'nxaa (cons 0 (list 'nxaa)))))
    1 (path-table?)
    (match? path-table=?
            (table (cons 'taesu (cons 717 (list 'taesu 'tyaayfu 'tnhgwf
                                                'nxaa)))
                   (cons 'ondtycr (cons 439 (list 'ondtycr 'nxaa)))
                   (cons 'nxaa (cons 0 (list 'nxaa)))
                   (cons 'hwuago (cons 926 (list 'hwuago 'audhlst 'ondtycr
                                                 'nxaa))))))
  (test no-aft-path
    (test-connection (cons 967 (list 'eimeihc 'oei 'pennw))
                     #f
                     (cons 829 (list 'fmii 'dhtae 'pennw))
                     'fmii
                     (table (cons 'pennw (cons 0 (list 'pennw)))))
    1 (path-table?)
    (match? path-table=?
            (table (cons 'pennw (cons 0 (list 'pennw)))
                   (cons 'fmii (cons 829 (list 'fmii 'dhtae 'pennw))))))
  (test no-path
    (test-connection (cons 777 (list 'hndl 'awflaia 'oncal 'eadn 'ern))
                     #f
                     #f
                     'okedoul
                     (table (cons 'ithelnh (cons 249 (list 'ithelnh 'oncal
                                                           'awflaia 'btib 
                                                           'ern)))
                            (cons 'ern (cons 0 (list 'ern)))))
    1 (path-table?)
    (match? path-table=?
            (table (cons 'ithelnh (cons 249 (list 'ithelnh 'oncal 'awflaia
                                                  'btib 'ern)))
                   (cons 'ern (cons 0 (list 'ern)))))))

(suite Floyd-Warshall
    ((two-level-path-table?
      (table-of symbol?
                (table-of symbol?
                          (pair-of number? (list-of symbol?))
                          symbol=?)
                symbol=?))
     (two-level-path-table=?
      (table-of= symbol=?
                 (table-of= symbol=? (pair-of= = (list-of= symbol=?))))))
  (test empty-graph
    (Floyd-Warshall (empty-graph))
    1 (two-level-path-table?)
    (match? two-level-path-table=? (table)))
  (test one-vertex-no-arcs
    (Floyd-Warshall (make-graph (set 'rpcina) (arcset)))
    1 (two-level-path-table?)
    (match? two-level-path-table=?
            (table (cons 'rpcina
                         (table (cons 'rpcina (cons 0 (list 'rpcina))))))))
  (test one-vertex-with-loop
    (Floyd-Warshall (make-graph (set 'tyotorl)
                                (arcset (make-arc 'tyotorl 'tyotorl 715))))
    1 (two-level-path-table?)
    (match? two-level-path-table=?
            (table (cons 'tyotorl
                         (table (cons 'tyotorl
                                      (cons 0 (list 'tyotorl))))))))
  (test small-graph
    (Floyd-Warshall (make-graph (set 'etun 'olosehv 'rlsb 'fhicsr)
                                (arcset (make-arc 'etun 'olosehv 801)
                                        (make-arc 'etun 'rlsb 485)
                                        (make-arc 'olosehv 'etun 921)
                                        (make-arc 'olosehv 'fhicsr 838)
                                        (make-arc 'rlsb 'etun 904)
                                        (make-arc 'rlsb 'fhicsr 24)
                                        (make-arc 'fhicsr 'olosehv 296)
                                        (make-arc 'fhicsr 'rlsb 707))))
    1 (two-level-path-table?)
    (match? two-level-path-table=?
            (table (cons 'etun
                         (table (cons 'etun (cons 0 (list 'etun)))
                                (cons 'olosehv
                                      (cons 801 (list 'olosehv 'etun)))
                                (cons 'rlsb (cons 485 (list 'rlsb 'etun)))
                                (cons 'fhicsr
                                      (cons 509 (list 'fhicsr 'rlsb
                                                      'etun)))))
                   (cons 'olosehv
                         (table (cons 'etun
                                      (cons 921 (list 'etun 'olosehv)))
                                (cons 'olosehv (cons 0 (list 'olosehv)))
                                (cons 'rlsb
                                      (cons 1406 (list 'rlsb 'etun
                                                       'olosehv)))
                                (cons 'fhicsr
                                      (cons 838 (list 'fhicsr 'olosehv)))))
                   (cons 'rlsb
                         (table (cons 'etun
                                      (cons 904 (list 'etun 'rlsb)))
                                (cons 'olosehv
                                      (cons 320 (list 'olosehv 'fhicsr
                                                      'rlsb)))
                                (cons 'rlsb (cons 0 (list 'rlsb)))
                                (cons 'fhicsr
                                      (cons 24 (list 'fhicsr 'rlsb)))))
                   (cons 'fhicsr
                         (table (cons 'etun
                                      (cons 1217 (list 'etun 'olosehv
                                                       'fhicsr)))
                                (cons 'olosehv
                                      (cons 296 (list 'olosehv 'fhicsr)))
                                (cons 'rlsb
                                      (cons 707 (list 'rlsb 'fhicsr)))
                                (cons 'fhicsr (cons 0 (list 'fhicsr))))))))
  (test smith-2.33
    (Floyd-Warshall (make-graph (set 1 2 3 4)
                                (arcset (make-arc 1 2 7)
                                        (make-arc 1 3 2)
                                        (make-arc 2 4 5)
                                        (make-arc 3 2 3)
                                        (make-arc 3 4 9)
                                        (make-arc 4 1 1)
                                        (make-arc 4 3 6))))
    1 ((table-of natural-number?
                 (table-of natural-number?
                           (pair-of number? (list-of natural-number?))
                           =)
                 =))
    (match? (table-of= = (table-of= = (pair-of= = (list-of= =))))
            (table (cons 1 (table (cons 1 (cons 0 (list 1)))
                                  (cons 2 (cons 5 (list 2 3 1)))
                                  (cons 3 (cons 2 (list 3 1)))
                                  (cons 4 (cons 10 (list 4 2 3 1)))))
                   (cons 2 (table (cons 1 (cons 6 (list 1 4 2)))
                                  (cons 2 (cons 0 (list 2)))
                                  (cons 3 (cons 8 (list 3 1 4 2)))
                                  (cons 4 (cons 5 (list 4 2)))))
                   (cons 3 (table (cons 1 (cons 9 (list 1 4 2 3)))
                                  (cons 2 (cons 3 (list 2 3)))
                                  (cons 3 (cons 0 (list 3)))
                                  (cons 4 (cons 8 (list 4 2 3)))))
                   (cons 4 (table (cons 1 (cons 1 (list 1 4)))
                                  (cons 2 (cons 6 (list 2 3 1 4)))
                                  (cons 3 (cons 3 (list 3 1 4)))
                                  (cons 4 (cons 0 (list 4)))))))))

;;; Test the procedures from the section on flow networks.

(suite capacity-graph? ()
  (test empty-graph
    (capacity-graph? (empty-graph))
    1 (boolean?)
    not)
  (test minimal-yes
    (capacity-graph? (make-graph (set 'paola 'daristh)
                                 (arcset (make-arc 'paola 'daristh 724))))
    1 (boolean?)
    true?)
  (test typical
    (capacity-graph? (make-graph (set 'usedh 'stiume 'eauf 'smeahn)
                                 (arcset (make-arc 'usedh 'stiume 620)
                                         (make-arc 'usedh 'eauf 396)
                                         (make-arc 'usedh 'smeahn 665)
                                         (make-arc 'stiume 'smeahn 311)
                                         (make-arc 'eauf 'smeahn 875))))
    1 (boolean?)
    true?)
  (test contains-self-loop
    (capacity-graph? (make-graph (set 'aptr 'noeolls 'wecn)
                                 (arcset (make-arc 'aptr 'aptr 10)
                                         (make-arc 'aptr 'wecn 77)
                                         (make-arc 'neoells 'aptr 440)
                                         (make-arc 'neoells 'wecn 116))))
    1 (boolean?)
    not)
  (test not-asymmetric
    (capacity-graph? (make-graph (set 'rndv 'aum 'tosfds 'inleed)
                                 (arcset (make-arc 'rndv 'aum 367)
                                         (make-arc 'rndv 'inleed 682)
                                         (make-arc 'tosfds 'rndv 284)
                                         (make-arc 'tosfds 'inleed 211)
                                         (make-arc 'inleed 'rndv 539)
                                         (make-arc 'inleed 'aum 989))))
    1 (boolean?)
    not)
  (test nonpositive-label
    (capacity-graph? (make-graph (set 'nshdtn 'ddu 'enrha)
                                 (arcset (make-arc 'ddu 'nshdtn 59)
                                         (make-arc 'ddu 'enrha 0)
                                         (make-arc 'enrha 'ddu 825))))
    1 (boolean?)
    not))

(suite flow-in-range? ()
  (test minimal-yes
    (flow-in-range? (make-graph (set 'ozdf 'ilum)
                                (arcset (make-arc 'ozdf 'ilum 839)))
                    (make-graph (set 'ozdf 'ilum)
                                (arcset (make-arc 'ozdf 'ilum 896))))
    1 (boolean?)
    true?)
  (test minimal-no
    (flow-in-range? (make-graph (set 'hyn 'rdoyeos)
                                (arcset (make-arc 'rdoyeos 'hyn 133)))
                    (make-graph (set 'hyn 'rdoyeos)
                                (arcset (make-arc 'rdoyeos 'hyn 131))))
    1 (boolean?)
    not)
  (test non-numeric-label
    (flow-in-range? (make-graph (set 'apiehg 'rcnardh 'nna 'heop)
                                (arcset (make-arc 'apiehg 'rcnardh null)
                                        (make-arc 'apiehg 'nna null)
                                        (make-arc 'rcnardh 'heop null)
                                        (make-arc 'nna 'heop null)))
                    (make-graph (set 'apiehg 'rcnardh 'nna 'heop)
                                (arcset (make-arc 'apiehg 'rcnardh 96)
                                        (make-arc 'apiehg 'nna 797)
                                        (make-arc 'rcnardh 'heop 519)
                                        (make-arc 'nna 'heop 30))))
    1 (boolean?)
    not)
  (test negative-label
    (flow-in-range? (make-graph (set 'sbebmd 'letnee 'mswi 'iihuuhc 'deia)
                                (arcset (make-arc 'sbebmb 'letnee 257)
                                        (make-arc 'sbebmb 'mswi 33)
                                        (make-arc 'letnee 'iihuuhc 348)
                                        (make-arc 'letnee 'deia -91)
                                        (make-arc 'mswi 'iihuuhc 33)
                                        (make-arc 'iihuuhc 'deia 381)))
                    (make-graph (set 'sbebmd 'letnee 'mswi 'iihuuhc 'deia)
                                (arcset (make-arc 'sbebmb 'letnee 613)
                                        (make-arc 'sbebmb 'mswi 659)
                                        (make-arc 'letnee 'iihuuhc 348)
                                        (make-arc 'letnee 'deia 654)
                                        (make-arc 'mswi 'iihuuhc 759)
                                        (make-arc 'iihuuhc 'deia 381))))
    1 (boolean?)
    not)
  (test zero-flow
    (flow-in-range? (make-graph (set 'rgedt 'oihpn 'hoiae 'ieides)
                                (arcset (make-arc 'rgedt 'oihpn 0)
                                        (make-arc 'rgedt 'hoiae 0)
                                        (make-arc 'oihpn 'hoiae 0)
                                        (make-arc 'oihpn 'ieides 0)
                                        (make-arc 'hoiae 'ieides 0)))
                    (make-graph (set 'rgedt 'oihpn 'hoiae 'ieides)
                                (arcset (make-arc 'rgedt 'oihpn 649)
                                        (make-arc 'rgedt 'hoiae 545)
                                        (make-arc 'oihpn 'hoiae 335)
                                        (make-arc 'oihpn 'ieides 507)
                                        (make-arc 'hoiae 'ieides 330))))
    1 (boolean?)
    true?)
  (test maximum-flow
    (flow-in-range? (make-graph (set 'nho 'lonnpim 'iongomg 'itp)
                                (arcset (make-arc 'nho 'lonnpim 590)
                                        (make-arc 'nho 'iongomg 201)
                                        (make-arc 'lonnpim 'iongomg 351)
                                        (make-arc 'lonnpim 'itp 239)
                                        (make-arc 'iongomg 'itp 552)))
                    (make-graph (set 'nho 'lonnpim 'iongomg 'itp)
                                (arcset (make-arc 'nho 'lonnpim 590)
                                        (make-arc 'nho 'iongomg 201)
                                        (make-arc 'lonnpim 'iongomg 351)
                                        (make-arc 'lonnpim 'itp 239)
                                        (make-arc 'iongomg 'itp 552))))
    1 (boolean?)
    true?)
  (test exceeds-maximum
    (flow-in-range? (make-graph (set 'oldtdgd 'sipy 'hetns)
                                (arcset (make-arc 'oldtdgd 'sipy 727)
                                        (make-arc 'sipy 'hetns 727)))
                    (make-graph (set 'oldtdgd 'sipy 'hetns)
                                (arcset (make-arc 'oldtdgd 'sipy 727)
                                        (make-arc 'sipy 'hetns 628))))
    1 (boolean?)
    not))

(suite net-flow ()
  (test no-arcs
    (net-flow 'ehe (make-graph (set 'adwtws 'ehe 'ope) (arcset)))
    1 (number?)
    zero?)
  (test mid-flow-zero
    (net-flow 'wrtse (make-graph (set 'derl 'niet 'wrtse 'raos 'igy)
                                 (arcset (make-arc 'derl 'niet 238)
                                         (make-arc 'derl 'wrtse 749)
                                         (make-arc 'niet 'wrtse 146)
                                         (make-arc 'niet 'raos 92)
                                         (make-arc 'wrtse 'raos 217)
                                         (make-arc 'wrtse 'igy 678)
                                         (make-arc 'raos 'igy 309))))
    1 (number?)
    zero?)
  (test at-origin
    (net-flow 'wiunn (make-graph (set 'wiunn 'altt 'sae 'ime)
                                 (arcset (make-arc 'wiunn 'altt 218)
                                         (make-arc 'wiunn 'sae 468)
                                         (make-arc 'altt 'ime 218)
                                         (make-arc 'sae 'ime 468))))
    1 (number?)
    (match? = 686))
  (test at-destination
    (net-flow 'riaret (make-graph (set 'ech 'heonhpe 'doae 'riaret)
                                  (arcset (make-arc 'ech 'heonhpe 798)
                                          (make-arc 'ech 'doae 352)
                                          (make-arc 'heonhpe 'doae 207)
                                          (make-arc 'heonhpe 'riaret 591)
                                          (make-arc 'doae 'riaret 559))))
    1 (number?)
    (match? = -1150)))

(suite flow-conserved? ()
  (test direct-flow
    (flow-conserved? (make-graph (set 'ssatis 'iehltl)
                                 (arcset (make-arc 'ssatis 'iehltl 529)))
                     'ssatis
                     'iehltl)
    1 (boolean?)
    true?)
  (test zero-flow
    (flow-conserved? (make-graph (set 'wkum 'nchbeao 'nfeaoo)
                                 (arcset (make-arc 'wkum 'nchbeao 0)
                                         (make-arc 'nchbeao 'wkum 0)
                                         (make-arc 'nchbeao 'nfeaoo 0)
                                         (make-arc 'nfeaoo 'nchbeao 0)))
                     'wkum
                     'nfeaoo)
    1 (boolean?)
    true?)
  (test implicit-zero-flow
    (flow-conserved? (make-graph (set 'frwem 'oenra 'ytcnn 'dte) (arcset))
                     'frwem
                     'dte)
    1 (boolean?)
    true?)
  (test larger-yes
    (flow-conserved? (make-graph (set 'ynwusn 'oilasi 'ind 'emdteil 'newte)
                                 (arcset (make-arc 'ynwusn 'oilasi 832)
                                         (make-arc 'ynwusn 'ind 587)
                                         (make-arc 'oilasi 'ind 222)
                                         (make-arc 'oilasi 'emdteil 610)
                                         (make-arc 'ind 'emdteil 494)
                                         (make-arc 'ind 'newte 315)
                                         (make-arc 'emdteil 'newte 1104)))
                     'ynwusn
                     'newte)
    1 (boolean?)
    true?)
  (test minimal-no
    (flow-conserved? (make-graph (set 'vneo 'tiegahn 'oeobaee)
                                 (arcset (make-arc 'vneo 'tiegahn 334)
                                         (make-arc 'tiegahn 'oeobaee 604)))
                     'vneo
                     'oeobaee)
    1 (boolean?)
    not)
  (test larger-no
    (flow-conserved? (make-graph (set 'ato 'hthon 'ygooalo 'wtoag 'sgg)
                                 (arcset (make-arc 'ato 'hthon 290)
                                         (make-arc 'ato 'ygooalo 471)
                                         (make-arc 'ato 'wtoag 755)
                                         (make-arc 'hthon 'sgg 290)
                                         (make-arc 'ygooalo 'wtoag 471)
                                         (make-arc 'wtoag 'sgg 1227)))
                     'ato
                     'sgg)
    1 (boolean?)
    not))

(suite flow-graph? ()
  (test small-yes
    (flow-graph? (make-graph (set 'kods 'tyeq 'erdt)
                             (arcset (make-arc 'kods 'tyeq 471)
                                     (make-arc 'tyeq 'erdt 471)))
                 (make-graph (set 'kods 'tyeq 'erdt)
                             (arcset (make-arc 'kods 'tyeq 755)
                                     (make-arc 'tyeq 'erdt 930)))
                 'kods
                 'erdt)
    1 (boolean?)
    true?)
  (test small-wrong-vertices
    (flow-graph? (make-graph (set 'eihnnt 'tsaery 'msa)
                             (arcset (make-arc 'eihnnt 'tsaery 509)
                                     (make-arc 'tsaery 'msa 509)))
                 (make-graph (set 'uutr 'raewh 'tunlhlp)
                             (arcset (make-arc 'uutr 'raewh 824)
                                     (make-arc 'raewh 'tunlhlp 528)))
                 'uutr
                 'raewh)
    1 (boolean?)
    not)
  (test small-unicorn-vertex
    (flow-graph? (make-graph (set 'sdd 'eteobtn 'wioea)
                             (arcset  (make-arc 'eteobtn 'wioea 212)
                                     (make-arc 'wioea 'sdd 212)))
                 (make-graph (set 'sdd 'eteobtn 'wioea)
                             (arcset (make-arc 'eteobtn 'wioea 625)
                                     (make-arc 'wioea 'sdd 489)))
                 'eteobtn
                 'nbri)
    1 (boolean?)
    not)
  (test small-same-origin-and-destination
    (flow-graph? (make-graph (set 'tmoady 'nhiel)
                             (arcset (make-arc 'tmoady 'nhiel 0)
                                     (make-arc 'nhiel 'tmoady 0)))
                 (make-graph (set 'tmoady 'nhiel)
                             (arcset (make-arc 'tmoady 'nhiel 427)
                                     (make-arc 'nhiel 'tmoady 662)))
                 'tmoady
                 'tmoady)
    1 (boolean?)
    not)
  (test small-violating-capacity
    (flow-graph? (make-graph (set 'edot 'wnhl 'nsn)
                             (arcset (make-arc 'edot 'wnhl 191)
                                     (make-arc 'wnhl 'nsn 191)))
                 (make-graph (set 'edot 'wnhl 'nsn)
                             (arcset (make-arc 'edot 'wnhl 525)
                                     (make-arc 'wnhl 'nsn 78)))
                 'edot
                 'nsn)
    1 (boolean?)
    not)
  (test small-unconserved
    (flow-graph? (make-graph (set 'oeoe 'annlues 'nlhdbhf)
                             (arcset (make-arc 'oeoe 'annlues 274)
                                     (make-arc 'annlues 'nlhdbhf 734)))
                 (make-graph (set 'oeoe 'annlues 'nlhdbhf)
                             (arcset (make-arc 'oeoe 'annlues 837)
                                     (make-arc 'annlues 'nlhdbhf 949)))
                 'oeoe
                 'nlhdbhf)
    1 (boolean?)
    not)
  (test zero-flow
    (flow-graph? (make-graph (set 'trsmeir 'rdsn 'mlacd 'addeese 'slcse)
                             (arcset (make-arc 'trsmeir 'rdsn 0)
                                     (make-arc 'trsmeir 'mlacd 0)
                                     (make-arc 'rdsn 'addeese 0)
                                     (make-arc 'mlacd 'rdsn 0)
                                     (make-arc 'mlacd 'slcse 0)
                                     (make-arc 'addeese 'mlacd 0)
                                     (make-arc 'addeese 'slces 0)))
                 (make-graph (set 'trsmeir 'rdsn 'mlacd 'addeese 'slcse)
                             (arcset (make-arc 'trsmeir 'rdsn 548)
                                     (make-arc 'trsmeir 'mlacd 279)
                                     (make-arc 'rdsn 'addeese 239)
                                     (make-arc 'mlacd 'rdsn 287)
                                     (make-arc 'mlacd 'slcse 371)
                                     (make-arc 'addeese 'mlacd 537)
                                     (make-arc 'addeese 'slces 487)))
                 'trsmeir
                 'slcse)
    1 (boolean?)
    true?)
  (test large-yes
    (flow-graph? (make-graph (set 'raicsoe 'afe 'anostt 'enao 'ihh)
                             (arcset (make-arc 'raicsoe 'afe 33)
                                     (make-arc 'raicsoe 'anostt 18)
                                     (make-arc 'afe 'enao 33)
                                     (make-arc 'anostt 'ihh 43)
                                     (make-arc 'enao 'anostt 25)
                                     (make-arc 'enao 'ihh 8)))
                 (make-graph (set 'raicsoe 'afe 'anostt 'enao 'ihh)
                             (arcset (make-arc 'raicsoe 'afe 544)
                                     (make-arc 'raicsoe 'anostt 85)
                                     (make-arc 'afe 'enao 39)
                                     (make-arc 'anostt 'ihh 366)
                                     (make-arc 'enao 'anostt 649)
                                     (make-arc 'enao 'ihh 382)))
                 'raicsoe
                 'ihh)
    1 (boolean?)
    true?))

(suite flow-value ()
  (test no-arcs
    (flow-value 'saacaet (make-graph (set 'saacaet 'eeerrre 'csyot 'blls
                                          'snnt 'eawaa 'sethca)
                                     (arcset)))
    1 (number?)
    zero?)
  (test zero-flow
    (flow-value 'gsans (make-graph (set 'gsans 'tpstae 'hhnoue 'neon)
                                   (arcset (make-arc 'gsans 'tpstae 0)
                                           (make-arc 'gsans 'hhnoue 0)
                                           (make-arc 'gsans 'neon 0)
                                           (make-arc 'tpstae 'hhnoue 0)
                                           (make-arc 'hhnoue 'neon 0))))
    1 (number?)
    zero?)
  (test typical
    (flow-value 'elqbe
                (make-graph (set 'elqbe 'duontli 'wteh 'irsdh 'whwrar)
                            (arcset (make-arc 'elqbe 'duontli 194)
                                    (make-arc 'elqbe 'wteh 157)
                                    (make-arc 'duontli 'wteh 140)
                                    (make-arc 'duontli 'irsdh 54)
                                    (make-arc 'wteh 'irsdh 297)
                                    (make-arc 'irsdh 'whwrar 351))))
    1 (number?)
    (match? = 351)))

(suite zero-flow ()
  (test minimal
    (zero-flow (make-graph (set 'gaeg 'tmhreuh)
                           (arcset (make-arc 'gaeg 'tmhreuh 942))))
    1 (graph?)
    (match? graph=? (make-graph (set 'gaeg 'tmhreuh)
                                (arcset (make-arc 'gaeg 'tmhreuh 0)))))
  (test larger-example
    (zero-flow (make-graph (set 'deogso 'ebmot 'lieo 'suhyoes)
                           (arcset (make-arc 'deogso 'ebmot 985)
                                   (make-arc 'deogso 'lieo 756)
                                   (make-arc 'deogso 'suhyoes 410)
                                   (make-arc 'lieo 'suhyoes 152)
                                   (make-arc 'suhyoes 'ebmot 13))))
    1 (graph?)
    (match? graph=? (make-graph (set 'deogso 'ebmot 'lieo 'suhyoes)
                                (arcset (make-arc 'deogso 'ebmot 0)
                                        (make-arc 'deogso 'lieo 0)
                                        (make-arc 'deogso 'suhyoes 0)
                                        (make-arc 'lieo 'suhyoes 0)
                                        (make-arc 'suhyoes 'ebmot 0))))))

(suite residual-network ()
  (test zero-flow
    (residual-network
      (make-graph (set 'aeiis 'tpgd 'ttiose 'yunhfpd)
                  (arcset (make-arc 'aeiis 'tpgd 219)
                          (make-arc 'aeiis 'ttiose 673)
                          (make-arc 'tpgd 'yunhfpd 230)
                          (make-arc 'ttiose 'yunhfpd 903)))
      (make-graph (set 'aeiis 'tpgd 'ttiose 'yunhfpd)
                  (arcset (make-arc 'aeiis 'tpgd 0)
                          (make-arc 'aeiis 'ttiose 0)
                          (make-arc 'tpgd 'yunhfpd 0)
                          (make-arc 'ttiose 'yunhfpd 0))))
    1 (graph?)
    (match? graph=? (make-graph (set 'aeiis 'tpgd 'ttiose 'yunhfpd)
                                (set (make-arc 'aeiis 'tpgd 219)
                                     (make-arc 'aeiis 'ttiose 673)
                                     (make-arc 'tpgd 'yunhfpd 230)
                                     (make-arc 'ttiose 'yunhfpd 903)))))
  (test full-capacity
    (residual-network
      (make-graph (set 'ehign 'sdoinde 'skagh 'tsso 'hetchts)
                  (arcset (make-arc 'ehign 'sdoinde 903)
                          (make-arc 'sdoinde 'skagh 305)
                          (make-arc 'sdoinde 'tsso 598)
                          (make-arc 'skagh 'tsso 108)
                          (make-arc 'skagh 'hetchts 197)
                          (make-arc 'tsso 'hetchts 706)))
      (make-graph (set 'ehign 'sdoinde 'skagh 'tsso 'hetchts)
                  (arcset (make-arc 'ehign 'sdoinde 903)
                          (make-arc 'sdoinde 'skagh 305)
                          (make-arc 'sdoinde 'tsso 598)
                          (make-arc 'skagh 'tsso 108)
                          (make-arc 'skagh 'hetchts 197)
                          (make-arc 'tsso 'hetchts 706))))
    1 (graph?)
    (match? graph=? (make-graph (set 'ehign 'sdoinde 'skagh 'tsso 'hetchts)
                                (arcset (make-arc 'sdoinde 'ehign 903)
                                        (make-arc 'skagh 'sdoinde 305)
                                        (make-arc 'tsso 'sdoinde 598)
                                        (make-arc 'tsso 'skagh 108)
                                        (make-arc 'hetchts 'skagh 197)
                                        (make-arc 'hetchts 'tsso 706)))))
  (test maximal-flow
    (residual-network
      (make-graph (set 'phen 'dtp 'spqtnoi 'eloeta 'manin)
                  (set (make-arc 'phen 'dtp 360)
                       (make-arc 'phen 'eloeta 820)
                       (make-arc 'dtp 'spqtnoi 698)
                       (make-arc 'spqtnoi 'manin 995)
                       (make-arc 'eloeta 'dtp 645)
                       (make-arc 'eloeta 'spqtnoi 590)))
      (make-graph (set 'phen 'dtp 'spqtnoi 'eloeta 'manin)
                  (arcset (make-arc 'phen 'dtp 360)
                          (make-arc 'phen 'eloeta 635)
                          (make-arc 'dtp 'spqtnoi 405)
                          (make-arc 'spqtnoi 'manin 995)
                          (make-arc 'eloeta 'dtp 45)
                          (make-arc 'eloeta 'spqtnoi 590))))
    1 (graph?)
    (match? graph=? (make-graph (set 'phen 'dtp 'spqtnoi 'eloeta 'manin)
                                (arcset (make-arc 'phen 'eloeta 185)
                                        (make-arc 'dtp 'phen 360)
                                        (make-arc 'dtp 'spqtnoi 293)
                                        (make-arc 'dtp 'eloeta 45)
                                        (make-arc 'spqtnoi 'dtp 405)
                                        (make-arc 'spqtnoi 'eloeta 590)
                                        (make-arc 'eloeta 'phen 635)
                                        (make-arc 'eloeta 'dtp 600)
                                        (make-arc 'manin 'spqtnoi 995)))))
  (test intermediate-flow
    (residual-network
      (make-graph (set 'edytbnt 'kds 'atrioh 'haoiixe 'ewhnoag)
                  (arcset (make-arc 'edytbnt 'kds 590)
                          (make-arc 'edytbnt 'atrioh 512)
                          (make-arc 'kds 'haoiixe 29)
                          (make-arc 'kds 'ewhnoag 642)
                          (make-arc 'atrioh 'kds 116)
                          (make-arc 'atrioh 'haoiixe 808)
                          (make-arc 'haoiixe 'ewhnoag 378)))
      (make-graph (set 'edytbnt 'kds 'atrioh 'haoiixe 'ewhnoag)
                  (arcset (make-arc 'edytbnt 'kds 65)
                          (make-arc 'edytbnt 'atrioh 410)
                          (make-arc 'kds 'haoiixe 15)
                          (make-arc 'kds 'ewhnoag 111)
                          (make-arc 'atrioh 'kds 61)
                          (make-arc 'atrioh 'haoiixe 349)
                          (make-arc 'haoiixe 'ewhnoag 364))))
    1 (graph?)
    (match? graph=?
            (make-graph (set 'edytbnt 'kds 'atrioh 'haoiixe 'ewhnoag)
                        (arcset (make-arc 'edytbnt 'kds 525)
                                (make-arc 'edytbnt 'atrioh 102)
                                (make-arc 'kds 'edytbnt 65)
                                (make-arc 'kds 'atrioh 61)
                                (make-arc 'kds 'haoiixe 14)
                                (make-arc 'kds 'ewhnoag 531)
                                (make-arc 'atrioh 'edytbnt 410)
                                (make-arc 'atrioh 'kds 55)
                                (make-arc 'atrioh 'haoiixe 459)
                                (make-arc 'haoiixe 'kds 15)
                                (make-arc 'haoiixe 'atrioh 349)
                                (make-arc 'haoiixe 'ewhnoag 14)
                                (make-arc 'ewhnoag 'kds 111)
                                (make-arc 'ewhnoag 'haoiixe 364)))))
  (test antiparallelism-transformed
    (residual-network
      (make-graph (set 'etavtr 'sso 'hman 'hfesesw 'pdno)
                  (arcset (make-arc 'etavtr 'sso 986)
                          (make-arc 'etavtr 'hman 640)
                          (make-arc 'sso 'hman 554)
                          (make-arc 'sso 'hfesesw 85)
                          (make-arc 'hman 'hfesesw 749)
                          (make-arc 'hman 'pdno 94)
                          (make-arc 'pdno 'sso 94)))
      (make-graph (set 'etavtr 'sso 'hman 'hfesesw 'pdno)
                  (arcset (make-arc 'etavtr 'sso 194)
                          (make-arc 'etavtr 'hman 640)
                          (make-arc 'sso 'hman 109)
                          (make-arc 'sso 'hfesesw 85)
                          (make-arc 'hman 'hfesesw 749)
                          (make-arc 'hman 'pdno 0)
                          (make-arc 'pdno 'sso 0))))
    1 (graph?)
    (match? graph=?
            (make-graph (set 'etavtr 'sso 'hman 'hfesesw 'pdno)
                        (arcset (make-arc 'etavtr 'sso 792)
                                (make-arc 'sso 'etavtr 194)
                                (make-arc 'sso 'hman 445)
                                (make-arc 'hman 'etavtr 640)
                                (make-arc 'hman 'sso 109)
                                (make-arc 'hman 'pdno 94)
                                (make-arc 'hfesesw 'sso 85)
                                (make-arc 'hfesesw 'hman 749)
                                (make-arc 'pdno 'sso 94))))))

(suite augment-flow ()
  (test minimal
    (augment-flow (make-graph (set 'edhesap 'towme)
                              (arcset (make-arc 'edhesap 'towme 0)))
                  (make-graph (set 'edhesap 'towme)
                              (arcset (make-arc 'edhesap 'towme 489)))
                  (list 'edhesap 'towme))
    1 (graph?)
    (match? graph=? (make-graph (set 'edhesap 'towme)
                                (arcset (make-arc 'edhesap 'towme 489)))))
  (test Cormen
    (augment-flow (make-graph (set 's 'v1 'v2 'v3 'v4 't)
                              (arcset (make-arc 's 'v1 11)
                                      (make-arc 's 'v2 8)
                                      (make-arc 'v1 'v3 12)
                                      (make-arc 'v2 'v1 1)
                                      (make-arc 'v2 'v4 11)
                                      (make-arc 'v3 'v2 4)
                                      (make-arc 'v3 't 15)
                                      (make-arc 'v4 'v3 7)
                                      (make-arc 'v4 't 4)))
                  (make-graph (set 's 'v1 'v2 'v3 'v4 't)
                              (arcset (make-arc 's 'v1 5)
                                      (make-arc 's 'v2 5)
                                      (make-arc 'v1 's 11)
                                      (make-arc 'v1 'v2 1)
                                      (make-arc 'v2 's 8)
                                      (make-arc 'v2 'v1 3)
                                      (make-arc 'v2 'v3 4)
                                      (make-arc 'v2 'v4 3)
                                      (make-arc 'v3 'v1 12)
                                      (make-arc 'v3 'v2 5)
                                      (make-arc 'v3 'v4 7)
                                      (make-arc 'v3 't 5)
                                      (make-arc 'v4 'v2 11)
                                      (make-arc 't 'v3 5)
                                      (make-arc 't 'v4 4)))
                  (list 's 'v2 'v3 't))
    1 (graph?)
    (match? graph=? (make-graph (set 's 'v1 'v2 'v3 'v4 't)
                                (arcset (make-arc 's 'v1 11)
                                        (make-arc 's 'v2 12)
                                        (make-arc 'v1 'v3 12)
                                        (make-arc 'v2 'v1 1)
                                        (make-arc 'v2 'v4 11)
                                        (make-arc 'v3 'v2 0)
                                        (make-arc 'v3 't 19)
                                        (make-arc 'v4 'v3 7)
                                        (make-arc 'v4 't 4)))))) 

(suite maximum-network-flow ()
  (test disconnected
    (maximum-network-flow (make-graph (set 'source 'sink) (arcset))
                    'source
                    'sink)
    1 (graph?)
    (match? graph=? (make-graph (set 'source 'sink) (arcset))))
  (test trivial
    (maximum-network-flow (make-graph (set 'ifn 'srgel)
                                (arcset (make-arc 'ifn 'srgel 100)))
                    'ifn
                    'srgel)
    1 (graph?)
    (match? graph=? (make-graph (set 'ifn 'srgel)
                                (arcset (make-arc 'ifn 'srgel 100)))))
  (test complicated
    (maximum-network-flow (make-graph (set 'mwetu 'ydn 'lnto 'niceuhe 'flfki)
                                (arcset (make-arc 'mwetu 'niceuhe 224)
                                        (make-arc 'mwetu 'flfki 179)
                                        (make-arc 'ydn 'lnto 528)
                                        (make-arc 'lnto 'niceuhe 248)
                                        (make-arc 'lnto 'flfki 117)
                                        (make-arc 'niceuhe 'ydn 606)))
                    'mwetu
                    'flfki)
    1 (graph?)
    (match? graph=? (make-graph (set 'mwetu 'ydn 'lnto 'niceuhe 'flfki)
                                (set (make-arc 'mwetu 'niceuhe 117)
                                     (make-arc 'mwetu 'flfki 179)
                                     (make-arc 'ydn 'lnto 117)
                                     (make-arc 'lnto 'niceuhe 0)
                                     (make-arc 'lnto 'flfki 117)
                                     (make-arc 'niceuhe 'ydn 117)))))

  ;; Kingston, exercise 12.10.

  (test kingston
    (maximum-network-flow (make-graph (set 'a 'b 'c 'd 'e 'f 'g 'h)
                                (arcset (make-arc 'a 'c 12)
                                        (make-arc 'a 'e 8)
                                        (make-arc 'c 'd 10)
                                        (make-arc 'c 'e 5)
                                        (make-arc 'd 'b 8)
                                        (make-arc 'd 'f 5)
                                        (make-arc 'e 'f 10)
                                        (make-arc 'e 'g 5)
                                        (make-arc 'f 'b 12)
                                        (make-arc 'f 'h 5)
                                        (make-arc 'g 'c 5)
                                        (make-arc 'h 'd 5)))
                    'a
                    'b)
    1 (graph?)
    (match? graph=? (make-graph (set 'a 'b 'c 'd 'e 'f 'g 'h)
                                (arcset (make-arc 'a 'c 12)
                                        (make-arc 'a 'e 8)
                                        (make-arc 'c 'd 10)
                                        (make-arc 'c 'e 2)
                                        (make-arc 'd 'b 8)
                                        (make-arc 'd 'f 2)
                                        (make-arc 'e 'f 10)
                                        (make-arc 'e 'g 0)
                                        (make-arc 'f 'b 12)
                                        (make-arc 'f 'h 0)
                                        (make-arc 'g 'c 0)
                                        (make-arc 'h 'd 0)))))

  ;; Even, Figure 5.1.

  (test even
    (maximum-network-flow (make-graph (set 's 'a 'b 'c 'd 't)
                                (arcset (make-arc 's 'a 15)
                                        (make-arc 's 'c 4)
                                        (make-arc 'a 'b 12)
                                        (make-arc 'b 'c 3)
                                        (make-arc 'b 't 7)
                                        (make-arc 'c 'd 10)
                                        (make-arc 'd 'a 5)
                                        (make-arc 'd 't 10)))
                    's
                    't)
    1 (graph?)
    (match? graph=? (make-graph (set 's 'a 'b 'c 'd 't)
                                (arcset (make-arc 's 'a 10)
                                        (make-arc 's 'c 4)
                                        (make-arc 'a 'b 10)
                                        (make-arc 'b 'c 3)
                                        (make-arc 'b 't 7)
                                        (make-arc 'c 'd 7)
                                        (make-arc 'd 'a 0)
                                        (make-arc 'd 't 7)))))

  ;; Sedgewick, _Algorithms_, figure 33.2

  (test sedgewick
    (maximum-network-flow (make-graph (set 'a 'b 'c 'd 'e 'f)
                                (arcset (make-arc 'a 'b 6)
                                        (make-arc 'a 'c 8)
                                        (make-arc 'b 'd 6)
                                        (make-arc 'b 'e 3)
                                        (make-arc 'c 'd 3)
                                        (make-arc 'c 'e 3)
                                        (make-arc 'd 'f 6)
                                        (make-arc 'e 'f 6)))
                    'a
                    'f)
    1 (graph?)
    (match? graph=? (make-graph (set 'a 'b 'c 'd 'e 'f)
                                (arcset (make-arc 'a 'b 6)
                                        (make-arc 'a 'c 6)
                                        (make-arc 'b 'd 3)
                                        (make-arc 'b 'e 3)
                                        (make-arc 'c 'd 3)
                                        (make-arc 'c 'e 3)
                                        (make-arc 'd 'f 6)
                                        (make-arc 'e 'f 6)))))
  ;; Manber, Figure 7.40

  (test manber
    (maximum-network-flow (make-graph (set 's 't 'u 'v 'w 'x 'y 'z)
                                (arcset (make-arc 's 'v 7)
                                        (make-arc 's 'x 5)
                                        (make-arc 's 'z 6)
                                        (make-arc 'u 't 7)
                                        (make-arc 'v 'u 4)
                                        (make-arc 'v 'w 1)
                                        (make-arc 'w 't 6)
                                        (make-arc 'w 'u 4)
                                        (make-arc 'x 'u 3)
                                        (make-arc 'x 'y 4)
                                        (make-arc 'y 't 3)
                                        (make-arc 'z 'w 5)))
                    's
                    't)
    1 (graph?)
    (^or (match? graph=? (make-graph (set 's 't 'u 'v 'w 'x 'y 'z)
                                     (arcset (make-arc 's 'v 5)
                                             (make-arc 's 'x 5)
                                             (make-arc 's 'z 5)
                                             (make-arc 'u 't 7)
                                             (make-arc 'v 'u 4)
                                             (make-arc 'v 'w 1)
                                             (make-arc 'w 't 5)
                                             (make-arc 'w 'u 1)
                                             (make-arc 'x 'u 2)
                                             (make-arc 'x 'y 3)
                                             (make-arc 'y 't 3)
                                             (make-arc 'z 'w 5))))
         (match? graph=? (make-graph (set 's 't 'u 'v 'w 'x 'y 'z)
                                     (arcset (make-arc 's 'v 5)
                                             (make-arc 's 'x 5)
                                             (make-arc 's 'z 5)
                                             (make-arc 'u 't 7)
                                             (make-arc 'v 'u 4)
                                             (make-arc 'v 'w 1)
                                             (make-arc 'w 't 6)
                                             (make-arc 'w 'u 0)
                                             (make-arc 'x 'u 3)
                                             (make-arc 'x 'y 2)
                                             (make-arc 'y 't 2)
                                             (make-arc 'z 'w 5))))
         (match? graph=? (make-graph (set 's 't 'u 'v 'w 'x 'y 'z)
                                     (arcset (make-arc 's 'v 5)
                                             (make-arc 's 'x 5)
                                             (make-arc 's 'z 5)
                                             (make-arc 'u 't 6)
                                             (make-arc 'v 'u 4)
                                             (make-arc 'v 'w 1)
                                             (make-arc 'w 't 6)
                                             (make-arc 'w 'u 0)
                                             (make-arc 'x 'u 2)
                                             (make-arc 'x 'y 3)
                                             (make-arc 'y 't 3)
                                             (make-arc 'z 'w 5))))))

  ;; Cormen, figure 26.3.

  (test cormen
    (maximum-network-flow (make-graph (set 's 'v1 'v2 'v3 'v4 't)
                                (arcset (make-arc 's 'v1 16)
                                        (make-arc 's 'v2 13)
                                        (make-arc 'v1 'v3 12)
                                        (make-arc 'v2 'v1 4)
                                        (make-arc 'v2 'v4 14)
                                        (make-arc 'v3 'v2 9)
                                        (make-arc 'v3 't 20)
                                        (make-arc 'v4 'v3 7)
                                        (make-arc 'v4 't 4)))
                    's
                    't)
    1 (graph?)
    (^or (match? graph=? (make-graph (set 's 'v1 'v2 'v3 'v4 't)
                                     (arcset (make-arc 's 'v1 11)
                                             (make-arc 's 'v2 12)
                                             (make-arc 'v1 'v3 12)
                                             (make-arc 'v2 'v1 1)
                                             (make-arc 'v2 'v4 11)
                                             (make-arc 'v3 'v2 0)
                                             (make-arc 'v3 't 19)
                                             (make-arc 'v4 'v3 7)
                                             (make-arc 'v4 't 4))))
         (match? graph=? (make-graph (set 's 'v1 'v2 'v3 'v4 't)
                                     (arcset (make-arc 's 'v1 10)
                                             (make-arc 's 'v2 13)
                                             (make-arc 'v1 'v3 12)
                                             (make-arc 'v2 'v1 2)
                                             (make-arc 'v2 'v4 11)
                                             (make-arc 'v3 'v2 0)
                                             (make-arc 'v3 't 19)
                                             (make-arc 'v4 'v3 7)
                                             (make-arc 'v4 't 4))))
         (match? graph=? (make-graph (set 's 'v1 'v2 'v3 'v4 't)
                                     (arcset (make-arc 's 'v1 12)
                                             (make-arc 's 'v2 11)
                                             (make-arc 'v1 'v3 12)
                                             (make-arc 'v2 'v1 0)
                                             (make-arc 'v2 'v4 11)
                                             (make-arc 'v3 'v2 0)
                                             (make-arc 'v3 't 19)
                                             (make-arc 'v4 'v3 7)
                                             (make-arc 'v4 't 4)))))))

;;; copyright (C) 2011, 2017 John David Stone

;;; This program is free software.  You may redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation -- either version 3 of the License, or (at
;;; your option) any later version.  A copy of the GNU General Public
;;; License is available on the World Wide Web at
;;;
;;;                http://www.gnu.org/licenses/gpl.html
;;;
;;; This program is distributed in the hope that it will be useful, but
;;; WITHOUT ANY WARRANTY -- without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; General Public License for more details.
